/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// HtmlPage.cpp: implementation of the HtmlPage class.
//
//////////////////////////////////////////////////////////////////////

#include "Common.h"
#include "UrlFile.h"
#include "UrlFolder.h"
#include "HtmlPage.h" // dep_chk_needed 
#include "HtmlSimpleText.h"
#include "HtmlTags.h" // dep_chk_unknown 

#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

HtmlPage::HtmlPage()
{
  m_pOutStream = NULL;
  m_bDeleteOutStream = false;
  
  m_pFile = NULL;
}

HtmlPage::HtmlPage(UrlFile* pFile)
{
  m_pFile = pFile;
  m_pOutStream = pFile->GetOutStream ();
  m_bDeleteOutStream = true;
}

HtmlPage::~HtmlPage()
{
  // close all open tags
  
  while (!m_stackBegunTags.empty())
    End();
  
  if (m_bDeleteOutStream)
    {
      assert (m_pOutStream);
      delete m_pOutStream;
    }
}

inline void RenderAndDeleteHtmlObject (HtmlObject* pObject,ostream* pOut,HtmlPage* pPage)
{
  assert (pObject);
  pObject->Render (pOut,pPage);
  delete pObject;
}

void HtmlPage::SetOutStream (ostream* pOutStream,bool bDeleteOutStream)
{
  assert (m_stackBegunTags.empty()); // otherwise will miss parts of the current tag
  assert (!m_pOutStream);
  assert (pOutStream);
  
  m_pOutStream = pOutStream;
  m_bDeleteOutStream = bDeleteOutStream;
  
  // write everything there is to write
  
  for_each (m_vectObjects.begin(),m_vectObjects.end(),RenderAndDeleteHtmlObject,m_pOutStream,this);
  m_vectObjects.clear();
}

void HtmlPage::AddObject(HtmlObject* pObject)
{
  if (!pObject)
    return; // nothing to be added
  
  if (m_pOutStream)
    RenderAndDeleteHtmlObject(pObject,m_pOutStream,this);
  else
    {
      if (!m_stackBegunTags.empty())
        {
          *(m_stackBegunTags.top()) << pObject;
        }
      else
        m_vectObjects.push_back(pObject);
    }
}

void HtmlPage::AddText(const string& strText)
{
	if (m_pOutStream)
		*m_pOutStream << strText;
	else
		AddObject (new HtmlSimpleText (strText));
}

void HtmlPage::AddText(const char* pText)
{
    assert (pText);

        if (m_pOutStream)
                *m_pOutStream << pText;
        else
                AddObject (new HtmlSimpleText (pText));
}

void HtmlPage::AddText(char token)
{
	if (m_pOutStream)
		*m_pOutStream << token;
	else
		AddObject (new HtmlSimpleText (token));
}

void HtmlPage::Begin (HtmlBinaryTag* pTag)
{
	if (m_pOutStream)
	{
		pTag->RenderStartTag (m_pOutStream,this);
	    pTag->RenderContents(m_pOutStream,this);
	}
	m_stackBegunTags.push(pTag);
}

void HtmlPage::End ()
{
	assert (!m_stackBegunTags.empty());

	HtmlBinaryTag* pTag = m_stackBegunTags.top();

	if (m_pOutStream)
	{
		pTag->RenderEndTag (m_pOutStream);
		delete pTag;
	}
	else
		AddObject (pTag); // to be rendered later	
	m_stackBegunTags.pop();
}

void HtmlPage::HeaderNoBody ()
{
	Begin(new html());

	AddObject (head() << (title(m_pFile ? m_pFile->GetDescription() : "")));
}

void HtmlPage::Header (const string& strText,bool bAlsoHeading)
{
	Begin(new html());

	AddObject (head() << (title(strText)));

	Begin(new body());

	if (bAlsoHeading)
		AddObject (h1(strText));
}

void HtmlPage::Header (bool bAlsoHeading)
{
	Header(m_pFile ? m_pFile->GetDescription() : "",bAlsoHeading);

}

void AddUrlFileAsLi(UrlFile* pFile,HtmlPage* pPage)
{
  string strDesc = pFile->GetDescription();
  
  *pPage << (li() << a(strDesc.empty() ? pFile->GetName() : strDesc).href(pFile));
}

void AddUrlFolderAsLi(UrlFolder* pFolder,HtmlPage* pPage)
{
  // if there is an index.html file, link to that
  // otherwise link to the folder
  
  UrlFile* pFile = pFolder->GetIndexFile ();

  pPage->Begin(new li());
  
  if (pFile)
    {
      string strDesc = pFile->GetDescription();
      *pPage << a(strDesc.empty() ? pFile->GetName() : strDesc).href(pFile);
    }
  else
    {
      string strDesc = pFolder->GetName() + "/";
      *pPage << a(strDesc).href(pFolder);
    }
  
  pPage->LinksToFiles(pFolder,true);

  pPage->End();
}

void HtmlPage::LinksToFiles (UrlFolder* pFolder,bool bOnlyFolders)
{
  if (pFolder->m_vectFolders.empty() &&
      (bOnlyFolders || pFolder->m_vectFiles.empty()))
      return; // nothing to add, so return to avoid unnecessary <ul>
      
  Begin(new ul());
  
  for_each(pFolder->m_vectFolders.begin(),pFolder->m_vectFolders.end(),AddUrlFolderAsLi,this);
  if (!bOnlyFolders)
    for_each(pFolder->m_vectFiles.begin(),pFolder->m_vectFiles.end(),AddUrlFileAsLi,this);
  //	for_each(pFolder->m_vectFolders.begin(),pFolder->m_vectFolders.end(),
  //		AddUrlFolderAsLi,this);
  
  End ();
}

//////////////////////////////////////////////////////////////////////

#include <fstream>

void test()
{
	ofstream out1("test1.html");

	HtmlPage page;

	page.SetOutStream(&out1);

	page.Begin(new html());

	//	(head() << title("Testing creating a html page"));
	page << (head() << title("Testing creating a html page"));

	page.Begin(new body());

	page << h1("Hello world");
	page << p("Hello everybody") << (p("What do you want to blow up today?") << img().src("test.jpg"));

	page << p("Hello everybody") << (p("What do you want to crash today?"));

	ofstream out2("test2.html");

	HtmlPage page2;

	page2.SetOutStream(&out2);

	page2.Header("This is a title");

	page2 << table();

	page2 << (table().border(1) << tr());

	page2 << (table().border(2) << tr() << td("ABC"));

	//HtmlTag& obj = (tr() << td("ABC"));

	page2 << ((table().border(4)) << (tr() << td("ABC")));

	page2 << ((table().border(5)) << (tr() << td("ABC") << td("DEF")));

	page2 << ((table().border(6)) << (tr() << td("ABC") 
							   << td("DEF")) 
					  << (tr() << td("GHI") 
						       << td("KJL")));


	
	
	page.End(); // body
	page.End(); // html

	/*
	HtmlPage page;
	HtmlTag tag;

	page << tag;
	page << HtmlTag();

	page << (tag << HtmlTag());

	tag << HtmlTag();

	page << tag;

	HtmlURL url;

	page << img().src(url).src(url);

	page << img().src(url);
*/
	

}



