/*  disasmviewer, part of exmhcg - executable program tuner
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



import java.io.*;
import java.util.*;

//import DisAsmData;
//import AsmCodeSnippet;
//import AsmInstr;

public class DisAsmDataReader extends DataInputStream
{
	DisAsmData data;	
	CodeSnippet snippetFallsThrough;	
	Hashtable hashIdTargets = new Hashtable();
	
	public DisAsmDataReader(InputStream in,DisAsmData d)
	{
		super(in);
		data = d;		
	}

	public void readHeader() throws Exception
		{
			/*
			 * UTF string      name
			 * UTF string      comment
			 *
			 * {
			 *   short         type (of extra header), until type = 0
			 *   [type==? => header info] 
			 *   [type>0  => RESERVED|
			 * }
			 */	 
			 
			String strName = readUTF();
			String strComment = readUTF();
			
			data.setName (strName);
			data.setComment (strComment);
			
			short nInfoType;
						
			while ((nInfoType = readShort()) != 0)
				{
				switch (nInfoType)
					{
						default:
							throw new IOException("DisAsmData::Header::InfoType reserved usage"); // if nInfoType has unknown value
					}
					
					
				}		
		}
	
	public void readCodeSnippet(CodeSnippet snippet) throws Exception
		{
			/*
			 * UTF string      id
			 * 
			 * short           bitfield of included information
			 *                 0x0001  fall through (possible that next CodeSnippet
			 *                         follows this one in execution order)
			 *                 0x0002  fall through target (this CodeSnippet
			 *                         follows previous in execution order)
			 *                 0x0004  internal jump target (this snippet is an
			 *                         target for a jump within this disassembly info file)
			 *                 0x0008  external jump(call) target (this snippet is an
			 *                         external target, like an function entry point)
			 *                 0xfff0  RESERVED
			 * { opbitfile & 0x08
			 *   int           next, number of known external callers/jumpers
			 *   { next times
			 *     UTF string  jdaf-file, relative path from this, where it can be found
			 *     UTF string  id, of external jumper
			 *     UTF string  name, preferred human readable name (if none given = empty, id is used) 
			 *   }
			 * }
			 *
			 */
			
			snippet.setId (readUTF());			
			
			short nIncluded = readShort();
			
			if ((nIncluded & 0xfff0) != 0)
				throw new IOException ("DisAsmData::CodeSnippet::Included reserved usage");
				
			if (((nIncluded & 0x02) != 0) != (snippetFallsThrough != null))
				throw new IOException ("DisAsmData::CodeSnippet : Fall through snippets not in order. (thisid="+
																snippet.strId+
																") (previd="+(snippetFallsThrough!=null?snippetFallsThrough.strId:"none")+")");
				
			if ((nIncluded & 0x02) != 0)	
				{
					snippet.prev = snippetFallsThrough;
					snippetFallsThrough.next = snippet;					
				}
			if ((nIncluded & 0x01) != 0)
				snippetFallsThrough = snippet;
			else
				snippetFallsThrough = null;
				
			if ((nIncluded & 0x04) != 0)	
				{
					snippet.vectInternalJumps = new Vector();					
					if (snippet.strId == null ||
							snippet.strId.length() == 0)
						throw new IOException("DisAsmData::CodeSnippet : Internal target must have an Id");
					if (hashIdTargets.put(snippet.strId,snippet) != null)
						throw new IOException("DisAsmData::CodeSnippet : Internal target must have an unique Id, '"+snippet.strId+"' already used");
				}

			if ((nIncluded & 0x08) != 0)	
				{
					snippet.vectExternalCalls = new Vector();					
					
					int nKnown = readInt();
					
					for (int i = 0; i < nKnown; i++)
						{
							String strJdafFile = readUTF();
							String strId = readUTF();
							String strName = readUTF();
							
							snippet.vectExternalCalls.add (new External (strJdafFile,strId,strName));
						}					
				}

		}


	public void readAsmCodeSnippet() throws Exception
		{
			/*
			 * CodeSnippet data (see readCodeSnippet)
			 * 
			 * short           bitfield of included information
			 *                 0x0001  offset
			 *                      2  length and code
			 *                 0xfffc  RESERVED
			 * 
			 * int             n, number of assembler instructions
			 * { bitfield & 0x01
			 *   int           offset
			 * }
			 *
			 * { n times
			 *   byte          bitfield of operands used
			 *                 0x01  op dest present
			 *                    2  op src  present
			 *                    4  op src  present
			 *                    8  has internal targets
			 *                   10  has external targets
			 *                   80  RESERVED
			 *   { bitfield & 0x02 (global bitfield, not opbitfield)
			 *     byte        length
			 *     bytes[length]  code (for instruction)
			 *   }
			 *   UTF string    instruction
			 *   { opbitfile & 0x01
			 *     [operand]   dest
			 *   }
			 *   { opbitfile & 0x02
			 *     [operand]   src
			 *   }
			 *   { opbitfile & 0x04
			 *     [operand]   src2
			 *   }
			 *   { opbitfile & 0x08
			 *     int           nint, number of known external targets
			 *     { nint times
			 *       UTF string  id, of internal target
			 *     }
			 *   }
			 *   { opbitfile & 0x10
			 *     int           next, number of known external targets
			 *     { next times
			 *       UTF string  jdaf-file, relative path from this, where it can be found
			 *       UTF string  id, of external target
			 *       UTF string  name, preferred human readable name (if none given = empty, id is used) 
			 *     }
			 *   }
			 * }
			 *   
			 * 
			 *
			 */	 
			 
			AsmCodeSnippet asmCodeSnippet = new AsmCodeSnippet();
			
			readCodeSnippet(asmCodeSnippet); // read general info

			short nIncluded = readShort();
			int nNumInstructions = readInt();
			
			if ((nIncluded & 0xfffc) != 0)
				throw new IOException ("DisAsmData::AsmCodeSnippet::Included reserved usage");
			if (nNumInstructions == 0)
				throw new IOException ("DisAsmData::AsmCodeSnippet::NumInstructions cannot be 0");
			
			if ((nIncluded & 0x01) != 0)
				{
					asmCodeSnippet.setOffset (readInt());									
				}
			for ( ; nNumInstructions > 0; nNumInstructions--)
				{
					byte bOperands = readByte();
		
					if ((bOperands & 0xe0) != 0)
						throw new IOException ("DisAsmData::AsmCodeSnippet::OperandSpecifier reserved usage");
						
					AsmInstr instr;
						
					if ((bOperands & 0x18) != 0)
						instr = new AsmBranchInstr();
					else
						instr = new AsmInstr();
					
					if ((nIncluded & 0x02) != 0)
						{
							byte nLength = readByte();
							
							byte b[] = new byte[nLength];
							
							readFully(b);
							instr.setCode (b);
						}
					instr.setInstr(readUTF());									
					
					if ((bOperands & 0x01) != 0)
						instr.setOpDest(readOperand());
					if ((bOperands & 0x02) != 0)
						instr.setOpSrc (readOperand());
					if ((bOperands & 0x04) != 0)
						instr.setOpSrc2(readOperand());
						
					if ((bOperands & 0x08) != 0)	
						{
							Vector v = ((AsmBranchInstr) instr).vectInternalTargets = new Vector();					
							
							int nKnown = readInt();
							
							for ( ; nKnown > 0; nKnown--)
								{
									String strId = readUTF();
									
									Object target;
									
									if ((target = hashIdTargets.get(strId)) != null)
										{
											v.add (target);								
											((CodeSnippet) target).vectInternalJumps.add (asmCodeSnippet);
										}
									else
										v.add (strId);
								}					
						}
					if ((bOperands & 0x10) != 0)	
						{
							Vector v = ((AsmBranchInstr) instr).vectExternalTargets = new Vector();					
							
							int nKnown = readInt();
							
							for ( ; nKnown > 0; nKnown--)
								{
									String strJdafFile = readUTF();
									String strId = readUTF();
									String strName = readUTF();
									
									v.add (new External (strJdafFile,strId,strName));
								}					
						}

					asmCodeSnippet.addInstr(instr);
				}	
			data.addSnippet (asmCodeSnippet);	
		}
	
	
	public Operand readOperand() throws Exception
		{
			Operand op = new Operand();
			
			op.setName (readUTF());
			
			return op;			
		}


	
	public void readData() throws Exception
		{
			/*
			 * [header]
			 *	 
			 * {
			 *   short         type (of extra header), until type = 0
			 *   [type==1 => asm code snippet] 
			 *   [type>1  => RESERVED|
			 * }
			 */	 

			snippetFallsThrough = null;
			
			System.out.println ("Read Header...");

			readHeader();
			
			System.out.println ("Read Data...");

			short nInfoType;
			
			while ((nInfoType = readShort()) != 0)
				{
					switch (nInfoType)
						{
							case 1:
								readAsmCodeSnippet();
								break;
							default:
								throw new IOException("DisAsmData::InfoType reserved usage"); // if nInfoType has unknown value
						}
				}		
			
			System.out.println ("Finished Reading...");

			if ((snippetFallsThrough != null))
				throw new IOException ("DisAsmData::CodeSnippet : Last fall through snippet missing target (id="+snippetFallsThrough.strId+")");		
				
			// fixup any internal targets that were not available when reading
			
			System.out.println ("Fixup Targets...");

			CodeSnippet snippet;
			AsmInstr instr;
			Vector v;
			Object to;
			
			for (Enumeration s = data.vectSnippets.elements(); s.hasMoreElements(); )
				if ((snippet = (CodeSnippet) s.nextElement()) instanceof AsmCodeSnippet)
//					{
//						System.out.println ("CodeSnippet: "+snippet.strId);
					for (Enumeration i = ((AsmCodeSnippet) snippet).vectInstrs.elements(); i.hasMoreElements(); )
						if ((instr = (AsmInstr) i.nextElement()) instanceof AsmBranchInstr)
//							{
//								System.out.println ("AsmBranchInstr: "+instr.strInstr);
							if ((v = ((AsmBranchInstr) instr).vectInternalTargets) != null)
								for (int ti = 0; ti < v.size(); ti++)
									if ((to = v.elementAt(ti)) instanceof String)
										{
											String strId = (String) to;
											
											Object target;													
											if ((target = hashIdTargets.get(strId)) != null)
												{
													v.setElementAt (target,ti); // replace the string Id with the real target
													((CodeSnippet) target).vectInternalJumps.add (snippet);
												}
											else
												throw new IOException("DisAsmData::AsmCodeSnippet : Internal target '"+strId+"'unknown");
										}												
//							}
//					}
			System.out.println ("Finished Fixup...");
		}		
}
