/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataAnalyseFunction.cpp: implementation of the DataAnalyseFunction class.
//
//////////////////////////////////////////////////////////////////////

#include "DataAnalyseFunction.h" 

#include "BranchTargetTraceSingle.h" 
#include "BranchTargetTrace.h" 
#include "BranchTargetCtrlTransfer.h" 

#include "OperandRegister.h" 
#include "OperandImmediate.h"
#include "OperandMemory.h" 

#include "CallFunction.h"
#include "JumpTarget.h"

#include "FlagsState.h"
#include "RegisterState.h"

#include "VariableOp.h" 
#include "VariableNConst.h" 

#include "VarContStackPtr.h" 
#include "VariableNSplit.h" 
#include "VariableNConst.h" 
	
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



DataAnalyseFunction::DataAnalyseFunction()
{

}

DataAnalyseFunction::~DataAnalyseFunction()
{
  
}

void DataAnalyseFunction::AddFunctionToAnalyse (CallFunction* pFunction)
{
  m_mapstackAnalyseFunctions.push(pFunction,WhyAnalyse(WHY_ANALYSE_INITIAL));
}

CallFunction* DataAnalyseFunction::NextFuncToAnalyse()
{
  for ( ; ; )
    {
      if (m_mapstackAnalyseFunctions.empty())
	return NULL;
      
      MapStackCallFunctions::map_value_type func = 
	m_mapstackAnalyseFunctions.pop();
      
      if (SetFunction (func.first))
	return func.first;
    }
}

bool DataAnalyseFunction::SetFunction (CallFunction* pFunction)
{
  // TODO: figure out if reanalysis really is needed
  // if reason only is WHY_ANALYSE_CALLER_CHANGED, and there are other callers as well
  // that make the combined registers no different from last analysis run, then
  // reanalysis is not necessary...

  // so perhaps return false
  



  m_pFunction = pFunction;

  // clear anything having with this function to do
  // FIX: fix before starting to do multiple passes

  /*   
       
    mapEntryStates.clear();

    mapIncomingState.clear ();
    mapOutgoingState->clear ();
    mapOutgoingStateBranch->clear ();
  */

  return true;
}

/*
  For each instruction (which is located for itself in a
  BranchTargetTraceSingle or BranchTargetCtrlTransfer), we keep track
  of what we know about incoming and outgoing information. For
  instructions that are ONLY reached as following another instruction,
  we do not keep track of the incoming stuff, as it is necessarily the
  same as the outgoing of the previous instruction. For
  control-transfer instructions (branches) there is of course the
  possibility of two outgoing ways. We keep track of those separately,
  as they may have different known states (eg flags, ecx, and the
  state of the stack, depending on if it is the call branch of a call
  instruction, or the following instruction, which would be the state
  as when all possibly called functions returned.

  The information is kept in three maps, using the addresses of the
  traces as keys.

  mapIncomingState
  mapOutgoingState
  mapOutgoingStateBranch

  and so the information is like

  IncomingState->
  * TraceSingle
  ->OutgoingState
  * TraceSingle
  ->OutgoingState
  * TraceSingle
  ->OutgoingState

  IncomingState->
  * CtrlTransfer
  ->OutgoingState and/or ->OutgoingState(for branch)

  IncomingState->
  * TraceSingle
  ->OutgoingState
  * TraceSingle
  ->OutgoingState
  * CtrlTransfer
  ->OutgoingState and/or ->OutgoingState(for branch)

  Note that each OutgoingState only keep track of what the instruction
  caused, and then also a pointer to the previous state so that they
  can be used to look up any incoming information, also if it was set
  several instructions ago, without copying everything several times.
  Because branches can have two outgoing states, we must take special
  care there when back-propagating info to the incomingState, so it is
  the least commonly known info of the two branches.

*/

typedef map<OutgoingState*,dword> MapOutgoingStateDepth;

//  const OutgoingState* m_pLast;

OutgoingState* FindCommonParent(VectOutgoingState& vectStates)
{
  // OutgoingStates usually have parents, and if the function only
  // have one place of entry (i.e. calls are only made to one
  // address), the EntryState there is a common parent. But usually
  // one could follow the traces a bit further and find later comman
  // parents, and we want the latest one, i.e.  the one closest to us.

  // our job is to find that one, think about the tree like this
  //                       
  //           ENTRY     ANOTHERENTRY 
  //             |             |    
  //             |             |
  //             o             +
  //            / \                           .
  //           /\  \                          .
  //          |  |  |         
  //          *   \ /       
  //               |
  //               *         
  //               
  // We want to join the two *s, then the parent we're looking for is
  // the o.  But if also the + can come to the current location, so we
  // want to join the *s and the +, then there is no common parent;
  //                         
  // As we want the last one, I think the simples way is to for one of
  // the entries (e.g. the first one), create a map of all known
  // parents and their depth. Then for each other entry walk the last
  // of parents back and see if we can find any match in the map, and
  // then clean the map of all entries with a smaller number
  // (i.e. closer to the first entry) than the one associated one, so
  // that when looking up further entries we do not find any parent
  // that is not a parent of everyone already looked at.

  assert (!vectStates.empty()); 

  if (vectStates.size() == 1)
    {
      // the common parent is trivial
      return vectStates.front();
    }
  
  // TODO: we should perhaps also make sure (i.e. assert) that there
  // are no dublets in the vectStates..

  MapOutgoingStateDepth mapStateDepth;
  
  VectOutgoingState::const_iterator iter = vectStates.begin();
  
  OutgoingState* pState = *iter;
  assert (pState);
  
  for (dword depth = 0; pState; depth++)
    {
      mapStateDepth.insert(MapOutgoingStateDepth::value_type(pState,depth));
      pState = pState->m_pLast;
    }

  dword common_depth = 0;
  
  ++iter;
  assert (iter != vectStates.end());

  for ( ; ; )
    {
      MapOutgoingStateDepth::iterator found;
      pState = *iter;

      assert (pState);

      do
	{
	  found = mapStateDepth.find(pState);
	  
	  if (found != mapStateDepth.end())
	    goto next_iteration;
	  pState = pState->m_pLast;
	}
      while (pState);
      // oups, found nothing, so there is no common parent...

      return NULL;

    next_iteration:;

      // aha, we found a common parent pCommonParent at depth
      // common_depth

      // if that depth is deeper than the previously known, then we'll
      // have to kill all with a smaller depth, so they cannot be
      // found at later iterations

      // but only do killing if there are more iterations...

      ++iter;
      if (iter == vectStates.end())
	return found->first; // we're done

      assert (found->second >= common_depth);
      
      if (found->second > common_depth)
	{
	  common_depth = found->second;

	  for (MapOutgoingStateDepth::iterator check = mapStateDepth.begin(); 
	       check != mapStateDepth.end(); )
	    {
	      if (check->second < common_depth)
		SET_MAP_ERASE_CONTINUE_ITER(mapStateDepth,check);
	      else
		++check;
	    }
	}
    }  
}

IncomingState* FindLoopPrevIncoming(OutgoingState* pSearch,
				    OutgoingState* pStopSearch,
				    IncomingState* pPrevIncoming,
				    SetIncomingState* pSetIncomingState)
{
  assert (pPrevIncoming);

  while (pSearch != pStopSearch)
    {
      assert (pSearch);

      if (pSearch == pPrevIncoming)
	return pPrevIncoming;

      if (pSetIncomingState &&
	  IS_TYPE(IncomingState*,pSearch))
	{	
	  SetIncomingState::const_iterator iter = 
	    pSetIncomingState->find((IncomingState*) pSearch);
	  
	  if (iter != pSetIncomingState->end())
	    return *iter;
	}

      pSearch = pSearch->m_pLast;
    }  
  return NULL;
}

void DataAnalyseFunction::AddOldIncomingState(CtrlTransferTarget* pCtrlTarget,
					      IncomingState* pOldIncomingState)
{
  pair<MapCTTSetIncomingState::iterator,bool> iter = 
    mapOldIncomingStates.insert(MapCTTSetIncomingState::value_type(pCtrlTarget,
								   SetIncomingState()));
  
  iter.first->second.insert(pOldIncomingState);
}

IncomingState* DataAnalyseFunction::CalcIncomingState (CtrlTransferTarget* pCtrlTarget)
{
  // MapCTTIncomingState::iterator incoming = mapIncomingState.find(pCtrlTarget);

  //printf ("Calc incoming...");
  
  /*
    if (incoming != mapIncomingState.end())
    pIncoming = &(incoming->second);
  */

  // handling is easy if there is only one target to us what possible
  // sources do we have? As we do things one function at a time,
  // multiple call targets are handled by the EntryState, so if we're
  // a call target anywhere (or several times), this counts for one,
  // then we could possiby have one or more jumps to us (from this
  // function, by definition (if it's a jump from another function,
  // it's treated as one big function.., perhaps this is not so wise,
  // we'll have to see if that is bad sometimes, like perhaps C++
  // exceptions throwing and being caught...) and then we can be the
  // fall-through target of the last instruction

  // if we several sources, we must figure out what is the common
  // informtion from all of them

  // collect all the states that may come into a list

  printf ("pCtrlTarget  %p\n",(void*) pCtrlTarget);
  
  IncomingState* pIncoming = new IncomingState();
  
  VectOutgoingState& vectSrcStates = pIncoming->m_vectSrcStates;
  
  SET_MAP_IF_FIND(MapCTTEntryState,mapEntryState,entry,pCtrlTarget)
    vectSrcStates.push_back(entry->second);
  
  // any fall through?
  
  BranchTarget* pTarget = pCtrlTarget->m_pTarget;
  
  if (pTarget->GetPrevTarget())
    {
      /*
	MapBTOutgoingState::iterator fall_through = mapOutgoingState->find(pTarget->GetPrevTarget());
	
	assert (fall_through != mapOutgoingState->end()); // if this fails, it means that a fall_through target is
	// about to be analysed before the previous target was analysed. That is not to happen!, because it is stupid => fix
	
	vectSrcStates.push_back(&(fall_through->second));
      */

      // actually, it may happen that a fall-through target is
      // analysed before the previous instruction, e.g. in case the
      // fall-through target as jumped to and the previous instruction
      // only can be reaced as a result of the fall-through target
      // being executed (like jumping into the middle of a loop), but
      // it should still be avoided to analyse in this order...

      SET_MAP_IF_FIND(MapBTOutgoingState,mapOutgoingState,
		      fall_through,pTarget->GetPrevTarget())
	vectSrcStates.push_back(fall_through->second);
    }
  
  // anybody branching to us??
  
  {
    JumpTarget* pJumpTarget = pTarget->GetJumpSource ();
    
    if (pJumpTarget)
      {
	// even plain uncoditional jumps should set their outgoing
	// state in the as the branch pointer, so it's easier here!
	
	FOR_EACH_ITER(VectFromTrace,pJumpTarget->m_vectFromTrace,branch_from)
	  {
	    SET_MAP_IF_FIND(MapBTOutgoingState,mapOutgoingStateBranch,
			    branch_state,*branch_from)
	      vectSrcStates.push_back(branch_state->second);
	    
	    // if *branch_state == NULL, it has not been analysed yet,
	    // so it's not included yet (we do not take everything as
	    // null, but as the combination of everything known so
	    // far...
	    //
	    // This enables variables that are not touched by a loop
	    // to fall through it, otherwise, they would be set to
	    // unknown becuase they are unknown at the end of the loop
	    // when the loop is begun to be analysed. However, this
	    // create some problem with loop counters and other
	    // variables that are mixed through loop iterations. So
	    // extra code is necessary to avoid these circular
	    // references, possibly slowly enlarging the possible
	    // values of variables by one for each analysis iteration
	    // of the loop. I.e., we'll have to find circular
	    // dependencies of variables, without killing everything
	    // of, see <SOME_OTHER_PLACE>
	  }
      }
  }
  
  assert (!vectSrcStates.empty()); // we must have at least one source, otherwise, how did we come here??
  
  if (vectSrcStates.size() == 1)
    {
      // only one source, so it is easy

      OutgoingState* pSrcState = vectSrcStates.front();
      assert (pSrcState);
      
      pIncoming->m_pLast = pSrcState;
      
      // pIncoming->pRegState = new RegisterState (/*pSrcState->pRegState*/);
      // = NULL, because nothing is needed

      // if we only have one source, it cannot be a loop, because a
      // loop need a way to be entered from also, and that would be
      // the second source. (However we might be the first analysis of
      // the loop, but that's no problem...)

      // so either this is the first analysis, or the previous target
      // has changed, so this must also change

      pair<MapCTTIncomingState::iterator,bool> prev = 
	mapIncomingState.insert (MapCTTIncomingState::value_type (pCtrlTarget,
								  pIncoming));
      
      if (!prev.second) // not inserted
	{
	  // somebody was there before us, check that something has changed
	  
	  printf ("(prev.first->second)->m_pLast != pSrcState, %p %p\n",
		  (void*) (prev.first->second)->m_pLast,(void*) pSrcState);

	  assert ((prev.first->second)->m_pLast != pSrcState);
	  
	  AddOldIncomingState(pCtrlTarget,prev.first->second);
	  
	  prev.first->second = pIncoming;
	}

      return pIncoming;
    }
  
  // multiple sources. combine them!
  
  OutgoingState* pCommonParent = FindCommonParent(vectSrcStates);
  
  pIncoming->m_pLast = pCommonParent;

  // Two problems, one easy, the other one very irritating:
  //  
  // * We'll have to figure out which variables have been touched by
  // any of the source states since they departed from the common
  // parent. All that have been touched must be merged.
  //
  // TODO: One possible optimisation would be not to do this (or the
  // actual merging) until the variable is actually requested, as if
  // the register is overwritten it is never used and merging was
  // unnecessary. Before optimising like this: figure out how often
  // this occur and if it therefore is worth the extra code. This
  // delaying of merging might have other consequences, perhaps
  // related to the next problem. Actually this optimization is
  // partially needed now anyway, since we must propagate any requests
  // for unknown registers back incase they were an argument to the
  // function, so we can get them there, because they may be
  // overwritten in the loop also..

  // * Problem II: loops: When first entering the code of a loop, the
  // outgoing state of the instruction branching back to us (looping)
  // is unknown, so that would completely destroy this state.  In
  // order for that not to happen, we do not care about NULL outgoing
  // states so that a reanalyse will be necessary when that state is
  // available. This is to facilitate tracking of registers (& stack)
  // which are not touched by the loop. So if the loop do not touch
  // them, then they will simply point to the same values as the
  // parent and there is no problem. Now come the nasty part: assume
  // the loop always enter with CX=0, and then increments CX once
  // every iteration. Then the fall-through will come with CX=0, and
  // the loop with CX=1, so we'll combine to CX=0,1, and then next
  // analyse run we'll get CX=0 and CX=1,2 from the loop, combining to
  // CX=0,1,2. Hmmm. This cannot happen, because we'll then reanalyse
  // until we've exhausted the available space 0-INT_MAX.. There are
  // also cases where perhaps AX and BX change places each
  // iteration. or perhaps AX=list[AX] (but this is perhaps not a
  // problem, unless the possible values in the list increase for
  // every iteration), and who knows what?, Alright, so when
  // reentering the loop analysis, we should check each variable, and
  // if it depends on some variable coming into the loop, and is not
  // at exactly the same place, then is should be set to unknown or
  // perhaps with some more intelligence in the program a directly
  // calculated range or so (but that is for later). Now, we get inte
  // trouble with the structure of OutgoingState keeping track of old
  // settings via pointers to old OutgoingStates, and then if somebody
  // reanalysed in the middle of the loop (perhaps another loop
  // there), then those pointers will be messed up, and we won't
  // detect these circular dependencies

  RegisterStateCombine* pRegStateCombine = new RegisterStateCombine ();
  
  FOR_EACH_ITER(VectOutgoingState,vectSrcStates,srcstate)
    {
      (*srcstate)->GetAffected(pCommonParent,pRegStateCombine->m_affected);

      
    }

  // special care is to be taken for registers that overlap, i.e.

  pRegStateCombine->m_affected.RemoveUnneedOverlap(); // remove AX if also EAX an so on

  // merging is only done when the actual register is requested. This
  // is so that any dependency-chains are not made longer than
  // necessary, and to avoid to set false function entry registers as
  // argument- carrying.  Suppose we after entry have en if-else
  // construct and that one (but not both) branches writes to eax (but
  // no read before write), so eax is used as register, but not
  // argument. Then at the point where the branches combine we'll have
  // eax as affected. And later somebody again write to eax. But no
  // read so eax is never argument. If we were to combine it here,
  // we'd mark it as incoming argument when the request goes bach to
  // the entry state. But instead we do not do the actual merge until
  // someone ask RegisterStateCombine for the register, and then we
  // have a real read request.
  //
  // NOTE: we must anyway be able to remove stuff from the entry state
  // if they were requested sometime but that later proved to be wrong
  // (due to keeping old incoming states (which is necessary), when
  // reanalysing if they have same registers set up for combine
  // (regardless if the old one have made requests that won't be made
  // again). Simply put: due to repeated analysis most outgoingstates
  // will when finished have correct dependencies, but (especially)
  // incomiong and entry states must be checked and unneeded stuff
  // cleared away, but we must then also differ between unneeded as in
  // not used and unneeded as in not touched (so that EAX stays
  // untouched during function call). (Lets say it's a big mess) We
  // might have to change MOV and similars (PUSHes and POPs) to
  // actually make a copy of the variable and not just a copy of the
  // pointer...

  pIncoming->pRegState = pRegStateCombine;

  //printf ("MULTIPLE source!\n");

  // alright then, here comes the loop analysis machete code, i.e. the
  // code to make sure that the analysis itself does not loop
  // indefinately1 to analyse the loop. The basic idea is to make more
  // analysis when nothing has changed. In order to make sure that not
  // so much changes (which is normally detected by seeing if the
  // pointers to the parent OutgoingStates have changed
  
  // ok, so figure out if it is a loop. It is if any of the sources
  // have passed an old version if incoming state for this node. Note
  // that such passage must have occured after it passed the
  // pCommonParent, so also assert that no node before the common
  // parent is looping

  // if we haven't been here before (i.e. have no old incoming states
  // for this node), we can't be a loop (yet)
  
  pair<MapCTTIncomingState::iterator,bool> prev = 
    mapIncomingState.insert (MapCTTIncomingState::value_type (pCtrlTarget,
							      pIncoming));
  
  if (prev.second) // inserted
    {
      // noone has been here before, so can't be loop
      
      assert (mapOldIncomingStates.find(pCtrlTarget) == 
	      mapOldIncomingStates.end());
      
      return pIncoming;
    }
  
  // ok, so someone was here before us...
  // right now, the last one to be here is prev->first->second
  
  assert (prev.first->second);
  
  MapCTTSetIncomingState::iterator oldset = 
    mapOldIncomingStates.find(pCtrlTarget);
  
  SetIncomingState* pSetOldIncoming = 
    (oldset != mapOldIncomingStates.end()) ? &oldset->second : NULL;
  
  // make sure loop is not BEFORE pCommonParent
  assert (!FindLoopPrevIncoming(pCommonParent,NULL,prev.first->second,
				pSetOldIncoming)); /* also works with pCommonParent being NULL */
  
  // ok, so now check the incoming states and see if any is a loop.
  // if any is a loop, figure out which variables are affected so that
  // they cannot be considered (i.e., are affected by the loop based
  // on some incoming values. They must be marked as unknown when
  // entering the loop

  FOR_EACH_ITER(VectOutgoingState,vectSrcStates,srcstate)
    {
      IncomingState* pOldIncoming;
      
      if ((pOldIncoming = FindLoopPrevIncoming(*srcstate,pCommonParent,
					       prev.first->second,
					       pSetOldIncoming)))
	{
	  assert (!FindLoopPrevIncoming(pOldIncoming->m_pLast,pCommonParent,
					prev.first->second,
					pSetOldIncoming)); // should only loop once
	  
	  // ok, so pOldIncoming was here before

	  // find any unhandlable variables, and 
	  
	  // for now we take eveything touched in loop as unhandlable
	  
	  (*srcstate)->GetAffected(pCommonParent,pRegStateCombine->m_unknown);
	}
    }

  // and make sure the stuff is only in either m_affected or m_unknown

  pRegStateCombine->m_unknown.RemoveUnneedOverlap(); // remove AX if also EAX an so on

  // TODO: this is not as it is to be, i), above code will change
  // anyway (so not everything in loop by default is unknown), ii)
  // unknown may contain only AX, while affected may have expanded to
  // EAX (due to other, non-loop source)

  FOR_EACH_ITER(SetReg,pRegStateCombine->m_unknown.m_setReg,reg)
    {
      assert (pRegStateCombine->m_affected.m_setReg.find(*reg) != 
	      pRegStateCombine->m_affected.m_setReg.end());
      pRegStateCombine->m_affected.m_setReg.erase(pRegStateCombine->m_affected.m_setReg.find(*reg));
    }
  
  // now, if we (pIncoming) have the same parent state as the incoming
  // state here before us (prev->first->second), and if we have the
  // same list of unhandlable variables, and any other variables
  // affected would be the same when looked up, then we can use the
  // old one (prev->first->second) and delete (pIncoming), so that
  // analysis may end...

  // and if we can use the old one, then analysis of the is not needed
  // (i.e. when we return NULL, no analysis is done)

  // can we use the old

  if (prev.first->second->HasSameContent(pIncoming))
    {
      // can use old!
      // new (pIncoming) not needed, kill it

      delete pIncoming;

      return NULL; // no need for reanalysis
    }

  // ok, so we need a new one, be replacing old and moving old to old-list
  AddOldIncomingState(pCtrlTarget,prev.first->second);
  prev.first->second = pIncoming;
  return pIncoming;

  /*
  assert ((prev->first->second)->m_pLast != pSrcState);
  
  AddOldIncomingState(pCtrlTarget,prev->first->second);
  
  prev->first->second = pIncoming;
  */
  return pIncoming;

  /*
  MapCTTVectIncomingState::iterator old = mapOldIncomingStates.find(pCtrlTarget);

  if (old != mapOldIncomingStates.end())
    {
      assert (FindLoopConstruct);
      
      
      
      
      
    }
      
  AddOldIncomingState(pCtrlTarget,prev->first->second);
  
  return pIncoming;
  */
}

template<class _Ttarget,class _Tstate/*,class _Tmap*/>
_Tstate* ReplaceState(const _Ttarget* pTarget,_Tstate* pNewState,
		      map<const _Ttarget*,_Tstate*>& map_ts)
{
  typedef map<const _Ttarget*,_Tstate*> Tmap;
  typedef typename Tmap::iterator Tmap_iter;
  typedef typename Tmap::value_type Tmap_vtype;
  
  Tmap_iter old = map_ts.find(pTarget);
  
  //printf (__ASSERT_FUNCTION);
  //printf ("\n");

  if (old != map_ts.end())
    {
      printf ("Found old...\n");
      
      if (old->second->HasSameContent(pNewState))
	{
	  printf ("No change...\n");
	  
	  // we've not changed, so we kan delete the new state and
	  // continue to use the old one
	  
	  // TODO: when fixing the memory leaks around here: 
	  // delete pNewState

	  return old->second;
	}
      pNewState->HasSameContent(old->second);

      printf ("Replace...\n");

      // TODO: remember old->second for deletion

      old->second = pNewState;

      return pNewState;
    }
  
  // there was no one before
  
  printf ("Insert...\n");

  map_ts.insert(Tmap_vtype(pTarget,pNewState));
  return pNewState;
}

void DataAnalyseFunction::AnalyseCtrlTransferTarget(CtrlTransferTarget* pCtrlTarget)
{
  // before analysis, all incoming states must get the info from
  // EntryStates, and any other outgoingstates that can give control
  // to them we can as well do it here, because it has to be done
  // before analyis

  //CtrlTransferTarget* pCtrlTarget = setstackToAnalyse.pop();
      
  // so now we'll analyse this target until we hit another target,
  // when we do, we'll just stop, and let the stack make sure that the
  // continued analysis is done (if needed)

  IncomingState* pIncoming = CalcIncomingState(pCtrlTarget);
  
  if (!pIncoming)
    {
      // no incoming because it was determined that nothing has
      // changed since last analysis
      return;
    }

  /*
      MapCTTIncomingState::iterator incoming = mapIncomingState.find(pCtrlTarget);
      
      IncomingState* pIncoming = NULL;
      
      if (incoming != mapIncomingState.end())
	pIncoming = &(incoming->second);
      */

      //if (pIncoming)
      //	printf ("Has incoming...\n");

  assert (pIncoming);
  
  //if (pIncoming)
  
  OutgoingState* pLastOutgoing = pIncoming;

   
  // ReplaceState(pCtrlTarget,pIncoming,mapIncomingState);  
  // mapIncomingState.insert (MapCTTIncomingState::value_type (pCtrlTarget,pIncoming));
  // pIncoming is invalid now

  BranchTarget* pTarget = pCtrlTarget->m_pTarget;
      
  CtrlTransferTarget* pNextTarget = NULL;

  OutgoingState* pOutgoingState;

  for ( ; ; )
    {
      pOutgoingState = new OutgoingState(pLastOutgoing);
	  
      /*
	MapCTTIncomingState mapIncomingState;
	MapBTOutgoingState mapOutgoingState;
      */
	  
      BranchTargetTraceSingle* pTraceSingle;
      BranchTargetTrace* pTraceMultiple;
      BranchTargetCtrlTransfer* pTraceCtrlTransfer;

      //printf (pLastOutgoing ? "OK\n": "NULL\n");
	  
      AnalyseCurInstr analyseCurInstr;

      analyseCurInstr.pVarDestMemOp = NULL;
      analyseCurInstr.pVarSrcMemOp  = NULL;

      if ((pTraceSingle = dynamic_cast<BranchTargetTraceSingle*>(pTarget)) != 
	  NULL)
	{
	  DetermineStateChange (analyseCurInstr,
				pLastOutgoing,
				pOutgoingState,
				pTraceSingle->GetInstruction());
	}
      else if ((pTraceCtrlTransfer = 
		dynamic_cast<BranchTargetCtrlTransfer*>(pTarget)) != NULL)
	{
	  OutgoingState* pOutgoingStateBranch = 
	    new OutgoingState(pLastOutgoing);
	  
	  BranchInstr* pBranchInstr = pTraceCtrlTransfer->GetBranchInstr();
	  
	  DetermineStateChangeCtrlTransfer (analyseCurInstr,
					    pLastOutgoing,
					    pOutgoingState,
					    pOutgoingStateBranch,
					    pBranchInstr);
	  
	  // TODO: sometimes a ctrl transfer instruction cannot follow
	  // both paths, like a JMP which cannot fall through. We
	  // should then not set the outgoingState in the
	  // mapOutgoingState map below (mapOutgoingState->insert ...)
	  
	  /*if (outgoingStateBranch.IsStateCleared())
	    {
	    mapOutgoingStateBranch->erase(pTarget);
	    }
	    else // if (outgoingStateBranch.HasAnything())
	    {
	    // even if this state has no new information, we set it because it
	    // is a branch and it will usually have something to convey anyway
	  */

	  bool bChangeToNew = (ReplaceState(pTarget,
					    pOutgoingStateBranch,
					    mapOutgoingStateBranch) == 
			       pOutgoingStateBranch);
	  //mapOutgoingStateBranch.insert(MapBTOutgoingState::value_type(pTarget,pOutgoingStateBranch));
	  /*}*/
	  //else
	  //{
	  //  mapOutgoingStateBranch->erase(pTarget);
	  //}	      

	  if (pBranchInstr->IsJump())
	    {
	      // target is local to this function
	      
	      FOR_EACH_ITER(VectTargets,pBranchInstr->m_vectTargets,target)
		{
		  assert (*target);
		  CtrlTransferTarget* pJumpTarget = 
		    dynamic_cast<CtrlTransferTarget*>(*target);
		  
		  if (pJumpTarget)
		    {
		      assert (pJumpTarget);
		      
		      if (mapIncomingState.find(pJumpTarget) == 
			  mapIncomingState.end() || // not analysed yet
			  bChangeToNew) // anything changed with this outgoing state */)
			m_setdequeToAnalyse.push_back(pJumpTarget);
		    }
		}
	    }
	  else
	    {
	      // target(s) is other function, reanalyse ??
	      //
	      // TODO: what happens if it is a recursive call (i.e. us
	      // again), should we be set up for reanalysis, or do we
	      // handle it ourselves this run?? (arguments update..)
	      // same possible problem as looping...

	      FOR_EACH_ITER(VectTargets,pBranchInstr->m_vectTargets,target)
		{
		  assert (*target);
		  
		}
      

	    }
	}
      else 
	{
	  assert ((pTraceMultiple = 
		   dynamic_cast<BranchTargetTrace*>(pTarget)) != NULL);
	  // for now, we require the instructions to be stored one and
	  // one in BranchTargetTraceSingle
	  assert (false);

	  // mapOutgoingStateBranch->m_bStateCleared = true;
	}

      /*if (outgoingState.IsStateCleared())
	{
	mapOutgoingState->erase(pTarget);
	pLastOutgoing = NULL; // &incoming_unknown_dummy; // instruction caused unknown state change, so we now know: nothing
	//printf ("CLEARED\n");
	}
	else // if (outgoingState.HasAnything())
	{
	// if the state has had anything changed, the interpretation here is that everything that
	// could have changed has been set in the state
      */
      if (analyseCurInstr.pVarDestMemOp)
	pOutgoingState->SetRegisterVar(reg_fakeMEMADDR_DEST,
				       analyseCurInstr.pVarDestMemOp);
      if (analyseCurInstr.pVarSrcMemOp)
	pOutgoingState->SetRegisterVar(reg_fakeMEMADDR_SRC,
				       analyseCurInstr.pVarSrcMemOp);
      
      pLastOutgoing = ReplaceState(pTarget,pOutgoingState,mapOutgoingState);
      
      // mapOutgoingState.insert (MapBTOutgoingState::value_type(pTarget,pOutgoingState));
      // pLastOutgoing = pOutgoingState;

      /*}*/
      //else
      //  {
      //    mapOutgoingState->erase(pTarget);
      //  }

      /*
	if (outgoingState.HasAnything())
	{
	// if the state has had anything changed, the interpretation here is that everything that
	// could have changed has been set in the state

	pLastOutgoing = &(mapOutgoingState->insert (MapBTOutgoingState::value_type(pTarget,outgoingState)).first->second);
	}
	else
	{
	mapOutgoingState->erase(pTarget);
	      
	if (outgoingState.IsStateCleared())
	pLastOutgoing = &incoming_unknown_dummy; // instruction caused unknown state change, so we now know: nothing
	}
      */
	  
      pTarget = pTarget->GetNextTarget();
      if (!pTarget || (pNextTarget = pTarget->GetCtrlTransferTarget()))
	break; // we hit end or another target, abort
    }
  /*
      if (pIncoming && incoming == mapIncomingState.end())
	mapIncomingState.insert (MapCTTIncomingState::value_type (pCtrlTarget,*pIncoming));
      */

  if (pNextTarget)
    {
      if (mapIncomingState.find(pNextTarget) == mapIncomingState.end() || // not analysed yet
	  pOutgoingState == pLastOutgoing /* something changed with this outgoing state */)
	m_setdequeToAnalyse.push_front(pNextTarget); // as it is directly following us, take care of it next (push front)
    }      
}

void DataAnalyseFunction::AnalyseFunction()
{
  // this is the engine for analysing a function until no more
  // information can be obtained.

  // analyse all branches within this function if analysis gives some
  // information that will help analysis in another part (re)analyse
  // that part (again)
  
  assert (m_setdequeToAnalyse.empty());

  // first find all targets in our function so we know were to analyse
  
  // m_pFunction->GetCtrlTransferTargets (setstackToAnalyse);
  // m_pFunction->GetCallTargets (setstackToAnalyse);

  // walk through the starting points (control transfer targets),
  // usually only one. Set up a stack state for each of them. This
  // really need more thought as we'll need some code to then make
  // them come together, especially if they are at different depths.
  // There is also the issue with programs calling the very next
  // instruction to get hold if eIP (instruction pointer), and then
  // actually never returning, because they just pop it. (Like the
  // startup code in ELF). We could of course hard-code to find CALL
  // relative-O and then POP, but if avoidable we'd rather handle it
  // in a more general way, allowing for a non-0-rel call and perhaps
  // some other instructions before the pop (e.g. for better
  // scheduling)
  //
  // We'll have similar problems with functions that take an variable
  // lngth list of arguments and treat them in a loop (like printf)
  //
  // It is very important that the stack is traced through the entire
  // function, and that we never loose control of where eSP points,
  // because after such a failure we can't trace the stack anymore,
  // and returns may go anywhere, so lost stack is a no-go for the
  // following program-modifying steps

  // so we set up a set (with hopefully only one member) of the states
  // that may be on calling the function. The members of this state
  // will also keep track of how deep into the stack the function
  // looks when reading arguments, and any registers that are used.

  // MapEntryStates mapEntryStates;
  
  const VectEntryPoints& entryPoints = m_pFunction->GetEntryPointsTargets ();
  
  assert (!entryPoints.empty());
  
  MapCTTEntryState mapOldEntryState;
  
  FOR_EACH_CONST_ITER(VectEntryPoints,entryPoints,entry)
    {
      assert (*entry);
      
      EntryState* pEntryState = new EntryState();
      
      pEntryState->InitState();

      pair<MapCTTEntryState::iterator,bool> prev =
	mapEntryState.insert (MapCTTEntryState::value_type(*entry,pEntryState));

      if (!prev.second)
	{
	  // there was an old one, move it to old list (so we after
	  // analysis can see if anything changed), and insert new one
	  // instead

	  mapOldEntryState.insert(MapCTTEntryState::value_type(*entry,prev.first->second));
	  
	  prev.first->second = pEntryState; // insert new one
	}
      
      m_setdequeToAnalyse.push_back (*entry);
    }
  
  assert (!mapEntryState.empty()); // there must be at least one entry position

  // TODO: setstackToAnalyse should be sorted (or have been sorted) so
  // that the analysis normally would flow in the correct order this
  // will avoid unnecessary recalculations. However, everything done
  // should be stable against the order it is done in (this has to
  // thought about carefully, and perhaps is not necessary, but would
  // be very good)

  // and then when we have reached the end of a trace, we must compare
  // the new outgoing state to the old one, and if information has
  // been gained, then we should see if we can improve on the
  // information known by the incoming state we should also be careful
  // about cases when information has been lost. This can happen
  // because we found another place that could call/jmp to us (thereby
  // perhaps destroying the combination of incoming info), or perhaps
  // because the combination code is not stable (I do not know how to
  // make it stable, i.e. such that when information about possible
  // variable values have strictly improved (so fewer values are
  // possible), we should get a combined value that is a subset (or at
  // most the same) as the previous value)

  // OutgoingState incoming_unknown_dummy; // used as dummy when incoming state is unknown...,
  // this is to avoid null pointers
  
  while (!m_setdequeToAnalyse.empty())
    {
      //CtrlTransferTarget* pCtrlTarget = setstackToAnalyse.pop();
      
      CtrlTransferTarget* pCtrlTarget = m_setdequeToAnalyse.pop_front();
      
      AnalyseCtrlTransferTarget(pCtrlTarget);
    }
  
  // so now when the function has been analysed, figure out what
  // registers and stack are arguments (read), return values (write),
  // or perhaps both

  // TODO: take care to not list registers restored via push / pop
  // sequences as arguments or return values (we must then make sure
  // that the value is not used anywhere in the function (argument)
  
  // all arguments will have been collected in the EntryState of the
  // function, and any stuff written can be found in the same way as
  // IncomingState figure out what has been touched between the common
  // parent and the direct sources.  In this case the entry states
  // will be the source states

  // TODO: first one should check the variable dependencies and remove
  // any unused ones.  could this remove stuff from the argument list?
  // (probably requires change of branches..)

  
  // MapCTTEntryState mapOldEntryState;
  
  FOR_EACH_CONST_ITER(VectEntryPoints,entryPoints,entry)
    {
      assert (*entry);
      
      MapCTTEntryState::iterator entry_state = mapEntryState.find (*entry);

      assert (entry_state != mapEntryState.end());

      RegisterStateEntry* pRegStateEntry = 
	(RegisterStateEntry*) entry_state->second->pRegState;
      
      pRegStateEntry->m_used.RemoveUnneedOverlap();
      
    }
  
  // and if argument/return list was changed, schedule any calling
  // functions for re-analysis

  
  
}

Variable* DataAnalyseFunction::GetVarFromMemOp(OutgoingState* pIncomingState,
					       Variable* pVarMemAddr,
					       dword dwSize)
{
  // ok, so it's needed to do something about the memory offset this
  // will have to be improved. For now cheat to get any simple stack
  // operations
  
  if (pVarMemAddr)
    {
      VarContStackPtr* pStackPtr;

      if ((pStackPtr = 
	   dynamic_cast<VarContStackPtr*>(pVarMemAddr->m_pVarContent)))
	{
	  // ok, so it is a stack pointer.
	  // It better be a stack pointer to this stack!

	  // TODO: make sure the RegisterState the stack pointer has
	  // is some parent of the current state

	  return pIncomingState ? 
	    pIncomingState->GetStackVar (pStackPtr->m_dwOffset,dwSize) : NULL;
      	}
    }
  
  // all informtion about where this point is lost (or badly unknown),
  // so we do not know where it comes from it is bad, bad not nearly
  // as bad as when writing to unknown location...  however, it
  // destroy any informtion of where values are used again, so, it
  // must be fixed

  // TODO: handle this...

  return NULL;
}

void DataAnalyseFunction::SetVarToMemOp(OutgoingState* pOutgoingState,
					Variable* pVarMemAddr,Variable* pVar)
{
  assert (pOutgoingState);

  // ok, so it's needed to do something about the memory offset this
  // will have to be improved. For now cheat to get any simple stack
  // operations
  
  if (pVarMemAddr)
    {
      VarContStackPtr* pStackPtr;

      if ((pStackPtr = 
	   dynamic_cast<VarContStackPtr*>(pVarMemAddr->m_pVarContent)))
	{
	  // ok, so it is a stack pointer.
	  // It better be a stack pointer to this stack!
	  
	  // TODO: make sure the RegisterState the stack pointer has
	  // is some parent of the current state

	  pOutgoingState->SetStackVar(pStackPtr->m_dwOffset,pVar);
	  return;
      	}
    }
  
  // all informtion about where this point is lost (or badly unknown),
  // this destroy all knowledge...  because it may point into memory,
  // stack, code, whereever...

  // TODO: handle this...
}

Variable* DataAnalyseFunction::GetVariableFromOp (OutgoingState* pIncomingState,
						  const Operand* pOp,
						  Variable* pVarMemAddr)
{
  assert (pOp);
  
  const OperandRegister* pOpReg;
  const OperandImmediate* pOpImm;
  const OperandMemory* pOpMem;
  
  if ((pOpReg = dynamic_cast<const OperandRegister*>(pOp)))
    {
      //printf ("Looking for: %02x,%s...\n",pOpReg->m_bReg,GetRegisterName(pOpReg->m_bReg));
     
      // return outgoingState.GetRegisterVar (pOpReg->m_bReg);
	
      Variable* pVar = pIncomingState ? pIncomingState->GetRegisterVar (pOpReg->m_bReg) : NULL;
      //if (pVar)
      //printf ("Got: %s\n",pVar->ValueString().c_str());
      //printf ("-------\n");
      return pVar;
    }
  else if ((pOpImm = dynamic_cast<const OperandImmediate*>(pOp)))
    {
      
      return pOpImm->GetConstVar ();
    }
  else if ((pOpMem = dynamic_cast<const OperandMemory*>(pOp)))
    {
      // ok, so it's needed to do something about the memory offset
      // this will have to be improved. For now cheat to get any
      // simple stack operations
      
      assert (pOpMem->m_wOpSize == 1 || 
	      pOpMem->m_wOpSize == 2 || 
	      pOpMem->m_wOpSize == 4); // others are valid, but not supported by GetStackVar yet
      
      return GetVarFromMemOp(pIncomingState,pVarMemAddr,pOpMem->m_wOpSize);
    }
  return NULL;
}

void DataAnalyseFunction::SetVariableToOp (OutgoingState* pOutgoingState,
					   const Operand* pOp,
					   Variable* pVar,
					   Variable* pVarMemAddr)
{
  assert (pOutgoingState);
  assert (pOp);
  assert (pVar);
  assert (pVar->GetSize() == pOp->GetOpSize());
  
  const OperandRegister* pOpReg;
  const OperandMemory* pOpMem;
  
  if ((pOpReg = dynamic_cast<const OperandRegister*>(pOp)))
    {
      
      pOutgoingState->SetRegisterVar (pOpReg->m_bReg,pVar);
    }
  else if ((pOpMem = dynamic_cast<const OperandMemory*>(pOp)))
    {
      // ok, so it's needed to do something about the memory offset
      // this will have to be improved. For now cheat to get any
      // simple stack operations

      SetVarToMemOp(pOutgoingState,pVarMemAddr,pVar);
    }
  else
    {
      assert (false); // there is no other valid destination op
      // only memory or register acceptable     
      
      // assert (!dynamic_cast<const OperandImmediate*>(pOp)); // couldn't set to immediate
    }
}
/*
Variable* VariableNCalcOpResult(Variable* ,Variable* )
{
	return NULL;
}
*/

Variable* CalcDestOpMemAddr(OutgoingState* pIncoming,const Operand* pOp)
{
  const OperandMemory* pOpMem = dynamic_cast<const OperandMemory*>(pOp);

  if (!pOpMem)
    return NULL;
  
  Variable* pVarBase  = pOpMem->m_bBaseReg  && pIncoming ? 
    pIncoming->GetRegisterVar (pOpMem->m_bBaseReg)  : NULL;
  Variable* pVarIndex = pOpMem->m_bIndexReg && pIncoming ? 
    pIncoming->GetRegisterVar (pOpMem->m_bIndexReg) : NULL;
  
  Variable* pVarMemAddr = CalcMemAddr(pVarBase,pVarIndex,pOpMem);
  
  return pVarMemAddr;
}

Variable* ZeroExtendVarToSize(Variable* pVar,dword nSize)
{
  if (!pVar || pVar->GetSize() == nSize)
    return pVar; // nothing to do
  
  // TODO: extend these operations (if called from another place than LEA)
  
  assert (nSize == 2 || nSize == 4);
  assert (pVar->GetSize() == 2 || pVar->GetSize() == 4);

  Variable16* pVar16;
  
  if ((pVar16 = dynamic_cast<Variable16*>(pVar)))
    {
      assert (nSize == 4);
      return new Variable32Split(new Variable16Const(0),pVar16);
    }
  Variable32* pVar32;
   
  if ((pVar32 = dynamic_cast<Variable32*>(pVar)))
    {
      assert (nSize == 2);
      return pVar32->GetHalfPart(0);
    }
  assert (false);
  return NULL; // should never be reached
}

void DataAnalyseFunction::DetermineStateChange (AnalyseCurInstr& analyseCurInstr,
						OutgoingState* pIncoming,
						OutgoingState* pOutgoingState,
						const Instruction* pInstruction)
{
  char instr[64];
  char op[64];

  pInstruction->InstrAsmString (instr);
  pInstruction->OpAsmString (op);

  printf ("INSTR: %s  %s\n",instr, op);


  if (pInstruction->m_nInstruction != instrPOP)
    {
      // because pop changes eSP before calculating the address..., so
      // we'll have to do this after that...

      analyseCurInstr.pVarDestMemOp = 
	CalcDestOpMemAddr(pIncoming,pInstruction->GetDestOp());
      analyseCurInstr.pVarSrcMemOp  = 
	CalcDestOpMemAddr(pIncoming,pInstruction->GetSrcOp());
    }

  if (pInstruction->m_nInstruction == instrMOV)
    {
      int nSize;

      Variable* pVarSrc = 
	GetVariableFromOpSrc (pIncoming,pInstruction,
			      analyseCurInstr.pVarSrcMemOp,nSize);

      //if (pVarSrc)
      //printf ("pVarSrc: %p(%p)\n",(void*) pVarSrc,(void*) pVarSrc->m_pVarContent);
      
      if (!pVarSrc)
	pVarSrc = VariableN_Unknown(nSize); // we must set the result anyway
      
      SetVariableToOpDest (pOutgoingState,pInstruction,pVarSrc,analyseCurInstr.pVarDestMemOp);
      
      // no need to set variable src and depend, because it is an
      // exact copy (hmm, is this wise?, because then we possibly get
      // several owners of the same variable)
    }
  else if (pInstruction->m_nInstruction == instrPUSH)
    {
      // we've got to get esp, subtract operand size and then get put
      // our argument into that stack position...

      int nSize;

      Variable* pVarSrc = 
	GetVariableFromOpSrc (pIncoming,pInstruction,
			      analyseCurInstr.pVarSrcMemOp,nSize);
      
      Variable* pVarESP = pIncoming ? 
	pIncoming->GetRegisterVar (regESP)  : NULL;
      
      pVarESP = CalcAddToVar(pVarESP,(dword) -4 /*add*/,4 /*size*/);
      
      pOutgoingState->SetRegisterVar(regESP,pVarESP);
      
      if (!pVarSrc)
	pVarSrc = VariableN_Unknown(nSize); // we must set the result anyway
      
      SetVarToMemOp(pOutgoingState,pVarESP,pVarSrc);
      
      // SetVariableToOp (pOutgoingState,pMemOpESP,pVarSrc);
    }
  else if (pInstruction->m_nInstruction == instrPOP)
    {
      // we've got to get esp, subtract operand size and then get put
      // our argument into that stack position...

      // Variable* pVarSrc = GetVariableFromOpSrc (pIncoming,pInstruction);
      
      // eSP += 4

      int nSize = pInstruction->GetDestOp()->GetOpSize();
      
      Variable* pVarESP = 
	pIncoming ? pIncoming->GetRegisterVar (regESP)  : NULL;

      Variable* pVarSrc = 
	GetVarFromMemOp (pIncoming,pVarESP,nSize);

      pVarESP = CalcAddToVar(pVarESP,(dword) 4 /*add*/,4 /*size*/);
      
      const OperandRegister* pOpReg = 
	dynamic_cast<const OperandRegister*>(pInstruction->GetDestOp());
      
      if (!pOpReg ||
	  pOpReg->m_bReg != regESP) // do not set it if dest is ESP, because, then SetVariableToOpDest will set it again
	pOutgoingState->SetRegisterVar(regESP,pVarESP);
      
      // this calculation of destination address must be after ESP is updated in pOutgoingState, becuase ESP may
      // be used in the address as base register
      
      analyseCurInstr.pVarDestMemOp = 
	CalcDestOpMemAddr(pOutgoingState,pInstruction->GetDestOp());
      
      if (!pVarSrc)
	pVarSrc = VariableN_Unknown(nSize); // we must set the result anyway
      
      SetVariableToOpDest (pOutgoingState,pInstruction,
			   pVarSrc,analyseCurInstr.pVarDestMemOp);

      // if (pVarSrc)
      //	SetVariableToOp (pOutgoingState,pMemOpESP,pVarSrc);
    }
  else if (pInstruction->m_nInstruction == instrNOP)
    {
      pOutgoingState->MakeSureHasRegisterState ();
      // nothing to be done...
    }
  else if (pInstruction->m_nInstruction == instrLEA)
    {
      assert (dynamic_cast<const OperandMemory*>(pInstruction->GetSrcOp()));
      assert (dynamic_cast<const OperandRegister*>(pInstruction->GetDestOp()));
      assert (analyseCurInstr.pVarDestMemOp == NULL);

      int nSize = pInstruction->GetDestOp()->GetOpSize();

      Variable* pVarSrc = analyseCurInstr.pVarSrcMemOp;

      if (pVarSrc)
	pVarSrc = ZeroExtendVarToSize(pVarSrc,nSize);
      else
	pVarSrc = VariableN_Unknown(nSize); // we must set the result anyway
      
      SetVariableToOpDest (pOutgoingState,pInstruction,
			   pVarSrc,NULL /* analyseCurInstr.pVarDestMemOp */);
      // nothing to be done...
    }
  else if ((pInstruction->m_nInstruction >= instrADD &&
	    pInstruction->m_nInstruction <= instrCMP) ||
	   pInstruction->m_nInstruction == instrTEST)
    {
      int nSize;
      
      Variable* pVarDest = 
	GetVariableFromOpDest(pIncoming,pInstruction,
			      analyseCurInstr.pVarDestMemOp,nSize);
      Variable* pVarSrc  = 
	GetVariableFromOpSrc (pIncoming,pInstruction,
			      analyseCurInstr.pVarSrcMemOp);
      
      Variable* pVarResult = NULL;
      
      StatusFlagsAffected flagsAffected;
      
      if ((pInstruction->m_nInstruction == instrSUB ||
           pInstruction->m_nInstruction == instrCMP ||
           pInstruction->m_nInstruction == instrXOR) &&
          pInstruction->OpDestSrcSameRegister ())
        {
          // we know it will be zero
          
          pVarResult = 
	    CalcOpResultZeroIdenticalArgs(nSize,
					  pInstruction->m_nInstruction,
					  flagsAffected);
        }
      else
        {
	  // we should take special care if the src and dest are the
	  // same (this can only happen with registers, as memory
	  // locations are either read or write (string instructions
	  // are an exception, but they are always src ESI dest EDI,
	  // which are different (usually - is there some time one
	  // might like to run a string instruction with src and dest
	  // the same???)

	  // like add with the same register will always produce an
	  // even result...
	  
          pVarResult = 
	    CalcOpResult(nSize,pInstruction->m_nInstruction,
			 pVarDest,pVarSrc,flagsAffected);
          
          
          
          /*
	    if (pVarDest)
	    pVarResult = ((VariableNBase*) pVarDest)->GetTreatment().CalcOpResult(pInstruction->m_nInstruction,pVarDest,pVarSrc);
	    else if (pVarSrc)
	    pVarResult = ((VariableNBase*) pVarSrc )->GetTreatment().CalcOpResult(pInstruction->m_nInstruction,pVarDest,pVarSrc);
	  */
/*
#define instrADD			(instrgINT_ARITH | 0x0000) 
#define instrOR				(instrgINT_ARITH | 0x0001) 
#define instrADC			(instrgINT_ARITH | 0x0002) 
#define instrSBB			(instrgINT_ARITH | 0x0003) 
#define instrAND			(instrgINT_ARITH | 0x0004) 
#define instrSUB			(instrgINT_ARITH | 0x0005) 
#define instrXOR			(instrgINT_ARITH | 0x0006) 
#define instrCMP			(instrgINT_ARITH | 0x0007) 

#define instrTEST			(instrgINT_ARITH | 0x0007) 
*/
	}

      assert (pVarResult); // we must set the destination with something as it has been affected
      
      if (pInstruction->m_nInstruction != instrCMP &&
          pInstruction->m_nInstruction != instrTEST)
        {
          SetVariableToOpDest (pOutgoingState,pInstruction,
			       pVarResult,analyseCurInstr.pVarDestMemOp);
        }
      pOutgoingState->SetFlags(flagsAffected);		
    }
  else if (pInstruction->m_nInstruction == instrNOT ||
	   pInstruction->m_nInstruction == instrNEG ||
	   pInstruction->m_nInstruction == instrINC ||
	   pInstruction->m_nInstruction == instrDEC)
    {
      int nSize;
      
      Variable* pVarDest = 
	GetVariableFromOpDest(pIncoming,pInstruction,
			      analyseCurInstr.pVarDestMemOp,nSize);
      
      Variable* pVarResult = NULL;
      
      StatusFlagsAffected flagsAffected;
      
      pVarResult = 
	CalcOpResult(nSize,pInstruction->m_nInstruction,
		     pVarDest,flagsAffected);
      
      assert (pVarResult); // we must set the destination with something as it has been affected     
      
      SetVariableToOpDest (pOutgoingState,pInstruction,
			   pVarResult,analyseCurInstr.pVarDestMemOp);
      pOutgoingState->SetFlags(flagsAffected);		
    }
  else
    {
      // we've got no idea what happened to the output state...  This
      // must be fixed, because it also mean wo do not know what
      // happened to the stack, and that is fatal! We also have no
      // idea what registers (and stack, and memory) are used as input
      // to this instruction, so it's impossible to remove anything
      // unneeded above this instruction. Conclusion: TODO: add
      // handlers for everything. At least, it is necessary to have
      // something binding together inputs and outputs so everything
      // untouched is recognised as such. Like a list of all
      // instructions that have no sideeffects except for the
      // registers/memory clearly stated in the instruction..

      pOutgoingState->SetCleared();
      //pOutgoingState->m_bStateCleared = true;
    }
}

// after a jump has been taken (or not taken), we know certain things
// about the flags for single flag jumps, we know the flag we jumped
// on.  for multiple flag jumps, we know the combination that was
// jumped on, and sometimes that also imply knowledge about all the
// flags involved

// first index is condition code, second index is for flags known to
// be zero on no jump, and one on jump resp
/*
static const wFlagsKnownAfterJump[8][2]
{
  { FLAGS_MASK_OF                                              ,FLAGS_MASK_OF
  { FLAGS_MASK_CF                                              ,FLAGS_MASK_CF
  { FLAGS_MASK_ZF                                              ,FLAGS_MASK_ZF
  { FLAGS_MASK_CForZF | FLAGS_MASK_CF | FLAGS_MASK_ZF          ,FLAGS_MASK_CForZF
  {FLAGS_MASK_SF                                              ,FLAGS_MASK_SF
  {FLAGS_MASK_PF                                              ,FLAGS_MASK_PF
  {FLAGS_MASK_SFneqOF                                         ,FLAGS_MASK_SFneqOF
  { FLAGS_MASK_ZForSFneqOF | FLAGS_MASK_ZF | FLAGS_MASK_SFneqOF,FLAGS_MASK_ZForSFneqOF



};
*/


#define JUMPS_DO_JMP      0x01 // we jump 
#define JUMPS_DO_NOT_JMP  0x02 // we do not jump
#define JUMPS_UNKNOWN     0x04 // we do now know what will happen
#define JUMPS_UNDEF       0x08 // set if the jump is undefined due to an undefined (processor and 
                               // instruction and operand content specific if set) flag

static const byte nCCFlag[] = 
{ 
  FLAGS_MASK_OF,
  FLAGS_MASK_CF,
  FLAGS_MASK_ZF,
  FLAGS_MASK_CForZF, 
  FLAGS_MASK_SF,
  FLAGS_MASK_PF,
  FLAGS_MASK_SFneqOF,
  FLAGS_MASK_ZForSFneqOF 
};
 
/*
word JumpsOnFlags(StatusFlagsAffected& ,word nInstr) 
{
  word wCC = INSTR_CC_GET_COND_BOTH(nInstr);
  bool bNeg = INSTR_CC_IS_NEG(nInstr);

  // wCC is the condition code, also known as the low 4 bits of the instruction

  
  // truth table for flag or flag condition (two depending on jumping on negated condition)

  static const byte bJumpTableFLorFL[2][4][4] = {
    {
      { JUMPS_DO_NOT_JMP, JUMPS_DO_JMP    , JUMPS_UNKNOWN   , JUMPS_UNDEF      },
      { JUMPS_DO_JMP    , JUMPS_DO_JMP    , JUMPS_DO_JMP    , JUMPS_DO_JMP     },
      { JUMPS_UNKNOWN   , JUMPS_DO_JMP    , JUMPS_UNKNOWN   , JUMPS_UNDEF      },
      { JUMPS_UNDEF     , JUMPS_DO_JMP    , JUMPS_UNDEF     , JUMPS_UNDEF      } },
    {
      { JUMPS_DO_JMP    , JUMPS_DO_NOT_JMP, JUMPS_UNKNOWN   , JUMPS_UNDEF      },
      { JUMPS_DO_NOT_JMP, JUMPS_DO_NOT_JMP, JUMPS_DO_NOT_JMP, JUMPS_DO_NOT_JMP },
      { JUMPS_UNKNOWN   , JUMPS_DO_NOT_JMP, JUMPS_UNKNOWN   , JUMPS_UNDEF      },
      { JUMPS_UNDEF     , JUMPS_DO_NOT_JMP, JUMPS_UNDEF     , JUMPS_UNDEF      } } };
  
  // first we check if the single flag (or special combination) is known

  word wCCFlag = nCCFlag[wCC >> 1];
  
  if (IsKnown(wCCflag))
    return (GetFlag(wCCflag) ^ bNeg) ? JUMPS_DO_JMP : JUMPS_DO_NOT_JMP;

  // it was not, so we have to take more action to figure out
  
  // special code is needed for the complicted conditions,
  // the easy ones (one flag ones) are handled in default

  switch (wCC)
    {
    case instrmCC_BE: // FLAGS_MASK_CForZF
      // so we did not know about CForZF directly

      // see if we can figure something out anyway,
      // the flags can have these conditions 0,1,?(unknown),x(undefined)
      // so then the truth table is (and it is symmetric)
      // 
      //      0 1 ? x
      //      -------
      //  0 | 0 1 ? x
      //  1 | 1 1 1 1
      //  ? | ? 1 ? x
      //  x | x 1 x x

      return bJumpTableFLorFL[bNeg][GetFlagCode(FLAGS_MASK_CF)][GetFlagCode(FLAGS_MASK_ZF)];

    case instrmCC_L: // FLAGS_MASK_SFneqOF // SF!=OF      = SF^OF
      
      //      0 1 ? x
      //      -------
      //  0 | 0 1 ? x
      //  1 | 1 0 ? x
      //  ? | ? ? ? x
      //  x | x x x x
      //
      // so the result is (any undef ? undef : (any unknown ? unknown : (SF ^ OF)))

      if (IsUndefined(FLAGS_MASK_SF) || IsUndefined(FLAGS_MASK_OF))
	return JUMPS_UNDEF;
      if (IsUnknown(FLAGS_MASK_SF) || IsUnknown(FLAGS_MASK_OF))
	return JUMPS_UNKNOWN;
      return (GetFlag(FLAGS_MASK_SF) ^ GetFlag(FLAGS_MASK_OF) ^ bNeg) ? JUMPS_DO_JMP : JUMPS_DO_NOT_JMP;

    case instrmCC_LE: // FLAGS_MASK_ZForSFneqOF // ZF|(SF!=OF) = ZF|(SF^OF)

      // first we combine SF^OF as above, then we get a truth table like CForZF for ZF and the combination

      byte wSFneqOF = (IsUndefined(FLAGS_MASK_SF) || IsUndefined(FLAGS_MASK_OF)) ? FLAG_UNDEFINED :
	(IsUnknown(FLAGS_MASK_SF) || IsUnknown(FLAGS_MASK_OF)) ? FLAG_UNKNOWN :
	  (GetFlag(FLAGS_MASK_SF) ^ GetFlag(FLAGS_MASK_OF));
      
      return bJumpTableFLorFL[bNeg][wSFneqOF][GetFlagCode(FLAGS_MASK_ZF)];
         
    default:
      // so the flag was not known
      // was it unknown or undefined?

      return IsUndefined(wCCflag) ? JUMPS_UNDEF : JUMPS_UNKNOWN;
    }
}
*/
 
void DataAnalyseFunction::DetermineStateChangeCtrlTransfer (AnalyseCurInstr& analyseCurInstr,
							    OutgoingState* pIncoming,
							    OutgoingState* pOutgoingState,
							    OutgoingState* pOutgoingStateBranch,
							    const BranchInstr* pInstruction)
{
  char instr[64];
  char op[64];
  
  pInstruction->InstrAsmString (instr);
  pInstruction->OpAsmString (op);
  
  printf ("INSTR: %s  %s\n",instr, op);
  
  
  assert (INSTR_GROUP(pInstruction->m_nInstruction) == instrgCTRL_TRANSFER);
  
  analyseCurInstr.pVarDestMemOp = 
    CalcDestOpMemAddr(pIncoming,pInstruction->GetDestOp());

  assert (!dynamic_cast<const OperandMemory*>(pInstruction->GetSrcOp())); // or we'll have to do analyseCurInstr.pVarSrcMemOp = ... also
  
  // printf ("ctrl: %d\n",pInstruction->m_nInstruction);
  
  if ((pInstruction->m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP_COND)
    {
      // printf ("cond\n");
      if (INSTR_IS_COND_COUNT(pInstruction->m_nInstruction))
	{
	  // printf ("count\n");
	  byte bReg = 
	    (pInstruction->m_nInstruction & instrmCOUNT_ECX ? regECX : regCX);
	  byte nBytes = 
	    (pInstruction->m_nInstruction & instrmCOUNT_ECX ? 4 : 2);
	  
	  switch (pInstruction->m_nInstruction & ~instrmCOUNT_ECX)
	    {
	    case instrJECXZ:  // branches on ECX = 0
	      pOutgoingState->SetRegisterVar      (bReg,VariableNNonZero(nBytes));
	      pOutgoingStateBranch->SetRegisterVar(bReg,VariableNZero(nBytes));
	      break;
	    case instrmLOOP:   // branches on eCX != 0
	      pOutgoingState->SetRegisterVar      (bReg,VariableNZero(nBytes));
	      pOutgoingStateBranch->SetRegisterVar(bReg,VariableNNonZero(nBytes));
	      break;
	    case instrmLOOPZ:  // branches on eCX != 0 && ZF == 1
	      pOutgoingStateBranch->SetRegisterVar(bReg,VariableNNonZero(nBytes));
	      pOutgoingStateBranch->SetFlags(StatusFlagsAffected(FLAGS_MASK_ZF,
								 FLAGS_MASK_ZF));
	      break;
	    case instrmLOOPNZ: // branches on eCX != 0 && ZF == 0
	      pOutgoingStateBranch->SetRegisterVar(bReg,VariableNNonZero(nBytes));
	      pOutgoingStateBranch->SetFlags(StatusFlagsAffected(FLAGS_MASK_ZF,0));
	      break;
	    }
	}
      else
	{
	  // conditional jump
	  // case instrmJCC:

	  word wCC = INSTR_CC_GET_COND_BOTH(pInstruction->m_nInstruction);
	  bool bNeg = INSTR_CC_IS_NEG(pInstruction->m_nInstruction);
	  
	  // printf ("Conditional jump: %d(%d,%d)\n",pInstruction->m_nInstruction,wCC,bNeg);
	  
	  word wCCFlag = nCCFlag[wCC >> 1];
	  
	  StatusFlagsAffected flagsAffected      (wCCFlag,bNeg ? wCCFlag : 0);
	  StatusFlagsAffected flagsAffectedBranch(wCCFlag,bNeg ? 0 : wCCFlag);
	  
	  if (wCCFlag == FLAGS_MASK_CForZF)
	    {
	      // if J branch, then we do not know which flag(s) was 1
	      // but if it did not branch, we know that both flags are 0
	      
	      (bNeg ? flagsAffectedBranch : flagsAffected).SetFlags(FLAGS_MASK_CF | FLAGS_MASK_ZF,0);
	    }
	  if (wCCFlag == FLAGS_MASK_ZForSFneqOF)
	    {
	      (bNeg ? flagsAffectedBranch : flagsAffected).SetFlags(FLAGS_MASK_ZF | FLAGS_MASK_SFneqOF,0);
	    }
	  
	  pOutgoingState      ->SetFlags(flagsAffected);
	  pOutgoingStateBranch->SetFlags(flagsAffectedBranch);
	}
      
      
      

      /*
#define instrmJCC           0x0100
#define instrmJECXZ         0x0200

#define instrmLOOP          0x0300
#define instrmLOOPZ         0x0400
#define instrmLOOPNZ        0x0500
      */

      /*
#define instrmJCC           0x0100
#define instrmJECXZ         0x0200

#define instrmLOOP          0x0300
#define instrmLOOPZ         0x0400
#define instrmLOOPNZ        0x0500
      */

    }
  else
    {
      pOutgoingState->SetCleared();
      pOutgoingStateBranch->SetCleared();
      /*
      pOutgoingState->m_bStateCleared = true;
      pOutgoingStateBranch->m_bStateCleared = true;
      */
    }





  // if it is an conditional, then we know the state of certain flags on exit
  




}

//////////////////////////////////////////////////////////////////////

void OutgoingState::SetCleared()
{
  assert (!pRegState);
  pRegState = new RegisterStateUnknown ();
}

void OutgoingState::MakeSureHasRegisterState ()
{
  if (!pRegState) 
    pRegState = new RegisterState (/*m_pLast ? m_pLast->pRegState : NULL*/);
}

void OutgoingState::SetRegisterVar (byte bReg,Variable* pVar)
{
  assert (pVar); 
  
  MakeSureHasRegisterState ();
  
  pRegState->SetRegisterVar (bReg,pVar);
}

void OutgoingState::SetStackVar (sdword dwOffset,Variable* pVar)
{
  assert (pVar); 
  
  MakeSureHasRegisterState ();

  pRegState->SetStackVar (dwOffset,pVar);
}

void OutgoingState::SetFlags(const StatusFlagsAffected& flagsAffected)
{
  MakeSureHasRegisterState ();
  
  if (flagsAffected.AnyAffectedOrUndefined())
    {
      FlagsStateAffected* pFlags = new FlagsStateAffected(flagsAffected);
      
      pRegState->SetFlags(pFlags);
    }
  else
    pRegState->SetFlags(NULL);
}

bool OutgoingState::HasSameContent(OutgoingState* pCompare)
{
  printf ("OutgoingState::HasSameContent   %p %p\n",(void*) m_pLast,(void*) pCompare->m_pLast);
  
  if (m_pLast != pCompare->m_pLast)
    return false;

  if (!pRegState || !pCompare->pRegState)
    return (!pRegState) == (!pCompare->pRegState);

  assert (!dynamic_cast<RegisterStateCombine*>(pRegState));
  assert (!dynamic_cast<RegisterStateCombine*>(pCompare->pRegState));
  
  return pRegState->HasSameContent(pCompare->pRegState);
}

//////////////////////////////////////////////////////////////////////

bool IncomingState::HasSameContent(IncomingState* pCompare)
{
  printf ("IncomingState::HasSameContent   %p %p\n",(void*) m_pLast,(void*) pCompare->m_pLast);

  if (m_pLast != pCompare->m_pLast)
    return false;

  // TODO: must check that ALL known parents are the same
  // this also mean that we DO NOT have to check the combined state,
  // because it only depend on the parent states (and is updated lazily,
  // i.e. only when something is requested)

  // 

  printf ("IncomingState::HasSameContent   %p %p\n",(void*) pRegState,(void*) pCompare->pRegState);
  
  if (!pRegState || !pCompare->pRegState)
    return (!pRegState) == (!pCompare->pRegState);
  
  //  RegisterState* pRegState;

  ASSERT_IS_TYPE_OR_NULL(RegisterStateCombine*,pRegState);
  ASSERT_IS_TYPE_OR_NULL(RegisterStateCombine*,pCompare->pRegState);
  
  return ((RegisterStateCombine*) pRegState)->HasSameContent((RegisterStateCombine*) pCompare->pRegState);
  
  
  /*
    printf ("IncomingState::HasSameContent   %d %d\n",m_vectSrcStates.size(),pCompare->m_vectSrcStates.size());

  if (m_vectSrcStates.size() != pCompare->m_vectSrcStates.size())
    return false;

  FOR_EACH_ITER(VectOutgoingState,m_vectSrcStates,state)
    {
      FOR_EACH_ITER(VectOutgoingState,pCompare->m_vectSrcStates,state_cmp)
	{
	  if (*state == *state_cmp)
	    goto next_state;
	}
      printf ("IncomingState::HasSameContent   %p not found\n",(void*) *state);
      // did not find...
      return false;
      
    next_state:;
    }
  */

  /*  
  // incoming state should either have combined stuff or no state at all (because only one parent, and
  // then none is needed)

  ASSERT_IS_TYPE_OR_NULL(RegisterStateCombine*,pRegState);
  ASSERT_IS_TYPE_OR_NULL(RegisterStateCombine*,pCompare->pRegState);

  assert ((!pRegState) == (!pCompare->pRegState)); // because we're suppose to be the same
  
  // might want some more asserts?? (i.e. check that both combines are the same...)

  return true;
  */
  /*
  if (!pRegState || !pCompare->pRegState)
    return (!pRegState) == (!pCompare->pRegState);
  
  return (RegisterStateCombine*) pRegState->HasSameContent((RegisterStateCombine*) pCompare->pRegState);
  */
}

/*
Variable* OutgoingState::GetRegisterVar (byte bReg) const
{
  return GetVar(bReg);
  
  //if (pRegState)
  //  pRegState->GetVar(bReg);
  //return NULL;
  
}

Variable* OutgoingState::GetStackVar (sdword dwOffset,dword dwSize) const
{
  return GetVar(bReg);
  
  //if (pRegState)
  //  pRegState->GetVar(bReg);
  //return NULL;
  
}
*/
/*
Variable* OutgoingStateInfo::GetRegisterVar (byte bReg) const
{
  return GetVar(bReg);
  
  //if (pRegState)
  //  pRegState->GetVar(bReg);
  //if (m_pLast)
  //  m_pLast->GetRegisterVar(bReg);
  //return NULL;
}
*/
#include "HtmlPage.h"
#include "HtmlTags.h" // dep_chk_unknown   

#include <typeinfo>

void OutgoingState::HtmlPlainDisAsm (HtmlPage& page,
				     const ReportDisAsmInfo& reportDisAsmInfo,
				     bool bBranch) const
{
  if (pRegState)
    {
      if (bBranch)
	{
	  //	  page << "----\n";
	}
      page << typeid(*pRegState).name() << '\n';
      pRegState->HtmlPlainDisAsm (page,reportDisAsmInfo);
    }
  /*
  const RegisterStateParentPtr* pParent = GetParent();

  printf (pParent ? "PARENT\n" : "NO PARENT\n");

  while(pParent)
    {
      page << "****\n";
      if (pParent->GetRegisterState())
	{
	  page << "====\n";
	  
	  pParent->GetRegisterState()->HtmlPlainDisAsm (page,reportDisAsmInfo);
	}
      pParent = pParent->GetParent();
    }
  */
}


//////////////////////////////////////////////////////////////////////

EntryState::EntryState()
{
  
}

void EntryState::InitState()
{
  // the entry state must have a stack state so the eSP register can
  // reference it TODO: we are also to setup the position 0 of the
  // stack so it have all possible values of the return addresses
  // (i.e., a list of what instructions are after the ones the can
  // call us here)
  
  pRegState = new RegisterStateEntry();
  
  // TODO: 16/32 bit, SP or ESP...
  
  VarContStackPtr* stack_ptr = new VarContStackPtr(pRegState,0);
  
  pRegState->SetRegisterVarContent(regESP,stack_ptr);
  
}




