/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataAnalyseFunction.h: interface for the DataAnalysisFunction class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__DATA_ANALYSE_FUNCTION_H__INCLUDED_)
#define __DATA_ANALYSE_FUNCTION_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h"
class DisAssembler; 
#include "RegisterState.h"
#include "setstack.h" 

struct StatusFlagsAffected;   

#include "Instruction.h"
#include "Operand.h"

#include <map>

using namespace std;

class HtmlPage;
class ReportDisAsmInfo;

class Variable;
class CallFunction;

class CtrlTransferTarget;
class BranchTarget;
class BranchInstr;

// Now for situations when one code trace jumps into another, we do
// not want to have to copy all register and variable information from
// the previous trace outgoing to this state's incoming. So we
// basically just want link to the previous state. But: if the
// incoming trace have several possible executional predecessors,
// there is a great possibility that some registers and variable on
// the stack have different values, but some are common. On example
// would be
// 
// SOME STATE
// if (something)
//   modify EAX,EBX like this
// else
//   modify EBX,ECX like that
// INCOMING STATE of following
//
// now the INCOMING STATE is mostly SOME STATE with some unknowns
// EAX,EBX,ECX, so we have to mask them out, while keeping the rest it
// could also be a loop or something. Possibly most registers would be
// touched, but perhaps the stack will be in a more consistent state
// (the goal for example being that registers etc pushed onto the
// stack for saving should be remembered when we come to the pop
// instruction (unless, of course), they were modified.
//
// So the appear the problem of choosing which base we should work on
// when combining different paths. Like above: SOME STATE will be part
// of both the if-branch end the else-branch, but which of if- and
// else- should INCOMING be based on (they both contain SOME STATE, so
// that is no problem) Well' the first simple rule is that if there is
// one fall-thrugh path (and there can only be one), that one is
// choosen. All others reach by jumps.  For the rest, we at least need
// to keep the code in such a way that it does not depend on what we
// base something on, because paths my be added while
// analysing. Anyway, the varable values will not depend on the path
// choosen.
//
// A special class for this with only the register and not the value
// is not needed (one could save something because the value will be
// unknown anyway), but we need a value which can know where (possibly
// several places) where it was set anyway since we'lll need to
// back-propagate information about what values are pointers.


class OutgoingState : public RegisterStateParentPtr
{
 public:
  OutgoingState() { m_pLast = NULL; pRegState = NULL; }
  OutgoingState(OutgoingState* pLast) { m_pLast = pLast; pRegState = NULL; }
  
 public:
  virtual ~OutgoingState() { }
  
 public:
  OutgoingState* m_pLast;

 public:
  void SetCleared();

 public:
  RegisterState* pRegState;
  
  // Variable* GetRegisterVar (byte bReg) const;
  // Variable* GetStackVar (sdword dwOffset,dword dwSize) const;
  
 public:
  void SetRegisterVar (byte bReg,Variable* pVar);
  void SetStackVar (sdword dwOffset,Variable* pVar);

  void SetFlags (const StatusFlagsAffected& flagsAffected);
  
  //protected:
  void MakeSureHasRegisterState ();

 public:
  bool HasSameContent(OutgoingState* pCompare);

 public:
  virtual const RegisterStateParentPtr* GetParent() const { return m_pLast; }
  virtual RegisterState* GetRegisterState() const { return pRegState; }

 public:
  //bool HasAnything () { return pRegState != NULL; }
  
 public:
  virtual void HtmlPlainDisAsm (HtmlPage& page,
				const ReportDisAsmInfo& reportDisAsmInfo,
				bool bBranch) const;


};

typedef vector<OutgoingState*> VectOutgoingState;

class IncomingState : public OutgoingState
{
 public:
  IncomingState() { }
  IncomingState(OutgoingState* pLast) : OutgoingState(pLast) { }
 public:
  virtual ~IncomingState() { }
  
 public:
  VectOutgoingState m_vectSrcStates;
  
 public:
  bool HasSameContent(IncomingState* pNewState);
};

class EntryState : public OutgoingState
{
 public:
  EntryState();

 public:
  void InitState();

};

struct AnalyseCurInstr
{
  Variable* pVarDestMemOp; // variable of calculated dest memory addr
  Variable* pVarSrcMemOp;  //                        src
};

#include "mapstack.h"
#include "setdeque.h"

#define WHY_ANALYSE_INITIAL         0x01
#define WHY_ANALYSE_CALLER_CHANGED  0x02
#define WHY_ANALYSE_CALLED_CHANGED  0x04

class WhyAnalyse
{
 public:
  WhyAnalyse(byte bWhy) { m_bWhy = bWhy; }
  
 public:
  byte m_bWhy;
  
 public:
  void combine(const WhyAnalyse& other) { m_bWhy |= other.m_bWhy; }  
};

typedef mapstack<CallFunction*,WhyAnalyse> MapStackCallFunctions;

typedef set<IncomingState*> SetIncomingState;

typedef map<const CtrlTransferTarget*,SetIncomingState> MapCTTSetIncomingState;

typedef map<const CtrlTransferTarget*,IncomingState*> MapCTTIncomingState;
typedef map<const BranchTarget*,OutgoingState*> MapBTOutgoingState;

typedef map<const CtrlTransferTarget*,EntryState*> MapCTTEntryState;

typedef setdeque<CtrlTransferTarget*> SetDequeCtrlTransferTargets;

class DataAnalyseFunction  
{
 public:
  DataAnalyseFunction();
  virtual ~DataAnalyseFunction();
  
 public:
  
  MapStackCallFunctions m_mapstackAnalyseFunctions;
  
  SetDequeCtrlTransferTargets m_setdequeToAnalyse;

 public:
  MapCTTEntryState mapEntryState;

  MapCTTSetIncomingState mapOldIncomingStates; // perhaps this should be integrated into mapIncomingState.. 

  MapCTTIncomingState mapIncomingState;
  MapBTOutgoingState mapOutgoingState;
  MapBTOutgoingState mapOutgoingStateBranch;
 
 public:
  CallFunction* m_pFunction;
  
 public:
  void AddFunctionToAnalyse (CallFunction* pFunction);
  CallFunction* NextFuncToAnalyse();
  
 public:
  bool SetFunction (CallFunction* pFunction);
  void AnalyseFunction ();

  void AnalyseCtrlTransferTarget(CtrlTransferTarget* pCtrlTarget);

 public:
  void AddOldIncomingState(CtrlTransferTarget* pCtrlTarget,
			   IncomingState* pOldIncomingState);
  
 public:
  IncomingState* CalcIncomingState (CtrlTransferTarget* pCtrlTarget);

 protected:
  void DetermineStateChange (AnalyseCurInstr& analyseCurInstr,
			     OutgoingState* pIncoming,
                             OutgoingState* pOutgoingState,
                             const Instruction* pInstruction);
  
  void DetermineStateChangeCtrlTransfer (AnalyseCurInstr& analyseCurInstr,
					 OutgoingState* pIncoming,
                                         OutgoingState* pOutgoingState,
                                         OutgoingState* pOutgoingStateBranch,
                                         const BranchInstr* pInstruction);
  
 protected:
  Variable* GetVarFromMemOp(OutgoingState* pIncomingState,
			    Variable* pVarMemAddr,dword dwSize);

  Variable* GetVariableFromOp (OutgoingState* pIncomingState,
			       const Operand* pOp,Variable* pVarMemAddr);
  
  Variable* GetVariableFromOpDest (OutgoingState* pIncomingState,
				   const Instruction* pInstruction,
				   Variable* pVarMemAddr,int& nSize) 
    { 
      const Operand* pOp = pInstruction->GetDestOp(); 
      assert(pOp); 
      nSize = pOp->GetOpSize(); 
      return GetVariableFromOp(pIncomingState,pOp,pVarMemAddr); 
    }
  Variable* GetVariableFromOpSrc (OutgoingState* pIncomingState,
				  const Instruction* pInstruction,
				  Variable* pVarMemAddr,int& nSize) 
    { 
      const Operand* pOp = pInstruction->GetSrcOp(); 
      assert(pOp); 
      nSize = pOp->GetOpSize(); 
      return GetVariableFromOp(pIncomingState,pOp,pVarMemAddr); 
    }
  Variable* GetVariableFromOpDest (OutgoingState* pIncomingState,
				   const Instruction* pInstruction,
				   Variable* pVarMemAddr) 
    { 
      return GetVariableFromOp(pIncomingState,
			       pInstruction->GetDestOp(),
			       pVarMemAddr); 
    }
  Variable* GetVariableFromOpSrc  (OutgoingState* pIncomingState,
				   const Instruction* pInstruction,
				   Variable* pVarMemAddr) 
    { 
      return GetVariableFromOp(pIncomingState,
			       pInstruction->GetSrcOp (),
			       pVarMemAddr);
    }
  Variable* GetVariableFromOpSrc2 (OutgoingState* pIncomingState,
				   const Instruction* pInstruction) 
    { 
      return GetVariableFromOp(pIncomingState,
			       pInstruction->GetSrc2Op(),
			       NULL); 
    }
  /*
  Variable* GetVariableFromOp (OutgoingStateInfo& outgoingState,const Operand* pOp);
  
  Variable* GetVariableFromOpDest (OutgoingStateInfo& outgoingState,const Instruction* pInstruction,int& nSize) 
    { 
      const Operand* pOp = pInstruction->GetDestOp(); 
      assert(pOp); 
      nSize = pOp->GetOpSize(); 
      return GetVariableFromOp(outgoingState,pOp); 
    }
  Variable* GetVariableFromOpDest (OutgoingStateInfo& outgoingState,const Instruction* pInstruction) 
    { return GetVariableFromOp(outgoingState,pInstruction->GetDestOp()); }
  Variable* GetVariableFromOpSrc  (OutgoingStateInfo& outgoingState,const Instruction* pInstruction) 
    { return GetVariableFromOp(outgoingState,pInstruction->GetSrcOp ()); }
  Variable* GetVariableFromOpSrc2 (OutgoingStateInfo& outgoingState,const Instruction* pInstruction) 
    { return GetVariableFromOp(outgoingState,pInstruction->GetSrc2Op()); }
*/  
  void SetVarToMemOp(OutgoingState* pOutgoingState,
		     Variable* pVarMemAddr,Variable* pVar);
  
  void SetVariableToOp (OutgoingState* pOutgoingState,
			const Operand* pOp,
			Variable* pVar,Variable* pVarMemAddr);
  void SetVariableToOpDest (OutgoingState* pOutgoingState,
			    const Instruction* pInstruction,
			    Variable* pVar,Variable* pVarMemAddr) 
    {
      SetVariableToOp(pOutgoingState,pInstruction->GetDestOp(),
		      pVar,pVarMemAddr); 
    } 
  void SetVariableToOpSrc  (OutgoingState* pOutgoingState,
			    const Instruction* pInstruction,
			    Variable* pVar,
			    Variable* pVarMemAddr) 
    { 
      SetVariableToOp(pOutgoingState,pInstruction->GetSrcOp (),
		      pVar,pVarMemAddr); 
    } 
  
  
  
};

#endif // !defined(__DATA_ANALYSE_FUNCTION_H__INCLUDED_)





