/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataPtrPESectionBody.cpp: implementation of the DataPtrPESectionBody class.
//
//////////////////////////////////////////////////////////////////////

#include "PEFileSectionBody.h" // dep_chk_needed 
#include "DataPtrPESectionBody.h" // dep_chk_needed   
#include "gen/DataPtrSrcError.h" // dep_chk_unknown   
#include <assert.h>

#include <list>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DataPtrPESectionBody::DataPtrPESectionBody(PEFileSectionBody* pDataSrc,
					   dword dwVirtualAddress)
{
  assert (pDataSrc);
  assert (pDataSrc->m_pData);
  assert (dwVirtualAddress - pDataSrc->m_dwVirtualAddress < 
	  pDataSrc->m_nSizeOfRawData);
  
  m_pDataSrc = pDataSrc;
  
  m_pCurrent = m_pDataSrc->m_pData + 
    dwVirtualAddress - m_pDataSrc->m_dwVirtualAddress;
}

DataPtrPESectionBody::~DataPtrPESectionBody()
{

}

dword DataPtrPESectionBody::GetRVA ()
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  assert (m_pCurrent >= m_pDataSrc->m_pData && 
	  m_pCurrent <= m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData);
  
  return m_pDataSrc->m_dwVirtualAddress+(m_pCurrent-m_pDataSrc->m_pData);
}

const byte* DataPtrPESectionBody::GetPtr (uint nBytes)
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  assert (m_pCurrent >= m_pDataSrc->m_pData && 
	  m_pCurrent <= m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData);
  
  if (((const byte*) m_pCurrent) + nBytes > 
      m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData)
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
  
  const byte* ptr = m_pCurrent;
  
  m_pCurrent = m_pCurrent + nBytes;
  
  return ptr;
}


const byte* DataPtrPESectionBody::GetPtr (dword dwVirtualAddress,uint nBytes)
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  const byte* pStart = 
    m_pDataSrc->m_pData + dwVirtualAddress - m_pDataSrc->m_dwVirtualAddress;
  
  if (pStart + nBytes > m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData)
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
  
  return pStart;
}

bool DataPtrPESectionBody::IsValidAddress (dword dwVirtualAddress)
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  return (dwVirtualAddress >= m_pDataSrc->m_dwVirtualAddress &&
	  dwVirtualAddress <  m_pDataSrc->m_dwVirtualAddress + m_pDataSrc->m_nSizeOfRawData);
}

bool DataPtrPESectionBody::IsValidAddress (dword dwVirtualAddress,dword dwSize)
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  return (dwVirtualAddress        >= m_pDataSrc->m_dwVirtualAddress &&
	  dwVirtualAddress+dwSize <= m_pDataSrc->m_dwVirtualAddress + m_pDataSrc->m_nSizeOfRawData);
}

const char* DataPtrPESectionBody::GetString ()
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  assert (m_pCurrent >= m_pDataSrc->m_pData && 
          m_pCurrent <= m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData);
  
  // we must first determine if string is withing bound
  //
  // cannot use strlen as it may wander of too far if string is not '\0'-terminated
  
  const byte* pText = m_pCurrent;
  const byte* pEnd = m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData;
  
  for ( ; ; )
    {
      if (pText >= pEnd)
        throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
      
      if (!(*pText))
        break;
      pText++;
    }
  
  pText++; // pText points to byte after ending '\0'
  swap (pText,m_pCurrent);
  
  // m_pCurrent now points after string
  
  return (const char*) pText;
}

const char* DataPtrPESectionBody::GetString (dword dwVirtualAddress)
{
  assert (m_pDataSrc);
  assert (m_pDataSrc->m_pData);
  
  // we must first determine if string is withing bound
  //
  // cannot use strlen as it may wander of too far if string is not '\0'-terminated

  const byte* pStart = m_pDataSrc->m_pData + 
    dwVirtualAddress - m_pDataSrc->m_dwVirtualAddress;
  
  if (pStart < m_pDataSrc->m_pData)
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
  
  const byte* pText = pStart;
  const byte* pEnd = m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData;
  
  for ( ; ; )
    {
      if (pText >= pEnd)
        throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
      
      if (!(*pText))
        break;
      pText++;
    }
  
  // m_pCurrent now points after string
  
  return (const char*) pStart;
}

DataPtrPESectionBody* DataPtrPESectionBody::GetRVADataPtr (dword dwVirtualAddress)
{
  if (dwVirtualAddress < m_pDataSrc->m_dwVirtualAddress ||
      dwVirtualAddress >= m_pDataSrc->m_dwVirtualAddress + m_pDataSrc->m_nSizeOfRawData)
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
  
  return new DataPtrPESectionBody(m_pDataSrc,dwVirtualAddress);
}


