/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataPtrSrc.cpp: implementation of the DataPtrSrc class.
//
//////////////////////////////////////////////////////////////////////

#include "DataPtrSrc.h" // dep_chk_needed   
#include "gen/DataPtrSrcError.h" // dep_chk_unknown   
#include <assert.h>

#include <list>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DataPtrSrc::DataPtrSrc(const byte* pStart,const byte* pEnd)
{
  SetSrc(pStart,pEnd);
}

DataPtrSrc::DataPtrSrc(const byte* pStart,dword dwLength,dword dwCurOffset)
{
  assert (dwCurOffset < dwLength);
  SetSrc(pStart,pStart+dwLength,dwCurOffset);
}

DataPtrSrc::DataPtrSrc()
{
  m_pStart = NULL;
  m_pEnd = NULL;
  
  m_pCurrent = NULL;
}

DataPtrSrc::~DataPtrSrc()
{

}

void DataPtrSrc::SetSrc(const byte* pStart,const byte* pEnd,dword dwCurOffset)
{
  assert (pStart && pEnd && pEnd >= pStart);
  assert (dwCurOffset <= (dword) (pEnd - pStart));
  
  m_pStart = pStart;
  m_pEnd = pEnd;
  
  m_pCurrent = m_pStart + dwCurOffset;
}

dword DataPtrSrc::GetOffset ()
{
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  return m_pCurrent-m_pStart;
}

void DataPtrSrc::SetCurrent (const byte* pCur)
{
  assert (pCur >= m_pStart && pCur <= m_pEnd);

  m_pCurrent = pCur;
}

void DataPtrSrc::SetOffset(dword dwRVA)
{
  assert (m_pStart + dwRVA <= m_pEnd);
  
  m_pCurrent = m_pStart + dwRVA;
}


const byte* DataPtrSrc::GetPtr (uint nBytes)
{
  assert (m_pCurrent);
  assert (m_pStart);
  
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  if (nBytes > (uint) (m_pEnd - m_pCurrent))
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(GetRVA());
  
  const byte* ptr = m_pCurrent;
  
  m_pCurrent = m_pCurrent + nBytes;
  
  return ptr;
}

const byte* DataPtrSrc::GetPtr (dword dwVirtualAddress,uint nBytes)
{
  assert (m_pStart);
  
  if (dwVirtualAddress > (dword) (m_pEnd - m_pStart))
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(dwVirtualAddress);
  
  if (dwVirtualAddress + nBytes > (dword) (m_pEnd - m_pStart))
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(dwVirtualAddress);
  
  const byte* pStart = m_pStart + dwVirtualAddress;

  return pStart;
}

uint DataPtrSrc::GetStringLength (const byte* pStart)
{
  const byte* pEnd = pStart;
  
  for ( ; ; )
    {
      if (pEnd >= m_pEnd)
        throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(m_pStart-m_pEnd);
      
      if (!(*pEnd))
        break;
      ++pEnd;
    }
  return pEnd - pStart;
}

const char* DataPtrSrc::GetStringPtr(uint* nLength)
{
  assert (m_pStart);

  const byte* pStart = m_pCurrent;
  
  *nLength = GetStringLength(m_pCurrent);
  m_pCurrent += *nLength + 1;
  return (const char*) pStart;
}

const char* DataPtrSrc::GetStringPtr(dword dwVirtualAddress,uint* nLength)
{
  assert (m_pStart);
  
  if (dwVirtualAddress > (dword) (m_pEnd - m_pStart))
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(dwVirtualAddress);
  
  *nLength = GetStringLength(m_pStart + dwVirtualAddress);
  return (const char*) (m_pStart + dwVirtualAddress);
}

const char* DataPtrSrc::GetStringPtr(dword dwVirtualAddress)
{
  assert (m_pStart);
  
  if (dwVirtualAddress > (dword) (m_pEnd - m_pStart))
    throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(dwVirtualAddress);
  
  GetStringLength(m_pStart + dwVirtualAddress); // make sure it is null terminated inside data source
  return (const char*) (m_pStart + dwVirtualAddress);
}

bool DataPtrSrc::IsValidAddress (dword dwVirtualAddress)
{
  assert (m_pEnd >= m_pStart);
  return (dwVirtualAddress < (dword) (m_pEnd - m_pStart));
}

//////////////////////////////////////////////////////////////////////

DataPtrSrcRVA::DataPtrSrcRVA(const byte* pStart,dword dwRVAOffset,
			     dword dwLength,dword dwRVACur)
  : DataPtrSrc(pStart,dwLength,dwRVACur-dwRVAOffset)
{
  m_dwRVAOffset = dwRVAOffset;
}

const byte* DataPtrSrcRVA::GetPtr (dword dwVirtualAddress,uint nBytes)
{
  return DataPtrSrc::GetPtr(dwVirtualAddress-m_dwRVAOffset,nBytes);
}

bool DataPtrSrcRVA::IsValidAddress (dword dwVirtualAddress)
{
  return DataPtrSrc::IsValidAddress(dwVirtualAddress-m_dwRVAOffset);
}

dword DataPtrSrcRVA::GetRVA ()
{
  return DataPtrSrc::GetRVA()+m_dwRVAOffset;
}

void DataPtrSrcRVA::SetRVA(dword dwRVA)
{
  DataPtrSrc::SetRVA(dwRVA-m_dwRVAOffset);
}


/*
const char* DataPtrSrc::GetString ()
{
	assert (m_pDataSrc);
	assert (m_pDataSrc->m_pData);

	assert (m_pCurrent >= m_pDataSrc->m_pData && 
		m_pCurrent <= m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData);

	// we must first determine if string is withing bound
	//
	// cannot use strlen as it may wander of too far if string is not '\0'-terminated

	const byte* pText = m_pCurrent;
	const byte* pEnd = m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData;
	
	for ( ; ; )
	{
		if (pText >= pEnd)
			throw "Attempt to read outside data source.";

		if (!(*pText))
			break;
		pText++;
	}

	pText++; // pText points to byte after ending '\0'
	swap (pText,m_pCurrent);

	// m_pCurrent now points after string

	return (const char*) pText;
}

const char* DataPtrSrc::GetString (dword dwVirtualAddress)
{
	assert (m_pDataSrc);
	assert (m_pDataSrc->m_pData);

	// we must first determine if string is withing bound
	//
	// cannot use strlen as it may wander of too far if string is not '\0'-terminated

	const byte* pStart = m_pDataSrc->m_pData + dwVirtualAddress - m_pDataSrc->m_dwVirtualAddress;

	if (pStart < m_pDataSrc->m_pData)
		throw "Attempt to read outside data source.";

	const byte* pText = pStart;
	const byte* pEnd = m_pDataSrc->m_pData + m_pDataSrc->m_nSizeOfRawData;
	
	for ( ; ; )
	{
		if (pText >= pEnd)
			throw "Attempt to read outside data source.";

		if (!(*pText))
			break;
		pText++;
	}

	// m_pCurrent now points after string

	return (const char*) pStart;
}

DataPtrSrc DataPtrSrc::GetRVADataPtr (dword dwVirtualAddress)
{
	if (dwVirtualAddress < m_pDataSrc->m_dwVirtualAddress ||
		dwVirtualAddress >= m_pDataSrc->m_dwVirtualAddress + m_pDataSrc->m_nSizeOfRawData)
		throw "Attempt to read outside data source.";

	return DataPtrSrc(m_pDataSrc,dwVirtualAddress);
}
*/

