/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DisAssembler_DataAnalysis.cpp: implementation of the DisAssembler class.
//
//////////////////////////////////////////////////////////////////////

#include "DisAssembler.h" 
#include "BranchTargetTraceSingle.h" 
#include "BranchTargetTrace.h"
#include "DataAnalyseFunction.h" 
#include "Executable.h"
#include "CallFunction.h"

#include "ReportDisAsmInfo.h"

#include <stdio.h>

#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void DisAssembler::DataAnalysis (UrlFolder* pParentFolder,
				 ReportDisAsmInfo& reportDisAsmInfo)
{
  //	LiveRegisterAnalysis liveRegister;
  
  m_pExecutable->InitVariableState();
  
  const SetCallFunctions* pSetCallFunctions = 
    m_pExecutable->GetSetCallFunctions();
  
  // This is a possibility for future improvement: when we are (able)
  // to handle larger programs memory will be used in large amounts
  // (which may be bad, depending on what the user has available),
  // this is then e possibility for the system to adapt to such
  // situations. The code is stored in traces, but more efficiently in
  // larger chunks (that is one trace between every jump, instead of a
  // minimal trace for every instruction). To ease handling, the
  // analyser may want it as minimal traces, but when a function is
  // not in the analyser, it may be compressed into larger traces, and
  // then unpacked when it has to be reanalysed. This also goes for
  // the information the analyser create. When a function is not _the_
  // one being analysed, the inner guts may be forgotten (memory
  // saving), but when it comes in for analysis again, all has to be
  // recreated, which necessarily did not have to happen if the
  // information was kept.
  //
  // For now, we do not care, but keep everyting. But not to destroy
  // this opportunity: have in mind not to construct the analyser such
  // that inner things about a function cannot be forgotten between
  // passes Things that may not be forgotten is of course anything
  // interfacing with other functions, like calls and returns
  
  FOR_EACH_CONST_ITER(SetCallFunctions,*pSetCallFunctions,i)
    {
      ConvertTracesToSingle(*i);
    }
  
  // now perform data analysis of functions until no further
  // information can be obtained
  // 
  // this means that we'll visit and analyse every function at least
  // once and when analysis of an function means that we have more
  // information for another function, that is put up on the list for
  // reanalysis
  
  // we'll keep the functions to be analysed in a stack and a set and
  // work until it's empty
   
  FOR_EACH_CONST_ITER(SetCallFunctions,*pSetCallFunctions,func)
    {
      m_pExecutable->m_analysis.AddFunctionToAnalyse (*func);
    }
  
  CallFunction* pFunc;
  
  while ((pFunc = m_pExecutable->m_analysis.NextFuncToAnalyse ()))
    {
      m_pExecutable->m_analysis.AnalyseFunction();
    }

  // and perhaps, would you be so kind to eject any information you
  // may have collected?

  reportDisAsmInfo.SetAnalyseFunction(&m_pExecutable->m_analysis);

  m_pExecutable->HtmlJavaDisAsmCreateUrls (pParentFolder,reportDisAsmInfo);
  m_pExecutable->HtmlJavaDisAsm (reportDisAsmInfo);
  
  reportDisAsmInfo.SetAnalyseFunction(NULL);
  
}

void DisAssembler::ConvertTracesToSingle (CallFunction* pFunction)
{
  VectCodeTraces vectCodeTraces;
  
  pFunction->GetCodeTraces (vectCodeTraces);
  
  for (VectCodeTraces::const_iterator i = vectCodeTraces.begin();
       i != vectCodeTraces.end (); i++)
    {
      BranchTarget* pTarget = (*i);
      
      while (pTarget)
        {
          BranchTargetTrace* pTraceMultiple = 
	    dynamic_cast<BranchTargetTrace*>(pTarget);
          
          if (pTraceMultiple)
            {
              BranchTargetTraceSingle* pFirst;
              BranchTargetTraceSingle* pLast;
              
              pFirst = pTraceMultiple->ConvertToTraceSingle(&pLast);
              
              RedirectPointers (pTraceMultiple,pFirst);
              if (pTarget == *i)
                {
                  /*
                    MapBranchTargetVA::iterator i1 = m_mapBranchTargetVA.find(pFirst);
                                        assert (i1 != m_mapBranchTargetVA.end());
                                        dword dwVA = i1->second;
                                        printf ("Retrofitting: %08x (%8x -> %8x)\n",dwVA,pTarget,pFirst);
                  */
                  
                  pFunction->RemoveCodeTrace(pTarget);
                  pFunction->AddCodeTrace(pFirst);
                }
              
              delete pTraceMultiple;
              pTarget = pLast;
            }
          pTarget = pTarget->GetNextTarget();
        }
    }
  
  // PrintConversionInfo();
}
/*
void DisAssembler::ConvertTracesToMultiple (CallFunction* pFunction)
{

}
*/

void DisAssembler::RedirectPointers (BranchTarget* pFrom,BranchTarget* pTo)
{
  MapBranchTargetVA::iterator i1 = m_mapBranchTargetVA.find(pFrom);
  
  assert (i1 != m_mapBranchTargetVA.end());
  
  dword dwVA = i1->second;
  
  MapDisAsmCodePart::iterator i2 = m_mapDisAsmCodePart.find(dwVA);
  
  i2->second.m_pTarget = pTo;
  
  m_mapBranchTargetVA.erase(i1);
  m_mapBranchTargetVA.insert(MapBranchTargetVA::value_type(pTo,dwVA));
  
  MapCodeTraceCallFunction::iterator i3 = 
    m_mapCodeTraceCallFunction.find(pFrom);
  
  if (i3 != m_mapCodeTraceCallFunction.end())
    {
      m_mapCodeTraceCallFunction.insert (MapCodeTraceCallFunction::value_type(pTo,i3->second));
      m_mapCodeTraceCallFunction.erase(i3);
    }
  /*
        





typedef map<dword,DisAsmCodePart>      MapDisAsmCodePart;

typedef map<dword,CallTarget*> MapVACallTarget;
typedef map<dword,JumpTarget*> MapVAJumpTarget;

typedef map<JumpTarget*,CallFunction*> MapUnsetJumpTargetCallFunction;

typedef map<const BranchTarget*,CallFunction*> MapCodeTraceCallFunction;

typedef map<const BranchTarget*,dword> MapBranchTargetVA;

	MapDisAsmCodePart   m_mapDisAsmCodePart;       // traces of code already (or about to be) handled
	MapVACallTarget		m_mapVACallTarget; // where our call targets are
	MapVAJumpTarget		m_mapVAJumpTarget; // where our jump targets are
	MapUnsetJumpTargetCallFunction m_mapUnsetJumpTargetCallFunction;	
	
	MapCodeTraceCallFunction m_mapCodeTraceCallFunction;

	MapBranchTargetVA m_mapBranchTargetVA;
*/
}






