/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ELF_File.cpp: implementation of the ELF_File class.
//
//////////////////////////////////////////////////////////////////////

#include "ELF_File.h"
#include "Elf32_Section.h"
#include "Elf32_ProgramBlock.h"
#include "Error.h"
#include "ErrorTracker.h"
#include "gen/ELF_File_DumpStructs.h"
#include "Executable.h"
#include "ExportedObject.h"
#include "DataPtrSrc.h"
#include "Misc.h"
#include "DisAssembler.h"
#include "ExportedObject.h" 

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _lseek lseek
#define _tell(fildes) (lseek(fildes,0,SEEK_CUR))
#endif
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ELF_File::ELF_File(const string& strFileName)
  : ExecutableImage(strFileName)
{
  m_pSectionHeaders = NULL;
  m_pProgramHeaders = NULL;
  
  m_pFileData = NULL;
  m_nSize = 0;

  e_entryInit = 0;			
  e_entryFini = 0;			
}

ELF_File::~ELF_File()
{
  free(m_pSectionHeaders);
  free(m_pProgramHeaders);
  
  free(m_pFileData);
  
  FOR_EACH_ITER(VectElf32_ProgramBlocks,m_vectProgramBlocks,block)
    delete *(block);
  
  FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
    delete *(sect);
}

//////////////////////////////////////////////////////////////////////

DataPtrSrc* ELF_File::GetFileData(Elf32_Off nOffset,Elf32_Off nSize)
{
  assert (m_pFileData);
  
  if (nOffset > m_nSize)
    throw GENERIC_ERROR("Attempt to get data beyond end of file.");
  if (nOffset + nSize > m_nSize)
    throw GENERIC_ERROR("Attempt to get data past end of file.");
  
  return new DataPtrSrc(m_pFileData+nOffset,m_pFileData+nOffset+nSize);
}

void ELF_File::GetFileData(void* pDest,Elf32_Off nOffset,Elf32_Off nSize)
{
  assert (m_pFileData);
  
  // we get as much data if possible first,
  // if there is an error, we set memory to zero and then throw an note
  
  if (nOffset <= m_nSize)
    {
      if (nOffset + nSize <= m_nSize)
        {
          memcpy (pDest,m_pFileData+nOffset,nSize);
          return; // all's well
        }
      
      Elf32_Off nCopy = m_nSize-nOffset;
      
      memcpy (pDest,m_pFileData+nOffset,nCopy);
      memset (((byte*) pDest)+nCopy,0,nSize-nCopy);
      
      throw GENERIC_ERROR("Attempt to get data past end of file.");
    }
  
  memset (pDest,0,nSize);
  
  throw GENERIC_ERROR("Attempt to get data beyond end of file.");
}

//////////////////////////////////////////////////////////////////////

void ELF_File::ReadFile (int file)
{
  m_nSize = _lseek (file,0,SEEK_END);
  
  printf ("%6x (%7d): File Length\n",m_nSize,m_nSize);

  m_pFileData = (byte*) malloc (m_nSize);

  VALIDATE(_lseek (file,0,SEEK_SET) == 0,
	   "Could not seek to beginning of file");
  
  VALIDATE(read (file,m_pFileData,m_nSize) == (int) m_nSize,
           "Failed to read file.");
  
  m_fileMap.SetFileLength(m_nSize);
  
  TRY_CATCH_EMIT_ERROR_THROW_GENERIC(
           GetFileData (&m_elfIdent,0,sizeof (m_elfIdent)),
                "Missing ELF ident.");
  
  m_fileMap.AddAreaType(0,sizeof (m_elfIdent),
			new GenericFileMapItem(GFMI_ELF_IDENT));
  
  VALIDATE1(m_elfIdent.dw_magic == ELF_ID_MAGIC,
	    "Wrong magic number: %x.",m_elfIdent.dw_magic);
  
  VALIDATE(m_elfIdent.b_class == ELF_ID_CLASS32,
	   "ELF ident class not 32 bit.");
  
  VALIDATE(m_elfIdent.b_data == ELF_ID_DATA2LSB,
	   "ELF ident data not little endian.");
  
  VALIDATE(m_elfIdent.b_version == ELF_EV_CURRENT,
	   "ELF ident version wrong.");
  
  TRY_CATCH_EMIT_ERROR_THROW_GENERIC(
           GetFileData (&m_elfHeader,sizeof (m_elfIdent),sizeof (m_elfHeader)),
                "Missing ELF header.");

  m_fileMap.AddAreaType(sizeof (m_elfIdent),sizeof (m_elfHeader),
			new GenericFileMapItem(GFMI_ELF_HEADER));
  
  VALIDATE(m_elfHeader.e_type == ELF_ET_REL ||
	   m_elfHeader.e_type == ELF_ET_EXEC ||
	   m_elfHeader.e_type == ELF_ET_DYN,
	   "ELF header type wrong.");  
  
  VALIDATE(m_elfHeader.e_machine == ELF_EM_386,
	   "ELF header machine wrong");
  
  VALIDATE(m_elfHeader.e_version == ELF_EV_CURRENT,
	   "ELF header version wrong.");
  
  VALIDATE(m_elfHeader.e_flags == 0,
	   "ELF header flags non-zero.");
  
  VALIDATE1(m_elfHeader.e_ehsize == 
	    sizeof (Elf32_Ident) + sizeof (Elf32_Header),
	    "ELF header size (%d) wrong.",m_elfHeader.e_ehsize);
  
  VALIDATE1(m_elfHeader.e_phentsize == sizeof (Elf32_ProgramHeader) ||
            (m_elfHeader.e_phentsize == 0 && m_elfHeader.e_phnum == 0),
	    "ELF header program header size (%d) wrong.",m_elfHeader.e_phentsize); // allow zero if no entries?
  
  VALIDATE1(m_elfHeader.e_shentsize == sizeof (Elf32_SectionHeader) ||
            (m_elfHeader.e_shentsize == 0 && m_elfHeader.e_shnum == 0),
	    "ELF header section header size (%d) wrong.",m_elfHeader.e_shentsize); // allow zero if no entries?
  
  VALIDATE(m_elfHeader.e_shstrndx == ELF_SHN_UNDEF ||
           m_elfHeader.e_shstrndx < m_elfHeader.e_shnum,
           "ELF header section string index out of bounds.");
  
  VALIDATE(!m_elfHeader.e_phoff == !m_elfHeader.e_phnum,
           "ELF header # program headers not consistent with offset.");
    
  VALIDATE(!m_elfHeader.e_shoff == !m_elfHeader.e_shnum,
           "ELF header # section headers not consistent with offset.");
  
  VALIDATE(m_elfHeader.e_shnum < ELF_SHN_LORESERVE,
           "ELF header # sections headers > reserved.");

  // read section and program headers

  uint nProgHeadsSize = sizeof (Elf32_ProgramHeader) * m_elfHeader.e_phnum;
  uint nSectHeadsSize = sizeof (Elf32_SectionHeader) * m_elfHeader.e_shnum;

  m_pProgramHeaders = new Elf32_ProgramHeader[nProgHeadsSize];
  m_pSectionHeaders = new Elf32_SectionHeader[nSectHeadsSize];

  TRY_CATCH_EMIT_ERROR_THROW_GENERIC(
           GetFileData (m_pProgramHeaders,m_elfHeader.e_phoff,nProgHeadsSize),
                "Missing ELF program headers.");

  m_fileMap.AddAreaType(m_elfHeader.e_phoff,nProgHeadsSize,
			new GenericFileMapItem(GFMI_ELF_PROGRAM_HEADER));
  
  TRY_CATCH_EMIT_ERROR_THROW_GENERIC(
           GetFileData (m_pSectionHeaders,m_elfHeader.e_shoff,nSectHeadsSize),
                "Missing ELF section headers.");

  m_fileMap.AddAreaType(m_elfHeader.e_shoff,nSectHeadsSize,
			new GenericFileMapItem(GFMI_ELF_SECTION_HEADER));
  
   // ok, so now read the data for the sections and programs (ehh, but
   // thats the name we get when stripping header off the name), so we
   // call them programblocks

  uint i;

  Elf32_ProgramHeader* pProgamHeader = m_pProgramHeaders;

  bool bHasHadLoadable = false;
  bool bHasHadInterp = false;
  bool bHasHadPhdr = false;

  for (i = 0; i < m_elfHeader.e_phnum; i++, pProgamHeader++)
    {
      printf ("ph %2d: 0x%08x - 0x%08x (0x%08x)\n",i,
              (m_pProgramHeaders+i)->p_offset,
              (m_pProgramHeaders+i)->p_offset+(m_pProgramHeaders+i)->p_filesz,
              (m_pProgramHeaders+i)->p_filesz);

      switch (pProgamHeader->p_type)
	{
	case ELF_PT_INTERP:
          VALIDATE(!bHasHadLoadable,
		   "Program header INTERP appearing after a program header LOADABLE.");
          VALIDATE(!bHasHadInterp,
		   "Two program header INTERP appearing.");
          bHasHadInterp = true;
          break;
	case ELF_PT_PHDR:
          VALIDATE(!bHasHadLoadable,
		   "Program header PHDR appearing after a program header LOADABLE.");
          VALIDATE(!bHasHadPhdr,
		   "Two program header INTERP appearing.");
          bHasHadPhdr = true;
          break;
	case ELF_PT_LOAD:
          bHasHadLoadable = true;
          break;
	}
      Elf32_ProgramBlock* pProgramBlock = new Elf32_ProgramBlock;
      
      DataPtrSrc* pData = NULL;
      
      TRY_CATCH_EMIT_ERROR_THROW_GENERIC(pData = GetFileData (pProgamHeader->p_offset,pProgamHeader->p_filesz),
                                         "Failure getting program body.");
      
      delete pData;
      
      pProgramBlock->InitData (pProgamHeader/*,pData*/);
      
      m_vectProgramBlocks.push_back(pProgramBlock);
      
      pProgramBlock->ValidateHeader();
      
      m_fileMap.AddAreaType(pProgamHeader->p_offset,pProgamHeader->p_filesz,
			    pProgramBlock);
    }
  
  // first we need to read the strings section if there is one (so we can get the names of the
  // other sections)

  Elf32_SectionStringTable* pNamesStringTable = NULL;
  
  if (m_elfHeader.e_shstrndx)
    {
      Elf32_SectionHeader* pStringSection = m_pSectionHeaders+m_elfHeader.e_shstrndx;
      
      VALIDATE(pStringSection->sh_type == ELF_SHT_STRTAB,
               "ELF header section string index, section is not string table.");
      VALIDATE(pStringSection->sh_flags == 0,
               "ELF header section string index, section flags not zero.");
      
      pNamesStringTable = new Elf32_SectionStringTable();

      DataPtrSrc* pData = NULL;
      
      TRY_CATCH_EMIT_ERROR_THROW_GENERIC(pData = GetFileData (pStringSection->sh_offset,pStringSection->sh_size),
                                         "Failure getting string section body.");
      
      pNamesStringTable->InitData(pStringSection,pData);
      pNamesStringTable->VerifyData();
    }
  
  Elf32_SectionHeader* pSectionHeader = m_pSectionHeaders;

  for (i = 0; i < m_elfHeader.e_shnum; i++, pSectionHeader++)
    {
      printf ("sh %2d: 0x%08x - 0x%08x (0x%08x) %s\n",i,
              (m_pSectionHeaders+i)->sh_offset,
              (m_pSectionHeaders+i)->sh_offset+(m_pSectionHeaders+i)->sh_size,
              (m_pSectionHeaders+i)->sh_size,
              pNamesStringTable ? pNamesStringTable->GetString(pSectionHeader->sh_name) : "");

      if (i == 0)
        {
          VALIDATE(iszero(pSectionHeader,sizeof(Elf32_SectionHeader)),
                   "Section header 0 not filled with 0s.");
        }
      
      Elf32_Section* pSection = NULL;
      
      if (i && i == m_elfHeader.e_shstrndx)
        {
          pSection = pNamesStringTable;
        }
      else
        {
          switch (pSectionHeader->sh_type)
            {
            case ELF_SHT_STRTAB:
              pSection = new Elf32_SectionStringTable;
              break;
            case ELF_SHT_DYNAMIC:
              pSection = new Elf32_SectionDynamic;
              break;
            case ELF_SHT_HASH:
              pSection = new Elf32_SectionHashTable;
              break;
            case ELF_SHT_REL:
            case ELF_SHT_RELA:
              pSection = new Elf32_SectionRelocation;
              break;
            case ELF_SHT_SYMTAB:
            case ELF_SHT_DYNSYM:
              pSection = new Elf32_SectionSymbolTable;
              break;
            case ELF_SHT_NOTE:
              pSection = new Elf32_SectionNote;
              break;
            default:
              pSection = new Elf32_Section;
              break;
            }
          DataPtrSrc* pData = NULL;
	  
          if (pSectionHeader->sh_type != ELF_SHT_NOBITS)
            TRY_CATCH_EMIT_ERROR_THROW_GENERIC(pData = GetFileData (pSectionHeader->sh_offset,
                                                                    pSectionHeader->sh_size),
                                               "Failure getting section body.");  
          
          pSection->InitData(pSectionHeader,pData);
        }
      m_vectSections.push_back(pSection);
      
      if (pNamesStringTable) // no need to get an empty name
        pSection->m_strName = pNamesStringTable->GetString(pSectionHeader->sh_name);
      else
	WARNING1(pSectionHeader->sh_name == 0,
		 "Section header %d has name index, but there is no section name string table.",i);
      
      // lets wee what program blocks we're part of
      
      FOR_EACH_ITER(VectProgramBlocks,m_vectProgramBlocks,progblock)
	{
	  if ((*progblock)->IsSectionInside(pSection))
	    (*progblock)->AddSection(pSection);
	}
      
      pSection->ValidateHeader();
      
      m_fileMap.AddAreaType(pSectionHeader->sh_offset,
			    pSectionHeader->sh_type == ELF_SHT_NOBITS ? 
			    0 : pSectionHeader->sh_size,
			    pSection);
    }

  // We should also be able to handle the case where the section
  // headers have been omitted from an executable, and extract all
  // necessary information from the dynamic block of the program
  // headers and if there are section headers and an dynamic block, we
  // should make sure they are saying the same things

  // so first get the dynamic block

  // setup links between sections
  
  pSectionHeader = m_pSectionHeaders;
  
  for (i = 0; i < m_elfHeader.e_shnum; i++, pSectionHeader++)
    {
      Elf32_Section* pSection = m_vectSections[i];

      switch (pSectionHeader->sh_type)
        {
        case ELF_SHT_DYNAMIC:
	  assert (dynamic_cast<Elf32_SectionDynamic*>(pSection));
	  ((Elf32_SectionDynamic*) pSection)->SetStringTable(GetStringSection(pSectionHeader->sh_link));
	  break;
        case ELF_SHT_HASH:
	  assert (dynamic_cast<Elf32_SectionHashTable*>(pSection));
	  ((Elf32_SectionHashTable*) pSection)->SetSymbolTable(GetSymbolSection(pSectionHeader->sh_link));
	  break;
        case ELF_SHT_REL:
        case ELF_SHT_RELA:
	  assert (dynamic_cast<Elf32_SectionRelocation*>(pSection));
	  ((Elf32_SectionRelocation*) pSection)->SetSymbolTable(GetSymbolSection(pSectionHeader->sh_link));
	  if (pSectionHeader->sh_info)
	    ((Elf32_SectionRelocation*) pSection)->SetRelocSection(GetSection(pSectionHeader->sh_info));
	  else
	    EMIT_GENERIC_ERROR("Relocation section does not specify what section is affected.");
	  break;
        case ELF_SHT_SYMTAB:
        case ELF_SHT_DYNSYM:
	  assert (dynamic_cast<Elf32_SectionSymbolTable*>(pSection));
	  ((Elf32_SectionSymbolTable*) pSection)->SetStringTable(GetStringSection(pSectionHeader->sh_link));
          break;
        default:
	  if (pSectionHeader->sh_type <= ELF_SHT_HI_SPEC_DEF)
            {
	      WARNING(pSectionHeader->sh_link == ELF_SHN_UNDEF,
                    "Section header link that is to be undefined (0) was not.");
            }
	  break;
        }
    }
  
    for (i = 0; i < m_elfHeader.e_shnum; i++, pSectionHeader++)
    {
      Elf32_Section* pSection = m_vectSections[i];
      
      if (pSection)
	pSection->VerifyData();
    }

    e_entryInit = 0;			
    e_entryFini = 0;			
    
    // TODO: ok, so find any dynamic block, and get / validate any
    // information this includes to use informtion to create 'shadow'
    // sections if we have the section's data, but not the actual
    // section header, so that we can read the information with the
    // section system (like above)
    
    for (i = 0; i < m_elfHeader.e_shnum; i++, pSectionHeader++)
      {
	Elf32_Section* pSection = m_vectSections[i];
	
	Elf32_SectionDynamic* pSectionDynamic; 
	
	if ((pSectionDynamic = dynamic_cast<Elf32_SectionDynamic*>(pSection)))
	{
	  // init and fini functions?
	  
	  pSectionDynamic->GetTag(ELF_DT_INIT,e_entryInit);
	  pSectionDynamic->GetTag(ELF_DT_FINI,e_entryFini);
	}
      }
}

//////////////////////////////////////////////////////////////////////

Elf32_Section* ELF_File::GetSection(Elf32_Word sect)
{
  VALIDATE(sect,
	   "Section header link null.");
  
  VALIDATE(sect < m_elfHeader.e_shnum,
	   "Section header link outside range.");
  
  assert (m_vectSections.size() > sect); // something bad has happened before, but we coped with that, but the section was not defined and so we get problems here
  
  VALIDATE(m_vectSections[sect],
	   "Section linked to by other section undefined");
  
  return m_vectSections[sect];
}

Elf32_SectionStringTable* ELF_File::GetStringSection(Elf32_Word sect)
{
  Elf32_Section* pSection = GetSection(sect);
  
  VALIDATE(pSection->m_header.sh_type == ELF_SHT_STRTAB,
	   "Section linked to wanted string section, but section was of other type.");
  
  assert (dynamic_cast<Elf32_SectionStringTable*>(pSection));
  
  return (Elf32_SectionStringTable*) pSection;
}

Elf32_SectionSymbolTable* ELF_File::GetSymbolSection(Elf32_Word sect)
{
  Elf32_Section* pSection = GetSection(sect);
  
  VALIDATE(pSection->m_header.sh_type == ELF_SHT_SYMTAB ||
	   pSection->m_header.sh_type == ELF_SHT_DYNSYM,
	   "Section linked to wanted string section, but section was of other type.");
  
  assert (dynamic_cast<Elf32_SectionSymbolTable*>(pSection));
  
  return (Elf32_SectionSymbolTable*) pSection;
}

//////////////////////////////////////////////////////////////////////

void ELF_File::PostReadFile ()
{
  ExecutableImage::PostReadFile();
  
  // so find out what are imports and exports
  
  // all symbols in the symbol table (at least the minimal dynamic
  // one) and be considered imports or exports. My interpretation of
  // the spec (can anyone tell me if this is correct), is that a
  // symbol with a SHN_UNDEF (0) or SHN_COMMON (0xfff2) section-field
  // is an import (i.e. not defined by this executable), and a symbol
  // otherwise is an export (possibly weak, that is: overridable)
  
  // and we must also take care of short-circuiting the plt so the
  // disassembler don't try to mess with that one
  
  // plan: read the symbol table. Find all exports.  Exports should
  // not be referenced by the relocations?

  // any one with SHN_UNDED or SHN_COMMON: import. But this does not
  // make it used.  an imported symbol is 'used' (or could at least be
  // used) when it is referenced by the relocation table. If the
  // relocation affect the GOT, then we should investigate further. If
  // it is an function, referenced by the plt, then the got should
  // have the address of the PUSH-instruction in the plt. Then the
  // information in this plt should reference back the got (jump
  // indirect address), and reloc-entry-offset
  // PUSH-instruction. Otherwise (i.e. not a plt-ed function) the got
  // entry should be zero? (as it hasn't been initialized..)

  // finally we should make sure that all plt entries have been
  // treated, so that none are left without possibility of relocation
  // (one could of course imagine two plt entries that point to the
  // same got entry, but I can think of no useful reason for
  // this. Because when unresolved, both will jump (*got) to the one
  // push of the plt that the got point to, and then push that one's
  // reloc offset, so they will indirectly point to the same place
    

  // here we go...

  // first simply inspect the symbol table, and locate any exports
  
  ReadImportExportSymbols ();
  
  
}


void ELF_File::ReadImportExportSymbols ()
{
    
  // go through the symbol table, anything not SHN_UNKNOWN or SHN_COMMON is an
  // export
  
  // first find outselves an symbol table
  // TODO: make this one bullet-proof, for now, just find the .dynsym section
  
  Elf32_Section* pSection = FindSection (".dynsym",ELF_SHT_DYNSYM);
  Elf32_SectionSymbolTable* pSymbols = 
    dynamic_cast<Elf32_SectionSymbolTable*>(pSection);
  
  if (!pSymbols)
    {
      printf ("No symbol table found. (besides, here is an TODO in the source...).\n");
      return;
    }
  
  FOR_EACH_ITER(VectSymbolItems,pSymbols->m_vectSymbolItems,symbol)
    {
      if ((*symbol)->m_wSection == ELF_SHN_UNDEF ||
          (*symbol)->m_wSection == ELF_SHN_COMMON)
        {
          // it's an import 
          
          printf ("Import: %s\n",(*symbol)->m_strName.c_str());
	  
          // 
          







        }
      else
        {
          // printf ("Export: (%08x) %s\n",(*symbol)->m_dwValue,(*symbol)->m_strName.c_str());

          // it's an export
	  
          // make sure the thing is inside the appointed section
          // (unless it's SHN_ABS, in which case it should be outside
          // the image?)

          if (ELF32_ST_TYPE((*symbol)->m_bInfo) == ELF_STT_SECTION ||
              ELF32_ST_TYPE((*symbol)->m_bInfo) == ELF_STT_FILE)
            {
              // we're not interested in section or file markers (for now..)

              continue;	      
            }

          if ((*symbol)->m_wSection == ELF_SHN_ABS)
            {
              // TODO: we should at least check that the address (and
              // size) is completely outside all sections one might
              // also consider including page-padding...
              
            }
          else
            {
              // first make sure it is an existing section number
	      
              WARNING1((*symbol)->m_wSection < m_elfHeader.e_shnum,
                       "Exported entry section number %d outside valid range.",(*symbol)->m_wSection);
              
              WARNING3(GetSection((*symbol)->m_wSection)->IsAddrInside((*symbol)->m_dwValue,(*symbol)->m_dwSize),
                       "Exported entry defined to be at address %d with size %d not inside specified section %d.",
                       (*symbol)->m_dwValue,(*symbol)->m_dwSize,(*symbol)->m_wSection);
            }
          
          // TODO: add logic for guessing and validating between
          // function and object type of export
	  
          byte bType = 0;
          
          switch (ELF32_ST_TYPE((*symbol)->m_bInfo))
            {
            case ELF_STT_OBJECT:
              bType = EXPORT_OBJECT_GIVEN;
	      
              // TODO: the map of the contents of the image (which
              // does not exist yet), should have the corresponding
              // bytes marked as an object, and a link to the export
              // so it can be relocated (and isn't forgotten)

              break;
            case ELF_STT_FUNC:
              // aha, a function
              // and an export
              // disasm.AddCallTarget((*symbol)->m_dwValue);
              bType = EXPORT_FUNC_GIVEN;
              break;
            case ELF_STT_NOTYPE:
              {
                // so, here we should guess depending on if it is in
                // the executable part
                
                Elf32_Section* pInSection = GetSection((*symbol)->m_wSection);
                
                VALIDATE((pInSection->m_header.sh_flags & ELF_SHF_ALLOC),
                         "Exported object in section that is not allocated in image.");
                
                WARNING((pInSection->m_header.sh_flags & (ELF_SHF_EXECINSTR | ELF_SHF_WRITE)),
                        "Exported object in section that is neither executable nor writeble "
                        "(so neither function nor writeable variable, must be read only variable).");
                
                bType = (pInSection->m_header.sh_flags & ELF_SHF_EXECINSTR) ? 
		  EXPORT_FUNC_GUESS : EXPORT_OBJECT_GUESS;
                break;
              }
            default:
              EMIT_GENERIC_ERROR1("Export with (for now) unsupported type %d. (FIXME)",ELF32_ST_TYPE((*symbol)->m_bInfo));
            }
          
          ExportedObject* pObject = 
	    new ExportedObject ((*symbol)->m_dwValue,(*symbol)->m_dwSize);
          pObject->m_vectNames.push_back((*symbol)->m_strName);
          pObject->m_bType = bType;
	  
          m_vectExportedObjects.push_back(pObject);
	  
          /*
            #define ELF_STT_NOTYPE		0		// NoType
            #define ELF_STT_OBJECT		1		// Object
            #define ELF_STT_FUNC		2		// Function
            #define ELF_STT_SECTION		3		// Section
            #define ELF_STT_FILE		4		// File
          */        
        }
    }
}

//////////////////////////////////////////////////////////////////////

void ELF_File::GetExportedCallTargets (DisAssembler& disasm)
{
  FOR_EACH_ITER(VectExportedObjects,m_vectExportedObjects,item)
    {
      if ((*item)->m_bType & (EXPORT_FUNC_MASK))
        {
          // we (at least) suspect it is a function, so add it for disassembly
	  
          disasm.AddCallTarget(*item);
        }
    }
}

//////////////////////////////////////////////////////////////////////

bool ELF_File::IsType (int file)
{
  assert (sizeof (Elf32_Ident) == EI_NIDENT);
  
  _lseek (file,0,SEEK_SET);
  
  unsigned long magic;
  
  if (read (file,&magic,sizeof (magic)) < (int) sizeof (magic))
    {
      // not even magic
      return false;
    }	
  
  if (magic != ELF_ID_MAGIC)
    {
      return false;
    }
  
  // good enough to assume ELF file
  // (ReadFile will be very strict anyway)
  
  return true;
}

//////////////////////////////////////////////////////////////////////

bool ELF_File::ProcessFile (UrlFolder* pFolder)
{
  // disassemble the file
  
  if (m_elfHeader.e_entry)
    m_pExecutable->m_pEntryFunction = new ExportedObject(m_elfHeader.e_entry);

  if (e_entryInit)
    m_pExecutable->m_pInitFunction = new ExportedObject(e_entryInit);
  if (e_entryFini)
    m_pExecutable->m_pFiniFunction = new ExportedObject(e_entryFini);
  if (!DisAssemble (pFolder))
    return false;
  
  return true;
}

//////////////////////////////////////////////////////////////////////

DataPtr*  ELF_File::GetDisAsmDataPtr (dword dwVirtualAddress)
{
  assert (m_pFileData);
  
  // ok, so any code we want to execute must come from a program
  // header, with execute rights, and LOADABLE
  
  // or, it could be in an section with executable code,

  // search the program headers

  FOR_EACH_ITER(VectProgramBlocks,m_vectProgramBlocks,progblock)
    {
      if ((*progblock)->m_header.p_type == ELF_PT_LOAD)
        {
          // are we inside this range?
	  
          if ((*progblock)->m_header.p_vaddr <= dwVirtualAddress &&
              (*progblock)->m_header.p_vaddr + (*progblock)->m_header.p_memsz > dwVirtualAddress)
            {
              if ((*progblock)->m_header.p_vaddr + (*progblock)->m_header.p_filesz <= dwVirtualAddress)
                throw GENERIC_ERROR("Attempt to get executable code from zero-extended program block. "
                                    "Valid?, but not supported by this program.");
	      
              if (!((*progblock)->m_header.p_flags & ELF_PT_X))
		throw GENERIC_ERROR("Attempt to get executable code from program block missing execution rights.");

	      
	      return new DataPtrSrcRVA(m_pFileData + (*progblock)->m_header.p_offset,
				       (*progblock)->m_header.p_vaddr,
				       (*progblock)->m_header.p_filesz,
				       dwVirtualAddress);
            }
        }
    }
  
  throw GENERIC_ERROR("Attempt to get executable code from outside loaded image.");
}

//////////////////////////////////////////////////////////////////////

#include "HtmlPage.h"
#include "HtmlTags.h"

void ELF_File::HTMLFileInfo (UrlFolder* pParent)
{
  UrlFile* pFilePE = pParent->AddFile ("elf_file.html",
				       "ELF file - executable");
  
  HtmlPage page(pFilePE);
  
  page.Header();

  page << h1("ELF Ident");
  DumpStructElf32_Ident(m_elfIdent,page);
  page << h1("Header");
  DumpStructElf32_Header(m_elfHeader,page);

  page << h1("Program Headers");
  page.Begin(table());
  page.Begin(tr());
  page << td("#");
  DumpStructTableHeaderElf32_ProgramHeader(page);
  page.End();

  uint i;

  for (i = 0; i < m_elfHeader.e_phnum; i++)
    {
      page.Begin(tr());
      page << td(strsprintf("%d",i));
      DumpStructTableElf32_ProgramHeader(*(m_pProgramHeaders+i),page);
      page.End();
    }     

  page.End();

  UrlFolder* pFolderSect = pParent->AddFolder ("section");

  page << h1("Section Headers");
  page.Begin(table());
  page.Begin(tr());
  page << td("#");
  DumpStructTableHeaderElf32_SectionHeader(page);
  page.End();

  for (i = 0; i < m_elfHeader.e_shnum; i++)
    {
      Elf32_Section* pSection = m_vectSections[i];

      page.Begin(tr());

      if (pSection)
	{
	  UrlFile* pFileSect = pSection->HTMLFileInfo(i,pFolderSect);
	  
	  page << (td() << a(strsprintf("%d",i)).href(pFileSect));
	}
      else
        page << td(strsprintf("%d",i));
      
      DumpStructTableElf32_SectionHeader(*(m_pSectionHeaders+i),page,
                                         pSection ? 
					 text(pSection->m_strName) : NULL);
      page.End();
    }
  
}






