/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ELF_File_Write.cpp: implementation of the ELF_File class.
//
//////////////////////////////////////////////////////////////////////

#include "ELF_File.h"
#include "Elf32_Section.h"
#include "Elf32_ProgramBlock.h"
#include "Error.h"
#include "ErrorTracker.h"
#include "gen/ELF_File_DumpStructs.h"
#include "Executable.h"
#include "ExportedObject.h"
#include "DataPtrSrc.h"
#include "Misc.h"

#include "CodeSection.h"
#include "DataDestRelocPtr.h"
#include "RelocatableDest.h"
#include "WriteProgramHeader.h"

#include "ExecutableWriter.h"

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _lseek lseek
#define _tell(fildes) (lseek(fildes,0,SEEK_CUR))
#endif
#include <stdio.h>

//////////////////////////////////////////////////////////////////////

ExecutableImage* ELF_File::GetWriteImage(const string& strFileName)
{
  ELF_File* pWriteELF = new ELF_File(strFileName);
  
  memset (&pWriteELF->m_elfIdent,0,sizeof (pWriteELF->m_elfIdent));
  memset (&pWriteELF->m_elfHeader,0,sizeof (pWriteELF->m_elfHeader));
  
  pWriteELF->m_elfIdent.dw_magic     = ELF_ID_MAGIC;
  pWriteELF->m_elfIdent.b_class      = ELF_ID_CLASS32;
  pWriteELF->m_elfIdent.b_data       = ELF_ID_DATA2LSB;
  pWriteELF->m_elfIdent.b_version    = ELF_EV_CURRENT;
    
  assert (m_elfHeader.e_type == ELF_ET_EXEC); // lighten this when capabilities expand

  pWriteELF->m_elfHeader.e_type      = m_elfHeader.e_type;
  pWriteELF->m_elfHeader.e_machine   = ELF_EM_386;
  pWriteELF->m_elfHeader.e_version   = ELF_EV_CURRENT;

  assert (m_pExecutable);
  
  pWriteELF->m_pExecutable = m_pExecutable;

  return pWriteELF;
}


//////////////////////////////////////////////////////////////////////
// This is for testing purposes.
// TO BE REMOVED
//
// It has the same role as GetWriteImage
// but we think up all the things in here insteas

ExecutableImage* CreateSimpleElfExecutable()
{
  ELF_File* pImage = new ELF_File("created_elf");

  memset (&pImage->m_elfIdent,0,sizeof (pImage->m_elfIdent));
  memset (&pImage->m_elfHeader,0,sizeof (pImage->m_elfHeader));

  pImage->m_elfIdent.dw_magic     = ELF_ID_MAGIC;
  pImage->m_elfIdent.b_class      = ELF_ID_CLASS32;
  pImage->m_elfIdent.b_data       = ELF_ID_DATA2LSB;
  pImage->m_elfIdent.b_version    = ELF_EV_CURRENT;

  pImage->m_elfHeader.e_type      = ELF_ET_EXEC;
  pImage->m_elfHeader.e_machine   = ELF_EM_386;
  pImage->m_elfHeader.e_version   = ELF_EV_CURRENT;

  Elf32_SectionRawData* pInterpSection = 
    new Elf32_SectionRawData(".interp",ELF_SHT_PROGBITS,ELF_SHF_ALLOC);
  pInterpSection->SetDataStr("/lib/ld-linux.so.2");

  pImage->m_vectSections.push_back(pInterpSection);

  Elf32_SectionNote* pNoteABI = new Elf32_SectionNote(".note.ABI-tag",
						      ELF_SHF_ALLOC);
  dword dwABIdata[] = { 0,0x2,0,0x1e };
  pNoteABI->AddNote("GNU",1,dwABIdata,sizeof (dwABIdata));
  pNoteABI->m_header.sh_flags = ELF_SHF_ALLOC;
  
  pImage->m_vectSections.push_back(pNoteABI);
  
  //  pImage->m_elfHeader

/*
  Elf32_Half e_type;			// File type
  Elf32_Half e_machine;			// Target machine
  Elf32_Word e_version;			// ELF version
  Elf32_Addr e_entry;			// Entry point
  Elf32_Off  e_phoff;			// PHT file offset
  Elf32_Off  e_shoff;			// SHT file offset
  Elf32_Word e_flags;			// Machine flags
  Elf32_Half e_ehsize;			// Header size
  Elf32_Half e_phentsize;		// PHT entry size
  Elf32_Half e_phnum;			// PHT entries
  Elf32_Half e_shentsize;		// SHT entry size
  Elf32_Half e_shnum;			// SHT entries
  Elf32_Half e_shstrndx;		// SHT strings
*/




  return pImage;
}

//////////////////////////////////////////////////////////////////////

#define count_of(array) (sizeof(array)/sizeof(array[0]))

struct FileSectionsLayoutInfo
{
  word wSection;
  byte wAlignment;      // 0 means default
  // byte bPartOfHeader;
};

const FileSectionsLayoutInfo infoELFFileSections[] =
{
  { 0, 4 }, // default for unlisted entries
  
  { REL_ELF_IDENT,  4 },
  { REL_ELF_HEADER, 4 },
  
  { REL_ELF_PRGM_HEADER_PHDR,     4 },   
  { REL_ELF_PRGM_HEADER_INTERP,   4 },   
  { REL_ELF_PRGM_HEADER_RX_LOAD1, 4 },     
  { REL_ELF_PRGM_HEADER_RX_LOAD2, 4 },    
  { REL_ELF_PRGM_HEADER_RW_LOAD,  4 },    
  { REL_ELF_PRGM_HEADER_DYNAMIC,  4 },   
  { REL_ELF_PRGM_HEADER_NOTE,     4 },   
  /*  
  // { REL_ELF_PRGM_HEADER_PHDR,          4,4 },   
  { REL_ELF_PRGM_HEADER_INTERP,   1 },   
  { REL_ELF_PRGM_HEADER_RX_LOAD1, 4 },     
  { REL_ELF_PRGM_HEADER_RX_LOAD2, 4 },    
  { REL_ELF_PRGM_HEADER_RW_LOAD,  4 },    
  { REL_ELF_PRGM_HEADER_DYNAMIC,  4 },   
  { REL_ELF_PRGM_HEADER_NOTE,     4 },   
  */

  { REL_ELF_INIT_SECTION, 4 },  
  { REL_ELF_FINI_SECTION, 4 },  
  { REL_ELF_PLT_SECTION,  4 },  
  { REL_ELF_TEXT_SECTION, 4 },  
 
                 
/*
    { RELATIVE_PE_OPT_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
        { RELATIVE_CODE_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
        { RELATIVE_DATA_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
        { RELATIVE_RSRC_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
        { RELATIVE_RELOC_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
*/	
  //	{ REL_ELF_TEXT_SECTION, 0,0 },
//	{ RELATIVE_DATA_SECTION, 0,0 },			 
//	{ RELATIVE_RSRC_SECTION, 0,0 },			 
//	{ RELATIVE_RELOC_SECTION, 0,0 },
};
/*
#define REL_ELF_PRGM_HEADER_PHDR        (0x0001 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_INTERP      (0x0002 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_RX_LOAD1    (0x0003 | REL_TYPE_ELF_PRGM_HEADER) // seldom-access (mainly on init)
#define REL_ELF_PRGM_HEADER_RX_LOAD2    (0x0004 | REL_TYPE_ELF_PRGM_HEADER) // .text .plt (and any in LOAD1 if that's omitted)
#define REL_ELF_PRGM_HEADER_RW_LOAD     (0x0005 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_DYNAMIC     (0x0006 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_NOTE        (0x0007 | REL_TYPE_ELF_PRGM_HEADER)
*/
//////////////////////////////////////////////////////////////////////

void ELF_File::DoFileLayout (word wRelocSection,const FileSectionsLayoutInfo& layoutInfo)
{
  // put the section at the next possible position in the file, which matches all requests
/*
  if (!m_destSections.HasSection(wRelocSection))
    return;
*/
  dword dwFileAlign = 1;
  dword dwImageAlign = 1;
  /*
    if (bForceNewImagePage)
    {
    dwFileAlign = dwImageAlign = 0x1000;
    }
    else */ if (layoutInfo.wAlignment)
      {
        dwFileAlign  = layoutInfo.wAlignment;
        dwImageAlign = layoutInfo.wAlignment;
      }
  
  assert (powerof2(dwFileAlign));
  assert (powerof2(dwImageAlign));
  
//  printf ("                        %08x  %08x\n",dwFileAlign,dwImageAlign);
//  printf ("%08x  %08x\n",m_fileLayout.dwFileNextFree,m_fileLayout.dwImageNextFree);
  
  assert ((m_fileLayout.dwFileNextFree % 0x1000) == 
	  (m_fileLayout.dwImageNextFree % 0x1000));
  
  m_fileLayout.dwFileNextFree  = next_multiple(m_fileLayout.dwFileNextFree , 
					       dwFileAlign);
  m_fileLayout.dwImageNextFree = next_multiple(m_fileLayout.dwImageNextFree,
					       dwImageAlign);
  
//  printf ("%08x  %08x\n",m_fileLayout.dwFileNextFree,m_fileLayout.dwImageNextFree);

  // we should always be congruent with image and file positions
  
  // so now we're aligned
  
  m_destSections.RelocateSection(wRelocSection | RELATIVE_FILE ,
				 m_fileLayout.dwFileNextFree);
  m_destSections.RelocateSection(wRelocSection | RELATIVE_IMAGE,
				 m_fileLayout.dwImageBase + 
				 m_fileLayout.dwImageNextFree);
  
  m_fileLayout.dwFileNextFree  += 
    m_destSections.GetSectionFileLength (wRelocSection);
  m_fileLayout.dwImageNextFree += 
    m_destSections.GetSectionImageLength(wRelocSection);
  
//  printf ("%08x  %08x\n",m_fileLayout.dwFileNextFree,m_fileLayout.dwImageNextFree);
  
}
/*
void ELF_File::DoFileLayout (word wRelocSectionStart)
{
  assert (infoELFFileSections[count_of(infoELFFileSections)-1].wSection >= wRelocSectionStart);

  uint i = 0;

  while (infoELFFileSections[i].wSection < wRelocSectionStart)
    i++;

  DoFileLayout(wRelocSectionStart,infoELFFileSections[i]);
}
*/
void ELF_File::DoFileLayout (word wRelocSectionStart,word wRelocSectionEnd)
{
  assert (wRelocSectionStart <= wRelocSectionEnd);
  
  MapSectionRelocatableDest::iterator begin = 
    m_destSections.mapSectionRelocatableDest.lower_bound(wRelocSectionStart);
  MapSectionRelocatableDest::iterator end = 
    m_destSections.mapSectionRelocatableDest.upper_bound(wRelocSectionEnd);
  
  uint i = 1;
  
  for (MapSectionRelocatableDest::iterator iter = begin; iter != end; ++iter)
    {
      while (i < count_of(infoELFFileSections) &&
             infoELFFileSections[i].wSection < iter->first)
        i++;
      
      DoFileLayout(iter->first,
		   (i < count_of(infoELFFileSections) &&
		    infoELFFileSections[i].wSection == 
		    iter->first) ? 
		   infoELFFileSections[i] : infoELFFileSections[0]);
    }
}

void ELF_File::CutFileAfter (word wRelocSectionKeep)
{
  // cut the file layout info after the specified section to allow for another try at positioning
  // the segments of the file
  
  assert (m_destSections.HasSection(wRelocSectionKeep));
  
  m_fileLayout.dwFileNextFree  = 
    m_destSections.GetSectionFileEnd (wRelocSectionKeep);
  m_fileLayout.dwImageNextFree = 
    m_destSections.GetSectionImageEnd(wRelocSectionKeep);
}

//////////////////////////////////////////////////////////////////////

Elf32_Section* ELF_File::FindSection (const char* pName,Elf32_Word type)
{
  FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
    {
      if ((*sect)->m_header.sh_type == type &&
          (*sect)->m_strName.compare(pName) == 0)
        {
          return *sect;
        }
    }
  return NULL;
}

#include <algorithm>

uint ELF_File::GetSectionIndex (Elf32_Section* pSection)
{
  VectElf32_Sections::iterator f = find(m_vectSections.begin(),
					m_vectSections.end(),pSection);
  
  if (f == m_vectSections.end())
    return 0;
  
  return distance(m_vectSections.begin(),f);
}

Elf32_ProgramBlock* ELF_File::AddProgramBlockLoadSectors(word wRelocMin,
							 word wRelocMax,
							 word wRelocProgramHeader)
{
  {
    FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
      {
        if ((*sect)->m_wRelocSect >= wRelocMin && 
	    (*sect)->m_wRelocSect <= wRelocMax)
          goto need_load_block;
      }
  }
  
  return NULL; // no sections in range
  
 need_load_block:
  
  Elf32_ProgramBlock* pLoad = new Elf32_ProgramBlock(ELF_PT_LOAD,ELF_PT_R);
  
  // MapSectionRelocatableDest::iterator end = m_destSections.mapSectionRelocatableDest.upper_bound(wRelocMax);
  
  FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
    {
      if ((*sect)->m_wRelocSect >= wRelocMin && 
	  (*sect)->m_wRelocSect <= wRelocMax)
        pLoad->AddLoadSection((*sect));
    }
  
  m_vectProgramBlocks.push_back(pLoad);
  
  m_destSections.AddRelocatableDest(wRelocProgramHeader,pLoad->WriteHeader());
  
  return pLoad;
}

#include <algorithm>

class SectionRelocNoCompare
{
public:
  bool operator()(Elf32_Section* lhs,Elf32_Section* rhs) 
  { 
    return lhs->m_wRelocSect < rhs->m_wRelocSect; 
  }
};

bool ELF_File::PrepareWrite()
{
  // now an elf file has an dual view of it's contents: sections and
  // program blocks in order not to confuse ourselves too much, we'll
  // keep to the sections.  actually, the sections contain the actual
  // data, and the program blocks are just hints to the dynamic linker
  // where to find the necessary information (so that the sectiion
  // headers actually can be omitted) so well do everything with
  // sections, and emit the section headers if we want to (they are at
  // the end enyway).  and so we just create any program headers
  // needed.  what of course tohe program headers require is that we
  // eject eny sections to be allocated in the image in such an order
  // that the memory access rights are the same for an contigous
  // block, but that's fairly easy

  // first we must collect and generate all sections needed then we
  // can fix any sections headers and program headers needed

  // so that we can keep some order on things (perhaps it could be
  // better with some more thinking and work, but this is what we get
  // for now) we have some fixed RELATIVE_xxx markers for relocation
  // of some known and used-by-this-program sections, and some unused
  // indices for whatever sections may hang around from the executable
  // (although it is unclear whether those should be copied at all if
  // they are unreferenced (as we may have (unintentionally) destroyed
  // the addresses)
  //
  // These known and unknown sections are ordered by the rights they
  // require which are (r, rx, r and rw) (the second r as .rodata seem
  // to usually be placed in that order)
  //
  // so they are like this:
  //
  // r    .hash  .dynsym  .dynstr  .rel.dyn  .rel.plt   and UNKNOWNS
  // rx   .init  .fini  .plt  .text  
  // r    .rodata
  // rw   .data  .dynamic  .got  .bss(last) UNKNOWNS
  // ---- stuff not to be in memory image
  //
  // There should be no problem putting the above stuff into one
  // program block (which would be loaded into continous memory with
  // rx rights, as our program will detect (and choke) on attempts to
  // execute code in non-x segments anyway)
  //
  // Then there is the problem that never or seldom-used parts should
  // if possible be located in separate loadable blocks so that they
  // can be paged out when no longer used (or infrequently used) This
  // separation will cost extra memory segements if implemented
  // cluelessly because of padding. Clueless = trimming executable
  // size. The cost for this memory cutting is to have some extra
  // zeros in the executable so that the addresses can be
  // aligned. (and it is really no catasrophy if we loose one page of
  // memory to reduce paging, because that page will eventually end up
  // in the disk cache, where it make muss less harm than travelling
  // back and forth to the cache together with other stuff).
  // 
  // of course we want an command-line switch for this behaviour
  //
  // --aggressively-avoid-paging (costing some cache perhaps)
  // --avoid-memory-padding      (costing some file size, that is to minimize memory footprint)
  // --minimize-file-size        (not liking the above swithes :-) )
  //
  // The sections that could benefit from this paging-out strategy are
  // .hash .dynsym .dynstr .rel.dyn .rel.plt which are never used
  // after execution, and anything in .init and .text which is also
  // only used at two specific times while executing started except
  // when doing lazy binding, which happens at most once for every
  // unbound symbol in the plt the UNKNOWNs should if they still exist
  // (see above) probably go into the executable page because we do
  // not know how they behave. (but if it saves a page we could as
  // well put them here)
  //
  // the plt (which is usually accessed from a lot of different
  // places) should be placed along with other code of the same kind
  // (some user defined miscellaneous functions that do simple tasks
  // that are used from all possible places in the code), but not in
  // the first read-seldom part as it is accessed (a lot), however,
  // the init and fini sections should go into that read-seldom
  // segment
  //
  // This lift another question into focus: the code and other data
  // should depend heavily on the alignment because of paging, that
  // is, stuff that is closely related (accessed) should be kept in
  // the same pages (and as few as possible) So it cannot be
  // responsibly created before we know where it's segment may start
  // (and this create some difficulties, BUT this program is not to be
  // your everyday link editor, it is to be the one that care about
  // the little details (and if it do not: it's a bug, or at least
  // mis-feature): you may say paging is not important: I say: stuff
  // is suppose to run on my 8- meg 486), and it has relatively more
  // hard drive than real memory, so file size is of less importance,
  // paging is of monumental)
  //
  // Code and data reordering for avoiding paging does not need a
  // switch, because the only switch would be: --maximize-swapping
  // hmm, one could of course, (if it turn out that the algoritm is
  // slow), consider an: --try-to-be-smart-about-swapping-but-quick
  // switch but that could be linked to an overall:
  // --speed-for-efficiency switch
  //
  // splitting the rx-segment into two is only useful when the
  // read-seldom part is a large part of that section, and requires an
  // additional program header and restructuring of the file so it has
  // to be done after an first attempt without it. For now, we do not
  // bother...
  //
  
  // we must create stuff in order...
  // so lets make some program headers
  /*
    #define REL_ELF_PRGM_HEADER_PHDR        (0x0001 | REL_TYPE_ELF_PRGM_HEADER)
    #define REL_ELF_PRGM_HEADER_INTERP      (0x0002 | REL_TYPE_ELF_PRGM_HEADER)
    #define REL_ELF_PRGM_HEADER_RX_LOAD1    (0x0003 | REL_TYPE_ELF_PRGM_HEADER)
    #define REL_ELF_PRGM_HEADER_RX_LOAD2    (0x0004 | REL_TYPE_ELF_PRGM_HEADER)
    #define REL_ELF_PRGM_HEADER_RW_LOAD     (0x0005 | REL_TYPE_ELF_PRGM_HEADER)
    #define REL_ELF_PRGM_HEADER_DYNAMIC     (0x0006 | REL_TYPE_ELF_PRGM_HEADER)
    #define REL_ELF_PRGM_HEADER_NOTE        (0x0007 | REL_TYPE_ELF_PRGM_HEADER)
  */
  // I do not know why one whould want to have the program headers in
  // the image, but let's ok that if the original executable had it
  
  // REL_ELF_PRGM_HEADER_INTERP
  // REL_ELF_PRGM_HEADER_DYNAMIC
  
  
  RelocatableDest* pDestImageIdent = new RelocatableDest();
  pDestImageIdent->AllocDataBuffer(sizeof (Elf32_Ident));
  
  DataDestRelocPtr destImageIdent(pDestImageIdent);
  
  destImageIdent.WriteData(&m_elfIdent,sizeof (Elf32_Ident));
  
  m_destSections.AddRelocatableDest(REL_ELF_IDENT,pDestImageIdent);
  
  
  RelocatableDest* pDestImageHeader = new RelocatableDest();
  pDestImageHeader->AllocDataBuffer(sizeof (Elf32_Header));
  
  DataDestRelocPtr destImageHeader(pDestImageHeader);
  
  m_elfHeader.e_ehsize    = sizeof (Elf32_Ident) + sizeof (Elf32_Header);
  
  destImageHeader.WriteData(&m_elfHeader,sizeof (Elf32_Header));
  
  m_elfHeader.e_entry = 0;
  
  destImageHeader.SetOffset(offsetof(Elf32_Header,e_entry));
  destImageHeader.WriteDWord(m_elfHeader.e_entry,REL_ELF_IMAGE_TEXT_SECTION);
  
  m_destSections.AddRelocatableDest(REL_ELF_HEADER,pDestImageHeader);
  
  // headers created, create program headers

  // word wRelocFirstLoadRX = REL_ELF_TEXT_SECTION;

/*
  WriteProgramHeader* pProgHdrLoadRX2 = CreateProgramHeader(ELF_PT_LOAD,ELF_PT_R | ELF_PT_X,wRelocFirstLoadRX);
  
  wRelocFirstPrgmHdr = REL_ELF_PRGM_HEADER_RX_LOAD2;
  m_destSections.AddRelocatableDest(REL_ELF_PRGM_HEADER_RX_LOAD2,pProgHdrLoadRX2->m_pRelocDest);
  m_elfHeader.e_phnum++;
*/

  // If we are going to emit any code (in .text, .init or .fini),
  // we'll have to create those sections so they are included in the tables
  
  Elf32_SectionRawData* pTextCodeSection = NULL;
  // Elf32_SectionRawData* pInitCodeSection = NULL;
  // Elf32_SectionRawData* pFiniCodeSection = NULL;
  
  if (1) // .text section needed (change 1 to real check)
    {
      pTextCodeSection = new Elf32_SectionRawData(".text",
						  ELF_SHT_PROGBITS,
						  ELF_SHF_ALLOC | 
						  ELF_SHF_EXECINSTR);
      m_vectSections.push_back(pTextCodeSection);    
    }
  
  // if it is an executable or shared file, we need a dynamic section, and symbol sections (and the string table is shared)
  
  Elf32_SectionDynamic* pDynamicSection = NULL;
  
  Elf32_SectionStringTable* pSymStrSection;
  Elf32_SectionSymbolTable* pSymTabSection;
  
  if (m_elfHeader.e_type == ELF_ET_EXEC ||
      m_elfHeader.e_type == ELF_ET_DYN)
    {
      Elf32_SectionHashTable*   pSymHashSection = 
	new Elf32_SectionHashTable(".hash");
      pSymStrSection  = new Elf32_SectionStringTable(".dynstr",ELF_SHF_ALLOC);
      pSymTabSection  = new Elf32_SectionSymbolTable(".dynsym",pSymStrSection,
						     pSymHashSection);
      
      pDynamicSection = new Elf32_SectionDynamic(pSymStrSection);
      
      Elf32_SectionGOT* pGOTSection = new Elf32_SectionGOT(".got");
      
      Elf32_SectionPLT* pPLTSection = new Elf32_SectionPLT(".plt");
      
      m_vectSections.push_back(pSymHashSection);    
      m_vectSections.push_back(pSymTabSection);    
      m_vectSections.push_back(pSymStrSection);    
      m_vectSections.push_back(pDynamicSection);    
      m_vectSections.push_back(pGOTSection);    
      m_vectSections.push_back(pPLTSection);    
      
      pDynamicSection->SetTag(ELF_DT_HASH  ,
			      DynamicValue(0,REL_ELF_IMAGE_HASH_SECTION));
      pDynamicSection->SetTag(ELF_DT_STRTAB,
			      DynamicValue(0,REL_ELF_IMAGE_DYNSYM_SECTION));
      pDynamicSection->SetTag(ELF_DT_SYMTAB,
			      DynamicValue(0,REL_ELF_IMAGE_DYNSTR_SECTION));
      pDynamicSection->SetTag(ELF_DT_PLTGOT,
			      DynamicValue(0,REL_ELF_IMAGE_GOT_SECTION));
      // REL_ELF_GOT_SECTION
      
      // pDynamicSection->SetTag(ELF_DT_REL   ,0x1234); // bogus
      // pDynamicSection->SetTag(ELF_DT_RELSZ ,0); // bogus
      // pDynamicSection->SetTag(ELF_DT_RELENT,sizeof (Elf32_Rel)); // bogus
      
      // pDynamicSection->SetTag(ELF_DT_INIT,DynamicValue(0,REL_ELF_IMAGE_TEXT_SECTION)); // bogus
      // pDynamicSection->SetTag(ELF_DT_FINI,DynamicValue(0,REL_ELF_IMAGE_TEXT_SECTION)); // bogus
      
      pDynamicSection->SetTag(ELF_DT_STRSZ ,0); // just allocate the tag (valuen set later)
      pDynamicSection->SetTag(ELF_DT_SYMENT,0); // just allocate the tag (valuen set later)
    }

  
  
  


  
  word wNextRelocUnalloc = REL_TYPE_ELF_UNALLOC_UNKNOWN1;
  word wNextRelocReadWrite = REL_TYPE_ELF_READ_WRITE_UNKNOWN1;
  word wNextRelocReadOnly = REL_TYPE_ELF_READ_ONLY_UNKNOWN1;
  
  // create the empty section header if one is needed (this is if we intend to emit any sections)
  
  if (!m_vectSections.empty())
    m_vectSections.insert(m_vectSections.begin(),new Elf32_SectionRawData());
  
    // if we output the sections, we should also output their names, if they have any
  // so create an string table for that
  
  Elf32_SectionStringTable* pNamesStringTable = 
    new Elf32_SectionStringTable(".shstrtab",0);
  
  m_vectSections.push_back(pNamesStringTable);    
  
    // fill the string table with the strings required
  
  {
    FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
      pNamesStringTable->RequestString((*sect)->m_strName);
  }
  
  // create the string table so we can get indexes into it
  
  pNamesStringTable->CreateSection();
  
  // create the sections
  
     // add all the sections contents to the executable
  FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
    {
      // and any that want to be allocated in the image to a corresponding program header

      // first figure out in what relocation slot it should go
      
      // look up our name in the string table and fill in appropriate index
      
      (*sect)->m_header.sh_name = 
	pNamesStringTable->GetStringIndex((*sect)->m_strName);
      
      word wReloc = 0;
      
      const Elf32_SpecialSectionRights* info = 
	GetSpecialSectionRights((*sect)->m_strName.c_str());

      if (info)
        {
          // so the type must agree, and the rights requested
	  
          assert ((*sect)->m_header.sh_type == info->type);
	  
          Elf32_Word wFlags = info->flags;
	  
          if (wFlags & ELF_SHF_ALLOC_IF_LOAD)
            {
              wFlags &= ~ELF_SHF_ALLOC_IF_LOAD;
              wFlags |= ELF_SHF_ALLOC; // because the relocation suggested is for allocated pieces
            }

          // for all the known blocks, the requested rights must be exact (except ELF_SHF_ALLOC_IF_LOAD)
	  
          Elf32_Word sectAccess = 
	    (*sect)->m_header.sh_flags & ELF_SHF_ACCESS_MASK;
	  
          assert (!(wFlags & ~ELF_SHF_ACCESS_MASK)); // so the next line does not require anything impossible
          
          assert (wFlags == sectAccess ||
                  (sectAccess == 0) && (info->flags & ELF_SHF_ALLOC_IF_LOAD));
	  
          if (wFlags == sectAccess)
            wReloc = info->wRelocBody;
        }
      else if ((*sect)->m_header.sh_type == 0)
        {
          // we only handle one type of section with SHT_NULL: the empty 0-index one

          assert (sect == m_vectSections.begin());
	  
          wReloc = REL_TYPE_ELF_SECTION_0;
        }          
      
      if (!wReloc)
        {
          // we'll have to figure out an appropriate reloc number based on the
          // requested rights of the section

          assert ((*sect)->m_header.sh_flags & ELF_SHF_ALLOC ||
                  (!((*sect)->m_header.sh_flags & ELF_SHF_WRITE) &&
                   !((*sect)->m_header.sh_flags & ELF_SHF_EXECINSTR))); // if write or exec, must be alloc
	  
          assert (!((*sect)->m_header.sh_flags & ELF_SHF_WRITE) || 
                  !((*sect)->m_header.sh_flags & ELF_SHF_EXECINSTR)); // we can't handle write-exec (self-modifying code?)
	  
          if (!((*sect)->m_header.sh_flags & ELF_SHF_ALLOC))
            wReloc = (wNextRelocUnalloc++); // not to be allocated
          else if (((*sect)->m_header.sh_flags & ELF_SHF_WRITE))
            {
              assert ((*sect)->m_header.sh_flags & ELF_SHF_ALLOC);
              wReloc = (wNextRelocReadWrite++);
            }
          else
            {
              assert (!((*sect)->m_header.sh_flags & 
			(ELF_SHF_WRITE | ELF_SHF_EXECINSTR)));
              wReloc = (wNextRelocReadOnly++);
            }     
        }

      assert (wReloc); // we must have found somewhere to put it

      // create an header
      
      (*sect)->SetWriteInfo(wReloc);
      (*sect)->CreateSection();

      //      if ((*sect)->m_wRelocSect != REL_TYPE_ELF_SECTION_0)
      //    {
      // no need to create section 0 contents, there are none
          








      /*
REL_TYPE_ELF_READ_ONLY_UNKNOWN1
REL_TYPE_ELF_READ_WRITE_UNKNOWN1
REL_TYPE_ELF_UNALLOC_UNKNOWN1
*/






/*
#define RELOC_SECT_HEADER_FROM_BODY(body) ((body) - REL_TYPE_ELF_BODY + REL_TYPE_ELF_SECT_HEADER)
#define RELOC_SECT_BODY_FROM_HEADER(hdr)  ((hdr)  + REL_TYPE_ELF_BODY - REL_TYPE_ELF_SECT_HEADER)
*/
/*
#define REL_TYPE_ELF_READ_ONLY_FIRST    (0x0001 | REL_TYPE_ELF_BODY) // first read-only
#define REL_ELF_INTERP_SECTION          (0x0001 | REL_TYPE_ELF_BODY)
#define REL_ELF_NOTE_ABI_TAG_SECTION    (0x0002 | REL_TYPE_ELF_BODY)
#define REL_ELF_HASH_SECTION            (0x0003 | REL_TYPE_ELF_BODY)
#define REL_ELF_DYNSYM_SECTION          (0x0004 | REL_TYPE_ELF_BODY)
#define REL_ELF_DYNSTR_SECTION          (0x0005 | REL_TYPE_ELF_BODY)
#define REL_ELF_RELDYN_SECTION          (0x0006 | REL_TYPE_ELF_BODY)
#define REL_ELF_RELPLT_SECTION          (0x0007 | REL_TYPE_ELF_BODY)
#define REL_TYPE_ELF_READ_ONLY_UNKNOWN1 (0x0008 | REL_TYPE_ELF_BODY) // first user-defined read-only
#define REL_TYPE_ELF_READ_ONLY_LAST     (0x003f | REL_TYPE_ELF_BODY) // last  user-defined read-only
	   				     
#define REL_TYPE_ELF_READ_EXECUTE_FIRST (0x0041 | REL_TYPE_ELF_BODY) // first read-execute
#define REL_ELF_INIT_SECTION            (0x0041 | REL_TYPE_ELF_BODY)
#define REL_ELF_FINI_SECTION            (0x0042 | REL_TYPE_ELF_BODY)
#define REL_ELF_PLT_SECTION             (0x0043 | REL_TYPE_ELF_BODY)
#define REL_ELF_TEXT_SECTION            (0x0044 | REL_TYPE_ELF_BODY)
#define REL_TYPE_ELF_READ_EXECUTE_LAST  (0x0044 | REL_TYPE_ELF_BODY) // last read-execute
	   				     
#define REL_TYPE_ELF_READ_ONLY2_FIRST   (0x0081 | REL_TYPE_ELF_BODY) // first read-only2 (to be placed after executable code)
#define REL_ELF_RODATA_SECTION          (0x0081 | REL_TYPE_ELF_BODY)
#define REL_TYPE_ELF_READ_ONLY2_LAST    (0x0081 | REL_TYPE_ELF_BODY) // last read-only2
	   				     
#define REL_TYPE_ELF_READ_WRITE_FIRST   (0x00c1 | REL_TYPE_ELF_BODY) // first read-write
#define REL_ELF_DATA_SECTION            (0x00c1 | REL_TYPE_ELF_BODY)
#define REL_ELF_DYNAMIC_SECTION         (0x00c2 | REL_TYPE_ELF_BODY)
#define REL_ELF_GOT_SECTION             (0x00c3 | REL_TYPE_ELF_BODY)
#define REL_ELF_BSS_SECTION             (0x00c4 | REL_TYPE_ELF_BODY)
#define REL_TYPE_ELF_READ_WRITE_UNKNOWN1 (0x00c5 | REL_TYPE_ELF_BODY) // first user-defined read-write
#define REL_TYPE_ELF_READ_WRITE_LAST     (0x00ff | REL_TYPE_ELF_BODY) // last  user-defined read-write
*/

    }
  
    // sort the sections in the section vector to they appear there in the same
    // order as they will in the section header table (arranged when DoFileLayout)

    SectionRelocNoCompare compare;
    
    sort(m_vectSections.begin(),m_vectSections.end(),compare);
    
    
    {
      FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
        {
          // fixup any links to other sections
	  
          Elf32_Section* pSection = (*sect);
	  
          switch (pSection->m_header.sh_type)
            {
            case ELF_SHT_DYNAMIC:
	      assert (dynamic_cast<Elf32_SectionDynamic*>(pSection));
                pSection->m_header.sh_link =
		  GetSectionIndex(((Elf32_SectionDynamic*) pSection)->GetStringTable());
                break;
            case ELF_SHT_HASH:
	      assert (dynamic_cast<Elf32_SectionHashTable*>(pSection));
	      pSection->m_header.sh_link = 
		GetSectionIndex(((Elf32_SectionHashTable*) pSection)->GetSymbolTable());
	      break;
            case ELF_SHT_REL:
            case ELF_SHT_RELA:
	      assert (dynamic_cast<Elf32_SectionRelocation*>(pSection));
	      pSection->m_header.sh_link = 
		GetSectionIndex(((Elf32_SectionRelocation*) pSection)->GetSymbolTable());
	      pSection->m_header.sh_info = 
		GetSectionIndex(((Elf32_SectionRelocation*) pSection)->GetRelocSection());
	      break;
            case ELF_SHT_SYMTAB:
            case ELF_SHT_DYNSYM:
	      assert (dynamic_cast<Elf32_SectionSymbolTable*>(pSection));
	      pSection->m_header.sh_link = 
		GetSectionIndex(((Elf32_SectionSymbolTable*) pSection)->GetStringTable());
              break;
            }
	  
	  
          word wReloc = (*sect)->m_wRelocSect;
	  
          m_destSections.AddRelocatableDest(wReloc,(*sect)->WriteSection());
          //  }
	  
          word wRelocHeader = RELOC_SECT_HEADER_FROM_BODY(wReloc);
	  
          m_destSections.AddRelocatableDest(wRelocHeader,
					    (*sect)->WriteHeader());
        }
    }




  
  // if there is an interpreter section,
  // make a program block linking to that
  
  Elf32_Section* pInterpSect;
  
  if ((pInterpSect = FindSection(".interp",ELF_SHT_PROGBITS)))
    {
      Elf32_ProgramBlock* pInterpBlock = new Elf32_ProgramBlock();
      pInterpBlock->SetSection(pInterpSect,ELF_PT_INTERP,ELF_PT_R);
      m_vectProgramBlocks.push_back(pInterpBlock);
      
      m_destSections.AddRelocatableDest(REL_ELF_PRGM_HEADER_INTERP,
					pInterpBlock->WriteHeader());
    }
  // and the ABI-tag note section
  Elf32_Section* pNoteABItagSect;
  
  if ((pNoteABItagSect = FindSection(".note.ABI-tag",ELF_SHT_NOTE)))
    {
      Elf32_ProgramBlock* pNoteABItagBlock = new Elf32_ProgramBlock();
      pNoteABItagBlock->SetSection(pNoteABItagSect,ELF_PT_NOTE,ELF_PT_R);
      m_vectProgramBlocks.push_back(pNoteABItagBlock);
      
      m_destSections.AddRelocatableDest(REL_ELF_PRGM_HEADER_NOTE_ABI,
					pNoteABItagBlock->WriteHeader());
    }
  // and the dynamic section

  if (pDynamicSection)
    {
      Elf32_ProgramBlock* pDynamicBlock = new Elf32_ProgramBlock();
      pDynamicBlock->SetSection(pDynamicSection,
				ELF_PT_DYNAMIC,ELF_PT_R | ELF_PT_W);
      m_vectProgramBlocks.push_back(pDynamicBlock);
      
      m_destSections.AddRelocatableDest(REL_ELF_PRGM_HEADER_DYNAMIC,
					pDynamicBlock->WriteHeader());
      
      assert (pSymStrSection);
      assert (pSymTabSection);
      pDynamicSection->SetTag(ELF_DT_STRSZ ,
			      pSymStrSection->m_header.sh_size);
      pDynamicSection->SetTag(ELF_DT_SYMENT,
			      pSymTabSection->m_header.sh_entsize);
    }
  
  /*Elf32_ProgramBlock* pRXLoad1 = */AddProgramBlockLoadSectors(REL_TYPE_ELF_READ_ONLY_FIRST,REL_TYPE_ELF_READ_ONLY2_LAST,REL_ELF_PRGM_HEADER_RX_LOAD1);;
  /*Elf32_ProgramBlock* pRXLoad2 = *///NULL;
  /*Elf32_ProgramBlock* pRWLoad  = */AddProgramBlockLoadSectors(REL_TYPE_ELF_READ_WRITE_FIRST,REL_TYPE_ELF_READ_WRITE_LAST,REL_ELF_PRGM_HEADER_RW_LOAD);;
  
  
  
  /*
#define REL_TYPE_ELF_READ_ONLY_FIRST    (0x0002 | REL_TYPE_ELF_BODY) // first read-only
#define REL_TYPE_ELF_READ_ONLY_LAST     (0x003f | REL_TYPE_ELF_BODY) // last  user-defined read-only

#define REL_TYPE_ELF_READ_EXECUTE_FIRST (0x0040 | REL_TYPE_ELF_BODY) // first read-execute
#define REL_TYPE_ELF_READ_EXECUTE_LAST  (0x005f | REL_TYPE_ELF_BODY) // last read-execute

#define REL_TYPE_ELF_READ_ONLY2_FIRST   (0x0060 | REL_TYPE_ELF_BODY) // first read-only2 (to be placed after executable code)
#define REL_TYPE_ELF_READ_ONLY2_LAST    (0x007f | REL_TYPE_ELF_BODY) // last read-only2

#define REL_TYPE_ELF_READ_WRITE_FIRST   (0x0080 | REL_TYPE_ELF_BODY) // first read-write
#define REL_TYPE_ELF_READ_WRITE_LAST    (0x00bf | REL_TYPE_ELF_BODY) // last  user-defined read-write
*/

/*
#define REL_ELF_PRGM_HEADER_PHDR        (0x0001 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_INTERP      (0x0002 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_NOTE_ABI    (0x0003 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_RX_LOAD1    (0x0004 | REL_TYPE_ELF_PRGM_HEADER) // seldom-access (mainly on init)
#define REL_ELF_PRGM_HEADER_RX_LOAD2    (0x0005 | REL_TYPE_ELF_PRGM_HEADER) // .text .plt (and any in LOAD1 if that's omitted)
#define REL_ELF_PRGM_HEADER_RW_LOAD     (0x0006 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_DYNAMIC     (0x0007 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_NOTE        (0x0008 | REL_TYPE_ELF_PRGM_HEADER)
*/

/*
#define REL_ELF_INTERP_SECTION          (0x0001 | REL_TYPE_ELF_BODY)
#define REL_ELF_NOTE_ABI_SECTION        (0x0002 | REL_TYPE_ELF_BODY)
*/
/*
struct Elf32_ProgramHeader // Dumpable
{
  Elf32_Word	p_type;			// Type
  Elf32_Off     p_offset;		// File offset
  Elf32_Addr	p_vaddr;		// Im. address		
  Elf32_Addr	p_paddr;		// Phys. address
  Elf32_Word	p_filesz;		// File size
  Elf32_Word	p_memsz;		// Image size
  Elf32_Word	p_flags;		// Flags
  Elf32_Word	p_align;		// Align
};

*/



//  VectElf32_ProgramBlocks

  // after having created the program headers we have to allocate the space in the executable for them so that
  // we know where the following sections will be so that writing any executable code can know how it is aligned
  // and not only on 16-byte boundaries or so, but how it is aligned on page boundaries so it can be arranged
  // so as to minimize paging

  m_fileLayout.dwImageBase = 0x08000000;
  m_fileLayout.dwFileNextFree = 0;
  m_fileLayout.dwImageNextFree = 0;
  
  DoFileLayout (REL_ELF_IDENT,REL_ELF_HEADER);
  DoFileLayout (REL_TYPE_FIRST_PRGM_HEADER,REL_TYPE_LAST_PRGM_HEADER);
  
  // we'll also create {hash, dynsym, dynstr}, plt, rel.plt and rel.dyn
  // as these come before any code (plt may be inside)
  
  DoFileLayout (REL_TYPE_ELF_READ_ONLY_FIRST,REL_TYPE_ELF_READ_ONLY_LAST);
  
  
  if (1) // any code to emit? (change to proper test) (we may have file without code)
    {
      assert (pTextCodeSection);
      assert (m_destSections.HasSection(REL_ELF_TEXT_SECTION)); // should already have been added
      
      CreateCodeSection(pTextCodeSection);
      
      // m_destSections.AddRelocatableDest(REL_ELF_TEXT_SECTION,CreateCodeSection ());	
    }
  



//  pProgHdrLoadRX2->SetSize (m_destSections.GetSectionImageLength(REL_ELF_TEXT_SECTION));

  // not necessary to layout section 0, as it has no contents (if we do, we'll probably have to fix some things at other places)
  // DoFileLayout (REL_TYPE_ELF_SECTION_0,REL_TYPE_ELF_SECTION_0);
  
  DoFileLayout (REL_TYPE_ELF_READ_EXECUTE_FIRST,
		REL_TYPE_ELF_READ_EXECUTE_LAST);
  
  DoFileLayout (REL_TYPE_ELF_READ_ONLY2_FIRST,REL_TYPE_ELF_READ_ONLY2_LAST);
  DoFileLayout (REL_TYPE_ELF_READ_WRITE_FIRST,REL_TYPE_ELF_READ_WRITE_LAST);
  DoFileLayout (REL_TYPE_ELF_UNALLOC_FIRST,REL_TYPE_ELF_UNALLOC_LAST);
  
  DoFileLayout (REL_TYPE_FIRST_SECTION_HEADER,REL_TYPE_LAST_SECTION_HEADER);
  
  m_elfHeader.e_phnum = m_vectProgramBlocks.size();
  m_elfHeader.e_shnum = m_vectSections.size();
  
  if ((m_elfHeader.e_phnum = m_vectProgramBlocks.size()))
    {
      m_elfHeader.e_phentsize = sizeof (Elf32_ProgramHeader);
      
      word wRelocFirstPrgmHdr = 
	m_destSections.GetSectionFirstInRange(REL_TYPE_FIRST_PRGM_HEADER,
					      REL_TYPE_LAST_PRGM_HEADER);
      
      assert (wRelocFirstPrgmHdr);
      
      destImageHeader.InsertRelocation(offsetof(Elf32_Header,e_phoff),
				       wRelocFirstPrgmHdr | RELATIVE_FILE);
    }
  if ((m_elfHeader.e_shnum = m_vectSections.size()))
    {
      m_elfHeader.e_shentsize = sizeof (Elf32_SectionHeader);
      
      word wRelocFirstSectHdr = 
	m_destSections.GetSectionFirstInRange(REL_TYPE_FIRST_SECTION_HEADER,
					      REL_TYPE_LAST_SECTION_HEADER);
      
      assert (wRelocFirstSectHdr);
      
      destImageHeader.InsertRelocation(offsetof(Elf32_Header,e_shoff),
				       wRelocFirstSectHdr | RELATIVE_FILE);
    }
  
  destImageHeader.SetOffset(offsetof(Elf32_Header,e_phentsize));
  destImageHeader.WriteWord(m_elfHeader.e_phentsize);
  destImageHeader.WriteWord(m_elfHeader.e_phnum);
  destImageHeader.WriteWord(m_elfHeader.e_shentsize);
  destImageHeader.WriteWord(m_elfHeader.e_shnum);
  
  m_elfHeader.e_shstrndx = GetSectionIndex(pNamesStringTable);  
  destImageHeader.SetOffset(offsetof(Elf32_Header,e_shstrndx));
  destImageHeader.WriteWord(m_elfHeader.e_shstrndx);
  
  { // because faulty scoping of for-variables of MSVC
    // detach ownership of RelocatableDest(s)
    FOR_EACH_ITER(VectElf32_ProgramBlocks,m_vectProgramBlocks,block)
      {
        (*block)->SetBlockRange();
        (*block)->m_pWriteHeader->ReleaseWrite();
      }
    FOR_EACH_ITER(VectElf32_Sections,m_vectSections,sect)
      {
        (*sect)->m_pWriteDest->ReleaseWrite();
        (*sect)->m_pWriteHeader->ReleaseWrite();
      }
  }
  
//  pProgHdrLoadRX2->ReleaseWrite();
  



    
    /*
    destImageDos.WriteWord(mzHeaderAndData->e_magic);
        destImageDos.SetOffset(offsetof(IMAGE_DOS_HEADER,e_lfanew));
        destImageDos.WriteDWord(0,RELATIVE_FILE_PE_HEADER);

        m_destSections.AddRelocatableDest(RELATIVE_MZ_HEADER,pDestImageDos);
*/

	// and sort the stuff into useful order
	// that is, calculate offsets for file and image taking image/file-alignment
	// into account

	// I've not found any direct instructions on whether the headers should also
	// go into the image or not. But it looks like it, so begin with those
  /*
    dword dwImageBase = 0x08000000;

    dword dwFileSize = 0;
        dword dwImageSize = 0;
        dword dwHeaderSize = 0;

        for (uint i = 0; i < count_of(infoELFFileSections); i++)
        {
            if (m_destSections.HasSection(infoELFFileSections[i].wSection))
                {
            dword dwFileAlign = 1;
                        dword dwImageAlign = 1;

                        if (infoELFFileSections[i].wAlignment)
                        {
                                dwFileAlign  = infoELFFileSections[i].wAlignment;
                                dwImageAlign = infoELFFileSections[i].wAlignment;
                        }

                        assert (powerof2(dwFileAlign));
                        assert (powerof2(dwImageAlign));

                        printf ("                        %08x  %08x\n",dwFileAlign,dwImageAlign);
                        printf ("%08x  %08x\n",dwFileSize,dwImageSize);

                        dwFileSize  = next_multiple( dwFileSize, dwFileAlign);
                        dwImageSize = next_multiple(dwImageSize,dwImageAlign);

                        printf ("%08x  %08x\n",dwFileSize,dwImageSize);

                        // so now we're aligned

                        m_destSections.RelocateSection(infoELFFileSections[i].wSection | RELATIVE_FILE ,dwFileSize);
                        m_destSections.RelocateSection(infoELFFileSections[i].wSection | RELATIVE_IMAGE,dwImageBase + dwImageSize);

                        dwFileSize  += m_destSections.GetSectionFileLength (infoELFFileSections[i].wSection);
                        dwImageSize += m_destSections.GetSectionImageLength(infoELFFileSections[i].wSection);

                        printf ("%08x  %08x\n",dwFileSize,dwImageSize);

                        if (infoELFFileSections[i].bPartOfHeader)
                                dwHeaderSize = dwFileSize;
                }
        }
  */
/*
	destImagePE.SetOffset(offsetof(IMAGE_PE_FILE_HEADER,wNumberOfSections));
	destImagePE.WriteDWord(peHeader.wNumberOfSections);

	peOptHeader.dwSizeOfImage = dwImageSize;
	peOptHeader.dwSizeOfHeaders = dwHeaderSize;

	destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,dwSizeOfImage));
	destImagePEopt.WriteDWord(peOptHeader.dwSizeOfImage);
	destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,dwSizeOfHeaders));
	destImagePEopt.WriteDWord(peOptHeader.dwSizeOfHeaders);
*/



  m_destSections.DoRelocations();
  
  return true;
}

//////////////////////////////////////////////////////////////////////

bool ELF_File::WriteFile()
{
  return ExecutableImage::WriteFile();
}

//////////////////////////////////////////////////////////////////////

void ELF_File::CreateCodeSection(Elf32_Section* pDestSection)
{
  //	CodeSection codeSection;
  
  // 	codeSection.AssureDestMemory(16); // 16 bytes maximum per instruction
  
  WriteRelocDest* pDest = pDestSection->m_pWriteDest;

  
  printf ("ELF_File::CreateCodeSection\n");
  
  if (m_strFileName != "created_elf")
    {
      ExecutableWriter writer;
      
      writer.m_pExecutable = m_pExecutable;
      
      writer.Write (pDest);

      pDest->TrimMemoryUsage();
      pDestSection->m_pWriteHeader->SetSize(pDestSection->m_header.sh_size = 
					    pDest->m_pRelocDest->m_dwSize);
      
      return;
    }
  
  printf ("Writing: created_elf\n");
  
  // we're 'created_elf', this is the small example ELF assebler file that we create
  // this code is not to be in a release version, because then we can handle properly compiled files!
  
  
  pDest->AssureDestMemory(1024);
  
  DataDestRelocPtr* pDestPtr = pDest->m_pDest;
  
  // pDestPtr->WriteByte (0xcc);   // INT3 

  /*
b9 00 00 00 00	MOV	ECX,0
bb 00 00 00 00	MOV	EBX,0

e8 12 00 00 00	CALL	func1
01 c3		ADD	EBX,EAX
41 81		INC	ECX
f9 00 00 10 00	CMP	ECX,0x100000
75 f0		JNZ	loop1

                AND     EBX,0xf7

b8 01 00 00 00	MOV	EAX,0x01
cd 80		INT	0x80
c3		RET


89 c8		MOV	EAX,ECX
d1 e0		SHL	EAX,1
c3		RET
  */
  /*
  pDestPtr->WriteByte (0xb9);   pDestPtr->WriteDWord(0x00000003);
  pDestPtr->WriteByte (0xbb);   pDestPtr->WriteDWord(0x00000036);

  //pDestPtr->WriteByte (0xe8);   pDestPtr->WriteDWord(0x00000019);

  pDestPtr->WriteByte (0x89);   pDestPtr->WriteByte (0xc8);
  pDestPtr->WriteByte (0x01);   pDestPtr->WriteByte (0xc0);
  //pDestPtr->WriteByte (0xd1);   pDestPtr->WriteByte (0xe0);
  pDestPtr->WriteByte (0x90);

  pDestPtr->WriteByte (0x01);   pDestPtr->WriteByte (0xc3);
  pDestPtr->WriteByte (0x41);   pDestPtr->WriteByte (0x81);
  pDestPtr->WriteByte (0xf9);   pDestPtr->WriteDWord(0x10000000);
  pDestPtr->WriteByte (0x75);   pDestPtr->WriteByte (0xf0);
  pDestPtr->WriteByte (0x81);
  pDestPtr->WriteByte (0xe3);   pDestPtr->WriteDWord(0x0000007f);
  pDestPtr->WriteByte (0xb8);   pDestPtr->WriteDWord(0x00000001);
  pDestPtr->WriteByte (0xcd);   pDestPtr->WriteByte (0x80);
  pDestPtr->WriteByte (0xc3);   
  pDestPtr->WriteByte (0x89);   pDestPtr->WriteByte (0xc8);
  pDestPtr->WriteByte (0xd1);   pDestPtr->WriteByte (0xe0);
  pDestPtr->WriteByte (0xc3);   
  */
  pDestPtr->WriteByte (0xb9);   pDestPtr->WriteDWord(0x00000003);
  pDestPtr->WriteByte (0xbb);   pDestPtr->WriteDWord(0x00000036);
  pDestPtr->WriteByte (0xe8);   pDestPtr->WriteDWord(0x00000019);
  pDestPtr->WriteByte (0x01);   pDestPtr->WriteByte (0xc3);
  pDestPtr->WriteByte (0x41);   pDestPtr->WriteByte (0x81);
  pDestPtr->WriteByte (0xf9);   pDestPtr->WriteDWord(100000000);
  pDestPtr->WriteByte (0x75);   pDestPtr->WriteByte (0xf0);
  pDestPtr->WriteByte (0x81);
  pDestPtr->WriteByte (0xe3);   pDestPtr->WriteDWord(0x0000007f);
  pDestPtr->WriteByte (0xb8);   pDestPtr->WriteDWord(0x00000001);
  pDestPtr->WriteByte (0xcd);   pDestPtr->WriteByte (0x80);
  pDestPtr->WriteByte (0xc3);   
  pDestPtr->WriteByte (0x89);   pDestPtr->WriteByte (0xc8);
  pDestPtr->WriteByte (0x40);
  pDestPtr->WriteByte (0x01);   pDestPtr->WriteByte (0xc0);
  //pDestPtr->WriteByte (0xd1);   pDestPtr->WriteByte (0xe0);
  pDestPtr->WriteByte (0xc3);   
  

  // so eject assembler for     MOV EBX,0    ; bb ?? 00 00 00 // exit code
  //                            MOV EAX,0    ; b8 01 00 00 00 // system call no (sys_exit)
  //                            INT 0x80     ; cd 80          // call kernel

  pDestPtr->WriteByte (0xbb);
  pDestPtr->WriteDWord(23);
  pDestPtr->WriteByte (0xb8);
  pDestPtr->WriteDWord(0x00000001);
  pDestPtr->WriteBytes(0xcd,0x080);
  
  // so eject assembler for     MOV EBX,0    ; bb ?? 00 00 00 // exit code
  //                            MOV EAX,0    ; b8 01 00 00 00 // system call no (sys_exit)
  //                            INT 0x80     ; cd 80          // call kernel

  pDestPtr->WriteByte (0xbb);
  pDestPtr->WriteDWord(23);
  pDestPtr->WriteByte (0xb8);
  pDestPtr->WriteDWord(0x00000001);
  pDestPtr->WriteBytes(0xcd,0x080);
  
  /*
    RelocatableDest* pReturn = codeSection.EndWrite();
    
    return pReturn;
  */
  pDest->TrimMemoryUsage();
  pDestSection->m_pWriteHeader->SetSize(pDestSection->m_header.sh_size = 
					pDest->m_pRelocDest->m_dwSize);
}

