/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Elf32_ProgramBlock.cpp: implementation of the Elf32_ProgramBlock class.
//
//////////////////////////////////////////////////////////////////////

#include "Elf32_ProgramBlock.h" 
#include "Elf32_Section.h" 
#include "Error.h"
#include "ErrorTracker.h"
#include "Misc.h"
#include "WriteProgramHeader.h"
#include "DataDestRelocPtr.h"
#include "RelocatableDest.h"

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _lseek lseek
#define _tell(fildes) (lseek(fildes,0,SEEK_CUR))
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Elf32_ProgramBlock::Elf32_ProgramBlock()
{
  memset (&m_header,0,sizeof(m_header));
  //m_pData = NULL;

  m_pWriteHeader = NULL;
}

Elf32_ProgramBlock::Elf32_ProgramBlock(Elf32_Word type,Elf32_Word flags)
{
  memset (&m_header,0,sizeof(m_header));
  //m_pData = NULL;

  m_header.p_type = type;
  m_header.p_flags = flags;

  m_pWriteHeader = NULL;
}

Elf32_ProgramBlock::~Elf32_ProgramBlock()
{
    delete m_pWriteHeader;
}

void Elf32_ProgramBlock::InitData (const Elf32_ProgramHeader* header/*,const void *pData*/)
{
  m_header = *header;
  
  //m_pData = (const byte*) pData;
  

}

bool Elf32_ProgramBlock::IsSectionInside (const Elf32_Section* sect)
{
  if (sect->m_header.sh_offset+sect->m_header.sh_size <=  m_header.p_offset ||
      sect->m_header.sh_offset                        >= (m_header.p_offset + 
							  m_header.p_filesz))
    return false;

  // we must be completely inside?
  
  VALIDATE(sect->m_header.sh_offset                        >= m_header.p_offset  &&
           sect->m_header.sh_offset+sect->m_header.sh_size <= m_header.p_offset + m_header.p_filesz,
           "Section partially inside an program header block.");
  
  if (sect->m_header.sh_addr) // section has virtual address in image
    {
      if (sect->m_header.sh_offset) // section has bits in file
	{
	  // the bits in file must match virtual address for program header
	  // and section
	  
	  VALIDATE(sect->m_header.sh_addr - sect->m_header.sh_offset ==
		   m_header.p_vaddr - m_header.p_offset,
		   "Section bits in file has different virtual address than same bits of program block.");
	}
    }
  
  return true;
}

void Elf32_ProgramBlock::AddSection (Elf32_Section* pSection)
{
  m_vectSections.push_back(pSection);
  
  pSection->AddProgramBlock(this);
}

void Elf32_ProgramBlock::ValidateHeader()
{
  if (m_header.p_type == ELF_PT_NULL)
    {
      EMIT_GENERIC_ERROR("Program Header is ignored entry. Contents undefined.");
      return; // ignored entry, so do not check any more
    }
  
  VALIDATE(m_header.p_filesz <= m_header.p_memsz,
	   "Program header file size larger than image size");

  if (m_header.p_align > 1)
    {
      // we have alignment requirements
      
      VALIDATE(powerof2(m_header.p_align),
	       "Program block alignment requiremnt not power of 2.");
      
      VALIDATE(!((m_header.p_offset - m_header.p_vaddr) % m_header.p_align),
	       "Program block virtual address and file offset not congruent modulo alignment specified.");
      /*
	VALIDATE(!(m_header.p_vaddr % m_header.p_align),
	"Program block virtual address not aligned as specified.");
	VALIDATE(!(m_header.p_offset % m_header.p_align),
			"Program block file offset aligned as specified.");
      */
	}
  VALIDATE(!((m_header.p_offset - m_header.p_vaddr) % 0x1000),
	   "Program block virtual address and file offset not congruent modulo 0x1000 (page size).");
  
  WARNING(m_header.p_paddr == 0 || m_header.p_paddr == m_header.p_vaddr,
	  "Program header physical address not zero or same as virtual address. It's value undefined according to spec, so this is just a warning.");
  
  VALIDATE(m_header.p_flags & ELF_PT_R,
	   "Uhh, program block does not have read flag??? (valid by spec, but useful?)");
  
  VALIDATE(!(m_header.p_flags & (ELF_PT_X)) ||
	   m_header.p_flags & ELF_PT_R,
	   "Program block requires execute rights, but does note require read, which is strange.");
  
  VALIDATE(!(m_header.p_flags & (ELF_PT_W)) ||
	   m_header.p_flags & ELF_PT_R,
	   "Program block requires write rights, but does note require read, which is strange.");
  
}

bool Elf32_ProgramBlock::ValidateCoExistence (FileMapItem* pOtherItem)
{
  // we can coexist with another section (okey'd by section, so we do not have to)
  
  // we can coexist with an program block

  Elf32_ProgramBlock* pProgramBlock;
  
  if ((pProgramBlock = dynamic_cast<Elf32_ProgramBlock*>(pOtherItem)))
    {
      // our requested access rights should bo compatible with the
      // other program block's (depending on who is inside who)
      
      // pProgramBlock
      
      
      return true; // it's ok
    }

  GenericFileMapItem* pGenericOtherItem;
  
  if ((pGenericOtherItem = dynamic_cast<GenericFileMapItem*>(pOtherItem)))
    {
      if (pGenericOtherItem->m_wType == GFMI_ELF_IDENT ||
	  pGenericOtherItem->m_wType == GFMI_ELF_HEADER ||
	  pGenericOtherItem->m_wType == GFMI_ELF_PROGRAM_HEADER)
        {
	  // the program headers may be in the program block
	  
	  return true;
        }
      
    }
  
  return false;
    // return FileMapItem::ValidateCoExistence(pOtherItem);
}

string Elf32_ProgramBlock::GetDescription()
{
  return "Program block";
}

//////////////////////////////////////////////////////////////////////

void Elf32_ProgramBlock::AddLoadSection (Elf32_Section* pSection)
{
  assert (pSection->m_header.sh_flags & ELF_SHF_ALLOC); // the section must want to be allocated
  
  AddSection(pSection);
  
  // now make our flags include the flags requested by the section

  if (pSection->m_header.sh_flags & ELF_SHF_EXECINSTR)
    m_header.p_flags |= ELF_PT_X;
  if (pSection->m_header.sh_flags & ELF_SHF_WRITE)
    m_header.p_flags |= ELF_PT_W;
}

void Elf32_ProgramBlock::SetSection (Elf32_Section* pSection,Elf32_Word type,Elf32_Word flags)
{
  AddSection(pSection);
  
  m_header.p_type = type;
  m_header.p_flags = flags;
}

RelocatableDest* Elf32_ProgramBlock::WriteHeader ()
{
  assert (!m_pWriteHeader);
  
  m_pWriteHeader = new WriteProgramHeader();
  
  m_pWriteHeader->AllocMemory(sizeof (Elf32_ProgramHeader));
  /*
    Elf32_ProgramHeader header;
    
    memset(&header,0,sizeof (header));
    
    header.p_type  = type;
    header.p_flags = flags;
  */
  m_pWriteHeader->m_pDest->WriteData(&m_header,sizeof (m_header));
  /*
    m_pWriteHeader->m_pDest->SetOffset(offsetof(Elf32_ProgramHeader,p_offset));
    m_pWriteHeader->m_pDest->WriteDWord(0,m_wRelocSect | RELATIVE_FILE);
    m_pWriteHeader->m_pDest->SetOffset(offsetof(Elf32_ProgramHeader,p_vaddr));
    m_pWriteHeader->m_pDest->WriteDWord(0,m_wRelocSect | RELATIVE_IMAGE);
    m_pWriteHeader->m_pDest->WriteDWord(0,m_wRelocSect | RELATIVE_IMAGE); // also se p_paddr to same as p_vaddr (although it's to be undefined (or at least unused))
  */
  /*
    assert (!m_vectSections.empty());
    
    word wRelocSect = m_vectSections.front()->m_wRelocSect;
    
    m_pWriteHeader->m_pDest->InsertRelocation(offsetof(Elf32_ProgramHeader,p_offset),wRelocSect | RELATIVE_FILE);
    m_pWriteHeader->m_pDest->InsertRelocation(offsetof(Elf32_ProgramHeader,p_vaddr),wRelocSect | RELATIVE_IMAGE);
    m_pWriteHeader->m_pDest->InsertRelocation(offsetof(Elf32_ProgramHeader,p_paddr),wRelocSect | RELATIVE_IMAGE);
  */
  return m_pWriteHeader->m_pRelocDest;
}

void Elf32_ProgramBlock::SetBlockRange()
{
  // so, make sure the bytes of all sections are included
  
  assert (!m_vectSections.empty()); // or we're not needed
  
  VectSections::iterator sect = m_vectSections.begin();
  
  assert (sect != m_vectSections.end());
  
  assert ((*sect)->m_pWriteDest);
  assert ((*sect)->m_pWriteDest->m_pRelocDest);

#define SECTION_IMAGE_ADDR  ((*sect)->m_pWriteDest->m_pRelocDest->m_relocInfo.m_dwImageOffset/*(*sect)->m_header.sh_addr*/)
#define SECTION_FILE_OFFSET ((*sect)->m_pWriteDest->m_pRelocDest->m_relocInfo.m_dwFileOffset/*(*sect)->m_header.sh_offset*/)
  
  m_header.p_vaddr  = SECTION_IMAGE_ADDR;
  m_header.p_memsz = (*sect)->m_header.sh_size;
  m_header.p_offset = SECTION_FILE_OFFSET;
  
  if ((*sect)->m_header.sh_type != ELF_SHT_NOBITS)
    m_header.p_filesz = m_header.p_memsz;
  
  m_header.p_align = (*sect)->m_header.sh_addralign;
  
  for (++sect ; sect != m_vectSections.end(); ++sect)
    {
      // we assume they come in order, (not in order would be allowed,
      // but requires more logic here, so if you want it, you have to
      // fix that)
      
      assert ((*sect)->m_pWriteDest);
      assert ((*sect)->m_pWriteDest->m_pRelocDest);
      
      assert (SECTION_IMAGE_ADDR > m_header.p_vaddr);
      
      // as all the sections should be loaded into memory
      // contiguously, make sure that is possible
      
      assert (SECTION_IMAGE_ADDR - m_header.p_vaddr ==
	      SECTION_FILE_OFFSET - m_header.p_offset);
      
      // and they should not overlap
      
      assert (SECTION_IMAGE_ADDR >= m_header.p_vaddr + m_header.p_memsz);
      
      // alignment should be ok
      
      if ((*sect)->m_header.sh_addralign > 1)
	{
	  // we have alignment requirements
	  
	  assert(powerof2((*sect)->m_header.sh_addralign));
	  assert(!(SECTION_IMAGE_ADDR % (*sect)->m_header.sh_addralign)); // address must be aligned
        }
      
      if ((*sect)->m_header.sh_addralign > m_header.p_align)
	(*sect)->m_header.sh_addralign = m_header.p_align;
      
      if ((*sect)->m_header.sh_type != ELF_SHT_NOBITS)
        {
	  assert (m_header.p_filesz == m_header.p_memsz); // or we have a section with bits following a section with no bits, which is (extremely) stupid, as the nobits would be stored anyway
	  
	  m_header.p_filesz = 
	    SECTION_IMAGE_ADDR + (*sect)->m_header.sh_size - m_header.p_vaddr;
        }
      m_header.p_memsz = 
	SECTION_IMAGE_ADDR + (*sect)->m_header.sh_size - m_header.p_vaddr;
    }
  
  m_header.p_paddr = m_header.p_vaddr; // p_paddr is undefined, but lets set it to p_vaddr
  
  if (m_header.p_type == ELF_PT_LOAD)
    m_header.p_align = 0x1000;
  
  // make sure alignment is corrent
  
  if (m_header.p_align > 1)
    {
      // we have alignment requirements
      
      assert(powerof2(m_header.p_align));
      assert(!((m_header.p_offset - m_header.p_vaddr) % m_header.p_align));
    }
  
  m_pWriteHeader->m_pDest->SetOffset(offsetof(Elf32_ProgramHeader,p_offset));
  m_pWriteHeader->m_pDest->WriteDWord(m_header.p_offset);
  m_pWriteHeader->m_pDest->WriteDWord(m_header.p_vaddr);
  m_pWriteHeader->m_pDest->WriteDWord(m_header.p_paddr);
  m_pWriteHeader->m_pDest->WriteDWord(m_header.p_filesz);
  m_pWriteHeader->m_pDest->WriteDWord(m_header.p_memsz);
  
  m_pWriteHeader->m_pDest->SetOffset(offsetof(Elf32_ProgramHeader,p_align));
  m_pWriteHeader->m_pDest->WriteDWord(m_header.p_align);
}


/*
  WriteProgramHeader* CreateProgramHeader(Elf32_Word type,Elf32_Word flags,word wRelocStarts)
  {
  }
*/






