/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Elf32_Section.cpp: implementation of the Elf32_Section class.
//
//////////////////////////////////////////////////////////////////////

#include "Elf32_Section.h" 
#include "Elf32_ProgramBlock.h"
#include "ErrorTracker.h"
#include "Error.h"
#include "Misc.h"
#include "BitField.h"
#include "DataPtrSrc.h"
#include "gen/ELF_File_DumpStructs.h"

#include "WriteRelocDest.h"
#include "WriteProgramHeader.h"
#include "DataDestRelocPtr.h"
#include "RelocatableDest.h"

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _lseek lseek
#define _tell(fildes) (lseek(fildes,0,SEEK_CUR))
#endif

#include <algorithm>

#ifdef _MSC_VER
// because the compiler wont give me 'min', so...
// even though it's suppose to come with <algorithm>

template <class _Tp>
inline const _Tp& min(const _Tp& __a, const _Tp& __b) {
  if (__b < __a) return __b; return __a;
}
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Elf32_Section::Elf32_Section()
{
  memset (&m_header,0,sizeof(m_header));
  m_pDataPtr = NULL;
  m_pWriteDest = NULL;
  
  m_wRelocSect = 0;
  
  m_pWriteHeader = NULL;
}

Elf32_Section::~Elf32_Section()
{
  delete m_pDataPtr;
  delete m_pWriteHeader;
}

void Elf32_Section::InitData (const Elf32_SectionHeader* header,
			      DataPtrSrc* pDataPtr)
{
  m_header = *header;
  
  assert(m_header.sh_type == ELF_SHT_NOBITS || pDataPtr);
  
  // if (m_header.sh_type != ELF_SHT_NOBITS)

  m_pDataPtr = pDataPtr;
}

void Elf32_Section::VerifyData()
{

}

void Elf32_Section::AddProgramBlock(Elf32_ProgramBlock* pProgramBlock)
{
  m_vectProgramBlocks.push_back(pProgramBlock);
}
/*
ELF_SHF_ALLOC_IF_LOAD
*/
/*
struct Elf32_SpecialSectionRights
{
	char name[10]; // must be first member
	word type;
	word flags;

    word wRelocBody;
};
*/
Elf32_SpecialSectionRights specialSectionRights[] =
{
  { ".bss",		ELF_SHT_NOBITS,		ELF_SHF_ALLOC | ELF_SHF_WRITE, 		REL_ELF_BSS_SECTION },
  { ".comment",		ELF_SHT_PROGBITS,	0 /* none */, 				REL_ELF_COMMENT },
  { ".data",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_WRITE, 		REL_ELF_DATA_SECTION },
  { ".data1",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_WRITE, 		0 },
  { ".debug",		ELF_SHT_PROGBITS,	0 /* none */, 				0 },
  { ".dynamic",		ELF_SHT_DYNAMIC,	ELF_SHF_ALLOC | ELF_SHF_WRITE /* I do not know if it is suppose to be write ?? */,
                                                                                        REL_ELF_DYNAMIC_SECTION },
  { ".dynstr",		ELF_SHT_STRTAB,		ELF_SHF_ALLOC, 				REL_ELF_DYNSTR_SECTION },
  { ".dynsym",		ELF_SHT_DYNSYM,		ELF_SHF_ALLOC, 				REL_ELF_DYNSYM_SECTION },
  { ".fini",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_EXECINSTR, 	REL_ELF_FINI_SECTION },
  { ".got",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_WRITE, 		REL_ELF_GOT_SECTION },
  { ".hash",		ELF_SHT_HASH,		ELF_SHF_ALLOC, 				REL_ELF_HASH_SECTION },
  { ".init",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_EXECINSTR, 	REL_ELF_INIT_SECTION },
  { ".interp",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC_IF_LOAD, 			REL_ELF_INTERP_SECTION },
  { ".line",		ELF_SHT_PROGBITS,	0 /* none */, 				0 },	 
  { ".note",		ELF_SHT_NOTE,		0 /* none */, 				0 },
  { ".plt",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_EXECINSTR, 	REL_ELF_PLT_SECTION },
  { ".rel*",		ELF_SHT_REL,		ELF_SHF_ALLOC_IF_LOAD, 			0 },
  { ".rela*",		ELF_SHT_RELA,		ELF_SHF_ALLOC_IF_LOAD, 			0 },
  { ".rodata",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC, 				REL_ELF_RODATA_SECTION },
  { ".rodata1",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC, 				0 },
  { ".shstrtab",	ELF_SHT_STRTAB,		0 /* none */, 				REL_ELF_SECT_HDR_STRING_TABLE },
  { ".strtab",		ELF_SHT_STRTAB,		ELF_SHF_ALLOC_IF_LOAD, 			0 },
  { ".symtab",		ELF_SHT_SYMTAB,		ELF_SHF_ALLOC_IF_LOAD, 			0 },
  { ".text",		ELF_SHT_PROGBITS,	ELF_SHF_ALLOC | ELF_SHF_EXECINSTR, 	REL_ELF_TEXT_SECTION },
};

/*
#define REL_ELF_NOTE_ABI_TAG_SECTION    (REL_ELF_INTERP_SECTION+1)
#define REL_ELF_HASH_SECTION            (REL_ELF_NOTE_ABI_TAG_SECTION+1)
#define REL_ELF_RELDYN_SECTION          (REL_ELF_DYNSTR_SECTION+1)
#define REL_ELF_RELPLT_SECTION          (REL_ELF_RELDYN_SECTION+1)
	   				     
// there is NO REL_TYPE_ELF_READ_EXECUTE_UNKNOWN1, because copying
// executable code unseen is not a good idea for us, as we have not
// checked it for jumps, data addresses, anything...
   	     
#define REL_ELF_COMMENT                 (REL_TYPE_ELF_UNALLOC_FIRST+1) // first read-write
#define REL_ELF_SECT_HDR_STRING_TABLE   (REL_ELF_COMMENT+1) // first user-defined read-write
#define REL_TYPE_ELF_UNALLOC_UNKNOWN1   (REL_ELF_SECT_HDR_STRING_TABLE+1) // first user-defined read-write
*/

int compare_spec_section_name(const void *elem1, const void *elem2)
{
  // both are strings if any string has a *, it means that the string
  // must match, but the other string might be longer
  
  const char *str1 = (const char*) elem1;
  const char *str2 = (const char*) elem2;
  
  for ( ; *str1 && *str2; str1++, str2++)
    {
      int diff = (int) *str1 - (int) *str2;
      if (diff)
	{
	  if (*str1 == '*' || *str2 == '*')
	    return 0;
	  return diff;
	}
    }
  int diff = (int) *str1 - (int) *str2;
  return diff;
}

#define count_of(array) (sizeof(array)/sizeof(array[0]))

const Elf32_SpecialSectionRights* GetSpecialSectionRights(const char* strName)
{
  const Elf32_SpecialSectionRights* rights;
  
  rights = (const Elf32_SpecialSectionRights*)
    bsearch(strName,
	    specialSectionRights,
	    count_of(specialSectionRights),
	    sizeof (specialSectionRights[0]),
                                compare_spec_section_name);
  
  return rights;
}

void Elf32_Section::SetSpecialSectionRights(const string& strName)
{
  m_strName = strName;

  const Elf32_SpecialSectionRights* rights = 
    GetSpecialSectionRights(strName.c_str());
  
  assert (!(rights->flags & ELF_SHF_ALLOC_IF_LOAD)); // or deal with it
  
  m_header.sh_type = rights->type;
  m_header.sh_flags = rights->flags;
}

void Elf32_Section::ValidateHeader()
{
  // if we have one of the above names, we should have to
  // corresponding rights set
  
  const Elf32_SpecialSectionRights* rights;
  
  if ((rights = GetSpecialSectionRights(m_strName.c_str())) != NULL)
    {
      WARNING2(m_header.sh_type == rights->type,
               "Section type does not match type required for given section name, %d != %d",m_header.sh_type,rights->type); // TODO: need to adjust type to be what it is suppose to be??, (changed VALIDATE to waring due to nasm,ld executable)
      
      Elf32_Word wFlags = rights->flags;
      
      if (wFlags & ELF_SHF_ALLOC_IF_LOAD)
        {
          wFlags &= ~ELF_SHF_ALLOC_IF_LOAD;
          
          // see if we're inside an LOADABLE program block
          
          FOR_EACH_CONST_ITER(VectProgramBlocks,m_vectProgramBlocks,pblock)
            {
              if ((*pblock)->m_header.p_type == ELF_PT_LOAD)
                {
                  wFlags |= ELF_SHF_ALLOC;
                  break;
                }
            }
        }
      
      assert (!(wFlags & ~ELF_SHF_ACCESS_MASK)); // so the next line does not require anything impossible
      
      WARNING2((m_header.sh_flags & ELF_SHF_ACCESS_MASK) == wFlags,
               "Section flags does not match flags required for given section name, %d != %d",(m_header.sh_flags & ELF_SHF_ACCESS_MASK),wFlags); // TODO: need to adjust type to be what it is suppose to be??, (changed VALIDATE to waring due to nasm,ld executable)
    }
  
  if (m_header.sh_addralign > 1)
    {
      // we have alignment requirements
      
      VALIDATE(powerof2(m_header.sh_addralign),
               "Section alignment requiremnt not power of 2.");
      
      VALIDATE(!(m_header.sh_addr % m_header.sh_addralign),
               "Section virtual address not aligned as specified.");
    }
  
  if (m_header.sh_entsize)
    {
      VALIDATE(!(m_header.sh_size % m_header.sh_entsize),
               "Section size is not multiple of entry size.");
    }
  
  WARNING(!(m_header.sh_flags & (ELF_SHF_WRITE | ELF_SHF_EXECINSTR)) ||
          m_header.sh_flags & ELF_SHF_ALLOC,
          "Section requires write or execute rights, but does note require alloc, which is strange.");
}

bool Elf32_Section::IsAddrInside(dword dwOffset,dword dwSize)
{
  if (!m_header.sh_addr)
    return false; // cannot be inside

  // if dwSize is zero, the size is not zero but unknown...
  
  if (dwOffset >= m_header.sh_addr &&
      dwOffset + dwSize <= m_header.sh_addr + m_header.sh_size)
    return true;
  
  return false;
}

bool Elf32_Section::ValidateCoExistence (FileMapItem* pOtherItem)
{
  // we cannot coexist with another section
  // we can coexist with an program block
  
  Elf32_ProgramBlock* pProgramBlock;
  
  if ((pProgramBlock = dynamic_cast<Elf32_ProgramBlock*>(pOtherItem)))
    {
      // our requested access rights should bo compatible with the
      // program block's
      
      // and we'd like to be completely inside the program block
      
      // pProgramBlock
      
      VALIDATE((pProgramBlock->m_header.p_offset <= m_header.sh_offset) &&
               (pProgramBlock->m_header.p_offset+
                pProgramBlock->m_header.p_filesz >= m_header.sh_offset+
                m_header.sh_size),
               "Section body not completely inside program block in file.");
      
      // a program block is always loaded into memory, so if the
      // section is here it should be too
      
      VALIDATE(m_header.sh_flags & ELF_SHF_ALLOC,
               "Section body inside program block not marked as loaded in memory (alloc).");
      
      VALIDATE((pProgramBlock->m_header.p_vaddr <= m_header.sh_addr) &&
               (pProgramBlock->m_header.p_vaddr+
                pProgramBlock->m_header.p_filesz >= m_header.sh_addr+
                m_header.sh_size),
               "Section body not completely inside program block in memory.");
      
      // we do not check ELF_SHF_ALLOC or ELF_PT_R flags
      
      // if a section want some right, the program block better
      // provide it
      
      VALIDATE(!(m_header.sh_flags & ELF_SHF_WRITE) ||
               pProgramBlock->m_header.p_flags & ELF_PT_W,
               "Program block enclosing section body does not give write access, but section needs it.");
      
      VALIDATE(!(m_header.sh_flags & ELF_SHF_EXECINSTR) ||
               pProgramBlock->m_header.p_flags & ELF_PT_X,
               "Program block enclosing section body does not give execute access, but section needs it.");
      
      return true; // it's ok
    }
  /*
    // This is not true: i forgot to check the NOBITS type, because if
    // that's set, then there is no size, and there can be no overlap
    
    if ((dynamic_cast<Elf32_Section*>(pOtherItem)))
    {
        // this is contrary to specification: it clearly state that no byte in the
        // file may belong to more than one section.
        // but as i found files violation this, we treat it with a warning, and accept
        // it

        EMIT_GENERIC_ERROR("Two Elf section bodies using the same in the executable file. "
            "This is violating the specification.");

        return true;
    }
*/

  return false;
  // return FileMapItem::ValidateCoExistence(pOtherItem);
}

string Elf32_Section::GetDescription()
{
  return "Section";
}

#include "HtmlPage.h"
#include "HtmlTags.h"

char tohex(byte value)
{
  return value < 10 ? '0' + value : 'a' - 10 + value;
}

#define FORMAT_RAW_DATA_SIZE_NEEDED(line,block) ((line)*3+(line)/(block)+3+1) /* line*2+line for hex and char, lne/blocks for block separators, +3 for ' | ' (normally only '| ' (if line divisible by block)), +1 for '\0'; */

void FormatRawData(char* pDest,const byte* pPtr,uint nLeft,uint nLine,uint nBlock)
{
#ifndef NDEBUG
  char* pDestEnd = pDest + FORMAT_RAW_DATA_SIZE_NEEDED(nLine,nBlock);
#endif
  
  const byte* p1 = pPtr;
  
  uint nWrite = min(nLeft,nLine);
  uint w = nWrite;
    
  for (uint b1 = nBlock; w ; w--, p1++)
    {
      *(pDest++) = tohex(*p1 >> 4);
      *(pDest++) = tohex(*p1 & 0x0f);
      if (!(--b1))
        {
	  *(pDest++) = ' ';
	  b1 = nBlock;
        }
    }
  
  if (nWrite < nLine)
    for (w = (nLine - nWrite) * 2 + nLine / nBlock - nWrite / nBlock; w; w--)
      *(pDest++) = ' ';
  
  if (nLine % nBlock)
    *(pDest++) = ' '; // no block separator was output last
  
  *(pDest++) = '|';
  *(pDest++) = ' ';
  
  p1 = pPtr;
  
  for (w = nWrite; w ; w--, p1++)
    {
      if (*p1 >= 0x20)
	*(pDest++) = *p1;
      else
	*(pDest++) = *p1 ? '.' : '@';
    }
  // maybe we should output spaces for padding if not full length ?
  *(pDest) = '\0';

  assert (pDest < pDestEnd); // < as there was no ++ when setting '\0'
}

#define MAX_FORMAT_RAW 0x8000

UrlFile* Elf32_Section::HTMLFileInfo(Elf32_Half nSectNo,UrlFolder* pParent)
{
  UrlFile* pFileSect = 
    pParent->AddFile (strsprintf("section%02d.html",nSectNo),
		      strsprintf("Section %d - %s",nSectNo,m_strName.c_str()));
  
  HtmlPage page(pFileSect);
  
  page.Header();
  
  DumpStructElf32_SectionHeader(m_header,page,text(m_strName));
  
  if (m_pDataPtr)
    {
      UrlFile* pFileRaw = 
	pParent->AddFile (strsprintf("section%02draw.html",nSectNo),
			  strsprintf("Section %d - %s - raw data",
				     nSectNo,m_strName.c_str()));
      
      page << (p() << a("Raw Data").href(pFileRaw));
      
      HtmlPage pageraw(pFileRaw);
      
      pageraw.Header();
      
      dword dwLineSize = m_header.sh_entsize ? m_header.sh_entsize : 16;
      
      const byte* ptr = (const byte*) m_pDataPtr->GetPtr(0,m_header.sh_size);
      
      pageraw.Begin(pre());
      
      const char* fmt_offset = 
	m_header.sh_flags & ELF_SHF_ALLOC ? "%06x %08x" : "%06x";
      
      char* fmtraw = new char[FORMAT_RAW_DATA_SIZE_NEEDED(dwLineSize,4)];
      
      uint nRawSize = min(m_header.sh_size,(uint) MAX_FORMAT_RAW);
      
      for (uint ind = 0; ind < nRawSize; ind += dwLineSize)
	{      
          pageraw << strsprintf(fmt_offset,
				m_header.sh_offset+ind,m_header.sh_addr+ind);
	  
          pageraw << " | ";
	  
          FormatRawData(fmtraw,ptr,nRawSize-ind,dwLineSize,4);
	  
          pageraw << fmtraw;
	  
          ptr += dwLineSize;
	  /*
	    const byte* p = ptr;

          uint nWrite = min(m_header.sh_size-ind,dwLineSize);

          uint w = nWrite;

          for ( ; w >= 4; w -= 4, p += 4)
          {
            sprintf (fmt,"%02x%02x%02x%02x ",*p,*(p+1),*(p+2),*(p+3));
            pageraw << fmt;
          }
          for ( ; w; w--, p++)
          {
            sprintf (fmt,"%02x",*p);
            pageraw << fmt;
          }

          for (w = 1 + (dwLineSize - nWrite) * 2 + dwLineSize / 4 - nWrite / 4; w; w--)
              pageraw << ' ';

          for ( ; nWrite; nWrite--, ptr++)
          {
              if (*ptr >= 0x20)
                pageraw << *ptr;
              else
                  pageraw << (*ptr ? '.' : '@');
          }
*/
          pageraw << '\n';      
	}
      
      delete[] fmtraw;
      
      if (nRawSize < m_header.sh_size)
        pageraw << 
	  strsprintf("\n... All not shown, only %d of %d bytes shown.",
		     nRawSize,m_header.sh_size);
      
      pageraw.End();
    }
  else
    page << (p("No raw data"));
  
  HTMLFileInfo(page);
  
  return pFileSect;
}

void Elf32_Section::HTMLFileInfo(HtmlPage& /* page */)
{
  
}

void Elf32_Section::SetWriteInfo (word wReloc)
{
  assert (m_wRelocSect == 0);

  m_wRelocSect = wReloc;
}
/*
void Elf32_Section::WriteSection ()
{
    CreateSection();

    m_header.sh_size = m_pWriteDest->m_pRelocDest->m_dwSize;
}
*/

RelocatableDest* Elf32_Section::WriteSection ()
{
  //   CreateSection();
  
  m_header.sh_size = m_pWriteDest->m_pRelocDest->m_dwSize;
  
  return m_pWriteDest->m_pRelocDest;
  
  //    return m_pWriteDest->ReleaseWrite();
}

/*
  void Elf32_Section::ReleaseWrite()
  {
  m_pWriteDest->ReleaseWrite();
  m_pWriteHeader->ReleaseWrite();
  }
*/
RelocatableDest* Elf32_Section::WriteHeader ()
{
  assert (!m_pWriteHeader);
  
  m_pWriteHeader = new WriteSectionHeader();
  
  m_pWriteHeader->AllocMemory(sizeof (Elf32_SectionHeader));
  /*
  Elf32_ProgramHeader header;

  memset(&header,0,sizeof (header));

  header.p_type  = type;
  header.p_flags = flags;
*/
  m_pWriteHeader->m_pDest->WriteData(&m_header,sizeof (m_header));
/*
  m_pWriteHeader->m_pDest->SetOffset(offsetof(Elf32_SectionHeader,sh_offset));
  m_pWriteHeader->m_pDest->WriteDWord(0,wRelocSect | RELATIVE_FILE);
*/
  m_pWriteHeader->m_pDest->InsertRelocation(offsetof(Elf32_SectionHeader,
						     sh_offset),
					    m_wRelocSect | RELATIVE_FILE);
  if (m_header.sh_flags & ELF_SHF_ALLOC)
    {
      /*
	m_pWriteHeader->m_pDest->SetOffset(offsetof(Elf32_SectionHeader,sh_addr));
	m_pWriteHeader->m_pDest->WriteDWord(0,wRelocSect | RELATIVE_IMAGE);
      */
      m_pWriteHeader->m_pDest->InsertRelocation(offsetof(Elf32_SectionHeader,
							 sh_addr),
						m_wRelocSect | RELATIVE_IMAGE);
    }
  
  return m_pWriteHeader->m_pRelocDest;
}


//////////////////////////////////////////////////////////////////////

Elf32_SectionRawData::Elf32_SectionRawData()
{
  m_pRawData = NULL;
}

Elf32_SectionRawData::Elf32_SectionRawData(const string& strName,
					   Elf32_Word type,Elf32_Word flags)
{
  m_pRawData = NULL;
  
  m_strName = strName;
  
  m_header.sh_type = type;
  m_header.sh_flags = flags;
}

void Elf32_SectionRawData::SetDataStr (const char* pStr)
{
  assert (!m_pRawData);
  
  m_header.sh_size = strlen(pStr)+1;
  m_header.sh_addralign = 1;
  
  m_pRawData = malloc (m_header.sh_size);
  
  memcpy (m_pRawData,pStr,m_header.sh_size);
}

Elf32_SectionRawData::~Elf32_SectionRawData()
{
  free (m_pRawData);
}

void Elf32_SectionRawData::InitData (const Elf32_SectionHeader* header,
				     DataPtrSrc* pDataPtr)
{
  Elf32_Section::InitData(header,pDataPtr);
  
  pDataPtr->ResetCurrent();
  
  m_pRawData = malloc (m_header.sh_size);
  
  pDataPtr->GetData(m_pRawData,m_header.sh_size);
}

void Elf32_SectionRawData::CreateSection()
{
  assert (!m_pWriteDest);
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(m_header.sh_size);
  
  m_pWriteDest->m_pDest->WriteData(m_pRawData,m_header.sh_size);    
  
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
}

//////////////////////////////////////////////////////////////////////

Elf32_SectionStringTable::Elf32_SectionStringTable()
{
  
}

Elf32_SectionStringTable::Elf32_SectionStringTable(const string& strName,
						   Elf32_Word flags)
{
  m_strName = strName;
  m_header.sh_type = ELF_SHT_STRTAB;
  m_header.sh_flags = flags;
}

void Elf32_SectionStringTable::VerifyData()
{
  // we now validate the actual data,
  // the first bytes are suppose to hold zero
  
  assert (m_pDataPtr);
  
  if (m_header.sh_size) // an empty string table is valid and need no check
    {
      VALIDATE(m_pDataPtr->GetByte(0) == 0,
               "String table first byte not zero");
      VALIDATE(m_pDataPtr->GetByte(m_header.sh_size-1) == 0,
               "String table last byte not zero");
    }
  
  // except for that, anything's allowed
}



void Elf32_SectionStringTable::HTMLFileInfo(HtmlPage& page)
{
  // simply walk through the table and display all strings
  
  Elf32_Section::HTMLFileInfo(page);
  
  page.Begin(pre());
  
  const char* ptr = (const char*) m_pDataPtr->GetPtr(0,m_header.sh_size);
  const char* endptr = (const char*) ptr + m_header.sh_size;

  uint nIndex = 0;
  
  char fmt[64];
  
  // m_setUsedIndices
  
  SetIndices::iterator iter = m_setUsedIndices.begin();
  
  while (ptr < endptr)
    {
      if (iter != m_setUsedIndices.end() && *iter == nIndex)
	{
          ++iter;
          sprintf (fmt,"0x%04x %5d+ ",nIndex,nIndex);
	}
      else
          sprintf (fmt,"0x%04x %5d  ",nIndex,nIndex);
      
      page << fmt << ptr << '\n';
      
      while (*ptr)
        {
          ptr++;
          nIndex++;
        }
      
      while (iter != m_setUsedIndices.end() && *iter < nIndex)
	{
          sprintf (fmt,"0x%04x %5d* ",*iter,*iter);
          // page << ("   " + min(*iter - nIndex,(uint) 3)); // spaces not seen unless fmt as raw
          page << fmt << (ptr + *iter - nIndex) << '\n';
          ++iter;
	}
      
      ptr++;  
      nIndex++;
    }
  
  page.End();
}

const char* Elf32_SectionStringTable::GetString(Elf32_Word nIndex)
{
  m_setUsedIndices.insert(/*(dword)*/ nIndex);
  
  if (nIndex == 0)
    return ""; // so we're also able to deal with empty string tables
  
  VALIDATE(nIndex < m_header.sh_size,
           "Attempt to get string from outside string table"); // would be caught be next statement also...
  
  return m_pDataPtr->GetStringPtr(nIndex); //  ((const char*) m_pDataPtr)+nIndex;
}
/*
class StringIndex
{
 public:
  StringIndex(const string& s,uint i) { str = s; ind = i; }

 public:
  string str;
  uint   index;
};

typedef set<Elf32_Word> SetIndices;
typedef vector<StringIndex> VectStringIndices;
*/
/*
Elf32_Word Elf32_SectionStringTable::InsertString(const string& str,bool bForceCreate)
{
    // this string will be needed in the to be string table

    uint l = str.length();

    if (!l) // is it the null string?
        return 0; // request fulfilled, even without string table (so ignore bForceCreate)

    // first see if we can fix the request with one of our existing ones;

    FOR_EACH_ITER(VectStringIndices,m_vectWantStrings,iter)
    {
        uint il = iter->str.length();

        if (il < l)
        {
            // the iter string may be the end of the requested one?

            if (il == 0)
            {
                if (iter->str.compare(str) == 0) // match
                {
                    if (bForceCreate && iter->index == (uint) -1)
                        CreateSection(iter);
                    return iter->index;
                }
            }
            else if (iter->index == (uint) -1) // may not replace string if it already has an index (i.e. has been written to string table)
            {
                if (iter->str.compare(str.c_str() + l - il) == 0)
                {
                    // so replace iter by string, because iter will still be able to fix
                    // any requests that might have required it

                    iter->str = str;

                    if (bForceCreate && iter->index == (uint) -1)
                        CreateSection(iter);
                    return iter->index; // fixed
                }
            }
#ifndef NDEBUG
            else
                printf ("StringTable : Performance hit, by requesting string after "
                        "part of string table has already been written (cannot merge "
                        "requests). Suppurted but not encouraged. Rewrite code to "
                        "request before CreateSection is called.");
#endif
        }
        else 
        {
            // iter string is longer (or equal) to request string

            if (str.compare(iter->str.c_str() + il - l) == 0)
            {
                if (bForceCreate && iter->index == (uint) -1)
                    CreateSection(iter);
                return iter->index + il - l; // fixed
            }
        }
    }

    if (m_vectWantStrings.empty())
        m_vectWantStrings.push_back(StringIndex("",0)); // we always have an empty string at the beginning

    m_vectWantStrings.push_back(StringIndex(str,(uint) -1));

    if (bForceCreate && m_vectWantStrings.back().index == (uint) -1)
        CreateSection();
    return m_vectWantStrings.back().index;
}
*/

void Elf32_SectionStringTable::RequestString(const string& str)
{
  // this string will be needed in the to be string table
  
  uint l = str.length();
  
  if (!l) // is it the null string?
    return; // request fulfilled, even without string table (so ignore bForceCreate)
  
  // first see if we can fix the request with one of our existing ones;
  
  FOR_EACH_ITER(VectStringIndices,m_vectWantStrings,iter)
    {
      uint il = iter->str.length();
      
      if (il < l)
        {
	  // the iter string may be the end of the requested one?
	  
	  if (il == 0)
            {
	      if (iter->str.compare(str) == 0) // match
		return;
            }
	  else if (iter->index == (uint) -1) // may not replace string if it already has an index (i.e. has been written to string table)
            {
	      if (iter->str.compare(str.c_str() + l - il) == 0)
                {
		  // so replace iter by string, because iter will
		  // still be able to fix any requests that might have
		  // required it
		  
		  assert(!m_pWriteDest);
		  iter->str = str;
		  return; // fixed
                }
            }
        }
      else 
        {
	  // iter string is longer (or equal) to request string
	  
	  if (str.compare(iter->str.c_str() + il - l) == 0)
            {
	      return;
	      // return iter->index + il - l; // fixed
            }
        }
    }
  
  assert(!m_pWriteDest);
  
  if (m_vectWantStrings.empty())
    m_vectWantStrings.push_back(StringIndex("",0)); // we always have an empty string at the beginning
  
  m_vectWantStrings.push_back(StringIndex(str,(uint) -1));
/*
  if (bForceCreate && m_vectWantStrings.back().index == (uint) -1)
  CreateSection();
  return m_vectWantStrings.back().index;
*/
}

Elf32_Word Elf32_SectionStringTable::GetStringIndex (const string& str)
{
  uint l = str.length();
  
  if (!l) // is it the null string?
    return 0; // request fulfilled, even without string table, so do not create one if unneeded
  
  assert (m_pWriteDest);
  
  // so we should be (part of) some string of the ones that were requested
  
  FOR_EACH_ITER(VectStringIndices,m_vectWantStrings,iter)
    {
      uint il = iter->str.length();
      
      if (il >= l && str.compare(iter->str.c_str() + il - l) == 0)
        {
	  return iter->index + il - l; // found
        }
    }
  assert (false); // ehhh, we were not requested!
  return 0; // so give them an empty one (for now), but fix request before creation of string table
  
  // return InsertString(str,true);
}
/*
void Elf32_SectionStringTable::CreateSection(VectStringIndices::iterator endcreate)
{
    // commit strings to string table up till string with iterator endcreate
    // this will fix the indices of those strings

    if (!m_pWriteDest)
    {
        // we're not created at all
        // fix that

        CreateSection();
        return;
    }

    assert (false);

    // if you came here, you added strings after CreateSection
    // or wanted to index of one before they were commited

    // then you have to write the code for appending strings!

    // or we should really rid the code of the possibility of doing this at all
    // as it precludes the ability to optimize the string table, and the
    // only thing we get in return is that you, the programmer, can be sloppy...

    // of course, one could of course argue for an version that have no chance at
    // all to optimize, because it simply insert any string needed, and then
    // does not require the extra memory to keep requested strings in a vector
    // but that's another story, suitable for debugging-symbol tables, which are
    // not our major concern here (our optimizations should make debugging very hard,
    // and if they don't, they are not clever enough :-) )

    // so, he he, we asserted false, and now we die

    *((char*) 0) = 0;

    // ho, ho, as you can see this is all commented out, so I fell for the optimize-only version
}
*/

void Elf32_SectionStringTable::CreateSection()
{
  if (m_pWriteDest)
    {
      return;
      /*
        // we've already been written (or at least part of us)
        // this is an request to commit the rest
	
        CreateSection (m_vectWantStrings.end());
        return;
      */
    }   
  
  // figure out how much memory is required for the requested strings...
  
  uint nLen = 0;
  
  FOR_EACH_ITER(VectStringIndices,m_vectWantStrings,iter)
    nLen += iter->str.length();
  nLen += m_vectWantStrings.size(); // each string also requires an null terminator
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(nLen);
  
  FOR_EACH_ITER(VectStringIndices,m_vectWantStrings,iter2) // iter2 because some compilers are unaware of scoping
    {
      iter2->index = m_pWriteDest->m_pDest->GetOffset();
        m_pWriteDest->m_pDest->WriteString(iter2->str);    
    }
  
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
}

//////////////////////////////////////////////////////////////////////

Elf32_SectionHashTable::Elf32_SectionHashTable(const string& strName)
{
  m_pSymbolSection = NULL;
  SetSpecialSectionRights(strName);
}

void Elf32_SectionHashTable::SetSymbolTable (Elf32_SectionSymbolTable* pSection)
{ 
  assert (!m_pSymbolSection); 
  m_pSymbolSection = pSection;
  
  VALIDATE(pSection->m_pHashTable == NULL,
	   "Symbol table linked to by hash table already has an hash table.");
  
  pSection->m_pHashTable = this;
}

void Elf32_SectionHashTable::VerifyData()
{
  // the hash table first has two entries with the number of
  // buckets and the number of chain entries, and then come the buckets
  // and then the chain
  
  // first of all, the sizes should agree
  
  assert (m_pDataPtr);
  
  m_pDataPtr->ResetCurrent();
  
  const Elf32_Hash* pHash = 
    (const Elf32_Hash*) m_pDataPtr->GetPtr(sizeof (Elf32_Hash));
  
  VALIDATE(sizeof (Elf32_Hash) + (pHash->h_buckets + pHash->h_chains) * sizeof (Elf32_Word) == m_header.sh_size,
	   "Hash table size does not agree with number of buckets and chains");
  
  const Elf32_Word* pBuckets = 
    (const Elf32_Word*) m_pDataPtr->GetPtr(sizeof (Elf32_Word) * 
					   pHash->h_buckets);
  const Elf32_Word* pChains = 
    (const Elf32_Word*) m_pDataPtr->GetPtr(sizeof (Elf32_Word) * 
					   pHash->h_chains);
  
  // now make sure that all chains do end (no infinite loops), and
  // that no two buckets point to the same chain (except perhaps for
  // chain containing 0)
  
  const Elf32_Word* pBucket = pBuckets;
  
  BitField bfUseChain(pHash->h_chains);
  
  for (Elf32_Word nb = 0; nb < pHash->h_buckets; nb++, pBucket++)
    {
      if (*pBucket)
	VALIDATE1(*pBucket < pHash->h_chains,
		  "Hash table bucket %d1 value larger than chain size.",nb);
      
      // now follow the chain
      
      Elf32_Word nChain = *pBucket;
      
      while (nChain)
        {
	  // the following will also find loops, so be careful if removing it
	  VALIDATE2(!bfUseChain.GetBit(nChain),
		    "Hash table bucket %d1 point into chain %d2, which has already been pointed to.",nb,nChain);
	  
	  bfUseChain.SetBit(nChain);
	  
	  VALIDATE1(*(pChains+nChain) < pHash->h_chains,
		    "Hash table chain %d1 value larger than chain size.",nChain);
	  
	  nChain = *(pChains+nChain);
        }
    }
  
  const Elf32_Word* pChain = pChains;
  
    for (Elf32_Word nc = 0; nc < pHash->h_chains; nc++, pChain++)
      {
        VALIDATE1(!*pChain || bfUseChain.GetBit(nc),
		  "Hash table chain %d1 unused, but not zero.",nc);
      }
}

uint BucketSearches(Elf32_Word* pHash,uint nItems,uint nBuckets,uint* pHits)
{
  memset (pHits,0,sizeof (uint) * nBuckets);

  for ( ; nItems; --nItems, ++pHash)
    (*(pHits + ((*pHash) % nBuckets)))++;
  
  uint nHits = 0;
  
  for ( ; nBuckets; --nBuckets, ++pHits)
    nHits += *pHits * (*pHits + 1);    
  
  return nHits;
}

void Elf32_SectionHashTable::CreateSection()
{
  // will be done when overloaded function below is called by symbol
  // table
}


void Elf32_SectionHashTable::CreateSection (Elf32_Word* pHash,uint nItems,uint nMaxBuckets,uint nExtraSearch/* = 3*/)
{
  // this uses a semi-scientific way of reducing number of buckets set
  // nExtraSearch to 0 to only allow for non-performance-hit
  // adjustements
  
  // build a hash table with maximum nMaxBuckets but if we get less
  // fill in with a smaller number of buckets, use that instead, the
  // fill in is counted as the number of compare operations that would
  // be required for a lookup of each item the hash represents once,
  // so a bucket with zero entries contribute with 0, one entry 1, two
  // entries 2+1=3, 3 entries 3+2+1=6
  
  // contrib = n*(n+1)/2 and as there is no rounding because of the
  // /2, we can omit it
  
  // then I've also added the rule of thumb: if saving one bucket
  // causes nExtraSearch more searches, it's better anyway (we save
  // memory)
  
  m_header.sh_entsize = sizeof (Elf32_Word);
  
  nExtraSearch *= 2; // to adjust for the 2 we never divide by
  
  if (nMaxBuckets > (nItems * (nItems + 1)) / nExtraSearch)
    {
      // worst case: all items in one bucket, will be found when
      // trying one bucket and that would beat anything larger than
      // what we set maxbuckets to, so let's set it to this
      
      nMaxBuckets = nItems * (nItems + 1);
    }

  assert (nMaxBuckets > 0);
  
  uint* nHits = new uint[nMaxBuckets];
  
  uint nBestBuckets = nMaxBuckets;
  uint nBestSearch = (uint) -1;
  
  uint nBadDirection = 0; // count how bad we look, and abort if we do not seem to ever be able to beat best value
  
  for (uint nBuckets = nMaxBuckets; nBuckets; nBuckets--)
    {
      uint nSearch = BucketSearches(pHash,nItems,nBuckets,nHits);
      
      if (nSearch > (nBestSearch + 
		     (nBestBuckets - nBuckets) * (1+2*nExtraSearch)))
        {
	  if (++nBadDirection > 3)
	    break; // we'll probably not find any better ordering anyway, break out
        }
      else
        {
	  nBadDirection = 0;
	  if (nSearch <= (nBestSearch + 
			  (nBestBuckets - nBuckets) * nExtraSearch))
            {
	      nBestBuckets = nBuckets;
	      nBestSearch = nSearch;
            }
        }
    }
  
  free(nHits);
  
  // so use nBestBuckets buckets
  
  assert (!m_pWriteDest);
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(sizeof (Elf32_Word) * (2 + nBestBuckets + nItems));
  
  m_pWriteDest->m_pDest->WriteDWord(nBestBuckets);
  m_pWriteDest->m_pDest->WriteDWord(nItems);
  
  Elf32_Word* pBucket = 
    (Elf32_Word*) m_pWriteDest->m_pDest->WritePtr(sizeof (Elf32_Word) * 
						  nBestBuckets);
  Elf32_Word* pChain  = 
    (Elf32_Word*) m_pWriteDest->m_pDest->WritePtr(sizeof (Elf32_Word) * 
						  nItems);
  
  // this code works for an input list where entry 0 has hash of zero
  // and is unused, it's just put into position zero and will later be
  // ignored, any other item having hash 0 will be regarded as used so
  // it does not work for entry 0 having hash zero because used string
  // (empty or non-empty) gave hash zero, or for any item with unused
  // entry just having set hash to zero (bucket zero, will trace all
  // those occurances)
  
  for (uint i = 0; i < nItems; i++)
    {
      // insert *(pHash) pointing to item i
      
      uint nBucket = *(pHash) % nBestBuckets;
      
      if (*(pBucket + nBucket))
        {
	  // follow the chain and set the last chain entry to us (as
	  // we're operating with increasing i, this is the way to do
	  // it to get an ever-increasing chain)
	  
	  nBucket = *(pBucket + nBucket);
	  
	  while (*(pChain + nBucket))
	    nBucket = *(pChain + nBucket);
	  
	  *(pChain + nBucket) = i;
        }
      else
        {
	  // previously unused, just set it
	  
	  *(pBucket + nBucket) = i;
        }
    }
  
  
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
}

unsigned long
elf_hash(const unsigned char *name)
{
  unsigned long h = 0, g;
  
  while(*name)
    {
      h = (h << 4) + (unsigned char) (*name++);
      if((g = h & 0xf0000000))
	h ^= g >> 24;
      h &= ~g;
    }
  return h;
}

//////////////////////////////////////////////////////////////////////

Elf32_SectionSymbolTable::Elf32_SectionSymbolTable(const string& strName,
						   Elf32_SectionStringTable* pStringSection,
						   Elf32_SectionHashTable* pHashTable)
{
  assert (pStringSection);
  assert (pHashTable);
  
  m_pStringSection = pStringSection;
  m_pHashTable = pHashTable; 
  pHashTable->m_pSymbolSection = this;
  
  SetSpecialSectionRights(strName);
  
  m_vectSymbolItems.push_back(new SymbolItem()); // insert the empty symbol item 0
}

Elf32_Word Elf32_SectionSymbolTable::GetSymbolIndex (const char* name)
{
  unsigned long nHash = elf_hash(name);
  
  assert (m_pHashTable);
  assert (m_pHashTable->m_pDataPtr);
  
  const Elf32_Hash* pHash = 
    (const Elf32_Hash*) m_pHashTable->m_pDataPtr->GetPtr(0,
							 sizeof (Elf32_Hash));
  
  //    VALIDATE(sizeof (Elf32_Hash) + (pHash->h_buckets + pHash->h_chains) * sizeof (Elf32_Word) == m_header.sh_size,
  //        "Hash table size does not agree with number of buckets and chains");
  
  assert (pHash->h_buckets);
  
  const Elf32_Word* pBuckets = 
    (const Elf32_Word*) m_pHashTable->m_pDataPtr->GetPtr(sizeof (Elf32_Hash),
							 sizeof (Elf32_Word) *
							 pHash->h_buckets);
  const Elf32_Word* pChains = 
    (const Elf32_Word*) m_pHashTable->m_pDataPtr->GetPtr(sizeof (Elf32_Hash) + 
							 sizeof (Elf32_Word) * 
							 pHash->h_buckets,
							 sizeof (Elf32_Word) *
							 pHash->h_chains);
  
  Elf32_Word nChain = *(pBuckets + (nHash % pHash->h_buckets));
  
  assert (nChain < pHash->h_chains);
  
  do
    {
      const Elf32_Symbol* pEntry = 
	(const Elf32_Symbol*) m_pDataPtr->GetPtr(nChain * 
						 sizeof (Elf32_Symbol),
						 sizeof (Elf32_Symbol));
      
      const char* pName = 
	pEntry->st_name ? m_pStringSection->GetString(pEntry->st_name) : NULL;
      
      if (strcmp(name,pName) == 0)
	break;
      
      nChain = *(pChains+nChain); // was validated in hash table verification
    }
  while (nChain);
  
  return nChain;
}

void Elf32_SectionSymbolTable::VerifyData()
{
  VALIDATE(m_pStringSection,
	   "Symbol table does not have an associated string table.");
  WARNING(m_pHashTable,
	  "Symbol table does not have an associated hash table.");
  
  m_pStringSection->VerifyData();
  if (m_pHashTable)
    m_pHashTable->VerifyData();
  
  VALIDATE(m_header.sh_entsize == sizeof (Elf32_Symbol),
	   "Symbol table entries size wrong.");

  uint nEntries = m_header.sh_size / m_header.sh_entsize;
  
  VALIDATE(m_header.sh_info <= nEntries,
	   "Symbol table header info (end of locals) larger than number of entries.");
  
  assert (!m_pHashTable || m_pHashTable->m_pDataPtr);
  
  const Elf32_Hash* pHash = m_pHashTable ? (const Elf32_Hash*) m_pHashTable->m_pDataPtr->GetPtr(0,sizeof (Elf32_Hash)) : NULL;
  
  VALIDATE(!m_pHashTable || pHash->h_chains == nEntries,
	   "Symbol table number of entries and hash table chain size does not agree.");
  
  assert (m_pDataPtr);
  
  m_pDataPtr->ResetCurrent();

  for (uint nEntry = 0; nEntry < nEntries; nEntry++)
    {
      const Elf32_Symbol* pEntry = 
	(const Elf32_Symbol*) m_pDataPtr->GetPtr(sizeof (Elf32_Symbol));
      
      if (nEntry == 0)
        {
          VALIDATE(iszero(pEntry,sizeof(Elf32_Symbol)),
                   "Symbol table entry 0 not filled with 0s.");
        }
      
      const char* pName = 
	pEntry->st_name ? m_pStringSection->GetString(pEntry->st_name) : NULL;
      
      VALIDATE1(!pName || !m_pHashTable || GetSymbolIndex(pName) == nEntry,
		"Symbol table entry %d1, cannot do reverse lookup with hash table (hash wrong or multiple items of same name)",nEntry);
      
      /* value & size not checked, but should be checked to see that they are
	 within bounds of the executable*/
      
      WARNING1(pEntry->st_other == 0,
	       "Symbol table entry %d1 member other not 0",nEntry);
      
      switch (ELF32_ST_TYPE(pEntry->st_info))
	{
	case ELF_STT_FILE:
          VALIDATE1(ELF32_ST_BIND(pEntry->st_info) == ELF_STB_LOCAL,
		    "Symbol table entry %d1, is file symbol, but not local.",nEntry);
          VALIDATE1(pEntry->st_shndx == ELF_SHN_ABS,
		    "Symbol table entry %d1, is file symbol, but section ref is not abs.",nEntry);
          WARNING1(nEntry == 1 || ELF32_ST_TYPE((pEntry-1)->st_info) == ELF_STT_FILE,
		   "Symbol table entry %d1, is file symbol, but does not precede other local symbols.",nEntry);
          break;
	}

      if (ELF32_ST_BIND(pEntry->st_info) == ELF_STB_LOCAL)
	{
          VALIDATE1(nEntry < m_header.sh_info,
		    "Symbol table entry %d1 is marked as local, but has index > end of locals (symbol table header info).",nEntry)
	    }
      else
	{
          /*ELF_STB_GLOBAL, or ELF_STB_WEAK , or other undef value*/
          VALIDATE1(nEntry >= m_header.sh_info,
		    "Symbol table entry %d1 is marked as non-local, but has index > end of locals (symbol table header info).",nEntry)
	    }
      
      m_vectSymbolItems.push_back(new SymbolItem(pName ? pName : "",*pEntry));
      
      /*
	Elf32_SectionStringTable* m_pStringSection;
	Elf32_SectionHashTable* m_pHashTable;
      */
    }
  
  // and now check the hash-table so that all the buckets have entries
  // that have the correct hash code
  
  if (m_pHashTable)
    {
      const Elf32_Word* pBuckets = 
	(const Elf32_Word*) 
	m_pHashTable->m_pDataPtr->GetPtr(sizeof (Elf32_Hash),
					 sizeof (Elf32_Word) * 
					 pHash->h_buckets);
      const Elf32_Word* pChains = 
	(const Elf32_Word*) 
	m_pHashTable->m_pDataPtr->GetPtr(sizeof (Elf32_Hash) + 
					 sizeof (Elf32_Word) * 
					 pHash->h_buckets,
					 sizeof (Elf32_Word) * 
					 pHash->h_chains);
      
      // now make sure that all chains do end (no infinite loops), and
      // that no two buckets point to the same chain (except perhaps
      // for chain containing 0)
      
      const Elf32_Word* pBucket = pBuckets;
      
      for (Elf32_Word nb = 0; nb < pHash->h_buckets; nb++, pBucket++)
        {
          if (*pBucket)
            {
              assert (*pBucket < pHash->h_chains);
              
              Elf32_Word nChain = *pBucket;
              
              do
                {
                  // now get the name of the entry specified, and make
                  // sure it's hash agree with the bucket number
		  
                  const Elf32_Symbol* pEntry = 
		    (const Elf32_Symbol*) 
		    m_pDataPtr->GetPtr(nChain * sizeof (Elf32_Symbol),
				       sizeof (Elf32_Symbol));
                  
                  const char* pName =
		    pEntry->st_name ? 
		    m_pStringSection->GetString(pEntry->st_name) : NULL;
                  
                  VALIDATE1(pName,
                            "Symbol table entry %d1 has no name (empty), but pointed to by hash table.",nChain);
                  
                  VALIDATE2(elf_hash(pName) % pHash->h_buckets == nb,
                            "Symbol table entry %d1 is pointed to by hash bucket %d2, but name hash is wrong for that.",nChain,nb);
		  
                  assert (*(pChains+nChain) < pHash->h_chains); // was validated in hash table verification
                  
                  nChain = *(pChains+nChain);
                }
              while (nChain);
            }
        }
    }
}

void Elf32_SectionSymbolTable::RequestStrings()
{
  FOR_EACH_ITER(VectSymbolItems,m_vectSymbolItems,symbol)
    {
      m_pStringSection->RequestString((*symbol)->m_strName); 
    }
}

void Elf32_SectionSymbolTable::CreateSection()
{
  assert (!m_pWriteDest);
  
  m_pWriteDest = new WriteRelocDest();
  
  assert (m_vectSymbolItems.size()); // at least one item
  
  m_pWriteDest->AllocMemory(m_vectSymbolItems.size() * sizeof (Elf32_Symbol));
  
  m_header.sh_entsize = sizeof (Elf32_Symbol);
  /*
    Elf32_SectionStringTable* m_pStringSection;
    Elf32_SectionHashTable* m_pHashTable;
  */
  
  assert (m_header.sh_info == 0); // kept from source file (reset here?)
  
  uint nItems = m_vectSymbolItems.size();
  uint nItem = 0;
  
  Elf32_Word* hashes = new uint[nItems];
  
  FOR_EACH_ITER(VectSymbolItems,m_vectSymbolItems,iter)
    {
      Elf32_Symbol symbol;
      
      memset (&symbol,0,sizeof (symbol));
      
      hashes[nItem] = elf_hash((*iter)->m_strName.c_str());
      
      symbol.st_name = m_pStringSection->GetStringIndex((*iter)->m_strName);
      symbol.st_value = (*iter)->m_dwValue;
      symbol.st_size = (*iter)->m_dwSize;
      symbol.st_info = (*iter)->m_bInfo;
      symbol.st_shndx = (*iter)->m_wSection;
      
      if (ELF32_ST_TYPE(symbol.st_info) == ELF_STB_LOCAL)
	{
	  assert (m_header.sh_info == nItem); // so that all locals come first
	  m_header.sh_info = nItem+1;
	}
      
      m_pWriteDest->m_pDest->WriteData(&symbol,sizeof (symbol));
      nItem++;
    }
  
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
  
  m_pHashTable->CreateSection(hashes,nItems,nItems > 25 ? nItems / 5 : 5);
  
  delete[] hashes;
}

void Elf32_SectionSymbolTable::HTMLFileInfo(HtmlPage& page)
{
  // simply walk through the table and display all strings
  
  Elf32_Section::HTMLFileInfo(page);
  
  assert (m_pDataPtr);
  
  uint nEntries = m_header.sh_size / m_header.sh_entsize;
  
  page.Begin(table()/*.border(0)*/);
  
  page.Begin(tr());
  page << td("#");
  DumpStructTableHeaderElf32_Symbol(page);
  page.End();

  assert (m_pStringSection);
  
  m_pDataPtr->ResetCurrent();
  
  for (uint nEntry = 0; nEntry < nEntries; nEntry++)
    {
      const Elf32_Symbol* pEntry = 
	(const Elf32_Symbol*) m_pDataPtr->GetPtr(sizeof (Elf32_Symbol));
      const char* pName = m_pStringSection->GetString(pEntry->st_name);
      
      page.Begin(tr());
      page << td(strsprintf("%d",nEntry));
      DumpStructTableElf32_Symbol(*pEntry,page,text(pName));
      page.End();
    }     
  
  page.End();
}

/*
dword TryFillBuckets(dword dwBuckets,dword* dwKeys,dword dwNumKeys)
{
    dword* pBucketUse = new dword[dwBuckets];
    memset (pBucketUse,0,sizeof (dword) * dwBuckets);

    for ( ; dwNumKeys; dwNumKeys--, dwKeys++)
        (*(pBucketUse+((*(dwKeys++)) % dwBuckets)))++;

    dword* pCnt = pBucketUse;

    dword dwHitCount = 0;
    // the larger, the worse. We return twice the number of compares you have to
    // do if you search for each symbol in the hash table

    for ( ; dwBuckets; dwNumKeys--, pCnt++)
        dwHitCount += (*pCnt)*(*pCnt-1);

    delete pBucketUse;

    return dwHitCount / 2;
}

dword OptimalHashNBuckets(dword* dwKeys,dword dwNumKeys,
                          float fTradeMemoryForSearchSpeed,
                          float fMaxBucketsPerEntry)
{
    // return a good bucket size so that the keys are spread good, but we do not
    // use too much memory

    // we get the keys as a list
    // it's strongly suggested that any doublet keys are removed before calling us,
    // because there is no chance anyway that we can eliminate hits in the same bucket
    // when the keys are identical

    // fTradeSearchForMemory tell how important it is to have a good hash table
    // 1 mean that we will accept one extra entry instead of 1 extra search
    // default is 0.1

    // fMaxBucketsPerEntry tell how many buckets we'll accept as maximum
    // default is 0.5

    assert (fTradeMemoryForSearchSpeed <= 1.0);
    assert (fMaxBucketsPerEntry <= 5.0); // no crazy values please

    dword dwMaxBuckets = 1 + dwNumKeys * fMaxBucketsPerEntry;
    if (dwMaxBuckets > dwNumKeys)
        dwMaxBuckets = dwNumKeys;


    dword dwBestBuckets = 1;
    dword dwBestHitCount = dwNumKeys*(dwNumKeys-1);

    for (dword dwBuckets = 1; dwBuckets < dwMaxBuckets; dwBuckets++)
    {
        dword dwHitCount = TryFillBuckets(dwBuckets,dwKeys,dwNumKeys);

        if (dwHitCount * fTradeMemoryForSearchSpeed < dwBuckets)
            return dwBuckets;

        if (dwHitCount < dwBestHitCount)
        {
            dwBestHitCount = dwHitCount;
            dwBestBuckets = dwBuckets;
        }
    }    
    return dwBestBuckets;
}
*/

#define DI_NUMBER_MASK  0x00ff

#define DI_D_IGN        0       // dummy
#define DI_D_VAL        0x0100  // d_un.d_ptr used
#define DI_D_PTR        0x0200  // d_un.d_ptr used
#define DI_D_MASK       0x0300

#define DI_IGN          0       // dummy
#define DI_MAND         0x01    // required
#define DI_OPT          0x02    // optional
#define DI_MAND_OPT_MASK 0x03

#define DI_EXEC_SHIFT   10
#define DI_SHRD_SHIFT   12

#define DI_EXEC_IGN     (DI_IGN  << DI_EXEC_SHIFT)
#define DI_EXEC_MAND    (DI_MAND << DI_EXEC_SHIFT)
#define DI_EXEC_OPT     (DI_OPT  << DI_EXEC_SHIFT)

#define DI_SHRD_IGN     (DI_IGN  << DI_SHRD_SHIFT)
#define DI_SHRD_MAND    (DI_MAND << DI_SHRD_SHIFT)
#define DI_SHRD_OPT     (DI_OPT  << DI_SHRD_SHIFT)

#define DI_MULT         0x4000 // may have multiple entries
#define DI_SNGL         0      // dummy (only one entry)

#define DI_MAND_REL_RELA 0x8000 // either REL or RELA is required

#define DI_REQ_NUM_MASK 0x3f

#define DI_REQUIRE      0x80    // does require
#define DI_SHLD_REQ     0x40    // should require (would be logical) (but spec does not say so)

struct DynamicEntryInfo
{
  word    m_wInfo; // the entry number, val/ptr, exec, shrd, mult
  
  byte    m_bRequires[2];
  
 //   word    m_wRelocInfo;
};
/*
#define REL_ELF_IMAGE_HASH_SECTION            	(RELATIVE_IMAGE | REL_ELF_HASH_SECTION)
#define REL_ELF_IMAGE_DYNSYM_SECTION            	(RELATIVE_IMAGE | REL_ELF_DYNSYM_SECTION)
#define REL_ELF_IMAGE_DYNSTR_SECTION            	(RELATIVE_IMAGE | REL_ELF_DYNSTR_SECTION)

#define REL_ELF_IMAGE_INIT_SECTION            	(RELATIVE_IMAGE | REL_ELF_INIT_SECTION)
#define REL_ELF_IMAGE_FINI_SECTION            	(RELATIVE_IMAGE | REL_ELF_FINI_SECTION)
#define REL_ELF_IMAGE_PLT_SECTION               	(RELATIVE_IMAGE | REL_ELF_PLT_SECTION)
#define REL_ELF_IMAGE_DYNAMIC_SECTION           (RELATIVE_IMAGE | REL_ELF_DYNAMIC_SECTION)
#define REL_ELF_IMAGE_GOT_SECTION              	(RELATIVE_IMAGE | REL_ELF_GOT_SECTION)

#define REL_ELF_IMAGE_RELDYN_SECTION                  (RELATIVE_IMAGE | REL_ELF_RELDYN_SECTION)
#define REL_ELF_IMAGE_RELPLT_SECTION              	(RELATIVE_IMAGE | REL_ELF_RELPLT_SECTION)
*/
DynamicEntryInfo dynamicEntryInfo[] =
{
    { ELF_DT_NULL     | DI_D_IGN | DI_EXEC_IGN  | DI_SHRD_IGN  | DI_MULT, { 0,0 } }, // this entry is mandatory, but that is handled elsewhere
    { ELF_DT_NEEDED   | DI_D_VAL | DI_EXEC_OPT  | DI_SHRD_OPT  | DI_MULT, { 0,0 } },
    { ELF_DT_PLTRELSZ | DI_D_VAL | DI_EXEC_OPT  | DI_SHRD_OPT,            { 0,0 } },
    { ELF_DT_PLTGOT   | DI_D_PTR | DI_EXEC_OPT  | DI_SHRD_OPT,            { 0,0 }/*, REL_ELF_IMAGE_GOT_SECTION*/ },
    { ELF_DT_HASH     | DI_D_PTR | DI_EXEC_MAND | DI_SHRD_MAND,           { 0,0 }/*, REL_ELF_IMAGE_HASH_SECTION*/ },
    { ELF_DT_STRTAB   | DI_D_PTR | DI_EXEC_MAND | DI_SHRD_MAND,           { 0,0 }/*, REL_ELF_IMAGE_DYNSTR_SECTION*/ },
    { ELF_DT_SYMTAB   | DI_D_PTR | DI_EXEC_MAND | DI_SHRD_MAND,           { 0,0 }/*, REL_ELF_IMAGE_DYNSYM_SECTION*/ },
    { ELF_DT_RELA     | DI_D_PTR | DI_EXEC_MAND | DI_SHRD_OPT | DI_MAND_REL_RELA, { DI_REQUIRE | ELF_DT_RELASZ, DI_REQUIRE | ELF_DT_RELAENT } },
    { ELF_DT_RELASZ   | DI_D_VAL | DI_EXEC_MAND | DI_SHRD_OPT | DI_MAND_REL_RELA, { DI_SHLD_REQ | ELF_DT_RELA,0 } }, 
    { ELF_DT_RELAENT  | DI_D_VAL | DI_EXEC_MAND | DI_SHRD_OPT | DI_MAND_REL_RELA, { DI_SHLD_REQ | ELF_DT_RELA,0 } },
    { ELF_DT_STRSZ    | DI_D_VAL | DI_EXEC_MAND | DI_SHRD_MAND,           { 0,0 } },
    { ELF_DT_SYMENT   | DI_D_VAL | DI_EXEC_MAND | DI_SHRD_MAND,           { 0,0 } },
    { ELF_DT_INIT     | DI_D_PTR | DI_EXEC_OPT  | DI_SHRD_OPT,            { 0,0 }/*, REL_ELF_IMAGE_INIT_SECTION*/ },
    { ELF_DT_FINI     | DI_D_PTR | DI_EXEC_OPT  | DI_SHRD_OPT,            { 0,0 }/*, REL_ELF_IMAGE_FINI_SECTION*/ },
    { ELF_DT_SONAME   | DI_D_VAL | DI_EXEC_IGN  | DI_SHRD_OPT,            { 0,0 } },
    { ELF_DT_RPATH    | DI_D_VAL | DI_EXEC_OPT  | DI_SHRD_IGN,            { 0,0 } },
    { ELF_DT_SYMBOLIC | DI_D_IGN | DI_EXEC_IGN  | DI_SHRD_OPT,            { 0,0 } },
    { ELF_DT_REL      | DI_D_PTR | DI_EXEC_MAND | DI_SHRD_OPT | DI_MAND_REL_RELA, { DI_REQUIRE | ELF_DT_RELSZ, DI_REQUIRE | ELF_DT_RELENT } },
    { ELF_DT_RELSZ    | DI_D_VAL | DI_EXEC_MAND | DI_SHRD_OPT | DI_MAND_REL_RELA, { DI_SHLD_REQ | ELF_DT_REL,0 } },
    { ELF_DT_RELENT   | DI_D_VAL | DI_EXEC_MAND | DI_SHRD_OPT | DI_MAND_REL_RELA, { DI_SHLD_REQ | ELF_DT_REL,0 } },
    { ELF_DT_PLTREL   | DI_D_VAL | DI_EXEC_OPT  | DI_SHRD_OPT,            { 0,0 } },
    { ELF_DT_DEBUG    | DI_D_PTR | DI_EXEC_OPT  | DI_SHRD_IGN,            { 0,0 } },
    { ELF_DT_TEXTREL  | DI_D_IGN | DI_EXEC_OPT  | DI_SHRD_OPT,            { 0,0 } }, 
    { ELF_DT_JMPREL   | DI_D_PTR | DI_EXEC_OPT  | DI_SHRD_OPT,            { DI_REQUIRE | ELF_DT_PLTRELSZ, DI_REQUIRE | ELF_DT_PLTREL }/*, REL_ELF_IMAGE_RELPLT_SECTION*/ },
};

uint rel_to_rela(uint i)
{
  assert ((i >= ELF_DT_REL && i <= ELF_DT_RELENT) ||
	  (i >= ELF_DT_RELA && i <= ELF_DT_RELAENT));
  
  if (i >= ELF_DT_RELA)
    return i - ELF_DT_RELA + ELF_DT_REL;
  else
    return i - ELF_DT_REL + ELF_DT_RELA;
}

Elf32_SectionDynamic::Elf32_SectionDynamic(Elf32_SectionStringTable* pStringSection)
{
  assert (pStringSection);
  
  m_pStringSection = pStringSection;
  
  // the following should be taken from the
  
  SetSpecialSectionRights(".dynamic");
}

void Elf32_SectionDynamic::VerifyData()
{
  assert (count_of(dynamicEntryInfo) == ELF_DT_NUM_NUMBERED);
  /*
    dyn_d_un m_value[ELF_DT_NUM_NUMBERED];
    bool m_bValueSet[ELF_DT_NUM_NUMBERED];
  */
  
  VALIDATE(m_pStringSection,
	   "Dynamic array does not have an associated string table.");
  
  m_pStringSection->VerifyData();
  
  VALIDATE(m_header.sh_entsize == sizeof (Elf32_Dynamic),
	   "Dynamic array entries size wrong.");
  
  uint nEntries = m_header.sh_size / m_header.sh_entsize;
  
  assert (m_pDataPtr);
  
  m_pDataPtr->ResetCurrent ();
  
  bool bHasHadEnd = false;
  
  for (uint nEntry = 0; nEntry < nEntries; nEntry++)
    {
      const Elf32_Dynamic* pEntry = 
	(const Elf32_Dynamic*) m_pDataPtr->GetPtr(sizeof (Elf32_Dynamic));
      
      switch (pEntry->d_tag)
        {
        case ELF_DT_NULL:
                bHasHadEnd = true;
		WARNING1(!pEntry->d_un.d_val,
			 "Dynamic array end marker (DT_NULL), at index %d1, has non-zero value member (allowed but strange).",nEntry);
		WARNING1(nEntry == nEntries - 1,
			 "Dynamic array prematurely ended with end marker (DT_NULL), at index %d1. (allowed but strange)",nEntry);
		break;
        default:
	  VALIDATE1(!bHasHadEnd,
		    "Dynamic array has non-end entry after end marker, at index %d.",nEntry);
	  break;
        }
      
      if (pEntry->d_tag < ELF_DT_NUM_NUMBERED)
	{   
	  assert ((dynamicEntryInfo[pEntry->d_tag].m_wInfo & DI_NUMBER_MASK) == pEntry->d_tag); // so nobody has messed up the table (that is the only reason for having the index also inside each entry in the table)
	  
	  if (dynamicEntryInfo[pEntry->d_tag].m_wInfo & DI_MULT)
            {
	      if (pEntry->d_tag == ELF_DT_NEEDED)
                {
		  m_vectLibNeeded.push_back(m_pStringSection->GetString(pEntry->d_un.d_val));
                }
	      
	      // entry of kind MULT is not added to list of entries
            }
	  else
            {
	      // single entry
	      
	      VALIDATE2(!HasTag(pEntry->d_tag),
			"Dynamic table has multiple entries of type %d (second at index %d).",pEntry->d_tag,nEntry);
	      
	      m_mapTagValues.insert(MapDynamicTagValue::value_type(pEntry->d_tag,pEntry->d_un));
	      
	      if (pEntry->d_tag == ELF_DT_RPATH)
                {
		  m_strRPath = m_pStringSection->GetString(pEntry->d_un.d_val);
		  // the entry in m_mapTagValues was set anyway, to detect multiple entries, it's value is insignificant and unused
                }
	      
	      //                m_bValueSet[pEntry->d_tag] = true;
	      //                m_value[pEntry->d_tag] = pEntry->d_un;
            }
	  
	  WARNING2((dynamicEntryInfo[pEntry->d_tag].m_wInfo & DI_D_MASK) ||
		   pEntry->d_un.d_ptr == 0 && pEntry->d_un.d_val == 0,
		   "Dynamic table has entry of type %d (at index %d), whose value/ptr is to be ignored but is non-zero (which is valid but strange).",pEntry->d_tag,nEntry)
	    }
      else
        {
	  WARNING2(0,
		   "Dynamic table has entry of type 0x%08x at entry %d. Do not know how to handle. Ignored.",pEntry->d_tag,nEntry);
	  
        }
      /*        
		dyn_d_un m_value[ELF_DT_NUM_NUMBERED];
		bool m_bValueSet[ELF_DT_NUM_NUMBERED];
      */
    }
  VALIDATE(bHasHadEnd,
	   "Dynamic array not ended with end marker (DT_NULL).");
  
  
  bool bExecutable = true; // otherwise shared lib
  
  for (uint i = 0; i < ELF_DT_NUM_NUMBERED; i++)
    {
      word bMandatory = 
	(dynamicEntryInfo[i].m_wInfo >> 
	 (bExecutable ? DI_EXEC_SHIFT : DI_SHRD_SHIFT)) & DI_MAND_OPT_MASK;
      
      switch (bMandatory)
	{
	case DI_IGN:
	  WARNING1(!HasTag(i),
		   "Dynamic table has entry of type 0x%08x, which is ignored (by spec).",i);
	  break;
	case DI_MAND:
	  // if it is mandatory, it must exist
	  // but must treat case of rel/rela, where actually only one is needed
	  
	  WARNING1(HasTag(i) || ((dynamicEntryInfo[i].m_wInfo & DI_MAND_REL_RELA) && HasTag(rel_to_rela(i))),
		   "Dynamic table is missing entry of type 0x%08x, which is mandatory.",i);
	  break;
	case DI_OPT:
	  break;
	  DEFAULT_ASSERT_FALSE;
	}
      
      if (HasTag(i))
	{
	  // let's see if we require any other types to be set
	  
	  for (uint j = 0; j < count_of(dynamicEntryInfo[0].m_bRequires); j++)
	    {
	      byte bReq = dynamicEntryInfo[i].m_bRequires[j];
	      
	      if (bReq & DI_REQUIRE)
		VALIDATE2(HasTag(bReq & DI_REQ_NUM_MASK),
			  "Dynamic table has entry of type %d, which requires entry of type %d, which is missing.",i,bReq & DI_REQ_NUM_MASK);
	      if (bReq & DI_SHLD_REQ)
		WARNING2(HasTag(bReq & DI_REQ_NUM_MASK),
			 "Dynamic table has entry of type %d, which implies entry of type %d, which is missing.",i,bReq & DI_REQ_NUM_MASK);
            }
	}
    }
  
  /*
#define DI_D_IGN        0       // dummy
#define DI_D_VAL        0x0100  // d_un.d_ptr used
#define DI_D_PTR        0x0200  // d_un.d_ptr used

#define DI_EXEC_IGN     0       // dummy
#define DI_EXEC_MAND    0x0400  // required
#define DI_EXEC_OPT     0x0800  // optional

#define 

#define DI_SHRD_IGN     0       // dummy
#define DI_SHRD_MAND    0x1000  // required
#define DI_SHRD_OPT     0x2000  // optional

#define DI_MULT         0x4000 // may have multiple entries
#define DI_SNGL         0      // dummy (only one entry)
*/

}

void Elf32_SectionDynamic::RequestStrings()
{
  FOR_EACH_ITER(VectStrings,m_vectLibNeeded,lib)
    {
      m_pStringSection->RequestString(*lib); 
    }
  m_pStringSection->RequestString(m_strRPath);  // there is no danger in requesting an empty string (it will not add any to the request table)
}

/*
  void Elf32_SectionDynamic::SetTag(Elf32_Sword tag,dword value)
  {
    assert (!HasTag(tag));

    m_mapTagValues.insert(MapDynamicTagValue::value_type(tag,value));
}
*/

bool Elf32_SectionDynamic::GetTag(Elf32_Sword tag,Elf32_Word& value)
{
  MapDynamicTagValue::iterator item = m_mapTagValues.find(tag);
  
  if (item == m_mapTagValues.end())
    return false;
  
  assert (sizeof (item->second.d_un) == sizeof (value));
  
  value = item->second.d_un.d_val;

  return true;
}


void Elf32_SectionDynamic::SetTag(Elf32_Sword tag,const DynamicValue& value)
{
  if (m_pWriteDest)
    {
      assert (HasTag(tag)); // so space has been allocated for it
      
      // it's already been written, so we'll have to find the entry
      // and change it

      // we'll simply calculate the offset

      uint nOffset = m_vectLibNeeded.size();
      
      MapDynamicTagValue::iterator item = m_mapTagValues.find(tag);
      
      assert (item != m_mapTagValues.end());
      
      item->second = value;
      
      nOffset += distance(m_mapTagValues.begin(),item);
      
      nOffset *= sizeof (Elf32_Dynamic);
      
      assert ((m_pWriteDest->m_pDest->SetOffset (nOffset),
               *((Elf32_Sword*) m_pWriteDest->m_pDest->WritePtr(sizeof (Elf32_Sword))) == tag));
      
      m_pWriteDest->m_pDest->SetOffset (nOffset + sizeof (Elf32_Sword));
      m_pWriteDest->m_pDest->WriteDWord(item->second.d_un.d_val);
    }
  else
    assert (!HasTag(tag));
  
  m_mapTagValues.insert(MapDynamicTagValue::value_type(tag,value));
  
}

void Elf32_SectionDynamic::CreateSection()
{
  assert (!m_pWriteDest);
  
  m_header.sh_entsize = sizeof (Elf32_Dynamic);
  
  // figure out how much memory is required for the requested strings...
  
  uint nItems = m_mapTagValues.size() + m_vectLibNeeded.size() + 1; // plus 1 to DT_NULL
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(nItems * sizeof (Elf32_Dynamic));
  
  FOR_EACH_ITER(VectStrings,m_vectLibNeeded,lib)
    {
      m_pWriteDest->m_pDest->WriteDWord(ELF_DT_NEEDED);
      m_pWriteDest->m_pDest->WriteDWord(m_pStringSection->GetStringIndex(*lib),
					REL_ELF_DYNSTR_SECTION);
    }
  
  if(!m_strRPath.empty())
    {
      uint nIndex = m_pStringSection->GetStringIndex(m_strRPath);
      
      m_mapTagValues.insert(MapDynamicTagValue::value_type(ELF_DT_RPATH,
							   DynamicValue(nIndex,
									REL_ELF_DYNSTR_SECTION)));
    }
  else
    assert (m_mapTagValues.find(ELF_DT_RPATH) == m_mapTagValues.end()); // cannot have an value (no rpath)
  
  FOR_EACH_ITER(MapDynamicTagValue,m_mapTagValues,tagval)
    {
      assert (sizeof (Elf32_Dynamic) == 2 * sizeof (dword)); // so writing below works
      assert (sizeof (tagval->first) == sizeof (dword));
      assert (sizeof (tagval->second.d_un.d_val) == sizeof (dword));
      
      m_pWriteDest->m_pDest->WriteDWord(tagval->first);
      m_pWriteDest->m_pDest->WriteDWord(tagval->second.d_un.d_val,
					tagval->second.wReloc);
    }
  
  m_pWriteDest->m_pDest->WriteDWord(ELF_DT_NULL);
  m_pWriteDest->m_pDest->WriteDWord(0);
  
  
  /*
    FOR_EACH_ITER(VectStringIndices,m_vectWantStrings,iter2) // iter2 because some compilers are unaware of scoping
    m_pWriteDest->m_pDest->WriteString(iter->str);
    
*/
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
  
  
}

void Elf32_SectionDynamic::HTMLFileInfo(HtmlPage& page)
{
  uint nEntries = m_header.sh_size / m_header.sh_entsize;
  
  assert (m_pDataPtr);
  
  page.Begin(table()/*.border(0)*/);
  
  page.Begin(tr());
  page << td("#");
  DumpStructTableHeaderElf32_Dynamic(page);
  page.End();
  
  m_pDataPtr->ResetCurrent ();
  
  for (uint nEntry = 0; nEntry < nEntries; nEntry++)
    {
      const Elf32_Dynamic* pEntry = 
	(const Elf32_Dynamic*) m_pDataPtr->GetPtr(sizeof (Elf32_Dynamic));
      
      page.Begin(tr());
      page << td(strsprintf("%d",nEntry));
      DumpStructTableElf32_Dynamic(*pEntry,page);
      page.End();
    }     
  
  page.End();
}

//////////////////////////////////////////////////////////////////////

Elf32_SectionNote::Elf32_SectionNote()
{

}

Elf32_SectionNote::Elf32_SectionNote(const string& strName,Elf32_Word flags)
  : Elf32_SectionRawData(strName,ELF_SHT_NOTE,flags)
{
  
}

void Elf32_SectionNote::VerifyData()
{
  assert (m_pDataPtr);
  
  // check that the note(s) satisfy the specification
  
  const Elf32_NoteHeader* pNote = 
    (const Elf32_NoteHeader*) m_pDataPtr->GetPtr(0,m_header.sh_size);
  const char* pEnd = ((const char*) pNote) + m_header.sh_size;
  
  while ((const char*) pNote < pEnd)
    {
      VALIDATE((const char*) (pNote + 1) < pEnd,
               "Note header passes end of note section. (beheaded?)");
      
      VALIDATE(pNote->n_type >= 0,
               "Note type must be non-negative according to specification.");
      
      const char* pName = (const char*) (pNote + 1);
      
      VALIDATE(pName + pNote->n_namesz <= pEnd,
               "Note name passes end of note section.");
      
      VALIDATE(pNote->n_namesz > 1,
               "Note name with 0 content, or content \"\0\" is reserved by specification.");
      
      VALIDATE(pNote->n_namesz == 0 ||
               *(pName + pNote->n_namesz-1) == 0,
               "Note name not null terminated at specified length.");
      
      WARNING(pNote->n_namesz == 0 ||
              strlen(pName) == pNote->n_namesz-1,
              "Note name null terminated before specified length.");
      
      const char* pDesc = pName + ((pNote->n_namesz + 3) & ~3);
      
      // there is a possibilty that pDesc is outside the size, if the
      // note has no desc and the name isn't padded because this is
      // the last note
      
      VALIDATE(pNote->n_descsz == 0 ||
               pDesc + pNote->n_descsz <= pEnd,
	       "Note desc passes end of note section.");
      
      pNote = (const Elf32_NoteHeader*) (pDesc + ((pNote->n_descsz + 3) & ~3));
    }
  
  /*
    // we now validate the actual data,
    // the first bytes are suppose to hold zero
    
    assert (m_pDataPtr);
    
    VALIDATE(*m_pDataPtr == 0,
    "String table first byte not zero");
    VALIDATE(*(m_pDataPtr+m_header.sh_size-1) == 0,
    "String table last byte not zero");
    
    // except for that, anything's allowed
  */
}

void Elf32_SectionNote::AddNote (const char* pStr,
				 Elf32_Word type,
				 const void* pDesc,
				 uint nDescLen)
{
  assert (!nDescLen || pDesc);
  assert (pStr);
  
  uint nStrLen = strlen (pStr)+1;
  
  uint nStartOffset = (m_header.sh_size + 3) & ~3; // round up to nearest 4-byte boundary
  
  uint nNeedSize = sizeof (Elf32_Word) * 3 + ((nStrLen + 3) & ~3) + nDescLen;
  
  m_pRawData = realloc (m_pRawData,nStartOffset+nNeedSize);
  
  byte* ptr = ((byte*) m_pRawData) + m_header.sh_size;
  
  for (uint i = (4 - m_header.sh_size) & 3; i; i--)
    *(ptr++) = 0;
  
  Elf32_Word* pDW = (Elf32_Word*) ptr;
  
  *(pDW++) = nStrLen;
  *(pDW++) = nDescLen;
  *(pDW++) = type;
  
  ptr = (byte*) pDW;
  
  memcpy (ptr,pStr,nStrLen);
  
  ptr += nStrLen;
  
  for (nStrLen = (4 - nStrLen) & 3; nStrLen; nStrLen--)
    *(ptr++) = 0;
  
  memcpy (ptr,pDesc,nDescLen);
  
  m_header.sh_size = nStartOffset+nNeedSize;
}


#define FMT_RAW_LINE_SIZE  16
#define FMT_RAW_BLOCK_SIZE 4

void Elf32_SectionNote::HTMLFileInfo(HtmlPage& page)
{
  assert (m_pDataPtr);
  
  // assume that we follow specification, so we do not have to check that everything is within bounds
  
  const Elf32_NoteHeader* pNote = (const Elf32_NoteHeader*) m_pDataPtr->GetPtr(0,m_header.sh_size);
  const char* pEnd = ((const char*) pNote) + m_header.sh_size;
  
  char* fmtraw = new char[FORMAT_RAW_DATA_SIZE_NEEDED(FMT_RAW_LINE_SIZE,FMT_RAW_BLOCK_SIZE)];
  
  while ((const char*) pNote < pEnd)
    {
      const char* pName = (const char*) (pNote + 1);
      const char* pDesc = pName + ((pNote->n_namesz + 3) & ~3);
      
      // there is a possibilty that pDesc is outside the size, if the note
      // has no desc and the name isn't padded because this is the last note
      
      page.Begin(table());
      
      page << (tr() << td("Name") << td(pName));
      page << (tr() << td("Type") << td(strsprintf("%d",pNote->n_type)));
      
      page.Begin(tr());
      page << td("Desc");
      
      if (pNote->n_descsz)
	{
	  page.Begin(td());
	  page.Begin(tt());
	  
	  const byte* ptr = (const byte*) pDesc;
	  
	  for (uint ind = 0; ind < pNote->n_descsz; ind += FMT_RAW_LINE_SIZE)
	    {      
	      if (ind)
		page << "<br>";
	      
	      FormatRawData(fmtraw,ptr,pNote->n_descsz-ind,
			    FMT_RAW_LINE_SIZE,FMT_RAW_BLOCK_SIZE);
	      page << fmtraw;
	      
	      ptr += FMT_RAW_LINE_SIZE;
	    }
	  page.End();
	  page.End();
	}
      else
	page << td("none");
      
      page.End();
      page.End();
      
      pNote = (const Elf32_NoteHeader*) (pDesc + ((pNote->n_descsz + 3) & ~3));
    }
  delete fmtraw;
  
  
  
  /*
    // simply walk through the table and display all strings

  Elf32_Section::HTMLFileInfo(page);

  page.Begin(pre());

  const char* ptr = (const char*) m_pDataPtr;
  const char* endptr = (const char*) m_pDataPtr + m_header.sh_size;

  uint nIndex = 0;

  char fmt[64];

  while (ptr <= endptr)
    {
      sprintf (fmt,"0x%04x %5d  ",nIndex,nIndex);

      page << fmt << ptr << '\n';

      while (*ptr)
        {
          ptr++;
          nIndex++;
        }      

      ptr++;  
      nIndex++;
    }

  page.End();
  */
}

//////////////////////////////////////////////////////////////////////

void Elf32_SectionRelocation::VerifyData()
{
  bool bRelA = (m_header.sh_type == ELF_SHT_RELA);
  uint nEntrySize = (bRelA ? sizeof (Elf32_Rela) : sizeof (Elf32_Rel));
  
  VALIDATE(m_header.sh_entsize == nEntrySize,
           "Relocation table entries size wrong.");
  
  uint nEntries = m_header.sh_size / m_header.sh_entsize;
  
  assert (m_pDataPtr);
  
  m_pDataPtr->ResetCurrent ();
  
  for (uint nEntry = 0; nEntry < nEntries; nEntry++)
    {
      const Elf32_Rela* pEntry = 
	(const Elf32_Rela*) m_pDataPtr->GetPtr(nEntrySize);
      
      if (ELF32_R_TYPE(pEntry->r_info) == ELF_R_386_RELATIVE)
        VALIDATE1(ELF32_R_SYM(pEntry->r_info) == 0,
                  "Relocation table entry %d type relative does not specify 0 as symbol table index.",nEntry);
      
      assert (!bRelA); // because RelocationItem can't handle them
      
      m_vectRelocationItems.push_back(new RelocationItem(*((const Elf32_Rel*) pEntry)));
    }
}

void Elf32_SectionRelocation::CreateSection()
{
  assert (!m_pWriteDest);
  
  // figure out how much memory is required for the requested strings...
  
  uint nItems = m_vectRelocationItems.size();
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(nItems * sizeof (Elf32_Rel));
  
  FOR_EACH_ITER(VectRelocationItem,m_vectRelocationItems,rel)
    {
      m_pWriteDest->m_pDest->WriteDWord((*rel)->m_dwOffset,
					(*rel)->m_wRelocSect);
      m_pWriteDest->m_pDest->WriteDWord((*rel)->m_dwInfo);
    }
}

void Elf32_SectionRelocation::HTMLFileInfo(HtmlPage& page)
{
  bool bRelA = (m_header.sh_type == ELF_SHT_RELA);
  uint nEntrySize = (bRelA ? sizeof (Elf32_Rela) : sizeof (Elf32_Rel));
  
  uint nEntries = m_header.sh_size / m_header.sh_entsize;
  
  assert (m_pDataPtr);
  
  page.Begin(table()/*.border(0)*/);
  
  page.Begin(tr());
  page << td("#");
  if (bRelA)
    DumpStructTableHeaderElf32_Rela(page);
  else
    DumpStructTableHeaderElf32_Rel(page);
  page.End();
  
  m_pDataPtr->ResetCurrent ();
  
  for (uint nEntry = 0; nEntry < nEntries; nEntry++)
    {
      const Elf32_Rela* pEntry = 
	(const Elf32_Rela*) m_pDataPtr->GetPtr(nEntrySize);
      
      page.Begin(tr());
      page << td(strsprintf("%d",nEntry));
      if (bRelA)
        DumpStructTableElf32_Rela(*pEntry,page);
      else
        DumpStructTableElf32_Rel(*pEntry,page);
      page.End();
    }     
  
  page.End();
}

//////////////////////////////////////////////////////////////////////

Elf32_SectionGOT::Elf32_SectionGOT(const string& strName)
{
  SetSpecialSectionRights(strName.c_str());
  
}

void Elf32_SectionGOT::VerifyData()
{
  VALIDATE(m_header.sh_entsize == sizeof (Elf32_Addr),
           "GOT table entries size wrong");
  
  
}


void Elf32_SectionGOT::CreateSection()
{
  assert (!m_pWriteDest);
  
  // figure out how much memory is required for the requested strings...
  
  m_header.sh_entsize = sizeof (Elf32_Addr);
  
  uint nItems = 3;
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(nItems * sizeof (Elf32_Addr));
  
  m_pWriteDest->m_pDest->WriteDWord(0,REL_ELF_IMAGE_DYNAMIC_SECTION);
  m_pWriteDest->m_pDest->WriteDWord(0);
  m_pWriteDest->m_pDest->WriteDWord(0);
  
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
}

void Elf32_SectionGOT::HTMLFileInfo(HtmlPage& /* page */)
{
  

}

//////////////////////////////////////////////////////////////////////

Elf32_SectionPLT::Elf32_SectionPLT(const string& strName)
{
  SetSpecialSectionRights(strName.c_str());
  
}

void Elf32_SectionPLT::VerifyData()
{
  VALIDATE(m_header.sh_entsize == 4, // why four?
           "PLT table entries size wrong");
  
  
}


void Elf32_SectionPLT::CreateSection()
{
  assert (!m_pWriteDest);
  
  // figure out how much memory is required for the requested strings...
  
  m_header.sh_entsize = 4;
  
  uint nItems = 1;
  
  m_pWriteDest = new WriteRelocDest();
  
  m_pWriteDest->AllocMemory(nItems * 16); // isch, hard-coded size
  
  m_pWriteDest->m_pDest->WriteBytes(0xff,0x35); // jmp [disp]
  m_pWriteDest->m_pDest->WriteDWord(4,REL_ELF_IMAGE_GOT_SECTION);
  m_pWriteDest->m_pDest->WriteBytes(0xff,0x25); // push [disp]
  m_pWriteDest->m_pDest->WriteDWord(8,REL_ELF_IMAGE_GOT_SECTION);
  for (int nops = 0; nops < 4; nops++)
    m_pWriteDest->m_pDest->WriteByte(0x90); // nop (by spec)
  /*
    for (int i = 1; i < ; i++)
    {
    m_pWriteDest->m_pDest->WriteBytes(0xff,0x25); // jmp [mem]
    m_pWriteDest->m_pDest->WriteDWord(8+i*4,REL_ELF_IMAGE_GOT_SECTION);
    m_pWriteDest->m_pDest->WriteByte(0x68); // push IMM
    m_pWriteDest->m_pDest->WriteDWord((i-1)*8,REL_ELF_IMAGE_GOT_SECTION);
    m_pWriteDest->m_pDest->WriteByte(-16-16*i); // jmp REL
    }
  */
  assert (m_pWriteDest->m_pDest->ReacedEnd()); // space exactly used
}

void Elf32_SectionPLT::HTMLFileInfo(HtmlPage& /* page */)
{

  
}

//////////////////////////////////////////////////////////////////////










