/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Elf32_Section.h: interface for the Elf32_Section class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__ELF32_SECTION_H__INCLUDED_)
#define __ELF32_SECTION_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h"
#include "ELF_File.h"
#include "FileMapItem.h"

class UrlFile;
class UrlFolder;
class HtmlPage;
class DataPtrSrc;
class WriteRelocDest;
class WriteSectionHeader;

//////////////////////////////////////////////////////////////////////

struct Elf32_SpecialSectionRights
{
  char name[10]; // must be first member
  word type;
  word flags;
  
  word wRelocBody;
};

const Elf32_SpecialSectionRights* GetSpecialSectionRights(const char* strName);

#include <vector>
#include <set>

using namespace std;

class Elf32_ProgramBlock;

typedef vector<Elf32_ProgramBlock*> VectProgramBlocks;

class Elf32_Section : public FileMapItem
{
 public:
  Elf32_Section();
  virtual ~Elf32_Section();

 public:
  Elf32_SectionHeader m_header;
  string m_strName;

  DataPtrSrc* m_pDataPtr;

  WriteRelocDest* m_pWriteDest;

  word m_wRelocSect;
  WriteSectionHeader* m_pWriteHeader;

 public:
  VectProgramBlocks m_vectProgramBlocks; // in which Program (part of elf file) are we  
  
 public:
  virtual void InitData (const Elf32_SectionHeader* header,
			 DataPtrSrc* pDataPtr);
  
 public:
  virtual void VerifyData();
  
 public:
  virtual void CreateSection()/* = 0*/ { };
  
  void SetWriteInfo (word wReloc);
  
  RelocatableDest* WriteHeader ();
  RelocatableDest* WriteSection (); 
  /* void WriteSection (); */

 public:
  void ValidateHeader();

 public:
  bool IsAddrInside(dword dwOffset,dword dwSize);
  
 public:
  void AddProgramBlock(Elf32_ProgramBlock* pProgramBlock);

 public:
  void SetSpecialSectionRights(const string& strName);  

 protected:
  virtual bool ValidateCoExistence (FileMapItem* pOtherItem);

 public:
  virtual string GetDescription();
  
 public:
  UrlFile* HTMLFileInfo(Elf32_Half nSectNo,UrlFolder* pParent);
  
 public:
  virtual void HTMLFileInfo(HtmlPage& page);
};

//////////////////////////////////////////////////////////////////////

// this is for storing contents of sections that we do not mess with
// we may (or may not) be able to validate the contents of the section,
// but they are mainly to be copied from the input file to the output
// file in raw form
// this of course require that they do not need any kind of relocations

class Elf32_SectionRawData : public Elf32_Section
{
 public:
  Elf32_SectionRawData();
  Elf32_SectionRawData(const string& strName,
		       Elf32_Word type,Elf32_Word flags);
  virtual ~Elf32_SectionRawData();

 public:
  void* m_pRawData;

 public:
  void SetDataStr (const char* pStr);

 public:
  virtual void InitData (const Elf32_SectionHeader* header,DataPtrSrc* pDataPtr);

 public:
  virtual void CreateSection();
};

//////////////////////////////////////////////////////////////////////

class StringIndex
{
 public:
  StringIndex(const string& s,uint i) { str = s; index = i; }
  
 public:
  string str;
  uint   index;
};

typedef set<Elf32_Word> SetIndices;
typedef vector<StringIndex> VectStringIndices;

class Elf32_SectionStringTable : public Elf32_Section
{
 public:
  Elf32_SectionStringTable();
  Elf32_SectionStringTable(const string& strName,Elf32_Word flags);
  
 public:
  SetIndices        m_setUsedIndices;  // so we know what substrings were requested (for info only)
  VectStringIndices m_vectWantStrings; // so we can merge multiple requests into one

 public:
  virtual void VerifyData();

  const char* GetString(Elf32_Word nIndex);

 protected:
  //Elf32_Word InsertString (const string& str,bool bForceCreate);
  // void CreateSection(VectStringIndices::iterator endcreate);
  
 public:
  // RequestString(const string& str) { InsertString(str,false); } // this string will be needed in the to be string table
  void RequestString(const string& str); // this string will be needed in the to be string table

  Elf32_Word GetStringIndex (const string& str);
  
  virtual void CreateSection();
 public:
  virtual void HTMLFileInfo(HtmlPage& page);
};

//////////////////////////////////////////////////////////////////////

class Elf32_SectionSymbolTable;
class Elf32_SectionHashTable;
class Elf32_SectionDynamic;
class Elf32_SectionRelocation;

class Elf32_SectionHashTable : public Elf32_Section
{
 public:
  Elf32_SectionHashTable() { m_pSymbolSection = NULL; }
  Elf32_SectionHashTable(const string& strName);
  
 public:
  Elf32_SectionSymbolTable* m_pSymbolSection;

 public:
  void SetSymbolTable (Elf32_SectionSymbolTable* pSection);
  Elf32_SectionSymbolTable* GetSymbolTable () 
    { assert (m_pSymbolSection); return m_pSymbolSection; }
  virtual void VerifyData();

 public:
  void CreateSection (Elf32_Word* hash,
		      uint nItems,uint nMaxBuckets,uint nExtraSearch = 3);

  virtual void CreateSection();
  
};

unsigned long elf_hash(const unsigned char *name);
inline unsigned long elf_hash(const char *name) 
{ return elf_hash((const unsigned char *) name); } // convert pointer argument to unsigned type (so we do not change the specified function)

/*
dword OptimalHashNBuckets(dword* dwKeys,dword dwNumKeys,
                          float fTradeMemoryForSearchSpeed = 0.1,
                          float fMaxBucketsPerEntry = 0.5);
*/
//////////////////////////////////////////////////////////////////////

struct DynamicValue
{
 public:
  DynamicValue(const dyn_d_un& un,word reloc = 0) 
  { d_un = un; wReloc = reloc; }
  DynamicValue(dword val,word reloc = 0) { assert(sizeof(d_un) == sizeof (val)); d_un.d_val = val; wReloc = reloc; }
  
 public:
  dyn_d_un d_un;
  word    wReloc;
};

typedef vector<string> VectStrings;
typedef map<Elf32_Sword,DynamicValue> MapDynamicTagValue;

class Elf32_SectionDynamic : public Elf32_Section
{
 public:
  Elf32_SectionDynamic() { m_pStringSection = NULL; }
  Elf32_SectionDynamic(Elf32_SectionStringTable* pStringSection);
  
 public:
  Elf32_SectionStringTable* m_pStringSection;
  
 public:
  void SetStringTable (Elf32_SectionStringTable* pSection) 
    { assert (!m_pStringSection); m_pStringSection = pSection; } 
  Elf32_SectionStringTable* GetStringTable () 
    { assert (m_pStringSection); return m_pStringSection; } // should have a string table
  
 public:
  MapDynamicTagValue m_mapTagValues; // only for singly defined values
  VectStrings   m_vectLibNeeded;
  string        m_strRPath;
  
 public:
  bool HasTag(Elf32_Sword tag) 
    { return m_mapTagValues.find(tag) != m_mapTagValues.end(); }
  
  void SetTag(Elf32_Sword tag,const DynamicValue& value);
  void SetTag(Elf32_Sword tag,dword value) 
    { SetTag(tag,DynamicValue(value)); }
  
  bool GetTag(Elf32_Sword tag,Elf32_Word& value);
  
 public:
  virtual void VerifyData();

 public:
  virtual void CreateSection();

 public:
  void RequestStrings();


 public:
  virtual void HTMLFileInfo(HtmlPage& page);
};

//////////////////////////////////////////////////////////////////////

class SymbolItem;

class RelocationItem
{
 public:
  RelocationItem(const Elf32_Rel& rel)  
    { 
      m_dwOffset = rel.r_offset; 
      m_dwInfo = rel.r_info; 
      /*m_dwAddend = 0; */ 
      m_pSymbol = NULL; 
      m_wRelocSect = 0; 
    }
  // RelocationItem(const Elf32_Rela& rel) { m_dwOffset = rel.r_offset; m_dwInfo = rel.r_info; m_dwAddend = rel.r_addend; m_pSymbol = NULL; }
  
 public:
  dword   m_dwOffset;
  dword   m_dwInfo;
  // sdword  m_dwAddend;

  word    m_wRelocSect;
  
 public:
  SymbolItem* m_pSymbol;
};

typedef vector<RelocationItem*> VectRelocationItem;

class Elf32_SectionRelocation : public Elf32_Section
{
 public:
  Elf32_SectionRelocation() 
    { m_pSymbolSection = NULL; m_pRelocSection = NULL; }
  
 public:
  Elf32_SectionSymbolTable* m_pSymbolSection;
  Elf32_Section* m_pRelocSection;
  
 public:
  VectRelocationItem m_vectRelocationItems;
  
 public:
  void SetSymbolTable (Elf32_SectionSymbolTable* pSection) 
    { assert (!m_pSymbolSection); m_pSymbolSection = pSection; } 
  void SetRelocSection (Elf32_Section* pSection) 
    { assert (!m_pRelocSection); m_pRelocSection = pSection; } 

  Elf32_SectionSymbolTable* GetSymbolTable () 
    { assert (m_pSymbolSection); return m_pSymbolSection; } 
  Elf32_Section* GetRelocSection () { return m_pRelocSection; } 

 public:
  virtual void VerifyData();
  
 public:
  virtual void CreateSection();

 public:
  virtual void HTMLFileInfo(HtmlPage& page);
};

//////////////////////////////////////////////////////////////////////

class SymbolItem
{
 public:
  SymbolItem() { m_dwValue = 0; m_dwSize = 0; m_bInfo = 0; m_wSection = 0; }
  
 public:
  SymbolItem(const string& strName,dword dwValue,dword dwSize,
	     byte bInfo,word wSection)
    {
      m_strName = strName;
      m_dwValue = dwValue;
      m_dwSize = dwSize;
      m_bInfo = bInfo;
      m_wSection = wSection;
    }

  SymbolItem(const string& strName,const Elf32_Symbol& symbol)
    {
      m_strName = strName;
      m_dwValue = symbol.st_value;
      m_dwSize = symbol.st_size;
      m_bInfo = symbol.st_info;
      m_wSection = symbol.st_shndx;
    }
  
 public:
  string  m_strName;
  dword   m_dwValue;
  dword   m_dwSize;
  byte    m_bInfo;
  word    m_wSection;
};

typedef vector<SymbolItem*> VectSymbolItems;

class Elf32_SectionSymbolTable : public Elf32_Section
{
 public:
  Elf32_SectionSymbolTable() { m_pStringSection = NULL; m_pHashTable = NULL; }
  Elf32_SectionSymbolTable(const string& strName,
			   Elf32_SectionStringTable* pStringSection,
			   Elf32_SectionHashTable* pHashTable);

 public:
  Elf32_SectionStringTable* m_pStringSection;
  Elf32_SectionHashTable* m_pHashTable;

 public:
  VectSymbolItems m_vectSymbolItems;
  
 public:
  void SetStringTable (Elf32_SectionStringTable* pSection) 
    { assert (!m_pStringSection); m_pStringSection = pSection; } 
  Elf32_SectionStringTable* GetStringTable () 
    { assert (m_pStringSection); return m_pStringSection; } // should have a string table

 public:
  virtual void VerifyData();

 public:
  Elf32_Word GetSymbolIndex (const char* name);

 public:
  virtual void CreateSection();

 public:
  virtual void HTMLFileInfo(HtmlPage& page);

 public:
  void RequestStrings();
};

//////////////////////////////////////////////////////////////////////

class Elf32_SectionNote : public Elf32_SectionRawData
{
 public:
  Elf32_SectionNote();
  Elf32_SectionNote(const string& strName,Elf32_Word flags);
  
 public:
  virtual void VerifyData();
  
 public:
  void AddNote (const char* pStr,Elf32_Word type,
		const void* pDesc,uint nDescLen);   
  void AddNote (const char* pStr,Elf32_Word type) 
    { AddNote(pStr,type,NULL,0); }
  
 public:
  virtual void HTMLFileInfo(HtmlPage& page);
};

//////////////////////////////////////////////////////////////////////

class Elf32_SectionGOT : public Elf32_Section
{
 public:
  Elf32_SectionGOT() { }
  Elf32_SectionGOT(const string& strName);

 public:
  virtual void VerifyData();

 public:
  virtual void CreateSection();

 public:
  virtual void HTMLFileInfo(HtmlPage& page);

};

//////////////////////////////////////////////////////////////////////

class Elf32_SectionPLT : public Elf32_Section
{
 public:
  Elf32_SectionPLT() { }
  Elf32_SectionPLT(const string& strName);
  
 public:
  virtual void VerifyData();
  
 public:
  virtual void CreateSection();

 public:
  virtual void HTMLFileInfo(HtmlPage& page);

};

//////////////////////////////////////////////////////////////////////

#endif // !defined(__ELF32_SECTION_H__INCLUDED_)
