/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ErrorTracker.h: interface for the ErrorTracker class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__ERROR_TRACKER_H__INCLUDED_)
#define __ERROR_TRACKER_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//////////////////////////////////////////////////////////////////////

/*
  Because this program uses multi-passes and sometimes it is hard to
  determine if an error is fatol or non-fatal when it occurs, we need
  a system to be able to track what errors have occured, and if
  something proves to be fatal, try to provide a link to the first
  error that caused the problem. NOTE: the errors I talk of here are
  not the errors in this executable, but problems that arise because
  our limited ability to cope with reading, modifying (or writing, but
  such errors should be rare here, caught before) an executable.

  For example: it is possible when disassembling that we find a 'crazy
  jump', that is one byte in the instruction segment that is to be
  interpreted in different ways depending on the fact that it was part
  of several instructions starting on different addresses. This is
  caused by jumps, and while doing a first stage of disassembly, we
  might try this and fail. But analysis might determine that one of
  the jumps actually never could be taken, so after analysis when we
  do disassembly again the error does not show up, and there will
  never be a problem, because the jump has been replaced bu an
  unreachable. However, if the jump has not been replaced by an
  unreachable, it's nice if when we eject that error, we might point
  back to the original one.

  Also, as there will be several passes of the disassembler, the same
  error condition might occur twice. In this case, we should just
  swallow it and do not bother the user with it again.
  
  Now, as we may be processing several executables, we somehow have to
  link the errors to the source file they are associated to, without
  having passing such references to all functions that we call.

  It's of course fairly easy for errors as they can be 'throw'n out of
  the processing function to a level where we know what the context
  is. But we also want to be able to issue warnings of different
  severity, such as: 'I detected this unnormal thing: reserved byte
  not 0, this may affect bla. bla...' And then just try to
  continue. If something bad happens, we might be able to link to this
  or not, but at least we've got the info out.

  We do this by setting the context when the context is switched. This
  is of course unnecessary work if there are no fault reports
  (errors), but propagating pointers are a big pain. The BIG minus of
  this method is that it partly spoils the object orientation by
  having a global error context. Note: this is an emergency step, I do
  not normally accept it a solution as it breaks object
  orientation. Better solution wanted! (Just as I do not wish do do
  all the .h and .cpp generation as I do now)
  
  The aim of the program is to be able to treat things exactly (that
  is, that an modified executable should do exactly what the original
  did). But sometimes some things must be assumed. The hope is to at
  least the notify that something was assumed, and that this might
  cause trouble.
  
  Also, I do extremely much dislake having any text or string to be
  output to the user encoded i source files. Especially distributed in
  a lot of source files, so instead we use a central place for such
  information. Ideally, it should be in a text file of some kind,
  which could be parsed and used when necessary. Such things have no
  place in an executable if they are large, as error messages and
  their explanations tend to be. (And I think we should try to give
  some explanations...)  
*/

//////////////////////////////////////////////////////////////////////

#include "GlobalDefs.h" 
class ErrorContext; 
#include "Error.h" 

#include <assert.h>

#include <stack>

using namespace std;

struct CurrentErrorContext
{
 public:
  CurrentErrorContext(ErrorContext* pContext,CurrentErrorContext* pParent)
  { m_pContext = pContext; m_bUsed = false; m_pParent = pParent; }
  
 public:
  CurrentErrorContext* m_pParent;
  
 public:
  ErrorContext* m_pContext;
  bool m_bUsed; // did we get an error with this one
  
 public:
  void SetUsed();
};

template<class __Ty>
struct less_ptr : binary_function<__Ty, __Ty, bool> {
  bool operator()(const __Ty& __X, const __Ty& __Y) const
  {return (*__X < *__Y); }
};

typedef vector<Error*> VectError;
typedef set<Error*, less_ptr<Error*> > SetError;

class ErrorTracker  
{
 public:
  ErrorTracker();
  virtual ~ErrorTracker();
  
 public:
  CurrentErrorContext* m_pCurrentErrorContext;

 public:
  SetError m_setError; // to avoid dublets
  VectError m_vectError;
  
 public:
  void PushContext(ErrorContext* pErrorContext); // after being sent here, we are responsible for deletion

 public:
  void PopContext();
  ErrorContext* GetCurrentContextSetUsed();
  
 public:
  void AddError(Error* pError);
};

extern ErrorTracker errorTracker;
/*
#define PUSH_ERROR_CONTEXT(context) {errorTracker.PushContext(new ErrorContext(context)); }
#define PUSH_ERROR_CONTEXT_FILE(context,file) {errorTracker.PushContext(new ErrorContext(context,file)); }
#ifndef NDEBUG			
#define POP_ERROR_CONTEXT(context) {errorTracker.PopContext(context); }
#else
#define POP_ERROR_CONTEXT() {errorTracker.PopContext(); }
#endif
*/
#define EMIT_ERROR(pError) { errorTracker.AddError(pError); }

#define CATCH_EMIT_ERROR  catch (Error* pError) { EMIT_ERROR(pError); }
#define TRY_CATCH_EMIT_ERROR(operation) { try { operation; } CATCH_EMIT_ERRORS }

#define CATCH_EMIT_ERROR_THROW_GENERIC(short_msg)  catch (Error* pError) { EMIT_ERROR(pError); throw GENERIC_ERROR(short_msg); }
#define TRY_CATCH_EMIT_ERROR_THROW_GENERIC(operation,short_msg) { try { operation; } CATCH_EMIT_ERROR_THROW_GENERIC(short_msg) }


// #define EMIT_WARNING_RVA(warning,rva) {errorTracker.AddWarning(new Warning(warning,rva));}
// #define THROW_WARNING_RVA(warning,rva) {throw new Warning(warning,rva);}

#endif // !defined(__ERROR_TRACKER_H__INCLUDED_)



