/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ExecutableImage.cpp: implementation of the ExecutableImage class.
//
//////////////////////////////////////////////////////////////////////

#include "ExecutableImage.h"
#include "RelocatableDest.h"
#include "DisAssembler.h"
#include "Executable.h" 
#include "ErrorTracker.h" 
#include "ErrorContext.h"
#include "ExportedObject.h" 
#include "DataPtr.h" 
#include "gen/DisAssemblerError.h"
#include "ReportDisAsmInfoOffset.h"

#include "BranchTargetTraceSingle.h"
#include "BranchTargetTrace.h" 
#include "BranchTargetCtrlTransfer.h"

#include "Variable.h"
#include "VarContStackPtr.h"
#include "OperandMemory.h"
#include "OperandRegister.h"

#include "BranchTargetCtrlTransfer.h"
#include "JumpTarget.h"

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ExecutableImage::ExecutableImage(const string& strFileName)
{
  m_strFileName = strFileName;
}

ExecutableImage::~ExecutableImage()
{

}


void ExecutableImage::PostReadFile ()
{
  m_pExecutable = new Executable();

}

bool ExecutableImage::WriteFile()
{
  // everything to be written has been prepared in	
  // PrepareWrite, so we just have to open the file
  // eject our contents, and close the file
  
#ifdef WIN32
  int file = open (m_strFileName.c_str(),
		   _O_BINARY | _O_WRONLY | _O_CREAT | _O_TRUNC,_S_IWRITE);
#else
  int file = open (m_strFileName.c_str(),
		   O_WRONLY | O_CREAT | O_TRUNC , S_IRWXU | S_IRGRP | S_IROTH);
#endif
  
  if (file == -1)
    return false;
  
  printf ("opened %s\n",m_strFileName.c_str());
  
  MapOffsetRelocatableDest mapOffsetRelocatableDest;
  
  m_destSections.GetSectionInFileOrder(mapOffsetRelocatableDest);
  
  dword dwLastWritten = 0;
  
  FOR_EACH_ITER(MapOffsetRelocatableDest,mapOffsetRelocatableDest,section)
    {
      assert (dwLastWritten <= section->first /* dwOffset */);
      
      if (dwLastWritten < section->first)
	{
	  //			write (file,,section->first-dwLastWritten); // fill with zeros

	  // at least on linux lseek fills file with zeros if jumping
	  // past end the win32 documentation doesn't tell
	  
	  if (lseek(file,section->first,SEEK_SET) != (int) section->first)
	    return false;
	}

      //		printf ("%4d %4d : ",dwLastWritten,section->first);
      //		for (uint i = 0; i < section->second->m_dwSize; i++)
      //		  printf ("%02x ",*(section->second->m_pData+i));
      //		printf ("\n");
      
      if (write (file,section->second->m_pData,section->second->m_dwSize) != 
	  (int) section->second->m_dwSize)
	return false;
      
      dwLastWritten = section->first + section->second->m_dwSize;
    }
  
  close (file);
  
  // and so voila, uno executable file (hopefully at least as
  // executable as file, too)
  
  return true;
}




void ExecutableImage::GetExportedCallTargets (DisAssembler& /* disasm */)
{

}

void ExecutableImage::GetCallTargets (DisAssembler& disasm)
{
  // add functions in reverse order of appearance (fini first, so it
  // gets popped last from the stack

  if (m_pExecutable->m_pFiniFunction)
    disasm.AddCallTarget (m_pExecutable->m_pFiniFunction); 
  
  GetExportedCallTargets(disasm);
  
  /////////////
  
  // add after exported functions, 
  // as disasm keeps a stack to use when getting the VAs to disasm (lifo), 
  // so we get disassembled first
  
  if (m_pExecutable->m_pEntryFunction)
    disasm.AddCallTarget (m_pExecutable->m_pEntryFunction); 
  if (m_pExecutable->m_pInitFunction)
    disasm.AddCallTarget (m_pExecutable->m_pInitFunction); 
}





bool ExecutableImage::DisAssemble (UrlFolder* pFolder)
{
  assert (m_pExecutable);
  
  DisAssembler disasm;
  
  disasm.SetExecutable (m_pExecutable);
  
  GetCallTargets(disasm);
  
  // we disasseble in a manner that resembles the way an processor
  // would execute (exe-cute, nice, huh?)  the code, that is: we begin
  // at the entry point (and all exports), and follow the code
  
  // we want to disasm so that we get a grip on the extent of
  // functions, but we must also be able to deal with functions with
  // multiple entry points and functions that do not return, but jump
  // to other functions

  // we do this by simple keeping track of entry points
  
  // however, later analysis of the code will benefit from a structure
  // of clear functions (entry,code,code,code,return)
  
  bool bDisAsmMore;
  
  int nDisAsmPass = 0;
  
  printf ("\nDisassembling...\n");
  
  do
    {
      nDisAsmPass++;
      
      PushErrorContext ec_disasm(CONTEXT_DISASSEMBLY);
      
      printf ("Disassembly pass %d...\n",nDisAsmPass);

      // first disasm as much as possible of executable, interleaved
      // with simple analysis of code (for example, finding targets of
      // absolute but fixed jumps)
      
      while (disasm.MoreToDisAsm ())
        {
          dword dwDisAsmVA = disasm.GetNextDisAsmVA ();
          
          // printf ("\n%08x\n\n",dwDisAsmVA);
          
          DataPtr* pDataPtr = NULL; 
          
          try
            {
              pDataPtr = GetDisAsmDataPtr (dwDisAsmVA);
              assert (pDataPtr); // or you should have thrown something
              disasm.DisAsm (dwDisAsmVA,pDataPtr);
            }
          catch (Error* pError)
            {
              pError->AddErrorInfoRVA(dwDisAsmVA);
              
              if (pDataPtr)
                {				
                  dword dwCodeEndVA = pDataPtr->GetRVA();
                  
                  pError->AddErrorInfoRVA(dwCodeEndVA);
                  
                  char* fmt = pDataPtr->FormatAsHexBytes(dwDisAsmVA,
							 dwCodeEndVA);
                  EMIT_ERROR(ERROR_DISASM_DECODING_INSTRUCTIONS_FAILED(fmt,dwDisAsmVA,dwCodeEndVA));
		  delete fmt;
                }
              errorTracker.AddError(pError);
            }
          delete pDataPtr;
          
          /*
            printf ("\nException(2): %s\n",str);
            
            printf ("%08x: ",dwDisAsmVA);
            
                                for (dword dwByteVA = dwDisAsmVA; dwByteVA < dwCodeEndVA; dwByteVA++)
                                {
                                byte bByte = pDataPtr->GetByte(dwByteVA);				
                                printf ("%02x ",bByte);
                                }
                                printf ("\n");
          */
          /*
            catch (...)
            {
                                printf ("%08x: Unknown Exception!!\n",dwDisAsmVA);
                                }
          */
          
          // now analyze code. It must be done while disassembling, as
          // we may find referneces to functions via pointers (and
          // hopefully, we'll find all...)
          
        }
      bDisAsmMore = false;
      
      // do the heavy analysis of the code, like figuring out possible
      // function pointers used at calls, switch statements using
      // jumps
      
      // we might also be able to remove some dead code dead code
      
      // this may lead to more code having to be disassembled, so
      // we're inside an loop
      
      // first, fixup all pointers between code traces
      
      disasm.DeleteUnusedTargets();
      // disasm.FixUpTargetPointers ();
      
#ifndef NDEBUG
      disasm.ConsistencyCheck ();
#endif
      
      ec_disasm.Pop();
      ec_disasm.Push(CONTEXT_DATA_ANALYSIS);
      
      printf ("Data analysis...\n");
      
      // analysis
      
      ReportDisAsmInfoOffset reportDisAsmInfo(disasm.GetMapCodeTraceCallFunction(),
                                              disasm.GetMapBranchTargetVA());
      
      printf ("Extra HTML & Java disasm...\n");
      
      UrlFolder* pFolderDisAsmExtra = pFolder->AddFolder("disasm_extra");
      
      disasm.DataAnalysis(pFolderDisAsmExtra,reportDisAsmInfo);
      
    }
  while (bDisAsmMore);
  
  printf ("Disassembly finished\n");
  
  
  
  UrlFolder* pFolderDisAsm = pFolder->AddFolder("disasm");
  
  printf ("HTML & Java disasm...\n");
  
  ReportDisAsmInfoOffset reportDisAsmInfo(disasm.GetMapCodeTraceCallFunction(),
                                          disasm.GetMapBranchTargetVA());
  
  m_pExecutable->HtmlJavaDisAsmCreateUrls (pFolderDisAsm,reportDisAsmInfo);
  m_pExecutable->HtmlJavaDisAsm (reportDisAsmInfo);
  
  
  /*	
        //	dwEntryPointVA = peOptHeader.BaseOfCode;
        
        MapHandledCode mapHandledCode; // all handled code is added to this list, to avoid multiple disassembly
        
        
        // add all exported functions to stack
        
        if (m_pExportedEntries)
                m_pExportedEntries->GetExportedFunctionVAs (stackDisAsmVA,pDataPtr);

        stackDisAsmVA.push(dwEntryPointVA);

        while (!stackDisAsmVA.empty())
        {
                dword dwStartVA = stackDisAsmVA.top();
                stackDisAsmVA.pop ();
                
                printf ("\n%08x\n\n",dwStartVA);

                // start disassembly at dwInstrVA

                DataPtr* pDataPtr = GetDataPtr (dwStartVA);

                // disassemble until we hit a jmp,call or ret (any branching, control transfer instruction)

//		BranchInstr* pPendingBranchInstr = NULL;
//		BranchTarget* pBranchTarget;
//
//		while (pBranchTarget = DisAsmCodeTrace (pDataPtr,&pPendingBranchInstr))
//		{
//
//
//		}

		InstructionDecoder decoder;

		BranchTarget* pBranchTarget = NULL;

		dword dwStartLastVA;

		for ( ; ; )
		{
			dwStartLastVA = pDataPtr->GetRVA();
			Instruction *pInstr = GetNextInstruction (decoder,pDataPtr);

			if (!pInstr)
			{
				// decoding failed
				break;
			}

			BranchInstr* branchInstr = NULL;

			if (branchInstr = dynamic_cast<BranchInstr*>(pInstr))
			{
				if (pBranchTarget)
					AddBranchTarget (pBranchTarget,dwStartVA,dwStartLastVA,mapHandledCode);

				pBranchTarget = new BranchTarget;

				pBranchTarget->AddInstruction (pInstr); // keep 

				dword dwEndThisVA = pDataPtr->GetRVA();

				AddBranchTarget (pBranchTarget,dwStartLastVA,dwEndThisVA,mapHandledCode);

				pBranchTarget = NULL;

				// sort out what we're pointing to

				// if we may continue, set the following instruction as a branch target

				if (branchInstr->CanTraceContinue ())
				{
					BT_VA_CT target;

					target.m_dwVA = dwEndThisVA;
					target.m_pCodeTrace = NULL;
					
					branchInstr->m_listBranchTargets.push_back(target);
				}

				if (branchInstr->IsTargetRelative ())
				{
					// the target is encoded using offset in immediate

					// so? get immediate

					dword dwTargetVA = dwEndThisVA+branchInstr->GetOffset();

					BT_VA_CT target;

					target.m_dwVA = dwTargetVA;
					target.m_pCodeTrace = NULL;

					stackDisAsmVA.push(dwTargetVA);
				}

				if (!branchInstr->IsConditional () &&
					!branchInstr->IsInterrupt ())
					break;
				// continue if we were conditional
				// or interrupt, otherwise continue with list
			}
			else
			{
				if (!pBranchTarget)
					pBranchTarget = new BranchTarget;

				pBranchTarget->AddInstruction (pInstr); // keep 
			}
		}
		if (pBranchTarget)
			AddBranchTarget (pBranchTarget,dwStartVA,dwStartLastVA,mapHandledCode);


		
	}

	printf ("\nDisasm finished...\n");
*/

  return true;
}


BranchTargetCtrlTransfer* FindReturnInstruction(BranchTarget* pStart)
{
  BranchTargetCtrlTransfer* pReturn = NULL;
  
  set<BranchTarget*> checked;
  stack<BranchTarget*> tocheck;
  
  tocheck.push(pStart); 
  
  while (!tocheck.empty())
    {
      BranchTarget* pCheck = tocheck.top();
      tocheck.pop();
      checked.insert(pCheck);
      
      BranchTargetCtrlTransfer* pCtrl = 
	dynamic_cast<BranchTargetCtrlTransfer*>(pCheck);
      
      if (pCtrl)
	{
	  // is it a return instruction
	  
	  BranchInstr* pCtrlInstr = pCtrl->GetBranchInstr();
	  
	  if (pCtrlInstr->m_nInstruction == instrRETN)
	    {
	      if (pReturn)
		{
		  printf ("Multiple return...");
		  return NULL; // we already got one
		}
	      
	      printf ("Found return...");
	      pReturn = pCtrl;
	    }
	  
	  FOR_EACH_ITER(VectTargets,pCtrlInstr->m_vectTargets,target)
	    {
	      JumpTarget* pJump = dynamic_cast<JumpTarget*>(*target);
	      
	      if (pJump)
		{
		  BranchTarget* pNext = pJump->m_pTarget;
		  
		  assert (pNext);
		  
		  if (pNext && checked.find(pNext) == checked.end())
		    tocheck.push(pNext);
		}
	    }
	  
	}
      
      BranchTarget* pNext = pCheck->GetNextTarget();
      
      if (pNext && checked.find(pNext) == checked.end())
	tocheck.push(pNext);
    }
  
  return pReturn;
}

void FixStackArguments(BranchTarget* pStart,Executable* pExecutable)
{
  set<BranchTarget*> checked;
  stack<BranchTarget*> tocheck;
  
  tocheck.push(pStart); 

  while (!tocheck.empty())
    {
      BranchTarget* pCheck = tocheck.top();
      tocheck.pop();
      checked.insert(pCheck);
      
      ////////////////////////////////////////////////////////

      // ok, so if the instruction has a memory operand that goes and
      // mess with any argument, i.e. a stack pointer with a positive
      // offset, decrease it by four to account for that an inlined
      // function does not have an return pointer on the stack
      
      BranchTargetTraceSingle* pTraceSingle;
      BranchTargetTrace* pTraceMultiple;
      BranchTargetCtrlTransfer* pTraceCtrlTransfer;
      
      Instruction* pInstr = NULL;
      
      if ((pTraceSingle = 
	   dynamic_cast<BranchTargetTraceSingle*>(pCheck)) != NULL)
	pInstr = pTraceSingle->GetInstruction();
      else if ((pTraceMultiple = 
		dynamic_cast<BranchTargetTrace*>(pCheck)) != NULL)
	assert (false);
      else if ((pTraceCtrlTransfer = 
		dynamic_cast<BranchTargetCtrlTransfer*>(pCheck)) != NULL)
	pInstr = pTraceCtrlTransfer->GetBranchInstr();
      
      assert (pInstr);

      MapBTOutgoingState::iterator state = 
	pExecutable->m_analysis.mapOutgoingState.find(pCheck);
      
      assert (state != pExecutable->m_analysis.mapOutgoingState.end()); // because every instruction should have been analysed
      
      Variable* pVarMemAddr;
      VarContStackPtr* pStackPtr;
      const OperandMemory* pOpMem;
      
      if ((pVarMemAddr = 
	   (state->second)->pRegState->GetRegisterNoFixUnknown(reg_fakeMEMADDR_DEST)) &&
	  (pStackPtr = dynamic_cast<VarContStackPtr*>(pVarMemAddr->m_pVarContent)) &&
	  ((sdword) pStackPtr->m_dwOffset) >= 0)
	{
	  assert (pStackPtr->m_dwOffset >= 4); // or we're messing with the return ptr
	  
	  pOpMem = dynamic_cast<const OperandMemory*>(pInstr->GetDestOp());
	  assert (pOpMem); // must have memory operand (or could be push, but we can't fix up a push...)
	  
	  printf ("Found argument ptr (dest) ... %d , %d -> %d\n",
		  (sdword) pStackPtr->m_dwOffset,pOpMem->m_dwDisp,pOpMem->m_dwDisp-4);
	  
	  (const_cast<OperandMemory*>(pOpMem))->m_dwDisp -= 4; // TODO: remove const_cast : this is really nasty
	}
      
      if ((pVarMemAddr = (state->second)->pRegState->GetRegisterNoFixUnknown(reg_fakeMEMADDR_SRC)) &&
	  (pStackPtr = dynamic_cast<VarContStackPtr*>(pVarMemAddr->m_pVarContent)) &&
	  ((sdword) pStackPtr->m_dwOffset) >= 0)
	{
	  assert (pStackPtr->m_dwOffset >= 4); // or we're messing with the return ptr
	  
	  pOpMem = dynamic_cast<const OperandMemory*>(pInstr->GetSrcOp());
	  assert (pOpMem); // must have memory operand (or could be push, but we can't fix up a push...)
	  
	  printf ("Found argument ptr (src) ... %d , %d -> %d\n",
		  (sdword) pStackPtr->m_dwOffset,pOpMem->m_dwDisp,pOpMem->m_dwDisp-4);
	  
	  (const_cast<OperandMemory*>(pOpMem))->m_dwDisp -= 4; // TODO: remove const_cast : this is really nasty
	}
      
      






      ////////////////////////////////////////////////////////

      BranchTarget* pNext = pCheck->GetNextTarget();
      
      if (pNext && checked.find(pNext) == checked.end())
	tocheck.push(pNext);
    }
}

bool ExecutableImage::TuneImage ()
{
  printf ("Tuning image...\n");
  
  // for now we only do one simple thing. Any function that only have
  // one caller, and the caller is also an internal function is
  // inlined...

  const SetCallFunctions* pSetCallFunctions = 
    m_pExecutable->GetSetCallFunctions ();
  
  
  FOR_EACH_CONST_ITER(SetCallFunctions,*pSetCallFunctions,func)
    {
      printf ("Attempting inline...\n");
      
      if ((*func)->m_vectEntryPoints.size() > 1)
	continue; // more than one entry point => more than one source
      
      printf ("1...\n");
      
      CallTarget* pEntryTarget = (*func)->m_vectEntryPoints.front();
      
      printf ("2...\n");
      
      if (pEntryTarget->m_vectFromTrace.size() > 1)
	continue; // more than one source
      
      printf ("3...\n");
      
      if (pEntryTarget->HasExportedAs())
	continue; // also exported

      printf ("DO INLINE\n");
      
      // so the plan is to change the calling instruction into running
      // right into our function, and then change the return
      // instruction to continue right after our function

      // let's be lazy: only try inline when there is only one return,
      // so we can simply omit it, and the entry point is the first
      // instruction, so there is no fall-thorugh to it.

      // we need pointer to the calling function.
      
      BranchTargetCtrlTransfer* pCallBranch = 
	pEntryTarget->m_vectFromTrace.front();
      BranchInstr* pCallInstr = pCallBranch->GetBranchInstr();
      
      assert (pCallInstr->IsCall());
      
      BranchTarget* pPrev = pCallBranch->GetPrevTarget(); // instr before instr call
      BranchTarget* pNext = pCallBranch->GetNextTarget(); // instr after instr call
      
      // no get hold of the called target
      
      BranchTarget* pCalledTarget = pEntryTarget->m_pTarget;
      
      if (pCalledTarget->GetPrevTarget())
	continue; // can't inline because has previous target
      
      // and now find the return instruction
      
      BranchTargetCtrlTransfer* pReturnTrace = 
	FindReturnInstruction(pCalledTarget);
      
      if (!pReturnTrace) // either none or several returns...
	continue; // can't inline

      BranchTarget* pBeforeReturn = pReturnTrace->GetPrevTarget();
      
      if (!pBeforeReturn)
	continue; // somebody _jumped_ to the return. we do not support that here
      
      // now see if anybody was about to jump to the call that we're transforming away

      if (pCallBranch->GetCallSource())
	continue; // can't fix if it is a call target
      
      // from here on we will go through with the change, because now we start to change things...
      
      FixStackArguments(pCalledTarget,m_pExecutable);

      JumpTarget* pCallJumpTarget = pCallBranch->GetJumpSource();
      
      if (pCallJumpTarget)
	{
	  printf ("Killed CALL was jump target...\n");
	  
	  JumpTarget* pNewJumpTarget = pCalledTarget->GetJumpSource();
	  if (!pNewJumpTarget)
	    {
	      printf ("Called target was not a jump target...\n");
	      pNewJumpTarget = new JumpTarget();
	      pCalledTarget->SetJumpSource(pNewJumpTarget);
	    }

	  FOR_EACH_ITER(VectFromTrace,pCallJumpTarget->m_vectFromTrace,from)
	    {
	      printf ("Changing 1 jumper...\n");
	      
	      BranchInstr* pJumpToUs = (*from)->GetBranchInstr();
	      
	      pJumpToUs->RemoveTarget(pCallJumpTarget);
	      pJumpToUs->AddTarget(pNewJumpTarget);
	      pNewJumpTarget->AddTargetSource(*from);
	    }
	}
      
      printf ("DOING INLINE\n");

      pPrev->ClearNextTarget (); // clear so we can insert
      pCallBranch->ClearNextTarget ();
      pBeforeReturn->ClearNextTarget();
      
      pPrev->SetNextTarget(pCalledTarget);
      pBeforeReturn->SetNextTarget(pNext);   

      m_pExecutable->RemoveCallFunction(*func); // we've been inlined, so we're no longer needed
      
      // pCalledTarget->m_pSource = pCallBranch->m_pSource; // if anybody jumped to us...

      /*
      VectCodeTraces vectCodeTraces;
      
      (*func)->GetCodeTraces(vectCodeTraces);
      
      FOR_EACH_CONST_ITER(VectCodeTraces,vectCodeTraces,trace)
	{
	  BranchTarget* pTarget = (*trace);
	  
	  while (pTarget)
	    {
      */

    }
  




  
  return true;
}





