/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ExportedEntries.cpp: implementation of the ExportedEntries class.
//
//////////////////////////////////////////////////////////////////////

#include "ExportedEntries.h" 

#include "DataDestRelocPtr.h" 
#include "DataPtrPESectionBody.h"  
#include "ForwardedExportedEntry.h"
#include "ExportedObject.h"

#include "gen/ExportedEntriesError.h" 


#include "RelocatableDest.h" 


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ExportedEntries::ExportedEntries()
{

}

ExportedEntries::~ExportedEntries()
{

}

struct ExportDirectoryTable
{
  dword dwCharacteristics;
  dword dwTimeDate;
  word  wMajorVersion;
  word  wMinorVersion;
  dword dwNameRVA;
  dword dwBase;
  dword dwNumberOfFunctions;
  dword dwNumberOfNames;
  dword dwFunctionsTableRVA;
  dword dwNamesTableRVA;
  dword dwNameOrdinalsTableRVA;
};

void ExportedEntries::ReadPEExports (DataPtrPESectionBody* pDataPtr)
{
  dword dwEDT_RVA = pDataPtr->GetRVA();
  const ExportDirectoryTable* pEDT = 
    (const ExportDirectoryTable*) pDataPtr->GetPtr(sizeof (ExportDirectoryTable));
  
  // we don't care about dwCharacteristics
  
  if (pEDT->dwCharacteristics)
    throw ERROR_EXP_ENTR_EXPORT_CHARACTERISTICS_NOT_ZERO()->RVA(dwEDT_RVA);
  
  // do not care about time/date (we can't reuse it, as we'll be
  // messing around with the file...)
  
  m_dwTimeDate = pEDT->dwTimeDate;
  m_wMajorVersion = pEDT->wMajorVersion;
  m_wMinorVersion = pEDT->wMinorVersion;
  
  m_strLibraryName = pDataPtr->GetString (pEDT->dwNameRVA);
  
  if (pEDT->dwNumberOfFunctions)
    {
      DataPtrPESectionBody* pDataPtrFunctionsTable = 
	pDataPtr->GetRVADataPtr(pEDT->dwFunctionsTableRVA);	
      
      for (unsigned int nOrdinalOffset = 0; 
	   nOrdinalOffset < pEDT->dwNumberOfFunctions; nOrdinalOffset++)
        {
          dword dwEntryRVA = pDataPtrFunctionsTable->GetDWord();
	  
          ExportedEntry* pExportedEntry;
	  
          if (pDataPtr->IsValidAddress (dwEntryRVA)) // RVA within section that has exports directory
            {
              pExportedEntry = 
		new ForwardedExportedEntry(pDataPtr->GetString (dwEntryRVA));
            }
          else
            {
              pExportedEntry = new ExportedObject(dwEntryRVA);
            }
	  
          dword dwOrdinal = pEDT->dwBase + nOrdinalOffset;

          m_mapExportedEntries.insert(MapOrdinalExportedEntry::value_type(dwOrdinal,pExportedEntry));
        }
      delete pDataPtrFunctionsTable;
    }
  
  if (pEDT->dwNumberOfNames)
    {
      DataPtrPESectionBody* pDataPtrNamesTable = 
	pDataPtr->GetRVADataPtr(pEDT->dwNamesTableRVA);	
      DataPtrPESectionBody* pDataPtrNameOrdinalsTable = 
	pDataPtr->GetRVADataPtr(pEDT->dwNameOrdinalsTableRVA);	
      
      for (unsigned int nName = 0; nName < pEDT->dwNumberOfNames; nName++)
        {
          dword dwNameRVA = pDataPtrNamesTable->GetDWord();
          dword dwOrdinalIndex = pDataPtrNameOrdinalsTable->GetWord();
	  
          // assuming pEDT->dwBase has NOT been added to dwOrdinalIndex
	  
          if (dwOrdinalIndex > pEDT->dwNumberOfFunctions)
            throw ERROR_EXP_ENTR_ORDINAL_OUTSIDE_RANGE(dwOrdinalIndex,
						       pEDT->dwNumberOfFunctions)->RVA(pDataPtrNameOrdinalsTable->GetRVA()-sizeof(word));
	  
          dword dwOrdinal = pEDT->dwBase + dwOrdinalIndex;
	  
          ExportedEntry* pExportedEntry = 
	    m_mapExportedEntries.find(dwOrdinal)->second;
	  
          pExportedEntry->m_vectNames.push_back (pDataPtr->GetString (dwNameRVA));			
        }
      delete pDataPtrNamesTable;
      delete pDataPtrNameOrdinalsTable;
    }
}

void ExportedEntries::GetExportedObjects (VectExportedObjects& vectorFunctionVA)
{
  for (MapOrdinalExportedEntry::iterator i = m_mapExportedEntries.begin();
       i != m_mapExportedEntries.end(); i++)
    {
      ExportedObject* pExportedObj = dynamic_cast<ExportedObject*>(i->second);
      
      if (pExportedObj)
	{
	  vectorFunctionVA.push_back(pExportedObj);
	}
    }
}
/*
void ExportedEntries::GetExportedFunctionVAs (vector<dword>& vectorFunctionVA,DataPtr* pCodeDataPtr)
{
  for (MapOrdinalExportedEntry::iterator i = m_mapExportedEntries.begin();
       i != m_mapExportedEntries.end(); i++)
    {
      ExportedObject* pExported = dynamic_cast<ExportedObject*>(i->second);

      if (pExported && 
          pCodeDataPtr->IsValidAddress (pExported->m_dwEntryRVA))
        {
          // inside code segment, assume it's a function

          vectorFunctionVA.push_back(pExported->m_dwEntryRVA);
        }
    }
}
*/
typedef map<string,word> MapNameOrdinalIndex;

RelocatableDest* ExportedEntries::WritePEExports ()
{
  ExportDirectoryTable EDT;

  EDT.dwCharacteristics = 0;
  EDT.dwTimeDate = m_dwTimeDate + 101; // 101 is a hack to avoid reuse (and bad binding of any calling executables)
  EDT.wMajorVersion = m_wMajorVersion;
  EDT.wMinorVersion = m_wMinorVersion;
  
  //	uint nNumFunctions = EDT.dwNumberOfFunctions;
  
  // figure out how many exported function names we have, and the length of them
  
  MapNameOrdinalIndex mapNameOrdinalIndex;
  
  dword dwStringsSize = m_strLibraryName.size()-1;
  
  EDT.dwBase = m_mapExportedEntries.begin()->first;
  
  if (m_mapExportedEntries.size())
    {
      //	EDT.dwNumberOfFunctions IS NOT= m_mapExportedEntries.size();
      EDT.dwNumberOfFunctions = ((--(m_mapExportedEntries.end()))->first - EDT.dwBase) + 1;
    }
  else
    EDT.dwNumberOfFunctions = 0;
  
  FOR_EACH_CONST_ITER(MapOrdinalExportedEntry,m_mapExportedEntries,func)
    {
      dword dwOrdinalIndex = func->first-EDT.dwBase;
      const ExportedEntry* pFunc = func->second;		
      
      const ForwardedExportedEntry* pForwardEntry;
      //		const ExportedObject* pEntry;
      
      if ((pForwardEntry = dynamic_cast<const ForwardedExportedEntry*>(pFunc)))
        {
          dwStringsSize += pForwardEntry->m_strForwardedName.size()+1;
        }
      else
        {
          ASSERT_IS_TYPE(const ExportedObject*,pFunc);
          //			pEntry = ASSIGN_AND_ASSERT_IS_TYPE(const ExportedObject*,pFunc);
          // m_dwEntryRVA
        }
      
      FOR_EACH_CONST_ITER(VectString,pFunc->m_vectNames,name)
        {
          dwStringsSize += name->size()+1;
          mapNameOrdinalIndex.insert(MapNameOrdinalIndex::value_type(*name,dwOrdinalIndex));
        }
    }
  
  EDT.dwNumberOfNames = mapNameOrdinalIndex.size();

  dword dwEDTSize = sizeof (ExportDirectoryTable);
  
  dword dwFunctionsTableSize = EDT.dwNumberOfFunctions * sizeof (dword);
  dword dwNamesTableSize     = EDT.dwNumberOfFunctions * sizeof (dword);
  dword dwOrdinalsTableSize  = EDT.dwNumberOfFunctions * sizeof (word);
  
  RelocatableDest* pDest = new RelocatableDest();
  
//  char* pImportTable = new char[dwIDTSize+2*dwLookUpTablesSize+dwLibNameStringSize+dwHintStringSize];
  
  pDest->AllocDataBuffer(dwEDTSize+dwFunctionsTableSize+dwNamesTableSize+dwOrdinalsTableSize+dwStringsSize);
  
//  RelocationInfo* pRelocation = new RelocationInfo();
  
  //  DataDestRelocPtr destPtr((void*) pImportTable,pImportTable+dwIDTSize+2*dwLookUpTablesSize+dwLibNameStringSize+dwHintStringSize,pRelocation);
  
  DataDestRelocPtr destPtr(pDest);

  //DataDestRelocPtr(void* pStart IFDEBUGPARAM(void* pEnd) ,RelocationInfo* pReloc);
  
#define EXP_DIR_PTR (0)
#define EXP_FUNCTIONS_PTR (EXP_DIR_PTR+dwEDTSize)
#define EXP_NAMES_PTR (EXP_FUNCTIONS_PTR+dwFunctionsTableSize)
#define EXP_ORDINAL_PTR (EXP_NAMES_PTR+dwNamesTableSize)
#define EXP_STRINGS_PTR (EXP_ORDINAL_PTR+dwOrdinalsTableSize)

  DataDestRelocPtr destPtrDirectory      (&destPtr,EXP_DIR_PTR       IFDEBUGPARAM(dwEDTSize           ) );
  DataDestRelocPtr destPtrFunctionsTable (&destPtr,EXP_FUNCTIONS_PTR IFDEBUGPARAM(dwFunctionsTableSize) );
  DataDestRelocPtr destPtrNamesTable     (&destPtr,EXP_NAMES_PTR     IFDEBUGPARAM(dwNamesTableSize    ) );
  DataDestRelocPtr destPtrOrdinalsTable  (&destPtr,EXP_ORDINAL_PTR   IFDEBUGPARAM(dwOrdinalsTableSize ) );
  DataDestRelocPtr destPtrStrings        (&destPtr,EXP_STRINGS_PTR   IFDEBUGPARAM(dwStringsSize       ) );
  
  destPtrDirectory.WriteDWord(EDT.dwCharacteristics);
  destPtrDirectory.WriteDWord(EDT.dwTimeDate);
  destPtrDirectory.WriteWord(EDT.wMajorVersion);
  destPtrDirectory.WriteWord(EDT.wMinorVersion);
  
  destPtrDirectory.WriteDWord(destPtrStrings.GetRVA(),RELATIVE_IMAGE_CURRENT);
  destPtrStrings.WriteString(m_strLibraryName);
  
  destPtrDirectory.WriteDWord(EDT.dwBase);
  destPtrDirectory.WriteDWord(EDT.dwNumberOfFunctions);
  destPtrDirectory.WriteDWord(EDT.dwNumberOfNames);
  
  destPtrDirectory.WriteDWord(destPtrFunctionsTable.GetRVA(),
			      RELATIVE_IMAGE_CURRENT);
  destPtrDirectory.WriteDWord(destPtrNamesTable.GetRVA(),
			      RELATIVE_IMAGE_CURRENT);
  destPtrDirectory.WriteDWord(destPtrOrdinalsTable.GetRVA(),
			      RELATIVE_IMAGE_CURRENT);
  
  // fill in Function table
  
  dword dwNextOrdinal = EDT.dwBase;
  
  FOR_EACH_CONST_ITER(MapOrdinalExportedEntry,m_mapExportedEntries,func2)
    {
      dword dwOrdinal = func2->first;
      const ExportedEntry* pFunc = func2->second;		
      
      const ForwardedExportedEntry* pForwardEntry;
      const ExportedObject* pEntry;
            
      for ( ; dwNextOrdinal < dwOrdinal; dwNextOrdinal++)
        destPtrFunctionsTable.WriteDWord(0);
      
      if ((pForwardEntry = dynamic_cast<const ForwardedExportedEntry*>(pFunc)))
        {
          destPtrFunctionsTable.WriteDWord(destPtrStrings.GetRVA(),
					   RELATIVE_IMAGE_CURRENT);
          destPtrStrings.WriteString(pForwardEntry->m_strForwardedName);
        }
      else
        {
          pEntry = ASSIGN_AND_ASSERT_IS_TYPE(const ExportedObject*,pFunc);
	  
          assert (pEntry->m_dwEntryRVA); // cannot be 0, as that would mark an unused export (and this is NOT unused) (so no exports in code section at rva 0 (att image_base directly when creating, and this should not happen)
          assert (pEntry->m_dwEntrySize == 0); // entry size only used by ELF files, so it should not be able to sneak in non-zero here (faulty code somewhere, go get it)
	  
          destPtrFunctionsTable.WriteDWord(pEntry->m_dwEntryRVA,
					   RELATIVE_IMAGE_CODE_SECTION);
        }
      dwNextOrdinal++;
    }
  
  assert (dwNextOrdinal == EDT.dwBase+EDT.dwNumberOfFunctions);
  
  // fill in name and ordinal tables
  
  FOR_EACH_CONST_ITER(MapNameOrdinalIndex,mapNameOrdinalIndex,name_ord)
    {
      const string& name = name_ord->first;
      dword dwOrdinalIndex = name_ord->second;
      
      destPtrNamesTable.WriteDWord(destPtrStrings.GetRVA(),
				   RELATIVE_IMAGE_CURRENT);
      destPtrStrings.WriteString(name);			
      
      destPtrOrdinalsTable.WriteWord(dwOrdinalIndex);
    }
  
  assert (destPtrDirectory.ReacedEnd());
  assert (destPtrFunctionsTable.ReacedEnd());
  assert (destPtrNamesTable.ReacedEnd());
  assert (destPtrOrdinalsTable.ReacedEnd());
  assert (destPtrStrings.ReacedEnd());
  
  return pDest;
}













