/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PEFile.cpp: implementation of the PEFile class.
//
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _lseek lseek
#define _tell(fildes) (lseek(fildes,0,SEEK_CUR))
#endif
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>

#include "PEFile.h" 
#include "MZFile.h"

#include "Misc.h" 

#include "PEFileSectionBody.h"
#include "DisAssembler.h" 
#include "gen/DataPtrSrcError.h" 
#include "Executable.h" 

#include "ImportedEntries.h"
#include "ExportedEntries.h"
#include "ExportedObject.h"
#include "GUIResources.h"
#include "ErrorContext.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

PEFile::PEFile(const string& strFileName)
  : ExecutableImage(strFileName)
{
  mzHeaderAndData = NULL;
  
  pSectionHeaders = NULL;
  
  pSectionBodies = NULL;
  m_pGUIResources = 0;
  
}

PEFile::~PEFile()
{
  free (mzHeaderAndData);
  free (pSectionHeaders);
  
  delete[] pSectionBodies;
  delete m_pGUIResources;
}

void PEFile::ReadFile (int file)
{
  long nLength = _lseek (file,0,SEEK_END);
  
  printf ("%6lx (%7ld): File Length\n",nLength,nLength);
  
  
  _lseek (file,0,SEEK_SET);
  
  IMAGE_MZ_HEADER mzheader;
  
  printf ("%6lx (%7ld): Reading MZ Header (%d)\n",_tell(file),_tell(file),sizeof (mzheader));
  
  VALIDATE(read (file,&mzheader,sizeof (mzheader)) == sizeof (mzheader),
           "Missing MZ header")
    
    // printf ("%d %d %c %c %d %d",mzheader.e_magic % 256,mzheader.e_magic / 256,mzheader.e_magic % 256,mzheader.e_magic / 256,sizeof(mzheader ),sizeof(mzheader.e_magic));
    
    VALIDATE1(mzheader.e_magic == IMAGE_MZ_SIGNATURE,
	      "Wrong magic number: %x\n",mzheader.e_magic);
  
  VALIDATE(mzheader.e_lfanew >= (int) sizeof (mzheader),
	   "PE header cannot start before end of MZ header\n");
  
  // allocate memory and store away stub prgm
  
  mzHeaderAndData = (IMAGE_MZ_HEADER*) malloc (mzheader.e_lfanew);
  
  memcpy (mzHeaderAndData,&mzheader,sizeof (mzheader));
  
  long nRead = mzheader.e_lfanew - sizeof (mzheader);
  
  //printf ("%6lx (%7ld): Reading Stub (%ld)\n",_tell(file),_tell(file),nRead);
  
  int r;
  VALIDATE1(((r = read (file,((char*) mzHeaderAndData)+sizeof (mzheader),nRead)) == nRead),
	    "Premature end of file (at %d) (before PE header)\n",r);
  
  // check PE signatue
  // read PE file header
  
  printf ("%6lx (%7ld): Reading signature and PE Header (%d)\n",_tell(file),_tell(file),sizeof (peHeader));
  
  VALIDATE(read (file,&peHeader,sizeof (peHeader)) == sizeof (peHeader),
	   "PE file header was found chopped off.\n");
  
  VALIDATE(peHeader.dwSignature == IMAGE_NT_SIGNATURE,
	   "PE signature wrong\n");
  
  VALIDATE(peHeader.wMachine == IMAGE_PE_MACHINE_I386 ||
	   peHeader.wMachine == IMAGE_PE_MACHINE_I486 ||
	   peHeader.wMachine == IMAGE_PE_MACHINE_I586,
	   "PE image not for 3/4/586 machine\n");
  
  VALIDATE(peHeader.wSizeOfOptionalHeader == sizeof (IMAGE_PE_OPTIONAL_HEADER),
	   "wSizeOfOptionalHeader has unknown value\n");
  
  printf ("%6lx (%7ld): Reading opt PE Header (%d)\n",_tell(file),_tell(file),sizeof (peOptHeader));
  
  VALIDATE(read (file,&peOptHeader,sizeof (peOptHeader)) == sizeof (peOptHeader),
	   "PE file opt. header was found chopped off.\n");
  
  VALIDATE(peOptHeader.wMagic == IMAGE_NT_OPT_MAGIC,
	   "PE optional header magic wrong\n");
  
  
  
  WARNING1(!peOptHeader.dwCheckSum,
	   "PE opt header contains a checksum: %d.\n",peOptHeader.dwCheckSum);
  
  //	VALIDATE((peOptHeader.SectionAlignment % 4096) == 0,
  //		"PE opt header section alignment not multiple of 4096");
  
  VALIDATE((peOptHeader.dwSectionAlignment >= 0x0020 && 
	    peOptHeader.dwSectionAlignment <= 0xffff),
	   "PE opt header section alignment not within range");
  VALIDATE(powerof2 (peOptHeader.dwSectionAlignment),
	   "PE opt header section alignment not power of 2");
  
  VALIDATE((peOptHeader.dwFileAlignment >= 0x0020 && 
	    peOptHeader.dwFileAlignment <= 0xffff),
	   "PE opt header file alignment not within range");
  VALIDATE(powerof2 (peOptHeader.dwFileAlignment),
	   "PE opt header file alignment not power of 2");
  
  VALIDATE(peOptHeader.dwNumberOfRvaAndSizes == 16,
	   "PE opt header NumberOfRvaAndSizes not 16 (valid but unexpected)");
  
  pSectionHeaders = 
    (IMAGE_SECTION_HEADER*) malloc (sizeof (IMAGE_SECTION_HEADER) * 
				    peHeader.wNumberOfSections);
  
  int i;
  
  for (i = 0; i < peHeader.wNumberOfSections; i++)
    {
      printf ("%6lx (%7ld): Reading Section Header (%d)\n",_tell(file),_tell(file),sizeof (IMAGE_SECTION_HEADER));
      
      VALIDATE((read (file,pSectionHeaders+i,sizeof (IMAGE_SECTION_HEADER)) == sizeof (IMAGE_SECTION_HEADER)),
	       "Premature end of file (reading section header)\n");
      
      VALIDATE1(!(pSectionHeaders+i)->dwVirtualAddress ||
		((pSectionHeaders+i)->dwVirtualAddress < peOptHeader.dwSizeOfImage &&
		 (pSectionHeaders+i)->dwVirtualAddress+(pSectionHeaders+i)->dwSizeOfRawData <= peOptHeader.dwSizeOfImage),
		"Section %d body outside image when loaded\n",i)
	}
  
  pSectionBodies = new PEFileSectionBody[peHeader.wNumberOfSections];
  
  unsigned int nOffsetSectHeaderEnd = _tell(file);
  
  for (i = 0; i < peHeader.wNumberOfSections; i++)
    {
      (pSectionBodies+i)->m_dwCharacteristics = (pSectionHeaders+i)->dwCharacteristics;
      if ((pSectionHeaders+i)->dwPointerToRawData == 0)
	{
	  unsigned int nSize = (pSectionHeaders+i)->dwSizeOfRawData;
	  
	  (pSectionBodies+i)->SetSize((pSectionHeaders+i)->dwVirtualAddress,
				      (pSectionHeaders+i)->Misc.dwVirtualSize,nSize);
	}
      else
	{		
	  VALIDATE1((pSectionHeaders+i)->dwPointerToRawData >= nOffsetSectHeaderEnd,
		    "Section body %d before end of section headers.\n",i);
	  
	  for (int j = 0; j < i; j++)
	    VALIDATE1((pSectionHeaders+i)->dwPointerToRawData >= (pSectionHeaders+j)->dwPointerToRawData+(pSectionHeaders+j)->dwSizeOfRawData ||
		      (pSectionHeaders+i)->dwPointerToRawData+(pSectionHeaders+i)->dwSizeOfRawData <= (pSectionHeaders+j)->dwPointerToRawData,
		      "Section body %d inside other section body.\n",i);
	  
	  unsigned int nSize = (pSectionHeaders+i)->dwSizeOfRawData;
	  
	  if ((pSectionHeaders+i)->Misc.dwVirtualSize > nSize)
	    nSize = (pSectionHeaders+i)->Misc.dwVirtualSize;		
	  
	  VALIDATE1((pSectionBodies+i)->Allocate((pSectionHeaders+i)->dwVirtualAddress,(pSectionHeaders+i)->Misc.dwVirtualSize,nSize),
		    "Failed to allocate memory %d bytes.\n",nSize);
	  
	  unsigned int nOffset = (pSectionHeaders+i)->dwPointerToRawData;
	  
	  pSectionBodies->m_dwFileOffset = (pSectionHeaders+i)->dwPointerToRawData;
	  
	  VALIDATE1(_lseek (file,nOffset,SEEK_SET) == (int) nOffset,
		    "Premature end of file (cannot seek to start of section body %d)\n",i);
	  
	  printf ("%6lx (%7ld): Reading Section Body (%d) (->%ld)\n",_tell(file),_tell(file),(pSectionHeaders+i)->dwSizeOfRawData,_tell(file)+(pSectionHeaders+i)->dwSizeOfRawData);
	  
	  VALIDATE1(read (file,(pSectionBodies+i)->m_pData,(pSectionHeaders+i)->dwSizeOfRawData) == (int) (pSectionHeaders+i)->dwSizeOfRawData,
		    "Premature end of file (reading section body %d)\n",i);
	}
    }
}
/*
	IMAGE_PE_FILE_HEADER peHeader;
	IMAGE_PE_OPTIONAL_HEADER peOptHeader;
*/
/*
typedef struct _IMAGE_PE_FILE_HEADER {
	unsigned short Machine;
	unsigned short NumberOfSections; 
	unsigned long  TimeDateStamp;
	unsigned long  PointerToSymbolTable; 
	unsigned long  NumberOfSymbols;
	unsigned short SizeOfOptionalHeader; 
	unsigned short Characteristics;
} IMAGE_PE_FILE_HEADER;

// Directory Entries
#define IMAGE_DIRECTORY_ENTRY_EXPORT         0
#define IMAGE_DIRECTORY_ENTRY_IMPORT         1
#define IMAGE_DIRECTORY_ENTRY_RESOURCE       2
#define IMAGE_DIRECTORY_ENTRY_EXCEPTION      3
#define IMAGE_DIRECTORY_ENTRY_SECURITY       4
#define IMAGE_DIRECTORY_ENTRY_BASERELOC      5
#define IMAGE_DIRECTORY_ENTRY_DEBUG          6
#define IMAGE_DIRECTORY_ENTRY_COPYRIGHT      7
#define IMAGE_DIRECTORY_ENTRY_GLOBALPTR      8
#define IMAGE_DIRECTORY_ENTRY_TLS            9
#define IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG    10

#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES 16

typedef struct _IMAGE_PE_DATA_DIRECTORY {
	unsigned long VirtualAddress;
	unsigned long Size;
} IMAGE_PE_DATA_DIRECTORY;

typedef struct _IMAGE_PE_OPTIONAL_HEADER
{  
	//   
	// Standard fields.   
	//
    unsigned short  Magic;  
        unsigned char   MajorLinkerVersion;  
        unsigned char   MinorLinkerVersion;
    unsigned long   SizeOfCode;  
        unsigned long   SizeOfInitializedData;
    unsigned long   SizeOfUninitializedData;  
        unsigned long   AddressOfEntryPoint;
    unsigned long   BaseOfCode; 
        unsigned long   BaseOfData;  
        //   
        // NT additional fields.
    //  
        unsigned long   ImageBase;  
        unsigned long   SectionAlignment;
    unsigned long   FileAlignment;  
        unsigned short  MajorOperatingSystemVersion;
    unsigned short  MinorOperatingSystemVersion; 
        unsigned short  MajorImageVersion;
    unsigned short  MinorImageVersion;  
        unsigned short  MajorSubsystemVersion;
    unsigned short  MinorSubsystemVersion;  
        unsigned long   Reserved1;  
        unsigned long   SizeOfImage;
    unsigned long   SizeOfHeaders;   
        unsigned long   CheckSum;  
        unsigned short  Subsystem;
    unsigned short  DllCharacteristics; 
        unsigned long   SizeOfStackReserve;
    unsigned long   SizeOfStackCommit; 
        unsigned long   SizeOfHeapReserve;
    unsigned long   SizeOfHeapCommit; 
        unsigned long   LoaderFlags;
    unsigned long   NumberOfRvaAndSizes;
    IMAGE_PE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
} IMAGE_PE_OPTIONAL_HEADER;
*/

#define PRINT_STRUCT_INFO(field,disp,str) {printf("%-35s " disp "\n",#field,(str)->field,(str)->field);}

#define PRINT_IMAGE_PE_FILE_HEADER_INFO(field,disp) {printf("%-35s " disp "\n",#field,peHeader.field,peHeader.field);}

#define PRINT_IMAGE_PE_OPTIONAL_HEADER(field,disp) {printf("%-35s " disp "\n",#field,peOptHeader.field,peOptHeader.field);}

char cDataDirName[][29] = {
  "Export Directory",
  "Import Directory",
  "Resource Directory",
  "Exception Directory",
  "Security Directory",
  "Base Relocation Table",
  "Debug Directory",
  "Description String",
  "Machine Value (MIPS GP)",
  "TLS Directory",
  "Load Configuration Directory",
  "-",
  "-",
  "-",
  "-",
  "-",
};

char mess[50000];

void PEFile::PrintInfo ()
{
  mess[0] = 38;
  mess[345] = 64;
  
  
  printf ("\nPE_FILE_HEADER: \n\n");
  
  PRINT_IMAGE_PE_FILE_HEADER_INFO(wMachine,"%d");
  PRINT_IMAGE_PE_FILE_HEADER_INFO(wNumberOfSections,"%d");
  PRINT_IMAGE_PE_FILE_HEADER_INFO(dwTimeDateStamp,"0x%08x %d");
  PRINT_IMAGE_PE_FILE_HEADER_INFO(dwPointerToSymbolTable,"0x%08x %d");
  PRINT_IMAGE_PE_FILE_HEADER_INFO(dwNumberOfSymbols,"%d");
  PRINT_IMAGE_PE_FILE_HEADER_INFO(wSizeOfOptionalHeader,"%d");
  PRINT_IMAGE_PE_FILE_HEADER_INFO(wCharacteristics,"0x%08x %d");
  
  printf ("\nPE_OPTIONAL_HEADER: \n\n");
  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMagic,"0x%08x %d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(bMajorLinkerVersion,"%d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(bMinorLinkerVersion,"%d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfCode,"0x%08x %d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfInitializedData,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfUninitializedData,"0x%08x %d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwAddressOfEntryPoint,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwBaseOfCode,"0x%08x %d"); 
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwBaseOfData,"0x%08x %d");  
  
  printf ("\n");
  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwImageBase,"0x%08x %d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSectionAlignment,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwFileAlignment,"0x%08x %d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMajorOperatingSystemVersion,"%d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMinorOperatingSystemVersion,"%d"); 
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMajorImageVersion,"%d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMinorImageVersion,"%d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMajorSubsystemVersion,"%d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wMinorSubsystemVersion,"%d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwReserved1,"%d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfImage,"%d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfHeaders,"%d");   
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwCheckSum,"%d");  
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wSubsystem,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(wDllCharacteristics,"0x%08x %d"); 
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfStackReserve,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfStackCommit,"0x%08x %d"); 
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfHeapReserve,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfHeapCommit,"0x%08x %d"); 
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwLoaderFlags,"0x%08x %d");
  PRINT_IMAGE_PE_OPTIONAL_HEADER(dwNumberOfRvaAndSizes,"%d");
  
  int i;
  
  for(i = 0; i < IMAGE_NUMBEROF_DIRECTORY_ENTRIES; i++)
    {
      if (peOptHeader.DataDirectory[i].dwVirtualAddress ||
	  peOptHeader.DataDirectory[i].dwSize)
	{
	  printf("DataDirectory[%2d].VirtualAddress    0x%08x %s\n",i,peOptHeader.DataDirectory[i].dwVirtualAddress,cDataDirName[i]);
	  printf("DataDirectory[%2d].Size              0x%08x\n",i,peOptHeader.DataDirectory[i].dwSize);
	}
    }
  
  for (i = 0; i < peHeader.wNumberOfSections; i++)
    {
      printf ("\nIMAGE_SECTION_HEADER %d: \n\n",i);
      
      char cName[IMAGE_SIZEOF_SHORT_NAME+1];
      
      cName[IMAGE_SIZEOF_SHORT_NAME] = 0;
      memcpy (cName,(pSectionHeaders+i)->bName,IMAGE_SIZEOF_SHORT_NAME);
      
      printf("%-30s %s\n","Name",cName);
      
      PRINT_STRUCT_INFO(Misc.dwPhysicalAddress,"0x%08x %d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(Misc.dwVirtualSize,"0x%08x %d",pSectionHeaders+i);		
      PRINT_STRUCT_INFO(dwVirtualAddress,"0x%08x %d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(dwSizeOfRawData,"0x%08x %d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(dwPointerToRawData,"0x%08x %d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(dwPointerToRelocations,"%d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(dwPointerToLinenumbers,"%d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(wNumberOfRelocations,"%d",pSectionHeaders+i);
      PRINT_STRUCT_INFO(wNumberOfLinenumbers,"%d",pSectionHeaders+i);
      
      printf("%-35s 0x%08x ","Characteristics",(pSectionHeaders+i)->dwCharacteristics);
      
      if ((pSectionHeaders+i)->dwCharacteristics & 0x00000020)
	printf ("Code ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x00000040)
	printf ("Initialized data ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x00000080)
	printf ("Uninitialized data ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x04000000)
	printf ("Cannot be cached ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x08000000)
	printf ("Not pageable ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x10000000)
	printf ("Shared ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x20000000)
	printf ("Executable ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x40000000)
	printf ("Readable ");
      if ((pSectionHeaders+i)->dwCharacteristics & 0x80000000)
	printf ("Writable ");
      
      printf ("\n");
    }
}

void test();

#define HTML_IMAGE_PE_FILE_HEADER_INFO(field,disp) \
sprintf(fmt,disp,peHeader.field,peHeader.field); \
page << (tr() << td(#field) << td(fmt));
// {printf("%-35s " disp "\n",#field,peHeader.field,peHeader.field);}

#define HTML_IMAGE_PE_OPTIONAL_HEADER(field,disp) \
sprintf(fmt,disp,peOptHeader.field,peOptHeader.field);\
page << (tr() << td(#field) << td(fmt));
// {printf("%-35s " disp "\n",#field,peOptHeader.field,peOptHeader.field);}

#define HTML_STRUCT_INFO(field,disp,str) \
sprintf(fmt,disp,(str)->field,(str)->field);\
page << (tr() << td(#field) << td(fmt));

// {printf("%-35s " disp "\n",#field,(str)->field,(str)->field);}

#include <fstream>

#include "HtmlPage.h" // dep_chk_needed 
#include "HtmlTags.h" // dep_chk_unknown 

void PEFile::HTMLFileInfo (UrlFolder* pParent)
{
  UrlFile* pFilePE = pParent->AddFile ("pe_file.html","PE file - executable");
  
  HtmlPage page(pFilePE);
  
  page.Header();
  
  // page << a("Disassembly").href(pDisAsm);
  
  char fmt[64];
  
  page << h1("PE_FILE_HEADER");
  
  page.Begin(table().border(1));
  
  HTML_IMAGE_PE_FILE_HEADER_INFO(wMachine,"%d");
  HTML_IMAGE_PE_FILE_HEADER_INFO(wNumberOfSections,"%d");
  HTML_IMAGE_PE_FILE_HEADER_INFO(dwTimeDateStamp,"0x%08x %d");
  HTML_IMAGE_PE_FILE_HEADER_INFO(dwPointerToSymbolTable,"0x%08x %d");
  HTML_IMAGE_PE_FILE_HEADER_INFO(dwNumberOfSymbols,"%d");
  HTML_IMAGE_PE_FILE_HEADER_INFO(wSizeOfOptionalHeader,"%d");
  HTML_IMAGE_PE_FILE_HEADER_INFO(wCharacteristics,"0x%08x %d");
  
  page.End();
  page << h1("PE_OPTIONAL_HEADER");
  page.Begin(table().border(1));
  
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMagic,"0x%08x %d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(bMajorLinkerVersion,"%d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(bMinorLinkerVersion,"%d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfCode,"0x%08x %d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfInitializedData,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfUninitializedData,"0x%08x %d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwAddressOfEntryPoint,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwBaseOfCode,"0x%08x %d"); 
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwBaseOfData,"0x%08x %d");  

  page << (tr() << td());
  
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwImageBase,"0x%08x %d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSectionAlignment,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwFileAlignment,"0x%08x %d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMajorOperatingSystemVersion,"%d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMinorOperatingSystemVersion,"%d"); 
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMajorImageVersion,"%d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMinorImageVersion,"%d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMajorSubsystemVersion,"%d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(wMinorSubsystemVersion,"%d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwReserved1,"%d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfImage,"%d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfHeaders,"%d");   
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwCheckSum,"%d");  
  HTML_IMAGE_PE_OPTIONAL_HEADER(wSubsystem,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(wDllCharacteristics,"0x%08x %d"); 
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfStackReserve,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfStackCommit,"0x%08x %d"); 
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfHeapReserve,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwSizeOfHeapCommit,"0x%08x %d"); 
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwLoaderFlags,"0x%08x %d");
  HTML_IMAGE_PE_OPTIONAL_HEADER(dwNumberOfRvaAndSizes,"%d");
  
  int i;
  
  for(i = 0; i < IMAGE_NUMBEROF_DIRECTORY_ENTRIES; i++)
    {
      if (peOptHeader.DataDirectory[i].dwVirtualAddress ||
          peOptHeader.DataDirectory[i].dwSize)
        {
          char fmt1[64];
          char fmt2[64];

          sprintf (fmt1,"DataDirectory[%2d].VirtualAddress",i);
          sprintf (fmt2,"0x%08x",peOptHeader.DataDirectory[i].dwVirtualAddress);

          page << (tr() << td(fmt1) << td(fmt2) << td(cDataDirName[i]));
          
          sprintf (fmt1,"DataDirectory[%2d].Size",i);
          sprintf (fmt2,"0x%08x",peOptHeader.DataDirectory[i].dwSize);
          
          page << (tr() << td(fmt1) << td(fmt2));
        }
    }
  
  page.End();
  
  if (pSectionHeaders)
    for (i = 0; i < peHeader.wNumberOfSections; i++)
      {
	sprintf (fmt,"IMAGE_SECTION_HEADER %d",i);
	
	page << h1(fmt);
	page.Begin(table().border(1));
	
	char cName[IMAGE_SIZEOF_SHORT_NAME+1];
	
	cName[IMAGE_SIZEOF_SHORT_NAME] = 0;
	memcpy (cName,(pSectionHeaders+i)->bName,IMAGE_SIZEOF_SHORT_NAME);
	
	page << (tr() << td("Name") << td(cName));
	
	HTML_STRUCT_INFO(Misc.dwPhysicalAddress,"0x%08x %d",pSectionHeaders+i);
	HTML_STRUCT_INFO(Misc.dwVirtualSize,"0x%08x %d",pSectionHeaders+i);		
	HTML_STRUCT_INFO(dwVirtualAddress,"0x%08x %d",pSectionHeaders+i);
	HTML_STRUCT_INFO(dwSizeOfRawData,"0x%08x %d",pSectionHeaders+i);
	HTML_STRUCT_INFO(dwPointerToRawData,"0x%08x %d",pSectionHeaders+i);
	HTML_STRUCT_INFO(dwPointerToRelocations,"%d",pSectionHeaders+i);
	HTML_STRUCT_INFO(dwPointerToLinenumbers,"%d",pSectionHeaders+i);
	HTML_STRUCT_INFO(wNumberOfRelocations,"%d",pSectionHeaders+i);
	HTML_STRUCT_INFO(wNumberOfLinenumbers,"%d",pSectionHeaders+i);
	
	string str;
	
	if ((pSectionHeaders+i)->dwCharacteristics & 0x00000020)
	  str +=  ("Code ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x00000040)
	  str += ("Initialized data ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x00000080)
	  str += ("Uninitialized data ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x04000000)
	  str += ("Cannot be cached ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x08000000)
	  str += ("Not pageable ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x10000000)
	  str += ("Shared ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x20000000)
	  str += ("Executable ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x40000000)
	  str += ("Readable ");
	if ((pSectionHeaders+i)->dwCharacteristics & 0x80000000)
	  str += ("Writable ");
	
	sprintf (fmt,"0x%08x",(pSectionHeaders+i)->dwCharacteristics);
	
	page << (tr() << td("Characteristics") << td(fmt) << td(str));
	
	page.End();
      }
  
  
  
  
  // test();

        /*
        ofstream page, page2;

        page.open(cFileName);
        page2.open("test.html");

        // page << "<html><head></head><body>Hejsan</body><html>\n";

    Cgicc cgi;
    
    // Output the HTTP headers for an HTML document, and the HTML 4.0 DTD info
    page << HTMLDoctype() << endl;
    page << html().set("lang", "en").set("dir", "ltr") << endl;

    // Set up the page's header and title.
    page << head() << endl;
    page << title() << "GNU cgicc v" << cgi.getVersion() << title() << endl;
    page << head() << endl;
    
    // Start the HTML body
    page << body().set("bgcolor","cc1177") << endl;

    // Print out a message
    page << h1("Hello, world from GNU cgicc") << endl;


    ////////////////////////////

        page2 << HTMLDoctype() << endl;
    page2 << html().set("lang", "en").set("dir", "ltr") << endl;

    // Set up the page's header and title.
    page2 << head() << endl;
    page2 << title() << "GNU cgicc v" << cgi.getVersion() << title() << endl;
    page2 << head() << endl;
    
    // Start the HTML body
    page2 << body().set("bgcolor","1177cc") << endl;

    // Print out a message
    page2 << h1("Hello, world from GNU cgicc") << endl;

    // Close the document
    page2 << body() << html();

    ////////////////////////////
        
        // Close the document
    page << body() << html();


*/

	/*
  HTMLPage page;

  page.Open (cFileName);

  page.Heading(1,"PE file info");
    


  page.Close ();
  */
}

bool PEFile::IsType (int file)
{
  // If it is a PE file, it has a MS-DOS MZ Header
  // and a PE header, is so: return true;
  
  _lseek (file,0,SEEK_SET);
  
  IMAGE_MZ_HEADER mzheader;
  
  if (read (file,&mzheader,sizeof (mzheader)) < (int) sizeof (mzheader))
    {
      // not even mz header		
      return false;
    }
  
  if (mzheader.e_magic != IMAGE_MZ_SIGNATURE)
    {
      // wrong magic (bad magic)
      return false;
    }
  
  // seek to location of PE header
  
  _lseek (file,mzheader.e_lfanew,SEEK_SET);
  
  unsigned long signature;
  
  if (read (file,&signature,sizeof (signature)) < (int) sizeof (signature))
    {
      // no signature at expected position		
      return false;
    }	
  
  if (signature != IMAGE_NT_SIGNATURE)
    {
      return false;
    }
  
  // good enough to assume PE file
  // (ReadFile will be very strict anyway)
  
  return true;
}
/*
  unsigned char PEFile::GetByte (unsigned long nVirtualAddress)
{
	for (int i = 0; i < nNumSections; i++)
	{
		if (nVirtualAddress < (pSectionBodies+i)->VirtualAddress ||
			nVirtualAddress >= (pSectionBodies+i)->VirtualAddress+(pSectionBodies+i)->Size)
			continue;
		return *((pSectionBodies+i)->pData+(nVirtualAddress-(pSectionBodies+i)->VirtualAddress));
	}
	throw "read outside initialized image";
}

unsigned short PEFile::GetWord (unsigned long nVirtualAddress)
{
	for (int i = 0; i < nNumSections; i++)
	{
		if (nVirtualAddress < (pSectionBodies+i)->VirtualAddress ||
			nVirtualAddress+1 >= (pSectionBodies+i)->VirtualAddress+(pSectionBodies+i)->Size)
			continue;
		return *((unsigned short*) ((pSectionBodies+i)->pData+(nVirtualAddress-(pSectionBodies+i)->VirtualAddress)));
	}
	throw "read outside initialized image";
}

unsigned int PEFile::GetDWord (unsigned long nVirtualAddress)
{
	for (int i = 0; i < nNumSections; i++)
	{
		if (nVirtualAddress < (pSectionBodies+i)->VirtualAddress ||
			nVirtualAddress+3 >= (pSectionBodies+i)->VirtualAddress+(pSectionBodies+i)->Size)
			continue;
		return *((unsigned int*) ((pSectionBodies+i)->pData+(nVirtualAddress-(pSectionBodies+i)->VirtualAddress)));
	}
	throw "read outside initialized image";
}
*/
DataPtrPESectionBody* PEFile::GetDataPtr (dword dwVirtualAddress)
{
  for (int i = 0; i < peHeader.wNumberOfSections; i++)
    {
      if (dwVirtualAddress < (pSectionBodies+i)->m_dwVirtualAddress ||
	  dwVirtualAddress >= (pSectionBodies+i)->m_dwVirtualAddress+(pSectionBodies+i)->m_nSizeOfRawData)
	continue;
      
      return new DataPtrPESectionBody((pSectionBodies+i),dwVirtualAddress);
    }
  throw ERROR_DATA_PTR_SRC_READING_OUTSIDE_SOURCE()->RVA(dwVirtualAddress);
}


/*
#ifdef WIN32
bool PEFile::DisAsm ()
{
	printf ("Beginning disassembly...\n\n");

	unsigned int IP = peOptHeader.AddressOfEntryPoint;

	for ( ; ; )
	{		
		unsigned int IPbegin = IP;
		DATAPTR dataptr = GetDataPtr (IP);
		byte* pBegin = dataptr.pData;

		Instr instr;

		try
		{
			instr.Decode (dataptr);

			IP += dataptr.pData - pBegin;
		}
		catch (char* str)
		{
			printf ("\nException: %s\n\n",str);

			printf ("%08x: ",IPbegin);

			for (; pBegin < dataptr.pData; pBegin++)
			{
				printf ("%02x ",*pBegin);
			}
			printf ("\n");

			char s[64];
			
			instr.AsmInstrString(s);

			printf ("\nSo far decoded:     %s\n",s);

			return false;
		}

		char str1[64];
		char* p1 = str1;

		printf ("%08x: ",IPbegin);

		for (byte* p = pBegin; p < dataptr.pData; p++)
		{
			p1 += sprintf (p1,"%02x ",*p);
		}
		
		char str2[64];

		instr.AsmInstrString(str2);

		printf ("%-30s%s\n",str1,str2);
	}
	return true;
}
#endif WIN32
*/
/*
void PEFile::DecodeInstruction (unsigned int *IP)
{
	unsigned char b1;
	unsigned int IPstart = *IP;

	bool b32 = true;

	for ( ; ; )
	{
		b1 = GetByte ((*IP)++);

		if (((b1 & 0x07) == 0x06 && (b1 & 0xe0) == 0x20) ||
			((b1 & 0xf0) == 0x60 && (b1 & 0x04) == 0x04))
		{
			printf ("prefix: %02x ",b1);
			switch (b1)
			{
			case 0x26: printf ("es:"); break;
			case 0x36: printf ("ss:"); break;
			case 0x2e: printf ("cs:"); break;
			case 0x3e: printf ("ds:"); break;
			case 0x64: printf ("fs:"); break;
			case 0x65: printf ("gs:"); break;
			case 0x66: printf ("opsize"); b32 = false; break;
			case 0x67: printf ("addrsize"); break;
			}
			printf ("\n");
		}
		else
			break;
	}
	if (b1 == 0x90)
	{
		printf ("NOP\n");
		return;
	}
	if ((b1 & 0xf0) == 0x50)
	{
		if ((b1 & 0x08) == 0x08)
			printf ("POP ");
		else
			printf ("PUSH ");
		if (b32)
			printf ("E");
		switch (b1 & 0x07)
		{
		case 0x00: printf ("AX"); break;
		case 0x01: printf ("CX"); break;
		case 0x02: printf ("DX"); break;
		case 0x03: printf ("BX"); break;
		case 0x04: printf ("SP"); break;
		case 0x05: printf ("BP"); break;
		case 0x06: printf ("SI"); break;
		case 0x07: printf ("DI"); break;
		}
		printf ("\n");
		return;
	}


	printf ("\n%08x (%d):  ",IPstart,IPstart);
	for (unsigned int IPtrace = IPstart; IPtrace < *IP; IPtrace++)
	{
		printf ("%02x ",GetByte (IPtrace));
	}
	printf ("\n");

	throw "Illegal instruction";
}
*/

void PEFile::PostReadFile ()
{
  ExecutableImage::PostReadFile();
  
  ReadImports ();
  ReadExports ();
  ReadResources();
}

bool PEFile::ProcessFile (UrlFolder* pFolder)
{
  // disassemble the file
  
  dword dwEntryPointVA = peOptHeader.dwAddressOfEntryPoint;
  
  m_pExecutable->m_pEntryFunction = new ExportedObject(dwEntryPointVA);
  
  if (!DisAssemble (pFolder))
    return false;
  
  return true;
}

void PEFile::ReadImports ()
{
  dword dwImportDirBaseRVA = 
    peOptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].dwVirtualAddress; 
  
  if (!dwImportDirBaseRVA)
    return; // there are no imports
  
  // ok, so read the import directory
  
  m_pExecutable->m_pImportedEntries = new ImportedEntries;
  
  DataPtrPESectionBody* pDataPtr = NULL;
  
  try
    {
      pDataPtr = GetDataPtr (dwImportDirBaseRVA);
      
      m_pExecutable->m_pImportedEntries->ReadPEImports (pDataPtr);
    }
  CATCH_EMIT_ERROR;
  delete pDataPtr;
}

void PEFile::ReadExports ()
{
  dword dwExportDirBaseRVA = peOptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].dwVirtualAddress; 
  
  if (!dwExportDirBaseRVA)
    return; // there are no imports
  
  // ok, so read the import directory
  
  m_pExecutable->m_pExportedEntries = new ExportedEntries;
  
  DataPtrPESectionBody* pDataPtr = NULL;
  
  try
    {
      pDataPtr = GetDataPtr (dwExportDirBaseRVA);
      m_pExecutable->m_pExportedEntries->ReadPEExports (pDataPtr);
    }
  CATCH_EMIT_ERROR;
  delete pDataPtr;
}

void PEFile::ReadResources ()
{
  // now, ideally, we're not interested in any resources, as our job
  // is to efficienyize the executable code. BUT the resources does
  // (deep down) in the resource tree have some RVAs, which mean that
  // if we move the resource image in the executable
  // (which we'll do as a result of changing and rewriting the machine
  // instructions) we'll have to patch the RVAs anyway. So at the same
  // time, as the block is not writable, we can also se if it can be
  // condensed, particularly by letting similar objects use the same
  // bits. And contrary to common belief, bits in memory are
  // precious. Especially when running on my low end machines with too
  // little of it :-)  And all machines seem to become classifiable as
  // this sooner or later...
  // Bloat: no thanks.

  dword dwResourceDirBaseRVA = 
    peOptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].dwVirtualAddress; 
  
  if (!dwResourceDirBaseRVA)
    return; // there are no imports
  
  // ok, so read the import directory
  
  m_pGUIResources = new GUIResources;
  
  DataPtrPESectionBody* pDataPtr = NULL;
  
  try
    {
      pDataPtr = GetDataPtr (dwResourceDirBaseRVA);
      
      m_pGUIResources->FindRelocations (pDataPtr);
    }
  CATCH_EMIT_ERROR;
  delete pDataPtr;
}

DataPtr* PEFile::GetDisAsmDataPtr (dword dwVirtualAddress)
{
  return GetDataPtr (dwVirtualAddress);
}

void PEFile::GetExportedCallTargets (DisAssembler& disasm)
{
  // Get the call targets that are known by the PE file structures
  
  if (m_pExecutable->m_pExportedEntries)
    {
      VectExportedObjects vectExportedObjects;
      
      //		DataPtr* pDataPtr = GetDataPtr (dwEntryPointVA); 
      
      m_pExecutable->m_pExportedEntries->GetExportedObjects (vectExportedObjects);
      
      FOR_EACH_ITER(VectExportedObjects,vectExportedObjects,object)
	{
	  DataPtrPESectionBody* pDataPtr = GetDataPtr ((*object)->m_dwEntryRVA);
	  
	  if (pDataPtr->m_pDataSrc->IsExecutable())
	    {
	      // section containing exported object has execution
	      // status so assume it's an function and try to
	      // disassemble
	      
	      disasm.AddCallTarget ((*object));
	    }
	  else
	    {
	      // has to be an variable...  deal with that
	      
	    }
	}
    }

  /////////////
  // add all targets from preprocessed mapfile
  /*
	FILE* in = fopen ("pvengine_targets.txt","rt");

	while (!feof(in))
	{
		char str[256];

		fgets (str,sizeof (str),in);

		dword dwOffset;
		
		sscanf(str,"%x ",&dwOffset);

//		printf ("Offset: %08x\n",dwOffset);	

		disasm.AddCallTarget (dwOffset);
	}
	fclose (in);	
*/
}



