/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PEFile_Write.cpp: implementation of the PEFile class.
//
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _lseek lseek
#define _tell(fildes) (lseek(fildes,0,SEEK_CUR))
#endif
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>

#include "PEFile.h" 
#include "MZFile.h"

#include "Misc.h" 

#include "RelocationInfo.h"
#include "RelocatableDest.h"
#include "DataDestRelocPtr.h"
#include "PEFileSectionBody.h"
#include "ReportDisAsmInfoOffset.h" 
#include "gen/DataPtrSrcError.h" 
#include "Executable.h"

#include "ExecutableWriter.h"

#include "ImportedEntries.h"
#include "ExportedEntries.h"
#include "ExportedObject.h"
#include "GUIResources.h"
#include "ErrorContext.h"

#include "CodeSection.h"

#include "BranchTargetTraceSingle.h" 
#include "BranchTargetTrace.h" 
#include "BranchTargetCtrlTransfer.h" 

#include "OperandMemory.h"

//////////////////////////////////////////////////////////////////////

ExecutableImage* PEFile::GetWriteImage(const string& strFileName)
{
  PEFile* pWritePE = new PEFile(strFileName);
  
  // DOS header
  
  // now we have stored all that was in the mz header, but do not
  // care, we just ignore it and create a new one (that means that any
  // stub program will be lost, but we do this because we (without
  // analysis) can't be sure that the stub program did not use
  // anything within the pe part you may want to change this if ficing
  // 16 bit support
  
  pWritePE->mzHeaderAndData = 
    (IMAGE_MZ_HEADER*) malloc (sizeof (IMAGE_MZ_HEADER));
  
  memset (pWritePE->mzHeaderAndData,0,sizeof (IMAGE_MZ_HEADER));
  
  pWritePE->mzHeaderAndData->e_magic = IMAGE_MZ_SIGNATURE;
  
  // PE header
  
  memcpy (&pWritePE->peHeader,&peHeader,sizeof (pWritePE->peHeader));
  
  // PE optional header (which is not optional at all)
  
  memcpy (&pWritePE->peOptHeader,&peOptHeader,sizeof (pWritePE->peOptHeader));
  
  memset (&pWritePE->peOptHeader.DataDirectory,0,
	  sizeof (pWritePE->peOptHeader.DataDirectory));
  
  
  assert (m_pExecutable);
  
  pWritePE->m_pExecutable = m_pExecutable;
  
  return pWritePE;
}

//////////////////////////////////////////////////////////////////////

#include <stddef.h>

#define count_of(array) (sizeof(array)/sizeof(array[0]))

struct SectionHeaderNames
{	
  word wSectionHeader;
  word wSrcSection;
  dword dwCharacteristics;
  byte bName[8];
};

const SectionHeaderNames infoSectionHeaders[] = 
{ 
  { RELATIVE_CODE_SECTION_HEADER  , RELATIVE_CODE_SECTION  , 0x60000020 ,".code\0\0" },
  { RELATIVE_DATA_SECTION_HEADER  , RELATIVE_DATA_SECTION  , 0xc0000040 ,".data\0\0" },
  { RELATIVE_RSRC_SECTION_HEADER  , RELATIVE_RSRC_SECTION  , 0x00000000 ,".rsrc\0\0" },
  { RELATIVE_RELOC_SECTION_HEADER , RELATIVE_RELOC_SECTION , 0x00000000 ,".reloc\0" },
};

struct FileSections
{
  word wSection;
  byte wAlignment; // 0 means default
  byte bPartOfHeader;
};

const FileSections infoPEFileSections[] =
{
  { RELATIVE_MZ_HEADER, 4,1 },
  { RELATIVE_PE_HEADER, 4,1 },				
  { RELATIVE_PE_OPT_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
  { RELATIVE_CODE_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
  { RELATIVE_DATA_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
  { RELATIVE_RSRC_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
  { RELATIVE_RELOC_SECTION_HEADER, 4,1 }, // will be aligned automatically by size of prevoius				
  
  { RELATIVE_CODE_SECTION, 0,0 },
  { RELATIVE_DATA_SECTION, 0,0 },			 
  { RELATIVE_RSRC_SECTION, 0,0 },			 
  { RELATIVE_RELOC_SECTION, 0,0 },
};

bool PEFile::PrepareWrite()
{
/*
	ImportedEntries*	 m_pImportedEntries;
	ExportedEntries* m_pExportedEntries;

	GUIResources*	 m_pGUIResources;
*/

  if (m_pExecutable->m_pImportedEntries)
    m_destSections.AddRelocatableDest(RELATIVE_IMPORT_DIRECTORY,m_pExecutable->m_pImportedEntries->WritePEImports());
  // write the active stuff : executable code when we come here, it
  // should already be decided in what order to write the stuff
  // actually, I expect that when this program has progressed a bit
  // furter in it's development, this will have to be moved to out of
  // this function as some optimizations depend on the exact wordings
  // of the instructions, and there is no use then to rewrite the
  // things inte here once more considering that we also before this
  // need the import table so we can reference imported functions
  
  m_destSections.AddRelocatableDest(RELATIVE_CODE_SECTION,CreateCodeSection ());	
  
  if (m_pExecutable->m_pExportedEntries)
    m_destSections.AddRelocatableDest(RELATIVE_EXPORT_DIRECTORY,
				      m_pExecutable->m_pExportedEntries->WritePEExports());
  
  // now we have to join all sections that are to be joined

  // DOS header
  
  // PE header
  
  //	memset (&peHeader,0,sizeof (peHeader));

  //	peHeader.Characteristics = ;
  peHeader.dwSignature = IMAGE_NT_SIGNATURE;
  peHeader.wMachine = IMAGE_PE_MACHINE_I386;
  peHeader.wNumberOfSections = 0;  // added to further down
  peHeader.dwTimeDateStamp = time(NULL);
  peHeader.dwPointerToSymbolTable = 0; // do now know
  peHeader.dwNumberOfSymbols = 0;      // do not know
  peHeader.wSizeOfOptionalHeader = sizeof (IMAGE_PE_OPTIONAL_HEADER); 
  //  peHeader.wCharacteristics = 0x01 || 0x80;  // FIXME
  
  // PE optional header (which is not optional at all)
  
  //	memset (&peOptHeader,0,sizeof (peOptHeader));
  
  peOptHeader.wMagic = IMAGE_NT_OPT_MAGIC;
  
  peOptHeader.dwSizeOfCode = 0;  // FIXME
  peOptHeader.dwSizeOfInitializedData = 0;  // FIXME
  peOptHeader.dwSizeOfUninitializedData = 0;  // FIXME
  peOptHeader.dwAddressOfEntryPoint = 0; // FIXME: important , relative
  peOptHeader.dwBaseOfCode = 0;  // FIXME
  peOptHeader.dwBaseOfData = 0;  // FIXME
  
  //	peOptHeader.dwImageBase = peOptHeader.dwImageBase; // relocate code to this base
  //	peOptHeader.dwSectionAlignment = peOptHeader.dwSectionAlignment;
  //	peOptHeader.dwFileAlignment = peOptHeader.dwFileAlignment;

  //	peOptHeader.wMajorOperatingSystemVersion = peOptHeader.wMajorOperatingSystemVersion;
  //    peOptHeader.wMinorOperatingSystemVersion = peOptHeader.wMinorOperatingSystemVersion; 
  //	peOptHeader.wMajorImageVersion = peOptHeader.wMajorImageVersion;
  //    peOptHeader.wMinorImageVersion = peOptHeader.wMinorImageVersion;  
  //	peOptHeader.wMajorSubsystemVersion = peOptHeader.wMajorSubsystemVersion;
  //   peOptHeader.wMinorSubsystemVersion = peOptHeader.wMinorSubsystemVersion;  

  peOptHeader.dwReserved1 = 0;
  
  peOptHeader.dwSizeOfImage = 0;
  peOptHeader.dwSizeOfHeaders = 0;
  
  peOptHeader.dwCheckSum = 0; // hmmm, algorithm unknown (will give trouble for any drivers)
  // but manipulating drivers with this program may be mighty stupid anyway (if something goes
  // wrong, the chance of the operating system hanging - or worse, is big anyway), so if anybody
  // want to do that, he (or she) will have to use EDITBIN /RELEASE )
  
  //	peOptHeader.wSubsystem = peOptHeader.wSubsystem;
  //  peOptHeader.wDllCharacteristics = peOptHeader.wDllCharacteristics; 
  //	peOptHeader.dwSizeOfStackReserve = peOptHeader.dwSizeOfStackReserve;
  //  peOptHeader.dwSizeOfStackCommit = peOptHeader.dwSizeOfStackCommit; 
  //	peOptHeader.dwSizeOfHeapReserve = peOptHeader.dwSizeOfHeapReserve;
  //  peOptHeader.dwSizeOfHeapCommit = peOptHeader.dwSizeOfHeapCommit; 
  //	peOptHeader.dwLoaderFlags = peOptHeader.dwLoaderFlags;
  //  peOptHeader.dwNumberOfRvaAndSizes = peOptHeader.dwNumberOfRvaAndSizes;
  
  // IMAGE_PE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
  

  /*
    #define IMAGE_DIRECTORY_ENTRY_EXPORT         0
    #define IMAGE_DIRECTORY_ENTRY_IMPORT         1
    #define IMAGE_DIRECTORY_ENTRY_RESOURCE       2
    #define IMAGE_DIRECTORY_ENTRY_EXCEPTION      3
    #define IMAGE_DIRECTORY_ENTRY_SECURITY       4
#define IMAGE_DIRECTORY_ENTRY_BASERELOC      5
  */
  
  //	m_destSections.
  
  RelocatableDest* pDestImageDos = new RelocatableDest();
  pDestImageDos->AllocDataBuffer(sizeof (IMAGE_MZ_HEADER));
  
  DataDestRelocPtr destImageDos(pDestImageDos);
  
  destImageDos.WriteWord(mzHeaderAndData->e_magic);
  destImageDos.SetOffset(offsetof(IMAGE_MZ_HEADER,e_lfanew));
  destImageDos.WriteDWord(0,RELATIVE_FILE_PE_HEADER);
  
  m_destSections.AddRelocatableDest(RELATIVE_MZ_HEADER,pDestImageDos);
  
  RelocatableDest* pDestImagePE = new RelocatableDest();
  pDestImagePE->AllocDataBuffer(sizeof (IMAGE_PE_FILE_HEADER));
  
  DataDestRelocPtr destImagePE(pDestImagePE);
  
  destImagePE.WriteData(&peHeader,sizeof (peHeader)); // there is nothing to be relocated
  
  m_destSections.AddRelocatableDest(RELATIVE_PE_HEADER,pDestImagePE);
  
  RelocatableDest* pDestImagePEopt = new RelocatableDest();
  pDestImagePEopt->AllocDataBuffer(sizeof (IMAGE_PE_OPTIONAL_HEADER));
  
  DataDestRelocPtr destImagePEopt(pDestImagePEopt);
  
  destImagePEopt.WriteData(&peOptHeader,sizeof (peOptHeader)); // there is nothing to be relocated
  
  // and now write any stuff that might need relocation
  
  // assert (m_destSections.HasSection() == has entry point);
  
  if (m_destSections.HasSection(RELATIVE_EXPORT_DIRECTORY))
    {
      destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,
					DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT]));
      destImagePEopt.WriteDWord(peOptHeader.dwBaseOfData,
				RELATIVE_IMAGE_EXPORT_DIRECTORY);
      destImagePEopt.WriteDWord(m_destSections.GetSectionFileLength(RELATIVE_EXPORT_DIRECTORY));
    }
  if (m_destSections.HasSection(RELATIVE_IMPORT_DIRECTORY))
    {
      destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,
					DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT]));
      destImagePEopt.WriteDWord(peOptHeader.dwBaseOfData,
				RELATIVE_IMAGE_IMPORT_DIRECTORY);
      destImagePEopt.WriteDWord(m_destSections.GetSectionFileLength(RELATIVE_IMPORT_DIRECTORY));
    }
  
  // merge export data section into data section perhaps import should
  // not be merged (is it needed after image has been loaded?)
  
  m_destSections.AppendToEnd (RELATIVE_DATA_SECTION,
			      RELATIVE_IMPORT_DIRECTORY,4); // align = 1 => byte aling, no padding
  m_destSections.AppendToEnd (RELATIVE_DATA_SECTION,
			      RELATIVE_EXPORT_DIRECTORY,4); // align = 1 => byte aling, no padding
  
  if (m_destSections.HasSection(RELATIVE_CODE_SECTION))
    {
      destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,
					dwAddressOfEntryPoint));
      destImagePEopt.WriteDWord(peOptHeader.dwAddressOfEntryPoint,
				RELATIVE_IMAGE_CODE_SECTION);
      destImagePEopt.WriteDWord(peOptHeader.dwBaseOfCode,
				RELATIVE_IMAGE_CODE_SECTION);
    }
  if (m_destSections.HasSection(RELATIVE_DATA_SECTION))
    {
      destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,
					dwBaseOfData));
      destImagePEopt.WriteDWord(peOptHeader.dwBaseOfData,
				RELATIVE_IMAGE_DATA_SECTION);
    }
  
  m_destSections.AddRelocatableDest(RELATIVE_PE_OPT_HEADER,pDestImagePEopt);
  
  uint i;
  
  // and create and section headers needed
  
  for (i = 0; i < count_of(infoSectionHeaders); i++)
    {
      if (m_destSections.HasSection(infoSectionHeaders[i].wSrcSection))
	{
	  // so create the section header
	  
	  IMAGE_SECTION_HEADER sectionheader;
	  memset (&sectionheader,0,sizeof (sectionheader));
	  
	  memcpy(sectionheader.bName,infoSectionHeaders[i].bName,
		 sizeof(sectionheader.bName));
	  // sectionheader.dwVirtualAddress = 0;
	  
	  sectionheader.Misc.dwVirtualSize = 
	    m_destSections.GetSectionImageLength(infoSectionHeaders[i].wSrcSection);
	  
	  sectionheader.dwSizeOfRawData = 
	    m_destSections.AlignSectionFileLength(infoSectionHeaders[i].wSrcSection,
										peOptHeader.dwFileAlignment);
	  // sectionheader.dwPointerToRawData = 0;
	  sectionheader.dwCharacteristics = 
	    infoSectionHeaders[i].dwCharacteristics;
	  
	  RelocatableDest* pDestImageSectHeader = new RelocatableDest();
	  pDestImageSectHeader->AllocDataBuffer(sizeof (IMAGE_SECTION_HEADER));
	  
	  DataDestRelocPtr destImageSectHeader(pDestImageSectHeader);
	  
	  destImageSectHeader.WriteData(&sectionheader,sizeof (sectionheader));
	  
	  destImageSectHeader.SetOffset(offsetof(IMAGE_SECTION_HEADER,
						 dwVirtualAddress));
	  destImageSectHeader.WriteDWord(0,
					 (infoSectionHeaders[i].wSrcSection | 
					  RELATIVE_IMAGE));
	  
	  destImageSectHeader.SetOffset(offsetof(IMAGE_SECTION_HEADER,
						 dwPointerToRawData));
	  destImageSectHeader.WriteDWord(0,
					 (infoSectionHeaders[i].wSrcSection | 
					  RELATIVE_FILE));
	  
	  m_destSections.AddRelocatableDest(infoSectionHeaders[i].wSectionHeader,
					    pDestImageSectHeader);
	  
	  peHeader.wNumberOfSections++;
	}
    }
  
  // collect all headers into two: mz and pe and collect section
  // headers into the header
  
  //	m_destSections.AppendToEnd (RELATIVE_PE_HEADER,RELATIVE_PE_OPT_HEADER,1); // align = 1 => byte aling, no padding
/*
  for (i = 0; i < count_of(infoSectionHeaders); i++)
	{
		assert (m_destSections.HasSection(infoSectionHeaders[i].wSectionHeader) ==
			m_destSections.HasSection(infoSectionHeaders[i].wSrcSection));

		if (m_destSections.HasSection(infoSectionHeaders[i].wSectionHeader))
		{
			m_destSections.AppendToEnd (RELATIVE_PE_HEADER,
							infoSectionHeaders[i].wSectionHeader,1); // align = 1 => byte aling, no padding
		}
	}
*/


  // and sort the stuff into useful order that is, calculate
  // offsets for file and image taking image/file-alignment into
  // account
  
  // I've not found any direct instructions on whether the headers
  // should also go into the image or not. But it looks like it, so
  // begin with those
  
  dword dwFileSize = 0;
  dword dwImageSize = 0;
  dword dwHeaderSize = 0;
  
  for (i = 0; i < count_of (infoPEFileSections); i++)
    {
      if (m_destSections.HasSection(infoPEFileSections[i].wSection))
	{
	  dword dwFileAlign = peOptHeader.dwFileAlignment;
	  dword dwImageAlign = peOptHeader.dwSectionAlignment;
	  
	  if (infoPEFileSections[i].wAlignment)
	    {
	      dwFileAlign  = infoPEFileSections[i].wAlignment;
	      dwImageAlign = infoPEFileSections[i].wAlignment;
	    }
	  
	  assert (powerof2(dwFileAlign));
	  assert (powerof2(dwImageAlign));
	  
	  //			printf ("                        %08x  %08x\n",dwFileAlign,dwImageAlign);
	  //			printf ("%08x  %08x\n",dwFileSize,dwImageSize);
	  
	  dwFileSize  = next_multiple( dwFileSize, dwFileAlign);
	  dwImageSize = next_multiple(dwImageSize,dwImageAlign);
	  
	  //			printf ("%08x  %08x\n",dwFileSize,dwImageSize);
	  
	  // so now we're aligned
	  
	  m_destSections.RelocateSection(infoPEFileSections[i].wSection | 
					 RELATIVE_FILE ,
					 dwFileSize);
	  m_destSections.RelocateSection(infoPEFileSections[i].wSection | 
					 RELATIVE_IMAGE,
					 dwImageSize);
	  
	  dwFileSize  += 
	    m_destSections.GetSectionFileLength (infoPEFileSections[i].wSection);
	  dwImageSize += 
	    m_destSections.GetSectionImageLength(infoPEFileSections[i].wSection);
			
			//			printf ("%08x  %08x\n",dwFileSize,dwImageSize);

	  if (infoPEFileSections[i].bPartOfHeader)
	    dwHeaderSize = dwFileSize;
	}
    }
  
  destImagePE.SetOffset(offsetof(IMAGE_PE_FILE_HEADER,wNumberOfSections));
  destImagePE.WriteDWord(peHeader.wNumberOfSections);
  
  peOptHeader.dwSizeOfImage = dwImageSize;
  peOptHeader.dwSizeOfHeaders = dwHeaderSize;
  
  destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,dwSizeOfImage));
  destImagePEopt.WriteDWord(peOptHeader.dwSizeOfImage);
  destImagePEopt.SetOffset(offsetof(IMAGE_PE_OPTIONAL_HEADER,dwSizeOfHeaders));
  destImagePEopt.WriteDWord(peOptHeader.dwSizeOfHeaders);
  
  


	/*
struct FileSections
{
	word wSection;
	byte wMinAlignment;
	byte wMaxAlignment; // 0 -> unlimited
};

const FileSections infoPEFileSections[] =
*/

	


  // create relocations if we want them relocations can be
  // created before we have commited any relocations pending,
  // but must be after we know where we want to put the stuff
  // (they can of course be patched to, but a) in a complicated
  // way and b), if compacted (by letting the bases of the
  // 4096-byte chunks begin at the next relocation), we might
  // need also to change size (because of moving relocations
  // between chunks)
	


  m_destSections.DoRelocations();
  
  // now reload the headers into the structs in the PEFile class, so
  // they can be used when generating the report for this newly
  // created (and soon written to disk) file
/*
	const RelocatableDest* pPE_Headers = m_destSections.GetRelocatableDest(RELATIVE_PE_HEADER);

	dword nOffset = 0;

	nOffset = pPE_Headers->GetData (&peHeader,   nOffset,sizeof (IMAGE_PE_FILE_HEADER));
	nOffset = pPE_Headers->GetData (&peOptHeader,nOffset,sizeof (IMAGE_PE_OPTIONAL_HEADER));

	uint nSectHeadersSize = sizeof (IMAGE_SECTION_HEADER) * peHeader.wNumberOfSections;
	pSectionHeaders = (IMAGE_SECTION_HEADER*) malloc (nSectHeadersSize);

	nOffset = pPE_Headers->GetData (pSectionHeaders,nOffset,nSectHeadersSize);
*/


//	IMAGE_PE_OPTIONAL_HEADER peOptHeader;

  // ok, so now we have all the sections (but not their location) but
  // to know their location, we need to know the header size so, first
  // we create the header, and fill in the pointers, as they are
  // easily changed by the relocation system, and then we cna place
  // the sections, and the relocations will be automagically fixed
  
  // the only thing being that we do not yet know about how big and so
  // cannot place the relocation section, so that will have to be last


  return true;
}

bool PEFile::WriteFile()
{
  return ExecutableImage::WriteFile();
}

RelocatableDest* PEFile::CreateCodeSection()
{
  CodeSection codeSection;
  
  // this is cheating:
  // just eject everything
  // we don't even care about the jumps
  // this is why ejecting the code must be multi-pass

  // we must first eject the code, and then fix all jumps (assuming
  // 32-bit relative jumps) then change any jumps that can be to 8-bit
  // relatives (only using 2 bytes instead of 6) and the code should
  // be that closely related stuff are close in the code and if jump
  // targets should be aligned on some border, that should be taken
  // care of, and making sure the same function (or more precisely,
  // usually used code within closely related functions), are placed
  // on the same page to minimize paging work by the processor from
  // swap file, and if you say you have a lot of memory anyway, you do
  // not have infinit amounts of cache, so there is a point to this
  // anyway (of course the cache has a finer granularity, but anyway)
  
  // the above is why this should not be called from inside the
  // preparewrite function, but from some other system
  
  // actually, the instruction placement is a very important part of
  // stream-lining the instruction flow, as e.g. the PII instruction
  // decoder likes no instructions longer than 7 bytes, and really
  // want them in order max7,max3,max3... bytes (a 7,3,3,7,3,3,7) to
  // be able to deal with them efficiently

  ExecutableWriter writer;
  
  writer.m_pExecutable = m_pExecutable;
  
  writer.Write (&codeSection);
  
  
  /*
  
  InstructionEncoder encoder;
  
  const SetCallFunctions* pSetCallFunctions = m_pExecutable->GetSetCallFunctions ();
  
  FOR_EACH_CONST_ITER(SetCallFunctions,*pSetCallFunctions,func)
    {
      VectCodeTraces vectCodeTraces;
      
      (*func)->GetCodeTraces(vectCodeTraces);
      
      FOR_EACH_CONST_ITER(VectCodeTraces,vectCodeTraces,trace)
	{
	  BranchTarget* pTarget = (*trace);
	  
	  while (pTarget)
	    {
	      // printf ("pTarget: %08x \n",pTarget);
	      
	      codeSection.AssureDestMemory(16); // 16 bytes maximum per instruction
	      
	      BranchTargetTraceSingle* pTraceSingle;
	      BranchTargetTrace* pTraceMultiple;
	      BranchTargetCtrlTransfer* pTraceCtrlTransfer;
	      
	      Instruction* pInstr = NULL;
	      
	      bool bHasImportedEntryFixup = false;
	      
	      if ((pTraceSingle = dynamic_cast<BranchTargetTraceSingle*>(pTarget)) != NULL)
		{
		  pInstr = pTraceSingle->GetInstruction();
		  //					DetermineStateChange (&pIncoming,outgoingState,
		  //						pTraceSingle->GetInstruction());
		}
	      else if ((pTraceMultiple = dynamic_cast<BranchTargetTrace*>(pTarget)) != NULL)
		{
		  assert (false);
		}
	      else if ((pTraceCtrlTransfer = dynamic_cast<BranchTargetCtrlTransfer*>(pTarget)) != NULL)
		{
		  //					DetermineStateChangeCtrlTransfer (&pIncoming,outgoingState,
		  //						pTraceCtrlTransfer->GetBranchInstr());
		  
		  // Here we really cheat, we brutally assume it's a pointer into
		  // the inport table and fixes it to point to the new position
		  
		  BranchInstr* pBranchInstr = pTraceCtrlTransfer->GetBranchInstr();
		  
		  Operand* pOp = pBranchInstr->m_pTargetOp;
		  
		  OperandMemory* pOpMem;
		  
		  if ((pOpMem = dynamic_cast<OperandMemory*>(pOp)))
		    {
		      assert (peOptHeader.dwImageBase >= 0x80); // so the address won't be shortened by the encoder and therefore not relocatable
		      
		      MapMovedImportAddressTableRVA::iterator fixup = mapMovedImportAddressTableRVA.find(pOpMem->m_dwDisp-peOptHeader.dwImageBase);
		      
		      assert (fixup != mapMovedImportAddressTableRVA.end());
		      
		      pOpMem->m_dwDisp = fixup->second+peOptHeader.dwImageBase;
		      
		      bHasImportedEntryFixup = true;
		    }
		  
		  pInstr = pBranchInstr;
		}
	      
	      assert (pInstr);
	      
	      word wDispImmLocs;
	      
	      encoder.Encode(pInstr,*(codeSection.m_pDest),&wDispImmLocs);
	      
	      if (bHasImportedEntryFixup)
		{
		  assert (DIL_DISP_LEN(wDispImmLocs) == 4);
		  
		  // there is a pointer to import table, somewhere,
		  // at offset DIL_DISP_LOC(wDispImmLocs)
		  // an total length was DIL_LENGTH(wDispImmLocs)
		  
		  codeSection.m_pDest->InsertRelocation(codeSection.m_pDest->GetOffset()-
							DIL_LENGTH(wDispImmLocs)+
							DIL_DISP_LOC(wDispImmLocs),RELATIVE_IMAGE_IMPORT_DIRECTORY);
		  
		}
	      
	      //				pTarget->HtmlPlainDisAsm(page,reportDisAsmInfo);
	      pTarget = pTarget->GetNextTarget();
	    }
	}
    }
  */
  
  /*
    const SetCallFunctions* GetSetCallFunctions () { return &m_setCallFunctions; }
    
    void GetCodeTraces (VectCodeTraces& vectCodeTraces);
  */
  
  
  
  RelocatableDest* pReturn = codeSection.EndWrite();
  
  return pReturn;
}









