/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RelocatableDest.cpp: implementation of the RelocatableDest class.
//
//////////////////////////////////////////////////////////////////////

#include "RelocatableDest.h" 
#include "Misc.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RelocatableDest::RelocatableDest()
{
  m_pData = NULL;
  m_dwSize = 0;
}

RelocatableDest::~RelocatableDest()
{
  free(m_pData);
}

void RelocatableDest::AllocDataBuffer(dword dwSize)
{
  assert (!m_pData == !m_dwSize);
  
  m_pData = (byte*) realloc (m_pData,dwSize);

  if (dwSize > m_dwSize)
    memset (m_pData+m_dwSize,0,dwSize-m_dwSize);
  
  m_dwSize = dwSize;
}

dword RelocatableDest::GetData(void* pDest,dword dwOffset,dword dwLength) const
{
  assert (dwOffset <= m_dwSize && dwOffset+dwLength <= m_dwSize);
  
  memcpy (pDest,m_pData+dwOffset,dwLength);
  
  return dwOffset+dwLength;
}

dword RelocatableDest::AlignLength (dword dwAlign)
{
  AllocDataBuffer (next_multiple(m_dwSize,dwAlign));
  return m_dwSize;
}

void RelocatableDest::AppendToEnd (word wRelocateSelf,
				   const RelocatableDest* pSrc,
				   dword& dwMoveFile,
				   dword& dwMoveImage,
				   dword dwAlign)
{
  assert (!(wRelocateSelf & RELATIVE_TYPE_MASK));
  assert (wRelocateSelf & RELATIVE_SECTION_MASK);
  
  // any relocations to self will be moved so they point into the
  // moved section (i.e. same as self)
  
  // nobody should have any relocations to us, as we do not know how
  // we've moved
  
  // assume the dest section is (or will be) aligned at least as much
  // as we want the source section to be. So we'll align to that.
  
  assert (powerof2 (dwAlign));
  assert (!(m_relocInfo.m_dwFileOffset & (dwAlign - 1)));
  assert (!(m_relocInfo.m_dwImageOffset & (dwAlign - 1)));
  assert (!(pSrc->m_relocInfo.m_dwFileOffset & (dwAlign - 1)));
  assert (!(pSrc->m_relocInfo.m_dwImageOffset & (dwAlign - 1)));
  
  // increase the current size, so when the src is appended, it will be aligned
  
  assert (((dwAlign - m_dwSize) & (dwAlign - 1)) < dwAlign);
  
  dwAlign = ((dwAlign - m_dwSize) & (dwAlign - 1));
  
  m_pData = (byte*) realloc (m_pData,m_dwSize+dwAlign+pSrc->m_dwSize);
  
  memset (m_pData+m_dwSize,0,dwAlign);
  
  m_dwSize += dwAlign; // no we're aligned
  
  memcpy (m_pData+m_dwSize,pSrc->m_pData,pSrc->m_dwSize);
  
  // add any relocations to our list, and fix any internal ones up
  
  assert ((pSrc->m_relocInfo.m_dwImageOffset == RELOCATION_IMAGE_NOT_INCLUDED) ||
          (m_relocInfo.m_dwImageOffset != RELOCATION_IMAGE_NOT_INCLUDED)); // an to-include-in-image section cannot be appended to an not-to-include-in-image section
  
  dwMoveFile  = m_relocInfo.m_dwFileOffset  - 
    pSrc->m_relocInfo.m_dwFileOffset + m_dwSize;
  dwMoveImage = m_relocInfo.m_dwImageOffset - 
    pSrc->m_relocInfo.m_dwImageOffset + m_dwSize;
  
  FOR_EACH_CONST_ITER(MapOffsetRelocInfo,
		      pSrc->m_relocInfo.m_mapOffsetRelocInfo,reloc)
    {
      assert (reloc->first + sizeof (dword) <= pSrc->m_dwSize);
      
      dword dwOffset = m_dwSize+reloc->first;
      word wRelocate = reloc->second.m_wRelocate;
      
      if ((wRelocate & RELATIVE_SECTION_MASK) == RELATIVE_CURRENT)
        wRelocate = (wRelocate & RELATIVE_TYPE_MASK) | wRelocateSelf;
      
      m_relocInfo.InsertRelocation(dwOffset,wRelocate);
    }
  m_dwSize += pSrc->m_dwSize;
}

dword RelocatableDest::Relocate(word wRelocate,dword dwMoveTo)
{
  word wType = (wRelocate & RELATIVE_TYPE_MASK);
  
  assert (wType == RELATIVE_FILE ||
	  wType == RELATIVE_IMAGE);
  assert (wRelocate & RELATIVE_SECTION_MASK);
  
  dword dwMove;
  if (wType == RELATIVE_FILE)
    {
      assert (m_relocInfo.m_dwFileOffset != RELOCATION_IMAGE_NOT_INCLUDED); // uhhh, if it is not suppose to go into the file, why does it exist, and besides, if it is not to be included, there can be no pointers
      
      dwMove = dwMoveTo - m_relocInfo.m_dwFileOffset;
      m_relocInfo.m_dwFileOffset = dwMoveTo;
    }
  else
    {
      assert (m_relocInfo.m_dwFileOffset != RELOCATION_IMAGE_NOT_INCLUDED); // if it is not to be included, there can be no pointers)
      
      dwMove = dwMoveTo - m_relocInfo.m_dwImageOffset;
      m_relocInfo.m_dwImageOffset = dwMoveTo;
    }
  
  //    Relocate(wType | RELATIVE_CURRENT,dwMove);
  
  return dwMove;
}

void RelocatableDest::DoRelocate(word wRelocateSelf,
				 const MapDoRelocation& mapDoRelocation)
{
  assert (!(wRelocateSelf & RELATIVE_TYPE_MASK));
  assert (wRelocateSelf & RELATIVE_SECTION_MASK);
  
  assert (m_pData || m_relocInfo.m_mapOffsetRelocInfo.empty());
  
  FOR_EACH_ITER(MapOffsetRelocInfo,m_relocInfo.m_mapOffsetRelocInfo,reloc)
    {
      assert (reloc->first + sizeof (dword) <= m_dwSize);
      
      word wRelocate = reloc->second.m_wRelocate;
      
      if ((wRelocate & RELATIVE_SECTION_MASK) == RELATIVE_CURRENT)
        wRelocate = (wRelocate & RELATIVE_TYPE_MASK) | wRelocateSelf;
      
      MapDoRelocation::const_iterator move = mapDoRelocation.find(wRelocate);
      
      if (move != mapDoRelocation.end())
        {
          assert (move->second.m_wAppendedTo == 0); // should have been fixed
          *((dword*) (m_pData+reloc->first)) += move->second.m_dwMove;
	  
          // printf ("%d : add %d\n",reloc->first,move->second.m_dwMove);
	  
          if (move->second.m_wChangeRelocation)
            reloc->second.m_wRelocate = move->second.m_wChangeRelocation;
        }
    }
}

/*
void RelocatableDest::Relocate(word wRelocate,dword dwMove)
{
	assert (wRelocate & RELATIVE_TYPE_MASK);
  	assert (wRelocate & RELATIVE_SECTION_MASK);

        assert (m_pData || m_relocInfo.m_mapOffsetRelocInfo.empty());

    FOR_EACH_ITER(MapOffsetRelocInfo,m_relocInfo.m_mapOffsetRelocInfo,reloc)
    {
        assert (reloc->first + sizeof (dword) <= m_dwSize);

        if (reloc->second.m_wRelocate == wRelocate)
        {
            *((dword*) (m_pData+reloc->first)) += dwMove;
        }
    }
}
*/


