/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TryDoInstructionEncoding.cpp: implementation of the TryDoInstructionEncoding class.
//
//////////////////////////////////////////////////////////////////////

#include "TryDoInstructionEncoding.h" 
#include "InstructionEncoder.h"
#include "InstructionDecoder.h"

#include "DataDestPtr.h"

#include "OperandRegister.h" 
#include "OperandMemory.h" 
#include "OperandImmediate.h" 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#define ENCODE_PREFIX_OP_SIZE	0x66
#define ENCODE_PREFIX_ADDR_SIZE	0x67

#define ENCODE_PREFIX_LOCK		0xf0
#define ENCODE_PREFIX_REPNE		0xf2
#define ENCODE_PREFIX_REPE		0xf3

#define ENCODE_TWO_BYTE_OPCODE_ESCAPE	0x0f

const byte bSegRegPrefixCode[6] = { 0x26,0x2e,0x36,0x3e,0x64,0x65 };
const byte bMMXSSEPrefixCode[4] = { 0x00 /* not used */,0x66,0xf3,0xf2 };

void OutputOpSizePrefix (DataDestPtr& dataPtr)
{
  dataPtr.WriteByte (ENCODE_PREFIX_OP_SIZE);
}

void OutputAddrSizePrefix (DataDestPtr& dataPtr)
{
  dataPtr.WriteByte (ENCODE_PREFIX_ADDR_SIZE);
}

void OutputSegRegPrefix (byte bSegReg,DataDestPtr& dataPtr)
{
  assert ((bSegReg & reggTYPE_MASK) == reggSEG);
  assert ((bSegReg & reggNAME_MASK) <= 5);
  /*
    #define regES				(reggSEG | 0x00)  26 
    #define regCS				(reggSEG | 0x01)  2e
    #define regSS				(reggSEG | 0x02)  36
    #define regDS				(reggSEG | 0x03)  3e
    #define regFS				(reggSEG | 0x04)  64
    #define regGS				(reggSEG | 0x05)  65
  */
  dataPtr.WriteByte (bSegRegPrefixCode[bSegReg & reggNAME_MASK]);
}

void OutputRepLockPrefix (byte bRepLock,DataDestPtr& dataPtr)
{
  assert (bRepLock == ENCODE_PREFIX_LOCK ||
	  bRepLock == ENCODE_PREFIX_REPE ||
	  bRepLock == ENCODE_PREFIX_REPNE);
  
  dataPtr.WriteByte (bRepLock);
}

void OutputMMXSSEPrefix (byte nPrefix,DataDestPtr& dataPtr)
{
  assert (nPrefix >= 1 && nPrefix <= 3);

  dataPtr.WriteByte (bMMXSSEPrefixCode[nPrefix]);
}

//////////////////////////////////////////////////////////////////////

bool PrefixDecision::ValidateSegment (const OperandMemory* pOpMem,
				      byte bDefaultSegReg)
{
  assert (!(m_bDecided & PD_SEGMENT)); // do not know about any instruction having two memory operands with overridable segments
  assert (!m_bSegmentReg);
  
  assert ((pOpMem->m_bSegmentReg & reggTYPE_MASK) == reggSEG);
  assert ((pOpMem->m_bSegmentReg & reggNAME_MASK) <= 5);
  
  assert (pOpMem->m_bSegmentReg);
  assert (bDefaultSegReg);
  
  m_bSegmentReg = pOpMem->m_bSegmentReg;
  if (m_bSegmentReg != bDefaultSegReg)
    m_bDecided |= PD_SEGMENT;
  return true;
}

bool PrefixDecision::ValidateAddressSize(bool bAddrSize32)
{
  if (m_bDecided & PD_ADDRSIZE32)
    return bAddrSize32 == !!(m_bSet & PD_ADDRSIZE32);
  
  m_bDecided |= PD_ADDRSIZE32;
  m_bSet |= (bAddrSize32 ? PD_ADDRSIZE32 : 0);
  return true;
}

void PrefixDecision::ValidateRepLock (byte bRepLock)
{
  assert (!(m_bDecided & PD_REPLOCK));
  
  assert (bRepLock == ENCODE_PREFIX_LOCK ||
	  bRepLock == ENCODE_PREFIX_REPE ||
	  bRepLock == ENCODE_PREFIX_REPNE);
  
  m_bRepLock = bRepLock;
  m_bDecided |= PD_REPLOCK;
  
  //	return true; // can only decide once (assert above), so always true
}

byte PrefixDecision::OutputPrefixesLength(bool bCodeSize32) const
{
  byte nLength = 0;
  
  if ((m_bDecided & PD_OPSIZE32) && 
      !!(m_bSet & PD_OPSIZE32) != bCodeSize32)
    nLength++;
  if ((m_bDecided & PD_ADDRSIZE32) && 
      !!(m_bSet & PD_ADDRSIZE32) != bCodeSize32)
    nLength++;
  if ((m_bDecided & PD_SEGMENT))
    {
      assert (m_bSegmentReg);
      nLength++;
    }
  if ((m_bDecided & PD_REPLOCK))
    nLength++;
  
  return nLength;
}

void PrefixDecision::OutputPrefixes(DataDestPtr& dataPtr,
				    bool bCodeSize32,bool bExtraSeg) const
{
  if ((m_bDecided & PD_OPSIZE32) && 
      !!(m_bSet & PD_OPSIZE32) != bCodeSize32)
    OutputOpSizePrefix (dataPtr);
  if ((m_bDecided & PD_ADDRSIZE32) && 
      !!(m_bSet & PD_ADDRSIZE32) != bCodeSize32)
    OutputAddrSizePrefix (dataPtr);
  if ((m_bDecided & PD_SEGMENT) || bExtraSeg)
    {
      assert (!!(m_bDecided & PD_SEGMENT) != bExtraSeg); // cannot have both at once (or bExtraSeg is not unnecessary)
      assert (m_bSegmentReg);
      OutputSegRegPrefix (m_bSegmentReg,dataPtr);
    }
  if ((m_bDecided & PD_REPLOCK))
    OutputRepLockPrefix (m_bRepLock,dataPtr);
}

//////////////////////////////////////////////////////////////////////

void ModRMDecision::SetDisplacement(const OperandMemory* pOpMem)
{
  assert (bMinDispSize == 0);
  
  dwDisp = pOpMem->m_dwDisp;
  bMinDispSize = pOpMem->m_bDispBytes;
}

byte ModRMDecision::OutputDisplacement (DataDestPtr& dataPtr,
					byte bLengthenTricks) const
{
  byte bSize = bMinDispSize & ~MODRM_MAY_ENLARGE_DISP_SIZE;
  
  if (!bSize && !(bLengthenTricks & LENGTHEN_TRICK_DISP))
    return 0;
  
  if (bSize == 4 || (bLengthenTricks & LENGTHEN_TRICK_DISP_TO_32))
    {
      dataPtr.WriteDWord (dwDisp);
      return 4;
    }
  else if (bSize == 2 || (bLengthenTricks & LENGTHEN_TRICK_DISP_TO_16))
    {
      dataPtr.WriteWord (dwDisp);
      return 2;
    }
  else
    {
      assert (bSize == 1 || (bLengthenTricks & LENGTHEN_TRICK_DISP_TO_8));
      dataPtr.WriteByte (dwDisp);
      return 1;
    }
}

//////////////////////////////////////////////////////////////////////

void ImmDecision::SetImmediate(const OperandImmediate* pOpImm,
			       bool bSignedExtended)
{
  if (m_bImmBytes)
    {
      assert (!bSignedExtended);
      
      assert (m_bScndImmBytes == 0);
      
      assert (pOpImm->m_bImmBytes == 1);
      assert (pOpImm->m_bImmHigh16Bytes == 0);
      
      m_bScndImm = pOpImm->m_dwImm;
      m_bScndImmBytes = 1;
      return;
    }
  assert (pOpImm->m_bImmHigh16Bytes == 0 || pOpImm->m_bImmHigh16Bytes == 2);

  m_dwImm = pOpImm->m_dwImm;
  m_wImmHigh16 = pOpImm->m_wImmHigh16;
  m_bImmBytes = pOpImm->m_bImmBytes;
  m_bImmHigh16Bytes = pOpImm->m_bImmHigh16Bytes;
  
  if (bSignedExtended)
    {
      assert (m_bImmBytes > 1);
      m_bImmBytes = 1;
    }
  return;
}

byte ImmDecision::OutputImmediatesLength() const
{
  return m_bImmBytes + m_bImmHigh16Bytes + m_bScndImmBytes;
}

byte ImmDecision::OutputImmediates(DataDestPtr& dataPtr) const
{
  assert (m_bImmHigh16Bytes == 0 || m_bImmHigh16Bytes == 2);
  
  if (!m_bImmBytes)
    {
      assert (!m_bScndImmBytes);
      return 0;
    }
  
  // output immediate
  
  if (m_bImmBytes == 1)
    dataPtr.WriteByte (m_dwImm);
  else if (m_bImmBytes == 2)
    dataPtr.WriteWord (m_dwImm);
  else
    {
      assert (m_bImmBytes == 4);

      dataPtr.WriteDWord (m_dwImm);
    }
  
  // output immediate segment selector
  
  if (m_bImmHigh16Bytes == 2)
    dataPtr.WriteWord (m_wImmHigh16);
  
  // output second immediate
  
  if (!m_bScndImmBytes)
    return m_bImmBytes;
  
  assert (m_bScndImmBytes == 1);
  dataPtr.WriteByte (m_bScndImm);
  
  // as any second immediate can only have length 1 (or be absent),
  // and if it was absent, we can't be here, so it's one, although a
  // compiler does usually not realize that, the very program this is
  // should be able to figure that out, and replace | (...) with | (1
  // << 4) is | 16

  return (m_bImmBytes /*+ m_bScndImmBytes*/) | (m_bScndImmBytes << 4);
}

//////////////////////////////////////////////////////////////////////

bool IsAddrSize32 (const OperandMemory* pOpMem)
{
  byte bBaseRegType = pOpMem->m_bBaseReg & reggTYPE_MASK;
  byte bIndexRegType = pOpMem->m_bIndexReg & reggTYPE_MASK;
  
  assert (!bBaseRegType || !bIndexRegType || (bBaseRegType == bIndexRegType)); // mixed address register sizes now allowed
  
  byte bRegType = bBaseRegType | bIndexRegType;
  
  assert ((bRegType == reggSTD32 || bRegType == reggSTD16) || // correct size
	  (bRegType == 0 && (pOpMem->DisplacementSize () == 4 ||
			     pOpMem->DisplacementSize () == 2))); // or only displacement
  
  if (bRegType == 0)
    {
      assert (pOpMem->DisplacementSize () == 4 ||
	      pOpMem->DisplacementSize () == 2);
      
      return (pOpMem->DisplacementSize () == 4);
    }
  return (bRegType == reggSTD32);
}

bool IsPureDisplacement (const OperandMemory* pOpMem)
{
  if (pOpMem->m_bBaseReg)
    return false;
  if (pOpMem->m_bIndexReg)
    return false;
  
  assert (pOpMem->m_bScale == 0);
  
  assert (pOpMem->DisplacementSize () == 4 ||
	  pOpMem->DisplacementSize () == 2);
  
  return true;
}

bool ValidateOpSIDIMemory (const OperandMemory* pOpMem,
			   byte bSIDI,
			   PrefixDecision& prefix)
{
  if (pOpMem->m_bBaseReg ||
      pOpMem->m_dwDisp ||
      pOpMem->m_bScale ||
      !pOpMem->m_bIndexReg)
    return false;
  
  byte bRegType = pOpMem->m_bIndexReg & reggTYPE_MASK;
  
  assert (bRegType == reggSTD16 || bRegType == reggSTD32);
  
  if (!prefix.ValidateAddressSize (bRegType == reggSTD32))
    return false;
  
  byte bRegName = pOpMem->m_bIndexReg & reggNAME_MASK;
  
  bool bDI = (bSIDI & (ext2_DEST_MEM_DI | ext2_SRC_MEM_DI)) != 0;

  if (bRegName != (bDI ? reggDI : reggSI))
    return false;
  
  if (bDI)
    return (pOpMem->m_bSegmentReg == regES);
  
  // so we are SI, overridable segment DS
  
  return prefix.ValidateSegment (pOpMem,regDS);
}

bool ValidateOperandSizeExtra(word wOpSize,byte bExtra)
{
  return wOpSize == OPSIZE_FROM_EXTRA(bExtra);
}

bool ValidateOperandSize (word wOpSize,
			  const InverseOpcodeInfo* pInverseOpcodeInfo,
			  PrefixDecision& prefix,
			  bool bIsLastOperand)
{
  assert (wOpSize);
  
  if ((pInverseOpcodeInfo->m_nArguments & arg_LAST_OPRND_OPSIZE_MASK) &&
      !(pInverseOpcodeInfo->m_nArguments & arg_SIGN_EXTEND_IMMEDIATE) &&
      bIsLastOperand)
    {
      if (pInverseOpcodeInfo->m_nArguments & arg_LAST_OPRND_DBL_OPSIZE)
	{
	  if (wOpSize & 0x01)
	    return false;
	  return ValidateOperandSize (wOpSize >> 1,
				      pInverseOpcodeInfo,prefix,false);
	}
      else if (pInverseOpcodeInfo->m_nArguments & arg_LAST_OPRND_HALF_OPSIZE)
	{
	  return ValidateOperandSize (wOpSize << 1,
				      pInverseOpcodeInfo,prefix,false);
	}
      assert ((pInverseOpcodeInfo->m_nArguments & arg_LAST_OPRND_OPSIZE_8) ||
	      (pInverseOpcodeInfo->m_nArguments & arg_LAST_OPRND_OPSIZE_16));
      return (wOpSize == 
	      (pInverseOpcodeInfo->m_nArguments & arg_LAST_OPRND_OPSIZE_8 ? 
	       1 : 2));
    }
  /*	if (pInverseOpcodeInfo->m_nArguments & arg_OPSIZE_8)
	return bByteSize == 1; */
  
  if (bIsLastOperand && 
      (pInverseOpcodeInfo->m_nArguments & arg_LAST_OP_MEM_OR_IMM_FAR_PTR))
    wOpSize -= 2; // segment
  
  switch (pInverseOpcodeInfo->m_nArguments & arg_OPSIZE_MASK)
    {
    case arg_OPSIZE_8:
      return wOpSize == 1;
    case arg_OPSIZE_16:
      return wOpSize == 2;
    case arg_OPSIZE_32:
      return wOpSize == 4;
    case arg_OPSIZE_16_32:
      {
	if (prefix.m_bDecided & PD_OPSIZE32)
	  return wOpSize == ((prefix.m_bSet & PD_OPSIZE32) ? 4 : 2);
	
	if (wOpSize != 2 && wOpSize != 4)
	  return false;
	
	prefix.m_bSet |= ((wOpSize == 4) ? PD_OPSIZE32 : 0);
	prefix.m_bDecided |= PD_OPSIZE32;
	return true;
      }
    case arg_OPSIZE_64:
      return wOpSize == 8;
    case arg_OPSIZE_128:
      return wOpSize == 16;
    case arg_OPSIZE_IN_EXTRA2:
      return wOpSize == OPSIZE_FROM_EXTRA(pInverseOpcodeInfo->m_nExtra2);
    case arg_OPSIZE_IN_EXTRAS_16_32:
      if (prefix.m_bDecided & PD_OPSIZE32)
	return ValidateOperandSizeExtra(wOpSize,
					((prefix.m_bSet & PD_OPSIZE32) ? 
					 pInverseOpcodeInfo->m_nExtra2 : 
					 pInverseOpcodeInfo->m_nExtra1));
      
      if (ValidateOperandSizeExtra(wOpSize,pInverseOpcodeInfo->m_nExtra1))
	{
	  prefix.m_bSet |= 0;
	  prefix.m_bDecided |= PD_OPSIZE32;
	  return true;
	}
      else if (ValidateOperandSizeExtra(wOpSize,pInverseOpcodeInfo->m_nExtra2))
	{
	  prefix.m_bSet |= PD_OPSIZE32;
	  prefix.m_bDecided |= PD_OPSIZE32;
	  return true;
	}
      return false;
      
/*
			if ((pInverseOpcodeInfo->m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL) ||
				((pInverseOpcodeInfo->m_nArguments & arg_OP_DEST_MASK) == arg_DEST_AS_REG_EXTRA) ||
				((pInverseOpcodeInfo->m_nArguments & arg_OP_SRC_MASK ) == arg_SRC_AS_REG_EXTRA ) ||
				((pInverseOpcodeInfo->m_nArguments & arg_OP_SRC2_MASK) == arg_SRC2_AS_REG_EXTRA))
			{
				// extra1 already taken, we're in extra2
				
				return ValidateOperandSizeExtra(bOpSize,pInverseOpcodeInfo->m_nExtra2);
			}
			else if (pInverseOpcodeInfo->m_nExtra2 &&
				!(pInverseOpcodeInfo->m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) &&
				((pInverseOpcodeInfo->m_nArguments & arg_OP_SRC_MASK ) != arg_SRC_AS_REG_EXTRA2))
			{
				// we may be in extra2 if it is not used by anoyone else,
				// if it is non-zero

				// in this case, it is dependent on the current operand size attribute

				if (prefix.m_bDecided & PD_OPSIZE32)
					return ValidateOperandSizeExtra(bOpSize,((prefix.m_bSet & PD_OPSIZE32) ? pInverseOpcodeInfo->m_nExtra2 : pInverseOpcodeInfo->m_nExtra1));

				if (ValidateOperandSizeExtra(bOpSize,pInverseOpcodeInfo->m_nExtra1))
				{
					prefix.m_bSet |= 0;
					prefix.m_bDecided |= PD_OPSIZE32;
					return true;
				}
				else if (ValidateOperandSizeExtra(bOpSize,pInverseOpcodeInfo->m_nExtra2))
				{
					prefix.m_bSet |= PD_OPSIZE32;
					prefix.m_bDecided |= PD_OPSIZE32;
					return true;
				}
				return false;
			}
			else
			{
				// else nExtra = info.m_nExtra1; already
				return ValidateOperandSizeExtra(bOpSize,pInverseOpcodeInfo->m_nExtra1);
			}
		}
*/
    default:
      assert (false);
    }
  assert (false); // should never be reached (info.m_nArguments should have opsize specified)
  return true; 
}


/*
#define arg_OPSIZE_8				0x00001000
#define arg_OPSIZE_16				0x00002000
#define arg_OPSIZE_32				0x00004000
#define arg_OPSIZE_16_32			0x00006000	// 16/32, depending in opsize-Prefix

#define arg_OPSIZE_IN_EXTRA			0x00008000

#define arg_OPSIZE_MASK				(arg_OPSIZE_8 | arg_OPSIZE_16 | arg_OPSIZE_32 | arg_OPSIZE_16_32 | arg_OPSIZE_IN_EXTRA)

#define arg_LAST_OPRND_OPSIZE_8		0x00010000	
#define arg_LAST_OPRND_OPSIZE_16	0x00020000	

#define arg_LAST_OPRND_OPSIZE_MASK	(arg_LAST_OPRND_OPSIZE_8 | arg_LAST_OPRND_OPSIZE_16)
*/

bool ValidateOpRegisterTypeNameExtra (const OperandRegister* pOpReg,byte bReg)
{
  assert (bReg & reggTYPE_MASK);
  
  if ((pOpReg->m_bReg & reggNAME_MASK) != (bReg & reggNAME_MASK))
    return false;
  
  assert (reggSTD8 < reggSTD16 &&
	  reggSTD16 < reggSTD32 &&
	  reggSTD32 < reggSEG &&
	  reggSTD32 < reggDBG &&
	  reggSTD32 < reggCTRL &&
	  reggSTD32 < reggTST);
  
  assert (reggSTD8_16_32 < reggSTD16_32 &&
	  reggSTD8_16_32 > reggSTD32 &&
	  reggSTD8_16_32 > reggSEG &&
	  reggSTD8_16_32 > reggDBG &&
	  reggSTD8_16_32 > reggCTRL &&
	  reggSTD8_16_32 > reggTST);
  
  byte bRegType = bReg & reggTYPE_MASK;
  
  if (bRegType == (pOpReg->m_bReg & reggTYPE_MASK))
    {
      assert (bRegType < reggSTD16_32);
      return true;
    }
  if (bRegType == reggSTD8_16_32)
    return (pOpReg->m_bReg & reggTYPE_MASK) <= reggSTD32;
  if (bRegType == reggSTD16_32)
    return ((pOpReg->m_bReg & reggTYPE_MASK) == reggSTD16) ||
      ((pOpReg->m_bReg & reggTYPE_MASK) == reggSTD32);
  return false;
}

bool ValidateOpRegisterStdType (const OperandRegister* pOpReg,
				byte bModRMRegType)
{
  byte bRegType = pOpReg->m_bReg & reggTYPE_MASK;
  
  if (bModRMRegType &= arg_MODRM_RT_MASK)
    {
      if (bModRMRegType == arg_MODRM_RT_FPU_REG)
	return bRegType == reggST;
      else if (bModRMRegType == arg_MODRM_RT_MMX_REG)
	return bRegType == reggMM;
      {
	assert (bModRMRegType == arg_MODRM_RT_SSE_REG);
	return bRegType == reggXMM;
      }
    }
  
  assert (reggSTD8 < reggSTD16 &&
	  reggSTD16 < reggSTD32 &&
	  reggSTD32 < reggSEG &&
	  reggSTD32 < reggDBG &&
	  reggSTD32 < reggCTRL &&
	  reggSTD32 < reggTST);
  
  assert (bRegType == reggSTD8 ||
	  bRegType == reggSTD16 ||
	  bRegType == reggSTD32 ||
	  bRegType == reggSEG ||
	  bRegType == reggDBG ||
	  bRegType == reggCTRL ||
	  bRegType == reggTST ||
	  bRegType == reggST ||
	  bRegType == reggMM ||
	  bRegType == reggXMM);
  
  // it had better be normal
  
  if (bRegType > reggSTD32)
    return false;
  return true;
}

bool ValidateOpRegisterStdTypeAndName (const OperandRegister* pOpReg,
				       byte bRegName,byte bModRMRegType = 0)
{
  if (!ValidateOpRegisterStdType(pOpReg,bModRMRegType))
    return false;
  
  return (bRegName == (pOpReg->m_bReg & reggNAME_MASK));
}

bool ValidateOpRegisterType (const OperandRegister* pOpReg,
			     bool bTypeSpecial,
			     byte bRegSpecialType,
			     byte bModRMRegType)
{
  if (bTypeSpecial)
    {
      byte bRegType = pOpReg->m_bReg & reggTYPE_MASK;
      
      // it's suppose to be special type
      
      if (bRegType != bRegSpecialType)
	return false;
      
      assert (bRegType != reggSEG ||
	      (pOpReg->m_bReg & reggNAME_MASK) < 6);
      assert (bRegType != reggCTRL ||
	      (pOpReg->m_bReg & reggNAME_MASK) == 0 ||
	      (pOpReg->m_bReg & reggNAME_MASK) == 2 ||
	      (pOpReg->m_bReg & reggNAME_MASK) == 3 ||
	      (pOpReg->m_bReg & reggNAME_MASK) == 4);
      return true;
    }
  else
    return ValidateOpRegisterStdType (pOpReg,bModRMRegType);
}

/* Addressing modes available:
Assuming the memory addressing has been simplified as much as possble (that is that
displacement has been reduced to smallest size necessary, the markings mean:
X encodable
+ encodable, swap base & index
1 encodable, with disp8 set to 0
2 encodable, with disp enlarged to disp16

16-bit: MOD/RM
                                                                
          index                
                                                                                        
          --   BX   BP   SI   DI    
                  
base  --  X  2 +012 + 12 +012 +012 
      BX  X012           X012 X012  
      BP  X 12           X012 X012   
      SI  X012 +012 +012             
      DI  X012 +012 +012             

32-bit: MOD/RM

scale1    index

         ---  EAX  ECX  EDX  EBX  ESP  EBP  ESI  EDI
                                           
base --- X  4 +014 +014 +014 +014      + 14 +014 +014                              
     EAX X014                                  
     ECX X014                                  
     EDX X014                                  
     EBX X014                                  
     ESP                                   
     EBP X 14                                  
     ESI X014
     EDI X014                                    

32-bit: with SIB

scale1   index

         ---  EAX  ECX  EDX  EBX  ESP  EBP  ESI  EDI
                                           
base --- X  4 X++4 X++4 X++4 X++4 +014 X +4 X++4 X++4 
     EAX X014 X014 X014 X014 X014 +014 X014 X014 X014 
     ECX X014 X014 X014 X014 X014 +014 X014 X014 X014 
     EDX X014 X014 X014 X014 X014 +014 X014 X014 X014 
     EBX X014 X014 X014 X014 X014 +014 X014 X014 X014 
     ESP X014 X014 X014 X014 X014      X014 X014 X014                                   
     EBP X 14 X+14 X+14 X+14 X+14 +014 X 14 X+14 X+14                                               
     ESI X014 X014 X014 X014 X014 +014 X014 X014 X014 
     EDI X014 X014 X014 X014 X014 +014 X014 X014 X014 

// so with scale 1, anything we can do with mod/rm, we
can also do with SIB,   and if scale is 2,4 or 8, SIB
is necessary anyway (so can't enlarge that way)

scale248 index

         ---  EAX  ECX  EDX  EBX  ESP  EBP  ESI  EDI
                                           
base --- X  4 X  4 X  4 X  4 X  4      X  4 X  4 X  4 
     EAX X014 X014 X014 X014 X014      X014 X014 X014 
     ECX X014 X014 X014 X014 X014      X014 X014 X014 
     EDX X014 X014 X014 X014 X014      X014 X014 X014 
     EBX X014 X014 X014 X014 X014      X014 X014 X014 
     ESP X014 X014 X014 X014 X014      X014 X014 X014                                   
     EBP X 14 X 14 X 14 X 14 X 14      X 14 X 14 X 14                                               
     ESI X014 X014 X014 X014 X014      X014 X014 X014 
     EDI X014 X014 X014 X014 X014      X014 X014 X014 

and for scale 2, 4 or 8, there are no possibilities
to swap base and index, but disp may be enlarged
(if smaller than 4)

we cannot scale with esp as index register

anytime we have a disp0 or disp1, we can also encode with disp2/4
(using more space)

  */

bool ValidateModRMMemOp(const OperandMemory* pOpMem,
			bool bAddrSize32,
			PrefixDecision& prefix,
			ModRMDecision& modrm)
{
  modrm.dwDisp = pOpMem->m_dwDisp;
  modrm.bMinDispSize = (bAddrSize32 ? 4 : 2);
  if (modrm.dwDisp >= -0x80 && modrm.dwDisp < 0x80)
    modrm.bMinDispSize = 1;
  if (!modrm.dwDisp)
    modrm.bMinDispSize = 0;	
  
  byte bBase = pOpMem->m_bBaseReg;
  byte bIndex = pOpMem->m_bIndexReg;
  
  assert (bIndex || !pOpMem->m_bScale);
  
  if (bAddrSize32)
    {
      // see how we would look like with sib byte:
      
      if (!pOpMem->m_bScale) // can only swap if scale is x1
	{
	  bool bNowSegPrefix = 
	    pOpMem->m_bSegmentReg != 
	    (IS_REG_NAME_SP_BP(bBase) ? regSS : regDS);
	  bool bSwapSegPrefix = 
	    pOpMem->m_bSegmentReg != 
	    (IS_REG_NAME_SP_BP(bIndex) ? regSS : regDS);
	  
	  if (bNowSegPrefix && !bSwapSegPrefix) // swap if it will kill segment override prefix (we do not care if it will enlarge disp from 0 bytes to four, as we dislike prefixes, but that would rarely happen)
	    {
	      swap(bBase,bIndex);
	    }
	  else if ((!bSwapSegPrefix || bNowSegPrefix)) // only if it won't create prefix
	    {
	      if (bBase == regEBP && bIndex != 0 && modrm.bMinDispSize == 0)
		swap (bBase,bIndex);
	      else if (bBase == 0 && modrm.bMinDispSize < 4)
		swap (bBase,bIndex);
	    }
	  
	  // cannot encode with ESP as index, swap necessary (even if it creates prefix)
	  if (bIndex == regESP)
	    {
	      if (bBase == regESP)
		return false; // can't encode this
	      swap (bBase,bIndex);
	    }
	}
      else
	{
	  if (bIndex == regESP)
	    return false; // can't encode this
	}
      
      // now encode us as modrm and sib (into bModRMExtraSIB and bSIB)
      // if we can do without SIB that will be noted into bModRM later
      
      byte bMod = (modrm.bMinDispSize == 4 ? 2 : modrm.bMinDispSize);
      
      byte bEncodeBase = (bBase & reggNAME_MASK);
      byte bEncodeIndex = (bIndex & reggNAME_MASK);
      
      if (!bBase)
	{
	  modrm.bMinDispSize = 4;
	  bMod = 0; // can't use modrm.bMinDispSize this time
	  bEncodeBase = reggBP; // no base encoded with code for regEBP
	}
      else if (bBase == regEBP && !modrm.bMinDispSize)
	{
	  modrm.bMinDispSize = 1;
	  bMod = 1;
	}
      
      if (!bIndex)
	bEncodeIndex = reggSP; // no index encoded with code for regESP
      
      modrm.bSIB = (pOpMem->m_bScale << 6) | (bEncodeIndex << 3) | bEncodeBase;
      modrm.bModRMExtraSIB = (bMod << 6) | 4; // set mod part, and 4 as rm for SIB use
      
      // now see if we could squezze us into mod/rm without sib
      // however, if that would require segment override SS otherwise
      // not required, rather take SIB than segment override (SIB
      // quicker decoded by cpu than segment override prefix)
      
      // swaps above should already have but index into base and
      // emptied base if that was possible without creating prefix so
      // there is no reason for us (who will not create a prefix) to
      // try any swaps, just see if we can encode
      
      if (!bIndex && bBase != regESP)
	{
	  modrm.bModRM = (bMod << 6) | bEncodeBase;
	}
      else
	modrm.bModRM = modrm.bModRMExtraSIB;
    }
  else
    {
      // 16 bit address
      
      // there is not much to do about it more than determine if it
      // can be encoded or not
      
      assert (pOpMem->GetScale () == 0);		
      
      modrm.bModRM = modrm.bMinDispSize << 6; // set mod part
      
      if (!bBase && bIndex)
	swap (bBase,bIndex);
      else if (bIndex && bBase > bIndex)
	{
	  // bBase is SI or DI, while index being BX or BP swapping
	  // others do not mather, as they do not encode anyway
	  
	  swap (bBase,bIndex);
	}
      
      switch ((bBase << 8) | bIndex)
	{
	case ((((word) regBX) << 8) | regSI): modrm.bModRM |= 0; break;
	case ((((word) regBX) << 8) | regDI): modrm.bModRM |= 1; break;
	case ((((word) regBP) << 8) | regSI): modrm.bModRM |= 2; break;
	case ((((word) regBP) << 8) | regDI): modrm.bModRM |= 3; break;
	  
	case (((word) regSI) << 8): modrm.bModRM |= 4; break;
	case (((word) regDI) << 8): modrm.bModRM |= 5; break;
	case (((word) regBP) << 8):
	  if (!modrm.bMinDispSize)
	    {
	      modrm.bMinDispSize = 1;
	      assert (!modrm.bModRM);
	      modrm.bModRM = (1 << 6) | 6; // mod | rm
	    }
	  else
	    modrm.bModRM |= 6; 
	  break;
	case (((word) regBX) << 8): modrm.bModRM |= 7; break;		
	case 0: // [disp16]
	  modrm.bMinDispSize = 2; // cannot use modrm.bMinDispSize as mod
	  modrm.bModRM = 0 | 6; // mod | rm 
	  break;
	default:
	  return false; // cannot encode
	  //			assert (false); // base/index combination impossible to encode, and should not have been set anywhere
	  break;
	}
    }
  
  if (!prefix.ValidateSegment(pOpMem,IS_REG_NAME_SP_BP(bBase) ? regSS : regDS))
    return false;
  
  modrm.bMinDispSize |= MODRM_MAY_ENLARGE_DISP_SIZE;
  
  return true;
}

//////////////////////////////////////////////////////////////////////

bool TryDoInstructionEncoding::ValidateOperandTypeSize (const Operand* pOp,
							dword dwOp,
							dword dwOpMask,
							bool bIsLastOperand)
{
  const OperandRegister* pOpReg;
  const OperandMemory* pOpMem;
  const OperandImmediate* pOpImm;

  if ((pOpReg = dynamic_cast<const OperandRegister*>(pOp)) != NULL)
    {
      if (dwOp == arg_OP_IS_MODRM_REG)
	{
	  if (!ValidateOpRegisterType (pOpReg,
				       (m_pInverseOpcodeInfo->m_nArguments & 
					arg_EXTRA_MODRM_REG_IS_SPECIAL) != 0,
				       m_pInverseOpcodeInfo->m_nExtra1,
				       (m_pInverseOpcodeInfo->m_nArguments >> 
					arg_MODRM_RT_REG_SHIFT)))
	    return false;
	  
	  if (m_pInverseOpcodeInfo->m_nArguments & arg_MODRM_RT_REG_MASK)
	    return true; // size ok, (fpu/mmx/sse)
	}
      else if (dwOp == arg_OP_IS_MODRM_MODRM)
	{	
	  if (m_pInverseOpcodeInfo->m_nArguments & arg_MODRM_ONLY_M_MODRM)
	    return false;
	  
	  if (!ValidateOpRegisterStdType (pOpReg,
					  (m_pInverseOpcodeInfo->m_nArguments >> 
					   arg_MODRM_RT_RM_SHIFT)))
	    return false;
	  
	  if (m_pInverseOpcodeInfo->m_nArguments & arg_MODRM_RT_RM_MASK)
	    return true; // size ok, (mmx/sse)
	}
      else if (dwOp == arg_OP_AS_REG_EXTRA)
	{
	  if (!ValidateOpRegisterTypeNameExtra (pOpReg,
						m_pInverseOpcodeInfo->m_nExtra1))
	    return false;
	  
	  if (IS_REG_TYPE_REAL(m_pInverseOpcodeInfo->m_nExtra1))
	    return true; // size is ok as the real register fitted with operand
	}
      else if (dwOp == arg_OP_AS_REG_EXTRA2)
	{
	  if (!ValidateOpRegisterTypeNameExtra (pOpReg,
						m_pInverseOpcodeInfo->m_nExtra2))
	    return false;
	  
	  if (IS_REG_TYPE_REAL(m_pInverseOpcodeInfo->m_nExtra2))
	    return true; // size is ok as the real register fitted with operand
	}
      else if (dwOp == arg_OP_LOW_3BITS_REG)
	{
	  if (!ValidateOpRegisterStdTypeAndName (pOpReg,
						 m_pInverseOpcodeInfo->m_nByte & 0x07))
	    return false;
	}
      else
	return false;
      
      // validate size (common to all) (below if-else-statements)
    }
  else if ((pOpMem = dynamic_cast<const OperandMemory*>(pOp)) != NULL)
    {
      if (dwOp == arg_OP_IS_MODRM_MODRM)
	{	
	  if (m_pInverseOpcodeInfo->m_nArguments & arg_MODRM_ONLY_R_MODRM)
	    return false;
	  
	  if (!m_prefix.ValidateAddressSize(IsAddrSize32(pOpMem)))
	    return false;

	  // further validation will be done when trying to encode modrm
	}
      else if (dwOp == arg_OP_IS_MEM_DISP)
	{
	  if (!IsPureDisplacement (pOpMem))
	    return false;
	  
	  if (!m_prefix.ValidateAddressSize(IsAddrSize32(pOpMem)))
	    return false;
	  
	  if (!m_prefix.ValidateSegment(pOpMem,regDS))
	    return false;
	}
      else if (m_pInverseOpcodeInfo->m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) // this must be last
	{
	  assert (dwOpMask != arg_OP_SRC2_MASK);
	  
	  byte bMaskSIDI = 
	    (dwOpMask == arg_OP_DEST_MASK) ? 
	    (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI) : 
	    (ext2_SRC_MEM_DI | ext2_SRC_MEM_SI);
	  
	  if (!(m_pInverseOpcodeInfo->m_nExtra2 & bMaskSIDI))
	    {
	      // arg_EXTRA2_DEST_SRC_SI_DI was last hope, but this
	      // operand not included (bad luck...)
	      
	      return false;
	    }
	  
	  if (!ValidateOpSIDIMemory (pOpMem,
				     (m_pInverseOpcodeInfo->m_nExtra2 & 
				      bMaskSIDI),
				     m_prefix))
	    return false;
	}
      else
	return false;
      
      // validate size (common to all) (below if-else-statements)
    }
  else if ((pOpImm = dynamic_cast<const OperandImmediate*>(pOp)) != NULL)
    {
      if (dwOp == arg_OP_IS_IMM)
	{
	  // there is nothing to validate
	  // specifically: operand size is validated elsewhere (below)
	  
	  assert (pOpImm->m_bImmHigh16Bytes == 0 || 
		  pOpImm->m_bImmHigh16Bytes == 2);
	  
	  if (!(m_pInverseOpcodeInfo->m_nArguments & 
		arg_LAST_OP_MEM_OR_IMM_FAR_PTR) !=
	      !pOpImm->m_bImmHigh16Bytes)
	    return false;
	  
	  if (m_pInverseOpcodeInfo->m_nArguments & 
	      arg_SIGN_EXTEND_IMMEDIATE)
	    {
	      assert (m_pInverseOpcodeInfo->m_nArguments & 
		      arg_LAST_OPRND_OPSIZE_8);
	      
	      // ok, so it's valid if it's value can be stored without
	      // being destroyed, i.e, is inside [-128,127]
	      
	      if (pOpImm->m_bImmHigh16Bytes ||
		  !pOpImm->SignExtended())
		return false;
	      
	      // but we still have to validate the op-size, as it
	      // should be equal to to other (dest) operand
	    }
	}
      else if (dwOp == arg_OP_CONST_1)
	{
	  return pOpImm->m_dwImm == 1 && !pOpImm->m_bImmHigh16Bytes;
	  // if the immediate is 1, size is not important (or actually correct)
	}
      else
	return false;
      
      // validate size (common to all) (below if-else-statements)
    }

//////////////////////////////////////////////////////////////////////
/*
	if (dwOp == arg_OP_IS_MODRM_REG)
	{
		if (!pOpReg)
			return false;

		if (!ValidateOpRegisterType (pOpReg,
				(m_info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL) != 0,
				m_info.m_nExtra1,
				(m_info.m_nArguments >> arg_MODRM_RT_REG_SHIFT)))
			return false;
	}
	else if (dwOp == arg_OP_IS_MODRM_MODRM)
	{	
		if (pOpReg)
		{
			if (m_info.m_nArguments & arg_MODRM_ONLY_M_MODRM)
				return false;

			if (!ValidateOpRegisterStdType (pOpReg,(m_info.m_nArguments >> arg_MODRM_RT_RM_SHIFT)))
				return false;
		}
		else if (pOpMem)
		{
			if (m_info.m_nArguments & arg_MODRM_ONLY_R_MODRM)
				return false;

			if (!m_prefix.ValidateAddressSize(IsAddrSize32(pOpMem)))
				return false;

			// further validation will be done when trying to encode modrm
		}
		else
			return false;
	}
	else if (dwOp == arg_OP_IS_IMM)
	{
		if (!pOpImm)
			return false;

		// there is nothing to validate

		// specifically operand size is validated elsewhere
	}
	else if (dwOp == arg_OP_IS_MEM_DISP)
	{
		if (!pOpMem)
			return false;

		if (!IsPureDisplacement (pOpMem))
			return false;

		if (!m_prefix.ValidateAddressSize(IsAddrSize32(pOpMem)))
			return false;

		if (!m_prefix.ValidateSegment(pOpMem,regDS))
			return false;
	}
	else if (dwOp == arg_OP_AS_REG_EXTRA)
	{
		if (!pOpReg)
			return false;

		if (!ValidateOpRegisterTypeNameExtra (pOpReg,m_info.m_nExtra1))
			return false;

		if (IS_REG_TYPE_REAL(m_info.m_nExtra1))
			return true; // size is ok as the real register fitted with operand
	}
	else if (dwOp == arg_OP_AS_REG_EXTRA2)
	{
		if (!pOpReg)
			return false;

		if (!ValidateOpRegisterTypeNameExtra (pOpReg,m_info.m_nExtra2))
			return false;

		if (IS_REG_TYPE_REAL(m_info.m_nExtra1))
			return true; // size is ok as the real register fitted with operand
	}
	else if (dwOp == arg_OP_LOW_3BITS_REG)
	{
		if (!pOpReg)
			return false;

		if (!ValidateOpRegisterStdTypeAndName (pOpReg,m_info.m_nInstruction & 0x07))
			return false;
	}
	else if (dwOp == arg_OP_CONST_1)
	{
		if (!pOpImm)
			return false;

		if (pOpImm->GetImmediate () != 1)
			return false;
	}
	else if (m_info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) // this must be last
	{
		if (!pOpMem)
			return false;

		assert (dwOpMask != arg_OP_SRC2_MASK);

		byte bMaskSIDI = (dwOpMask == arg_OP_DEST_MASK) ? (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI) : (ext2_SRC_MEM_DI | ext2_SRC_MEM_SI);

		if (!(m_info.m_nExtra2 & bMaskSIDI))
		{
			// arg_EXTRA2_DEST_SRC_SI_DI  was last hope, but this operand not included (bad luck...)

			return false;
		}

		if (ValidateOpSIDIMemory (pOpMem,m_info,(m_info.m_nExtra2 & bMaskSIDI),m_prefix))
			return false;
	}
	else
	{
		assert (false); // we
		return false;
	}
*/
	// now see if it has the right size
  
  if (!ValidateOperandSize (pOp->GetOpSize (),m_pInverseOpcodeInfo,m_prefix,bIsLastOperand))
    return false;
  
  return true;
}

TryDoInstructionEncoding::TryDoInstructionEncoding(bool bCodeSize32)
{
  m_bCodeSize32 = bCodeSize32;
}

void TryDoInstructionEncoding::Reset()
{
  m_prefix.Reset();
  m_modrm.Reset();
  m_imm.Reset();

  m_bShortestLength = 0;
}
/*
bool CanEncodeType(word wFlagOpMayBe,const Operand* pOp)
{
	if (dynamic_cast<const OperandRegister*>(pOp))
		return !!(wFlagOpMayBe & INVERSE_OPCODE_OP_MAY_BE_REG);
	if (dynamic_cast<const OperandMemory*>(pOp))
		return !!(wFlagOpMayBe & INVERSE_OPCODE_OP_MAY_BE_MEM);
	if (dynamic_cast<const OperandImmediate*>(pOp))
		return !!(wFlagOpMayBe & INVERSE_OPCODE_OP_MAY_BE_IMM);

	assert (!pOp);

//	There is not much use in running this test, as we already have the right instruction,
//	and instructions usually have the same number of operands in the same places, so
//	this will not differentiate between different possible encodings

	return !wFlagOpMayBe; // if op is nothing, then we should not want to encode anything
}
*/

word OpTypeMayBe(const Operand* pOp)
{
  if (dynamic_cast<const OperandRegister*>(pOp))
    return INVERSE_OPCODE_OP_MAY_BE_REG;
  if (dynamic_cast<const OperandMemory*>(pOp))
    return INVERSE_OPCODE_OP_MAY_BE_MEM;
  assert (dynamic_cast<const OperandImmediate*>(pOp));
  
  return INVERSE_OPCODE_OP_MAY_BE_IMM;
}

void TryDoInstructionEncoding::CalcNeedOpTypeMask(const Instruction* pInstr,
						  word& wNeedMask, // bits that has to be one
						  word& wNeedNotMask) // bits that has to be zero
{
  wNeedMask = 0;
  wNeedNotMask = 0;
  
  const Operand* pOp;
  
  if ((pOp = pInstr->GetDestOp()))
    wNeedMask |= OpTypeMayBe(pOp) << INVERSE_OPCODE_OP_DEST_SHIFT;
  else
    wNeedNotMask |= INVERSE_OPCODE_DEST_MASK;
  if ((pOp = pInstr->GetSrcOp()))
    wNeedMask |= OpTypeMayBe(pOp) << INVERSE_OPCODE_OP_SRC_SHIFT;
  else
    wNeedNotMask |= INVERSE_OPCODE_SRC_MASK;
  if ((pOp = pInstr->GetSrc2Op()))
    wNeedMask |= OpTypeMayBe(pOp) << INVERSE_OPCODE_OP_SRC2_SHIFT;
  else
    wNeedNotMask |= INVERSE_OPCODE_SRC2_MASK;
}

bool TryDoInstructionEncoding::TryEncoding (const InverseOpcodeInfo* pInverseOpcodeInfo,
					    const Instruction* pInstr)
{
  assert (pInverseOpcodeInfo);
  assert (pInstr);
  
#ifndef NDEBUG
  // this should have been done before we were called, but lets check
  // it...
  
  TRY_DO_INSTR_ENCODE_PREPARE(pInstr);
  assert (TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(pInverseOpcodeInfo));
#endif
  
/*
	// we first determine if we have a chance to decode all operands

	const Operand* pOpDest = pInstr->GetDestOp();
	const Operand* pOpSrc = pInstr->GetSrcOp();
	const Operand* pOpSrc2 = pInstr->GetSrc2Op();

	// not that this would not be working nicely together with swapping operands as being done below for XCHG, if
	// the operands were of different types. Currently, we're only swapping is both registers

	if (!CanEncodeType((pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_DEST_MASK) >> INVERSE_OPCODE_OP_DEST_SHIFT,pOpDest) ||
		!CanEncodeType((pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_SRC_MASK) >> INVERSE_OPCODE_OP_SRC_SHIFT ,pOpSrc) ||
		!CanEncodeType((pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_SRC2_MASK) >> INVERSE_OPCODE_OP_SRC2_SHIFT,pOpSrc2))
		return false;
*/
  Reset();
  
  m_pInverseOpcodeInfo = pInverseOpcodeInfo;
/*
	// find out info about instruction

	// if anything mis-matches, abort

	m_info = infoByte[m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_2_BYTE ? 1 : 0][m_pInverseOpcodeInfo->m_nByte];

	assert (((m_info.m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP ||
		     (m_info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)) ==
			!!(m_pInverseOpcodeInfo->m_nFlags & (INVERSE_OPCODE_TAG_GROUP_MOD_11 | 
												 INVERSE_OPCODE_TAG_GROUP_MOD_MEM)));

	if ((m_info.m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
	{
		// we're part of a group, find group and us

		int nGroup;

		if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_GROUP_MOD_MEM)
			nGroup = m_info.m_nExtra1;
		else
		{
			assert (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_GROUP_MOD_11);
			nGroup = m_info.m_nExtra2;
		}
		if (m_info.m_nInstruction & instrmGROUP_MOD_11)
			m_info.m_nExtra2 = 0;

		OpcodeInfo info2 = infoGroups[nGroup][(m_pInverseOpcodeInfo->m_nGroupNNN)];

		assert ((info2.m_nInstruction == instrSUBGRP_COPROC) ==
						!!(m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_SUBGROUP));

		if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_SUBGROUP)
			info2 = infoGroups[info2.m_nExtra1][m_pInverseOpcodeInfo->m_nSubGroupN];
		
		assert (m_info.m_nExtra2 == 0 ||
						info2.m_nExtra2 == 0); // otherwise m_info will be jumbled
		
		m_info.m_nArguments |= info2.m_nArguments;
		m_info.m_nInstruction = info2.m_nInstruction;
		m_info.m_nExtra1 = info2.m_nExtra1;
		m_info.m_nExtra2 |= info2.m_nExtra2;

		assert ((m_info.m_nArguments & arg_OP_DEST_MASK) != arg_DEST_IS_MODRM_REG); // as group got NNN from REG, we can't get register from there too
		assert ((m_info.m_nArguments & arg_OP_SRC_MASK ) != arg_SRC_IS_MODRM_REG ); // as group got NNN from REG, we can't get register from there too
	}

	assert ((m_info.m_nInstruction == instrPRE_MMX_SSE ||
			 m_info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11) ==
			!!(m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_MMX_SSE_PREFIX));

	if (m_info.m_nInstruction == instrPRE_MMX_SSE ||
		m_info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
	{
		// bSSEPrefix contain any Prefix

		byte bSSEPrefix = m_pInverseOpcodeInfo->m_nSubGroupN;

		assert (m_info.m_nExtra2 & (1 << bSSEPrefix));

		assert (!(m_info.m_nArguments & ~arg_NEED_MODRM_BYTE));

		int nOffset = (m_info.m_nExtra1 << 1) + bSSEPrefix;

		assert (!(m_pInverseOpcodeInfo->m_nFlags & (INVERSE_OPCODE_TAG_GROUP_MOD_MEM |
																						 INVERSE_OPCODE_TAG_GROUP_MOD_11)) ||
						((m_info.m_nExtra2 & ext2_SSE_PREFIX_NO_66) && 
						 !(m_info.m_nExtra2 & ext2_SSE_PREFIX_F3_F2))); // so that nOffset += 2 is valid

		if (m_info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11 &&
			m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_GROUP_MOD_11)
			nOffset += 2;

		assert (nOffset < ext_MMX_SSE_NUM_GROUPS * 2);

		m_info = infoMMX_SSE[nOffset]; // index to be multiplied by two
	}

#ifndef NDEBUG
	if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_LOCK_REP_PREFIX)
	{
		if ((pInstr->m_nInstruction & instrgMASK) == instrgSTRING)
		{
			if (m_info.m_nInstruction == instrCMPS ||
				m_info.m_nInstruction == instrSCAS)
				assert ((m_info.m_nInstruction | instrgmSTRING_REPE) == pInstr->m_nInstruction ||
						(m_info.m_nInstruction | instrgmSTRING_REPNE) == pInstr->m_nInstruction);
			else
				assert ((m_info.m_nInstruction | instrgmSTRING_REP) == pInstr->m_nInstruction);
		}
		else
		{
			assert (((pInstr->m_nInstruction & instrgMASK) == instrgINT_ARITH ||
					 (pInstr->m_nInstruction & instrgMASK) == instrgMEMORY) &&
					(m_info.m_nInstruction | instrgmINT_ARITH_MEMORY_LOCK) == pInstr->m_nInstruction);
		}
	}
	else
		assert (m_info.m_nInstruction == pInstr->m_nInstruction); // or we should not have been called
#endif
*/

	// now we have to determine if we can encode all operands...

	// first see if we got the right number and sort of operands (dest,src,src2)

/*
	Removed, as it's tested by CanEncodeType above

	if (!(dwDestOp) != (!pOpDest) ||
		!(dwSrcOp)  != (!pOpSrc))
	{
		// could be ok if string instruction...

		if (!(m_info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI))
			return false; // brutal mismatch (wrong operands)

		if (!(dwDestOp ||
			  (m_info.m_nExtra2 & (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI))) != (!pOpDest) ||
			!(dwSrcOp ||
			  (m_info.m_nExtra2 & (ext2_SRC_MEM_DI | ext2_SRC_MEM_SI))) != (!pOpSrc))
			return false; // brutal mismatch (wrong operands)
	}

	if (!(dwSrc2Op) != (!pOpSrc2))
		return false; // brutal mismatch (wrong operands)
*/
  // some reordering of operands for instructions that are
  // symmetric and that can be better encoded in one way than
  // another XCHG is the only I can think of
  // 
  // this does not affect ability to find two byte instruction
  // even if one-byte may be used we will however not find both
  // variants of the two-byte instruction (but we would not have
  // done that anyway, without extra code for that))
  
  const Operand* pOpDest = pInstr->GetDestOp();
  const Operand* pOpSrc = pInstr->GetSrcOp();
  const Operand* pOpSrc2 = pInstr->GetSrc2Op();
  
  if (pInstr->m_nInstruction == instrXCHG)
    {
      assert (pOpDest && pOpSrc);
      
      const OperandRegister* pOpDestReg;
      
      // NOTE: It's okay to swap dest and src when registers, because
      // it may enchance encoding but is it a good idea to swap when
      // one operand is mem and another is reg then we should also do
      // this for TEST instruction (and perhaps some more, which would
      // probably lead to more trouble than when we always create
      // instructions with operands in the right order
      //
      // CanEncodeType above removes usefulness of swapping operands
      // of different types
      //
      // I do not know. Figure it out...
      //
      // Actually, we should probably always swap to give smallest possible
      // encoding, but at the same time, when possible, also give the
      // opportunity to lengthen the instruction...      
      
      if ((pOpDestReg = 
	   dynamic_cast<const OperandRegister*>(pOpDest)) != NULL &&
	  (pOpDestReg->m_bReg == regEAX || pOpDestReg->m_bReg == regAX) &&
	  dynamic_cast<const OperandRegister*>(pOpSrc))
	swap (pOpDest,pOpSrc);
    }	
  
  if (INSTR_GROUP(pInstr->m_nInstruction) == instrgCTRL_TRANSFER &&
      INSTR_IS_COND_COUNT(pInstr->m_nInstruction))
    m_prefix.ValidateAddressSize(pInstr->m_nInstruction & instrmCOUNT_ECX);
  
  // ok, so now we know that we have operands in the right places, but
  // are they correct ?
  
  // begin by investigating type of pDest
  /*
    assert (arg_HAS_OP_DEST == ((argb_MODRM__IS_REG | argb_MODRM__IS_MODRM | argb__IS_IMM | argb__IS_MEM_DISP | argb__AS_REG_EXTRA                       | argb_LOW_3BITS__REG) << 0));
    assert (arg_HAS_OP_SRC == ((argb_MODRM__IS_REG  | argb_MODRM__IS_MODRM | argb__IS_IMM | argb__IS_MEM_DISP | argb__AS_REG_EXTRA | argb__AS_REG_EXTRA2 | argb_LOW_3BITS__REG | argb__CONST_1) << 1));
    assert (arg_HAS_OP_SRC2 == ((argb__IS_IMM | argb__AS_REG_EXTRA) << 2));
  */
  
  dword dwDestOp = 
    (m_pInverseOpcodeInfo->m_nArguments & arg_OP_DEST_MASK) >> 
    arg_OP_DEST_SHIFT;
  dword dwSrcOp  = 
    (m_pInverseOpcodeInfo->m_nArguments & arg_OP_SRC_MASK ) >> 
    arg_OP_SRC_SHIFT;
  dword dwSrc2Op = 
    (m_pInverseOpcodeInfo->m_nArguments & arg_OP_SRC2_MASK) >> 
    arg_OP_SRC2_SHIFT;
  /*
    assert (!dwDestOp == !(pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_DEST_MASK));
    assert (!dwSrcOp  == !(pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_SRC_MASK));
    assert (!dwSrc2Op == !(pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_SRC2_MASK));
*/
  
  if (pOpDest)
    {
      // begin by verifying op-size, as that is easier and will fail
      // in at least half the cases anyway
      
      //		if (!ValidateOperandSize (pOpDest->GetOpSize (),m_info,m_prefix,!dwSrcOp))
      //			return false;
      if (!ValidateOperandTypeSize (pOpDest,
				    dwDestOp,arg_OP_DEST_MASK,!dwSrcOp))
	return false;
    }
  if (pOpSrc)
    {
      //		if (!ValidateOperandSize (pOpSrc->GetOpSize (),m_info,m_prefix,!dwSrc2Op))
      //			return false;
      if (!ValidateOperandTypeSize (pOpSrc,dwSrcOp,arg_OP_SRC_MASK,!dwSrc2Op))
	return false;
    }
  if (pOpSrc2)
    {
      //		if (!ValidateOperandSize (pOpSrc2->GetOpSize (),m_info,m_prefix,true))
//			return false;
      if (!ValidateOperandTypeSize (pOpSrc2,dwSrc2Op,arg_OP_SRC2_MASK,true))
	return false;
    }
  /*
	const AnyInstruction* pAnyInstr;
	
	if ((pAnyInstr = dynamic_cast<const AnyInstruction*>(pInstr)) &&
	pAnyInstr->GetRepLock())
	{
		m_prefix.ValidateRepLock(pAnyInstr->GetRepLock());
	}
*/
/*
	if (m_pInverseOpcodeInfo->m_nFlags & (INVERSE_OPCODE_TAG_LOCK_PREFIX | 
										  INVERSE_OPCODE_TAG_REPE_PREFIX |
										  INVERSE_OPCODE_TAG_REPNE_PREFIX))
	{
		if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_LOCK_PREFIX)
		{
			assert (((pInstr->m_nInstruction & instrgMASK) == instrgINT_ARITH ||
					 (pInstr->m_nInstruction & instrgMASK) == instrgMEMORY) &&
					(pInstr->m_nInstruction & instrgmINT_ARITH_MEMORY_LOCK));
			m_prefix.ValidateRepLock(ENCODE_PREFIX_LOCK);
		}
		else if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_REPE_PREFIX)
		{
			assert (((pInstr->m_nInstruction & instrgMASK) == instrgSTRING) &&
				    (pInstr->m_nInstruction & (instrgmSTRING_REP | instrgmSTRING_REPE)));
			m_prefix.ValidateRepLock(ENCODE_PREFIX_REPE);
		}
		else
		{
			assert (((pInstr->m_nInstruction & instrgMASK) == instrgSTRING) &&
				    (pInstr->m_nInstruction & instrgmSTRING_REPNE));

			assert (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_REPNE_PREFIX);
			m_prefix.ValidateRepLock(ENCODE_PREFIX_REPNE);
		}
	}
*/
  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_LOCK_REP_PREFIX)
    {
      if ((pInstr->m_nInstruction & instrgMASK) == instrgSTRING)
	{
	  if (pInstr->m_nInstruction & 
	      (instrgmSTRING_REP | instrgmSTRING_REPE))
	    m_prefix.ValidateRepLock(ENCODE_PREFIX_REPE);
	  else
	    {
	      assert ((pInstr->m_nInstruction & instrgmSTRING_REPNE));
	      m_prefix.ValidateRepLock(ENCODE_PREFIX_REPNE);
	    }
	}
      else
	{
	  assert (((pInstr->m_nInstruction & instrgMASK) == instrgINT_ARITH ||
		   (pInstr->m_nInstruction & instrgMASK) == instrgMEMORY) &&
		  (pInstr->m_nInstruction & instrgmINT_ARITH_MEMORY_LOCK));
	  assert (dynamic_cast<const OperandMemory*>(pOpDest));
	  //			if (!dynamic_cast<const OperandMemory*>(pOpDest))
	  //				return false;
	  m_prefix.ValidateRepLock(ENCODE_PREFIX_LOCK);
	}
    }
  
  //	m_prefix.ValidateRepLock(pInstr->GetRepLock());
  
  // everything matched!  Encoding is possible!
  
  // prepare encoding
  
  if (m_pInverseOpcodeInfo->m_nArguments & arg_NEED_MODRM_BYTE)
    {
      byte bReg = 0;
      
      if (dwDestOp == arg_OP_IS_MODRM_REG)
	bReg = ((const OperandRegister*) pOpDest)->m_bReg & reggNAME_MASK;
      else if (dwSrcOp == arg_OP_IS_MODRM_REG)
	bReg = ((const OperandRegister*) pOpSrc)->m_bReg & reggNAME_MASK;
      else if (m_pInverseOpcodeInfo->m_nFlags & 
	       (INVERSE_OPCODE_TAG_GROUP_MOD_11 | 
		INVERSE_OPCODE_TAG_GROUP_MOD_MEM))
	bReg = m_pInverseOpcodeInfo->m_nGroupNNN;
      
      const Operand* pOpModRM = NULL;
      
      if (dwDestOp == arg_OP_IS_MODRM_MODRM)
	pOpModRM = pOpDest;
      else if (dwSrcOp == arg_OP_IS_MODRM_MODRM)
	pOpModRM = pOpSrc;
      
      const OperandMemory* pOpMem;
      const OperandRegister* pOpReg;
      
      if ((pOpReg = dynamic_cast<const OperandRegister*>(pOpModRM)) != NULL)
	{
	  m_modrm.bModRMExtraSIB = m_modrm.bModRM = 
	    (0xc0 | (bReg << 3) | pOpReg->m_bReg & reggNAME_MASK);
	}
      else if ((pOpMem = dynamic_cast<const OperandMemory*>(pOpModRM)) != NULL)
	{
	  if (!ValidateModRMMemOp(pOpMem,
				  m_prefix.IsAddrSize32(),m_prefix,m_modrm))
	    return false;
	  
	  m_modrm.bModRM |= (bReg << 3);
	  m_modrm.bModRMExtraSIB |= (bReg << 3);
	}
      else
	{
	  assert (!pOpModRM);
	  
	  byte bMod = 0;
	  byte bRM = 0;
	  
	  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_SUBGROUP)
	    bRM = (m_pInverseOpcodeInfo->m_nSubGroupN/* & INVERSE_OPCODE_SUBGROUP_N_MASK*/);
	  
	  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_GROUP_MOD_11)
	    bMod = 3;
	  
	  m_modrm.bModRMExtraSIB = m_modrm.bModRM = 
	    ((bMod << 6) | (bReg << 3) | bRM);
	}
    }
  
  if (dwDestOp == arg_OP_IS_MEM_DISP)
    m_modrm.SetDisplacement ((const OperandMemory*) pOpDest);
  else if (dwSrcOp == arg_OP_IS_MEM_DISP)
    m_modrm.SetDisplacement ((const OperandMemory*) pOpSrc);
  
  if (dwDestOp == arg_OP_IS_IMM)
    m_imm.SetImmediate ((const OperandImmediate*) pOpDest);
  else if (dwSrcOp == arg_OP_IS_IMM)
    m_imm.SetImmediate ((const OperandImmediate*) pOpSrc,
			!!(m_pInverseOpcodeInfo->m_nArguments & 
			   arg_SIGN_EXTEND_IMMEDIATE));
  
  if (dwSrc2Op == arg_OP_IS_IMM)
    m_imm.SetImmediate ((const OperandImmediate*) pOpSrc2,
			!!(m_pInverseOpcodeInfo->m_nArguments & 
			   arg_SIGN_EXTEND_IMMEDIATE));
  
  CalcShortestLength(); // this
  
  return true;
}

void TryDoInstructionEncoding::CalcShortestLength()
{
  int nLength = 0;
  
  nLength += m_prefix.OutputPrefixesLength(m_bCodeSize32);
  
  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_MMX_SSE_PREFIX)
    {
      assert (!(m_prefix.m_bDecided & (PD_OPSIZE32 | PD_REPLOCK))); // cannot have those Prefixes already (they must be undecided (end therefore unneeded))
      if (m_pInverseOpcodeInfo->m_nSubGroupN/* & INVERSE_OPCODE_SUBGROUP_N_MASK*/)
	nLength++;
    }
  
  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_2_BYTE)
    nLength++;
  nLength++; // opcode
  
  if (m_pInverseOpcodeInfo->m_nArguments & arg_NEED_MODRM_BYTE)
    {
      nLength++; // modrm
      
      if (m_prefix.IsDecidedAndSetAddrSize32() &&
	  m_modrm.bModRM == m_modrm.bModRMExtraSIB)
	nLength++; // SIB
    }
  
  nLength += m_modrm.OutputDisplacementMinLength();
  nLength += m_imm.OutputImmediatesLength();
  
  m_bShortestLength = nLength;
}

#define LENGTHEN_CAN_EMIT_EXTRA_SIB        (m_prefix.IsDecidedAndSetAddrSize32() && m_modrm.bModRM != m_modrm.bModRMExtraSIB)
#define LENGTHEN_CAN_EMIT_EXTRA_SEG_PREFIX (!(m_prefix.m_bDecided & PD_SEGMENT) && m_prefix.m_bSegmentReg)
#define LENGTHEN_CAN_EXPAND_DISP_FROM_0    (m_modrm.bMinDispSize == (MODRM_MAY_ENLARGE_DISP_SIZE | 0))
#define LENGTHEN_CAN_EXPAND_DISP_FROM_1    (m_modrm.bMinDispSize == (MODRM_MAY_ENLARGE_DISP_SIZE | 1))

word TryDoInstructionEncoding::GetLengths ()
{
  assert (m_bShortestLength); 
  
  word wLengths = 1 << (m_bShortestLength - 1);
  
  if (m_prefix.IsDecidedAddrSize32())
    {
      if (LENGTHEN_CAN_EMIT_EXTRA_SIB)
	wLengths = wLengths | (wLengths << 1); // we can encode with or without sib
      
      if (LENGTHEN_CAN_EXPAND_DISP_FROM_0)
	wLengths = wLengths | (wLengths << 1) | 
	  (wLengths << (m_prefix.IsAddrSize32() ? 4 : 2)); // we can also encode with dispsize = 1 or 2/4
      else if (LENGTHEN_CAN_EXPAND_DISP_FROM_1)
	wLengths = wLengths | 
	  (wLengths << (m_prefix.IsAddrSize32() ? 3 : 1)); // we can also encode with dispsize = 2/4
    }
  
  if (LENGTHEN_CAN_EMIT_EXTRA_SEG_PREFIX)
    wLengths = wLengths | (wLengths << 1); // we can decode with an extra unnecessary segment prefix (= default seg prefix)
  
  return wLengths;
}

bool TryDoInstructionEncoding::DoEncodingLength (DataDestPtr& dataPtr,
						 byte bLength,
						 word* pDispImmLocs,
						 byte* pLengthenTricksUsed) const
{
  // see if we can fix bLength
  
  // ok, so we want  bLength  , but only have m_bShortestLength
  
  assert (bLength);
  assert (m_bShortestLength);
  
  if (bLength < m_bShortestLength)
    return false; // impossible
  
  // prefer to emit only sib
  
  if (bLength == m_bShortestLength)
    {
      if (pLengthenTricksUsed)
	*pLengthenTricksUsed = 0;
      DoEncoding(dataPtr,pDispImmLocs);
      return true;
    }
  
  byte bLengthenTricks = 0;
  
  byte bMissing = bLength - m_bShortestLength;
  
  if (bMissing == 1 && // only missing one byte
      LENGTHEN_CAN_EMIT_EXTRA_SIB)
    {
      bLengthenTricks |= LENGTHEN_TRICK_ISSUE_SIB;
    }
  else
    {
      // only sib was not enough, then first do as much as possible
      // with displacement
      
      if (m_prefix.IsDecidedAddrSize32())
	{
	  byte bEnlargeTo16_32 = m_prefix.IsAddrSize32() ? 4 : 2;
	  
	  if (LENGTHEN_CAN_EXPAND_DISP_FROM_0 &&
	      bMissing >= bEnlargeTo16_32)
	    {
	      // enlarge from disp0 -> disp2/4
	      bLengthenTricks |= (m_prefix.IsAddrSize32() ? 
				  LENGTHEN_TRISK_DISP_0_TO_32 : 
				  LENGTHEN_TRISK_DISP_0_TO_16);
	      bMissing -= bEnlargeTo16_32;
	    }
	  else if (LENGTHEN_CAN_EXPAND_DISP_FROM_1 &&
		   bMissing >= (bEnlargeTo16_32-1))
	    {
	      // enlarge from disp1 -> disp2/4
	      bLengthenTricks |= (m_prefix.IsAddrSize32() ? 
				  LENGTHEN_TRISK_DISP_8_TO_32 : 
				  LENGTHEN_TRISK_DISP_8_TO_16);
	      bMissing -= (bEnlargeTo16_32-1);
	    }
	  else if (LENGTHEN_CAN_EXPAND_DISP_FROM_0)
	    {
	      // enlarge from disp0 -> disp1
	      bLengthenTricks |= LENGTHEN_TRICK_DISP_0_TO_8;
	      bMissing--;
	    }
	  
	  // then do sib if necessary and possible
	  if (bMissing &&
	      LENGTHEN_CAN_EMIT_EXTRA_SIB)
	    {
	      bLengthenTricks |= LENGTHEN_TRICK_ISSUE_SIB;
	      bMissing--;
	    }
	}
      
      // last, try expanding with prefix
      if (bMissing &&
	  LENGTHEN_CAN_EMIT_EXTRA_SEG_PREFIX)
	{
	  bLengthenTricks |= LENGTHEN_TRICK_ISSUE_SEG;
	  bMissing--;
	}
      
      if (bMissing)
	return false; // we can't do this length
    }
  
  // now encode with our tricks...
  
  DoEncoding(dataPtr,pDispImmLocs,bLengthenTricks);
  
  if (pLengthenTricksUsed)
    *pLengthenTricksUsed = bLengthenTricks;
  
  return true; // we can't do this length
}

void TryDoInstructionEncoding::DoEncoding (DataDestPtr& dataPtr,
					   word* pDispImmLocs,
					   byte bLengthenTricks) const
{
  assert (m_bShortestLength); 
  
  assert (sizeof (void*) <= sizeof (dword)); // so it fits in dwStart
  
  dword dwStart = (dword) dataPtr.GetCurrent();
  
  // the code assume that the order of any Prefixes does not affect
  // execution speed on any processor (it can't affect size (but
  // perhaps hardware bugs) :-)) if it does, rewrite! to take that
  // into account
  //
  // also investigate behaviour of fast encoders

  m_prefix.OutputPrefixes(dataPtr,
			  m_bCodeSize32,
			  !!(bLengthenTricks & LENGTHEN_TRICK_ISSUE_SEG));
  
  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_MMX_SSE_PREFIX)
    {
      assert (!(m_prefix.m_bDecided & (PD_OPSIZE32 | PD_REPLOCK))); // cannot have those Prefixes already (they must be undecided (end therefore unneeded))
      if (m_pInverseOpcodeInfo->m_nSubGroupN/* & INVERSE_OPCODE_SUBGROUP_N_MASK*/)
	OutputMMXSSEPrefix(m_pInverseOpcodeInfo->m_nSubGroupN/* & INVERSE_OPCODE_SUBGROUP_N_MASK*/,dataPtr);
    }
  
  if (m_pInverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_2_BYTE)
    dataPtr.WriteByte (ENCODE_TWO_BYTE_OPCODE_ESCAPE); // two byte opcode Prefix
  
  dataPtr.WriteByte (m_pInverseOpcodeInfo->m_nByte); // the op-code
  
  if (m_pInverseOpcodeInfo->m_nArguments & arg_NEED_MODRM_BYTE)
    {
      byte bModRM = 
	(bLengthenTricks & LENGTHEN_TRICK_ISSUE_SIB) ? 
	m_modrm.bModRMExtraSIB : m_modrm.bModRM;
      
      if (bLengthenTricks & LENGTHEN_TRICK_DISP_TO_8)
	{
	  assert ((bModRM >> 6) == 0);
	  bModRM = (1 << 6) | (bModRM & 0x3f);
	}
      else if (bLengthenTricks & 
	       (LENGTHEN_TRICK_DISP_TO_16 | LENGTHEN_TRICK_DISP_TO_32))
	{
	  assert (((bLengthenTricks & 
		    (LENGTHEN_TRISK_DISP_0_TO_16 | 
		     LENGTHEN_TRISK_DISP_0_TO_32)) && (bModRM >> 6) == 0) ||
		  ((bLengthenTricks & 
		    (LENGTHEN_TRISK_DISP_8_TO_16 | 
		     LENGTHEN_TRISK_DISP_8_TO_32)) && (bModRM >> 6) == 1));
	  bModRM = (2 << 6) | (bModRM & 0x3f);
	}
      
      dataPtr.WriteByte (bModRM);
      
      assert (!m_prefix.IsDecidedAndSetAddrSize32() ||
	      ((m_modrm.bModRM == m_modrm.bModRMExtraSIB) + 
	       !!(bLengthenTricks & LENGTHEN_TRICK_ISSUE_SIB)) <= 1); // can't have two reasons to emit sib
      
      if (m_prefix.IsDecidedAndSetAddrSize32() &&
	  m_modrm.bModRM == m_modrm.bModRMExtraSIB ||
	  bLengthenTricks & LENGTHEN_TRICK_ISSUE_SIB)
	dataPtr.WriteByte (m_modrm.bSIB);
    }
#ifndef NDEBUG
  else
    {
      assert (!(bLengthenTricks & LENGTHEN_TRICK_ISSUE_SIB));
      assert (!(bLengthenTricks & LENGTHEN_TRICK_DISP));
    }
#endif
  
  dwStart = (dword) dataPtr.GetCurrent() - dwStart;
  
  byte bDispLen = m_modrm.OutputDisplacement(dataPtr,bLengthenTricks);
  byte bImmsLen = m_imm.OutputImmediates(dataPtr);
  
  printf ("%d %d %d\n",dwStart,bDispLen,bImmsLen);

  if (pDispImmLocs)
    {
      assert (dwStart < 16); // so it fits in four bits
      
      *pDispImmLocs = dwStart | (bDispLen << 4) | (bImmsLen << 8);
    }
}

/* FIXME: handle  

  arg_LAST_OPRND_SIGN_EXTEND_IMMEDIATE

*/

