/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VerifyEncoderDecoder.cpp: 
//
//////////////////////////////////////////////////////////////////////

#include "VerifyEncoderDecoder.h" 

#include "BranchInstr.h" 
#include "AnyInstruction.h" 

#include "Error.h" 

#include "OperandImmediate.h" 
#include "OperandRegister.h" 
#include "OperandMemory.h" 

#include "gen/Instruction_Names.h"

#include <stdio.h>

//////////////////////////////////////////////////////////////////////

#define RETURN_IF_FALSE(expr) { if (!(expr)) return false; }

bool IsSameOperand(const Operand* pOp1,const Operand* pOp2)
{
  RETURN_IF_FALSE (!pOp1 == !pOp2);
  
  if (!pOp1)
    return true;
  
  // now, they better be of same type and content
  
  const OperandImmediate* pOpImm1;
  const OperandImmediate* pOpImm2;
  
  if ((pOpImm1 = dynamic_cast<const OperandImmediate*>(pOp1)) != NULL &&
      (pOpImm2 = dynamic_cast<const OperandImmediate*>(pOp2)) != NULL)
    {
      RETURN_IF_FALSE (pOpImm1->m_bImmBytes);
      RETURN_IF_FALSE (pOpImm1->m_bImmBytes       == pOpImm2->m_bImmBytes);
      RETURN_IF_FALSE (pOpImm1->m_dwImm           == pOpImm2->m_dwImm);
      RETURN_IF_FALSE (pOpImm1->m_bImmHigh16Bytes == pOpImm2->m_bImmHigh16Bytes);
      if (pOpImm1->m_bImmHigh16Bytes)
	RETURN_IF_FALSE (pOpImm1->m_wImmHigh16      == pOpImm2->m_wImmHigh16);
      return true;
    }
  
  const OperandRegister* pOpReg1;
  const OperandRegister* pOpReg2;
  
  if ((pOpReg1 = dynamic_cast<const OperandRegister*>(pOp1)) != NULL &&
      (pOpReg2 = dynamic_cast<const OperandRegister*>(pOp2)) != NULL)
    {
      RETURN_IF_FALSE (pOpReg1->m_bReg == pOpReg2->m_bReg);
      return true;
    }
  
  const OperandMemory* pOpMem1;
  const OperandMemory* pOpMem2;
  
  if ((pOpMem1 = dynamic_cast<const OperandMemory*>(pOp1))!= NULL &&
      (pOpMem2 = dynamic_cast<const OperandMemory*>(pOp2)) != NULL)
    {
      RETURN_IF_FALSE (pOpMem1->m_bSegmentReg == pOpMem2->m_bSegmentReg);
      RETURN_IF_FALSE (pOpMem1->m_wOpSize     == pOpMem2->m_wOpSize);
      RETURN_IF_FALSE (pOpMem1->m_bScale      == pOpMem2->m_bScale);
      if (pOpMem1->m_bScale == 0) // 1<<m_bScale is scale, so this is x1
	{
	  // if scale is one, base and index may be swapped
	  // allow this as it is done silently by Encoder
	  
	  RETURN_IF_FALSE (((pOpMem1->m_bBaseReg  == pOpMem2->m_bBaseReg) && 
			    (pOpMem1->m_bIndexReg == pOpMem2->m_bIndexReg)) || 
			   ((pOpMem1->m_bBaseReg  == pOpMem2->m_bIndexReg) && 
			    (pOpMem1->m_bIndexReg == pOpMem2->m_bBaseReg)));
	}
      else
	{
	  RETURN_IF_FALSE (pOpMem1->m_bBaseReg == pOpMem2->m_bBaseReg);
	  RETURN_IF_FALSE (pOpMem1->m_bIndexReg == pOpMem2->m_bIndexReg);
	}
      RETURN_IF_FALSE (pOpMem1->m_dwDisp == pOpMem2->m_dwDisp);
      
      // make sure same address-size:
      
      RETURN_IF_FALSE (pOpMem1->m_bBaseReg || pOpMem1->m_bIndexReg ||
		       pOpMem1->m_bDispBytes == pOpMem2->m_bDispBytes);
      return true;
    }
  
  RETURN_IF_FALSE (false); // pOp1 and pOp2 are of different types
}

void VerifySameOperand(const Operand* pOp1,const Operand* pOp2)
{
  assert (!pOp1 == !pOp2);
  
  if (!pOp1)
    return;
  
  // now, they better be of same type and content
  
  const OperandImmediate* pOpImm1;
  const OperandImmediate* pOpImm2;
  
  if ((pOpImm1 = dynamic_cast<const OperandImmediate*>(pOp1)) != NULL &&
      (pOpImm2 = dynamic_cast<const OperandImmediate*>(pOp2)) != NULL)
    {
      assert (pOpImm1->m_bImmBytes);
      assert (pOpImm1->m_bImmBytes       == pOpImm2->m_bImmBytes);
      assert (pOpImm1->m_dwImm           == pOpImm2->m_dwImm);
      assert (pOpImm1->m_bImmHigh16Bytes == pOpImm2->m_bImmHigh16Bytes);
      if (pOpImm1->m_bImmHigh16Bytes)
	assert (pOpImm1->m_wImmHigh16      == pOpImm2->m_wImmHigh16);
      return;
    }
  
  const OperandRegister* pOpReg1;
  const OperandRegister* pOpReg2;
  
  if ((pOpReg1 = dynamic_cast<const OperandRegister*>(pOp1)) != NULL &&
      (pOpReg2 = dynamic_cast<const OperandRegister*>(pOp2)) != NULL)
    {
      assert (pOpReg1->m_bReg == pOpReg2->m_bReg);
      return;
    }
  
  const OperandMemory* pOpMem1;
  const OperandMemory* pOpMem2;
  
  if ((pOpMem1 = dynamic_cast<const OperandMemory*>(pOp1))!= NULL &&
      (pOpMem2 = dynamic_cast<const OperandMemory*>(pOp2)) != NULL)
    {
      assert (pOpMem1->m_bSegmentReg == pOpMem2->m_bSegmentReg);
      assert (pOpMem1->m_wOpSize     == pOpMem2->m_wOpSize);
      assert (pOpMem1->m_bScale      == pOpMem2->m_bScale);
      if (pOpMem1->m_bScale == 0) // 1<<m_bScale is scale, so this is x1
	{
	  // if scale is one, base and index may be swapped
	  // allow this as it is done silently by Encoder
	  
	  assert (((pOpMem1->m_bBaseReg  == pOpMem2->m_bBaseReg) && 
		   (pOpMem1->m_bIndexReg == pOpMem2->m_bIndexReg)) || 
		  ((pOpMem1->m_bBaseReg  == pOpMem2->m_bIndexReg) && 
		   (pOpMem1->m_bIndexReg == pOpMem2->m_bBaseReg)));
	}
      else
	{
	  assert (pOpMem1->m_bBaseReg == pOpMem2->m_bBaseReg);
	  assert (pOpMem1->m_bIndexReg == pOpMem2->m_bIndexReg);
	}
      assert (pOpMem1->m_dwDisp == pOpMem2->m_dwDisp);
      
      // make sure same address-size:
      
      assert (pOpMem1->m_bBaseReg || pOpMem1->m_bIndexReg ||
	      pOpMem1->m_bDispBytes == pOpMem2->m_bDispBytes);
      return;
    }
  
  assert (false); // pOp1 and pOp2 are of different types
}

//////////////////////////////////////////////////////////////////////

void VerifySameInstruction(const Instruction* pInstr1,
			   const Instruction* pInstr2)
{
  // we do not care how the instruction was packaged, but if some package
  // contain something another can't, then we care
  
  assert (pInstr1);
  assert (pInstr2);
  
  assert (pInstr1->m_nInstruction == pInstr2->m_nInstruction);
  
  assert (((pInstr1->m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER) ==
	  (dynamic_cast<const BranchInstr*>(pInstr1) != NULL)); // ctrl transfers in branch package
  
  assert (!dynamic_cast<const BranchInstr*>(pInstr1) ==
	  !dynamic_cast<const BranchInstr*>(pInstr2)); // branches must be packaged in same
  /*
    byte bSpecial1 = 0;
    byte bSpecial2 = 0;
    
	const AnyInstruction* pAny;
	
	if ((pAny = dynamic_cast<const AnyInstruction*>(pInstr1)) != NULL)
	bSpecial1 = pAny->m_bSpecial;
	if ((pAny = dynamic_cast<const AnyInstruction*>(pInstr2)) != NULL)
		bSpecial2 = pAny->m_bSpecial;

	assert (bSpecial1 == bSpecial2);
*/
  if (pInstr1->m_nInstruction == instrXCHG)
    {
      // this instruction may have the operands the other way around
      
      assert ((IsSameOperand(pInstr1->GetDestOp(),pInstr2->GetDestOp()) &&
	       IsSameOperand(pInstr1->GetSrcOp() ,pInstr2->GetSrcOp() )) ||
	      (IsSameOperand(pInstr1->GetDestOp(),pInstr2->GetSrcOp() ) &&
	       IsSameOperand(pInstr1->GetSrcOp() ,pInstr2->GetDestOp())));
    }
  else
    {
      VerifySameOperand (pInstr1->GetDestOp(),pInstr2->GetDestOp());
      VerifySameOperand (pInstr1->GetSrcOp() ,pInstr2->GetSrcOp() );
      VerifySameOperand (pInstr1->GetSrc2Op(),pInstr2->GetSrc2Op());
    }
}

//////////////////////////////////////////////////////////////////////


void TestCoders::Test ()
{
  for (int nCodeSize32 = 0; nCodeSize32 < 2; nCodeSize32++)
    {
      m_decoder.m_bCodeSize32 = (nCodeSize32 == 1);
      m_encoder.m_bCodeSize32 = (nCodeSize32 == 1);
      
      DoTest();
    }
}

void TestCoders::VerifyDecodeIntoSame (const Instruction* pInstr,
				       const byte* cCode,
				       const byte bLength)
{
  DataPtrSrc code(cCode,cCode+bLength);
  
  Instruction* pCheck = NULL;
  
  try
    {
      pCheck = m_decoder.Decode (code);
    }
  catch (Error* pError)
    {
      pError->PrintShortMessage();
      //			printf ("\nException(2): %s\n",str);
      assert (false); // could not decode what we encoded: encoding/decoding error
      delete pError;
    }
  catch (...)
    {
      assert (false); // this should not happen (we should now about all possible error conditions)
      throw;
    }
  
  assert (pCheck); // should always be, Decode either throws exception or returns non-NULL
  
  assert (code.GetRVA() == bLength); // should be same length
  
  VerifySameInstruction(pInstr,pCheck);			
  delete pCheck;
}

//////////////////////////////////////////////////////////////////////

#define IA 16807
#define IM 2147483647
#define IQ 127773
#define IR 2836

#define IRNDMAX ((2147483647-1)<<1)

long Rnd::r()
{
  long k;

  k = m_state/IQ;
  m_state = IA*(m_state - k*IQ) - IR*k;
  
  if (m_state < 0)
    m_state += IM;
  
  return m_state;
}

word Rnd::rword(word max)
{
  // return a number in the range [0,max) max is power of 2
  
  // we can get random numbers between in the range [1,IM-1] 
  
  assert (!((max - 1) ^ ((max >> 1) | (max - 1) >> 1)));
  
  return (r() >> 15) & (max - 1);
}

word Rnd::rword(word max,byte maxlog2)
{
  // return a number in the range [0,max)
  
  // we can get random numbers between in the range [1,IM-1] 
  
  assert ((1 << maxlog2) >= max && (1 << (maxlog2-1)) < max);
  
  for ( ; ; )
    {
      word _r = (word) (r() >> (31 - maxlog2));
      assert (_r <= (1 << maxlog2));
      if (_r < max)
        return _r;      
    }
}

byte Rnd::rbyte_mask(byte mask)
{
  assert (mask);
  
  for ( ; ; )
    {
      byte _r = (byte) (r() >> 28);
      assert (_r < 8);
      if (mask & (1 << _r))
        return _r;
    }
}

//////////////////////////////////////////////////////////////////////

void InstrArgInfo::AddMemOtherSize (word wSize)
{ 
  assert (wSize);
  if (wMemOtherSize[0] == 0)
    wMemOtherSize[0] = wSize;
  else
    {
      assert (wMemOtherSize[1] == 0);
      wMemOtherSize[1] = wSize;
    }
}

void InstrArgInfo::GenerateRegisterOps(VectorInstrArgOperand& vectOperands,
				       Rnd& rnd,
				       bool bWrongOp,
				       byte bRegType,byte nNum,byte bMask)
{
  while (nNum--)
    {
      assert (bMask);
      
      byte bName = rnd.rbyte_mask(bMask);
      bMask ^= (1 << bName);
      PushOperand(vectOperands,new OperandRegister(bRegType | bName),bWrongOp);
      
      assert (GetRegisterName (bRegType | bName));
    }
}

void InstrArgInfo::PushOperand(VectorInstrArgOperand& vectOperands,
			       Operand* pOperand,bool bWrongOp)
{
  vectOperands.push_back(InstrArgOperand(pOperand,bWrongOp));
}

void InstrArgInfo::GenerateMemoryOps(VectorInstrArgOperand& vectOperands,
				     Rnd& rnd,bool bWrongOp,
				     byte nNum,word nOpSize,
				     bool bSIDI)
{
  // as there are so many possibilities of generating addresses with
  // the registers as pointers we don't bother to check that wo do not
  // make any duplicates
  
  if (bSIDI)
    {
      for (int i = 0; i < 3; i++)
	{
	  // 16-bit:
	  
	  {
	    OperandMemory* pMemOp = new OperandMemory();
	    
	    pMemOp->m_wOpSize = nOpSize;
	    pMemOp->m_bIndexReg = regSI;
	    if ((pMemOp->m_bSegmentReg = rnd.rbyte(10,4)) < 6)
	      pMemOp->m_bSegmentReg += reggSEG; // so 1..6 becomes 0..5
	    else
	      pMemOp->m_bSegmentReg = regDS;
	    
	    PushOperand(vectOperands,pMemOp,bWrongOp);
	    
	    pMemOp = new OperandMemory();
	    
	    pMemOp->m_wOpSize = nOpSize;
	    pMemOp->m_bIndexReg = regDI;
	    if ((pMemOp->m_bSegmentReg = rnd.rbyte(10,4)) < 6)
	      pMemOp->m_bSegmentReg += reggSEG; // so 1..6 becomes 0..5
	    else
	      pMemOp->m_bSegmentReg = regES;
	    
	    PushOperand(vectOperands,pMemOp,bWrongOp);
	  }
	  
	  // 32-bit:
	  
	  {
	    OperandMemory* pMemOp = new OperandMemory();
	    
	    pMemOp->m_wOpSize = nOpSize;
	    pMemOp->m_bIndexReg = regESI;
	    if ((pMemOp->m_bSegmentReg = rnd.rbyte(10,4)) < 6)
	      pMemOp->m_bSegmentReg += reggSEG; // so 1..6 becomes 0..5
	    else
	      pMemOp->m_bSegmentReg = regDS;
	    
	    PushOperand(vectOperands,pMemOp,bWrongOp);
	    
	    pMemOp = new OperandMemory();
	    
	    pMemOp->m_wOpSize = nOpSize;
	    pMemOp->m_bIndexReg = regEDI;
	    if ((pMemOp->m_bSegmentReg = rnd.rbyte(10,4)) < 6)
	      pMemOp->m_bSegmentReg += reggSEG; // so 1..6 becomes 0..5
	    else
	      pMemOp->m_bSegmentReg = regES;
	    
	    PushOperand(vectOperands,pMemOp,bWrongOp);
	  }
	}
    }
  
  while (nNum--)
    {
      // for each num we generate two: one with 16 bit addressing and
      // one with 32 bit adressing
      
      // 16-bit:
      
      {
	OperandMemory* pMemOp = new OperandMemory();
	
	pMemOp->m_wOpSize = nOpSize;
	
	switch (pMemOp->m_bDispBytes = rnd.rbyte(3,2))
	  {
	  case 1: pMemOp->m_dwDisp = (sbyte) rnd.rbyte(); break;
	  case 2: pMemOp->m_dwDisp = (sword) rnd.rword(); break;
	  }
	// pMemOp->m_bScale = 0;
	
	for ( ; ; )
	  {
	    if ((pMemOp->m_bBaseReg = rnd.rbyte(10,4)) < 8)
	      pMemOp->m_bBaseReg += reggSTD16;
	    else
	      pMemOp->m_bBaseReg = 0;
	    if ((pMemOp->m_bIndexReg = rnd.rbyte(10,4)) < 8)
	      pMemOp->m_bIndexReg += reggSTD16;
	    else
	      pMemOp->m_bIndexReg = 0;
	    
	    if ((pMemOp->m_bBaseReg == regeBX || 
		 pMemOp->m_bBaseReg == regeBP || 
		 pMemOp->m_bBaseReg == regeSI || 
		 pMemOp->m_bBaseReg == regeDI) &&
		(pMemOp->m_bIndexReg == regeBX || 
		 pMemOp->m_bIndexReg == regeBP || 
		 pMemOp->m_bIndexReg == regeSI || 
		 pMemOp->m_bIndexReg == regeDI))
	      break;
	    // do again, because impossible registers chosen
	    if (rnd.rbyte() < 10) // accept even impossible registers
	      break;
	  }
	
	if ((pMemOp->m_bSegmentReg = rnd.rbyte(10,4)) < 6)
	  pMemOp->m_bSegmentReg += reggSEG; // so 1..6 becomes 0..5
	else
	  pMemOp->m_bSegmentReg = 
	    (IS_REG_NAME_SP_BP(pMemOp->m_bBaseReg) ? regSS : regDS);
	
	if (pMemOp->m_bDispBytes < 2 && 
	    !pMemOp->m_bBaseReg &&
	    !pMemOp->m_bIndexReg)
	  pMemOp->m_bDispBytes = 2;
	
	PushOperand(vectOperands,pMemOp,bWrongOp);
      }
      
      // 32-bit:
      
      {
	OperandMemory* pMemOp = new OperandMemory();
	
	pMemOp->m_wOpSize = nOpSize;
	
	switch (pMemOp->m_bDispBytes = rnd.rbyte(3,2))
	  {
	  case 1: pMemOp->m_dwDisp = (sbyte) rnd.rbyte(); break;
	  case 2: pMemOp->m_dwDisp = (sdword) rnd.rdword(); 
	    pMemOp->m_bDispBytes = 4; break;
	  }
	pMemOp->m_bScale = rnd.rbyte(4);
	
	if ((pMemOp->m_bBaseReg = rnd.rbyte(16)) < 8)
	  pMemOp->m_bBaseReg += reggSTD32;
	else
	  pMemOp->m_bBaseReg = 0;
	if ((pMemOp->m_bIndexReg = rnd.rbyte(16)) < 8)
	  pMemOp->m_bIndexReg += reggSTD32;
	else
	  {
	    pMemOp->m_bIndexReg = 0;
	    pMemOp->m_bScale = 0;
	  }
	
	if ((pMemOp->m_bSegmentReg = rnd.rbyte(10,4)) < 6)
	  pMemOp->m_bSegmentReg += reggSEG; // so 1..6 becomes 0..5
	else
	  pMemOp->m_bSegmentReg = 
	    (IS_REG_NAME_SP_BP(pMemOp->m_bBaseReg) ? regSS : regDS);
	
	if (pMemOp->m_bDispBytes < 4 && 
	    !pMemOp->m_bBaseReg &&
	    !pMemOp->m_bIndexReg)
	  pMemOp->m_bDispBytes = 4;
	
	PushOperand(vectOperands,pMemOp,bWrongOp);
      }
    }
}

void InstrArgInfo::GenerateOperands(VectorInstrArgOperand& vectOperands,
				    Rnd& rnd,dword dwArg)
{
  if (dwArg & ARG_INFO_IMM_8)
    {
      PushOperand(vectOperands,new OperandImmediate(0),false);
      PushOperand(vectOperands,new OperandImmediate(1),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(8)),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte()),false);
    }
  if (dwArg & ARG_INFO_IMM_16)
    {
      PushOperand(vectOperands,new OperandImmediate(0,2),false);
      PushOperand(vectOperands,new OperandImmediate(1,2),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(),2),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),2),false);
    }
  if (dwArg & ARG_INFO_IMM_32)
    {
      PushOperand(vectOperands,new OperandImmediate(0,4),false);
      PushOperand(vectOperands,new OperandImmediate(1,4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(),4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rdword(),4),false);
    }
  if (dwArg & ARG_INFO_IMM_HI16_16_32)
    {
      PushOperand(vectOperands,new OperandImmediate(0,0,2),false);
      PushOperand(vectOperands,new OperandImmediate(1,0,2),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(),0,2),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),0,2),false);
      
      PushOperand(vectOperands,new OperandImmediate(0,0,4),false);
      PushOperand(vectOperands,new OperandImmediate(1,0,4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(),0,4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),0,4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rdword(),0,4),false);
      
      PushOperand(vectOperands,new OperandImmediate(0,rnd.rword(),2),false);
      PushOperand(vectOperands,new OperandImmediate(1,rnd.rword(),2),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(),rnd.rword(),2),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),rnd.rword(),2),false);
      
      PushOperand(vectOperands,new OperandImmediate(0,rnd.rword(),4),false);
      PushOperand(vectOperands,new OperandImmediate(1,rnd.rword(),4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rbyte(),rnd.rword(),4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),rnd.rword(),4),false);
      PushOperand(vectOperands,new OperandImmediate(rnd.rdword(),rnd.rword(),4),false);
    }
  if (dwArg & ARG_INFO_REG_STD_8)
    GenerateRegisterOps(vectOperands,rnd,false,reggSTD8,8/*4*/);
  if (dwArg & ARG_INFO_REG_STD_16)
    GenerateRegisterOps(vectOperands,rnd,false,reggSTD16,8/*4*/);
  if (dwArg & ARG_INFO_REG_STD_32)
    GenerateRegisterOps(vectOperands,rnd,false,reggSTD32,8/*4*/);
  if (dwArg & ARG_INFO_REG_SEG)
    GenerateRegisterOps(vectOperands,rnd,false,reggSEG,6/*3*/,reggSEG_MASK);
  if (dwArg & ARG_INFO_REG_DBG_CTRL_TST)
    {
      GenerateRegisterOps(vectOperands,rnd,false,reggDBG,8/*4*/);
      GenerateRegisterOps(vectOperands,rnd,false,reggCTRL,4/*2*/,reggCTRL_MASK);
      GenerateRegisterOps(vectOperands,rnd,false,reggTST,8/*4*/);
    }
  if (dwArg & ARG_INFO_REG_ST)
    GenerateRegisterOps(vectOperands,rnd,false,reggST,8/*4*/);
  if (dwArg & ARG_INFO_REG_MM)
    GenerateRegisterOps(vectOperands,rnd,false,reggMM,8/*4*/);
  if (dwArg & ARG_INFO_REG_XMM)
    GenerateRegisterOps(vectOperands,rnd,false,reggXMM,8/*4*/);
  if (dwArg & ARG_INFO_MEM_8)
    GenerateMemoryOps(vectOperands,rnd,false,15,1,!!(dwArg & ARG_INFO_MEM_SI_DI));
  if (dwArg & ARG_INFO_MEM_16)
    GenerateMemoryOps(vectOperands,rnd,false,15,2,!!(dwArg & ARG_INFO_MEM_SI_DI));
  if (dwArg & ARG_INFO_MEM_32)
    GenerateMemoryOps(vectOperands,rnd,false,15,4,!!(dwArg & ARG_INFO_MEM_SI_DI));
  if (dwArg & ARG_INFO_MEM_64)
    GenerateMemoryOps(vectOperands,rnd,false,15,8);
  if (dwArg & ARG_INFO_MEM_128)
    GenerateMemoryOps(vectOperands,rnd,false,15,16);
  if (dwArg & ARG_INFO_MEM_OTHER)
    {
      
      assert (wMemOtherSize[0]);
      GenerateMemoryOps(vectOperands,rnd,false,15,wMemOtherSize[0]);
      if (wMemOtherSize[1])
	GenerateMemoryOps(vectOperands,rnd,false,15,wMemOtherSize[1]);
    }
  
  if (dwArg & ARG_INFO_NO_OPERAND)
    PushOperand(vectOperands,NULL,false);
}

void InstrArgInfo::GenerateWrongOperands(VectorInstrArgOperand& vectOperands,
					 Rnd& rnd,
					 dword dwArg)
{
  // used to generate operands that should not be accepted. As they
  // generally are much larger in number, we can't make so many
  // without making too much combinations when running all for each of
  // dest,src,src2 against each other
  
  if (dwArg & ARG_INFO_IMM_8)
    PushOperand(vectOperands,new OperandImmediate(rnd.rbyte()),true);
  if (dwArg & ARG_INFO_IMM_16)
    PushOperand(vectOperands,new OperandImmediate(rnd.rword(),2),true);
  if (dwArg & ARG_INFO_IMM_32)
    PushOperand(vectOperands,new OperandImmediate(rnd.rdword(),4),true);
  if (dwArg & ARG_INFO_IMM_HI16_16_32)
    {
      PushOperand(vectOperands,new OperandImmediate(rnd.rword(),rnd.rword(),2),true);
      PushOperand(vectOperands,new OperandImmediate(rnd.rdword(),rnd.rword(),4),true);
    }
  if (dwArg & ARG_INFO_REG_STD_8)
    PushOperand(vectOperands,new OperandRegister(reggSTD8 | rnd.rbyte(8)),true);
  if (dwArg & ARG_INFO_REG_STD_16)
    PushOperand(vectOperands,new OperandRegister(reggSTD16 | rnd.rbyte(8)),true);
  if (dwArg & ARG_INFO_REG_STD_32)
    PushOperand(vectOperands,new OperandRegister(reggSTD32 | rnd.rbyte(8)),true);
  if (dwArg & ARG_INFO_REG_SEG)
    PushOperand(vectOperands,new OperandRegister(reggSEG | rnd.rbyte(6,3)),true);
  if (dwArg & ARG_INFO_REG_DBG_CTRL_TST)
    {
      PushOperand(vectOperands,new OperandRegister(reggDBG | rnd.rbyte(8)),true);
      PushOperand(vectOperands,new OperandRegister(reggCTRL | rnd.rbyte_mask(reggCTRL_MASK)),true);
      PushOperand(vectOperands,new OperandRegister(reggTST | rnd.rbyte(8)),true);
    }
  if (dwArg & ARG_INFO_REG_ST)
    PushOperand(vectOperands,new OperandRegister(reggST | rnd.rbyte(8)),true);
  if (dwArg & ARG_INFO_REG_MM)
    PushOperand(vectOperands,new OperandRegister(reggMM | rnd.rbyte(8)),true);
  if (dwArg & ARG_INFO_REG_XMM)
    PushOperand(vectOperands,new OperandRegister(reggXMM | rnd.rbyte(8)),true);
  if (dwArg & ARG_INFO_MEM_8)
    GenerateMemoryOps(vectOperands,rnd,true,2,1);
  if (dwArg & ARG_INFO_MEM_16)
    GenerateMemoryOps(vectOperands,rnd,true,2,2);
  if (dwArg & ARG_INFO_MEM_32)
    GenerateMemoryOps(vectOperands,rnd,true,2,4);
  if (dwArg & ARG_INFO_MEM_64)
    GenerateMemoryOps(vectOperands,rnd,true,2,8);
  if (dwArg & ARG_INFO_MEM_128)
    GenerateMemoryOps(vectOperands,rnd,true,2,16);
}

void InstrArgInfo::GenerateOperands(VectorInstrArgOperand& vectOperandsDest,
				    VectorInstrArgOperand& vectOperandsSrc,
				    VectorInstrArgOperand& vectOperandsSrc2,
				    dword dwRndSeed)
{
  Rnd rnd(dwRndSeed);
  
  GenerateOperands     (vectOperandsDest,rnd,dwDestArg);
  if (dwDestArg != ARG_INFO_NO_OPERAND)
    GenerateWrongOperands(vectOperandsDest,rnd,~dwDestArg);
  
  GenerateOperands     (vectOperandsSrc ,rnd,dwSrcArg);
  if (dwSrcArg != ARG_INFO_NO_OPERAND)
    GenerateWrongOperands(vectOperandsSrc,rnd,~dwSrcArg);
  
  GenerateOperands     (vectOperandsSrc2,rnd,dwSrc2Arg);
  if (dwSrc2Arg != ARG_INFO_NO_OPERAND)
    GenerateWrongOperands(vectOperandsSrc2,rnd,~dwSrc2Arg);
}

//////////////////////////////////////////////////////////////////////

dword ArgRegFromType(byte bReg)
{
  switch (REG_TYPE(bReg))
    {
    case reggSTD8:	return ARG_INFO_REG_STD_8;	
    case reggSTD16:	return ARG_INFO_REG_STD_16;			
    case reggSTD32:	return ARG_INFO_REG_STD_32;			
      
    case reggSEG:	return ARG_INFO_REG_SEG;				
      
    case reggDBG:				
    case reggCTRL:				
    case reggTST:	return ARG_INFO_REG_DBG_CTRL_TST;				

    case reggST:	return ARG_INFO_REG_ST;				
    case reggMM:	return ARG_INFO_REG_MM;				
    case reggXMM:	return ARG_INFO_REG_XMM;				

    case reggSTD8_16_32:   return ARG_INFO_REG_STD_8 | ARG_INFO_REG_STD_16 | ARG_INFO_REG_STD_32;	
    case reggSTD16_32: return ARG_INFO_REG_STD_16 | ARG_INFO_REG_STD_32;	
    }
  assert (false);
  return 0;
}

dword ArgRegFromRT(dword dwRT)
{
  switch (dwRT)
    {
    case arg_MODRM_RT_FPU_REG: return ARG_INFO_REG_ST;
    case arg_MODRM_RT_MMX_REG: return ARG_INFO_REG_MM;
    case arg_MODRM_RT_SSE_REG: return ARG_INFO_REG_XMM;
    }
  assert (false);
  return 0;
}

dword ArgMultFromOpsize(dword nArguments,bool bLastOp)
{
  dword dwMult = 0;

  if (nArguments & arg_OPSIZE_8)
    dwMult |= 1;
  if (nArguments & arg_OPSIZE_16)
    dwMult |= 2;
  if (nArguments & arg_OPSIZE_32)
    dwMult |= 4;
  if (nArguments & arg_OPSIZE_64)
    dwMult |= 8;
  if (nArguments & arg_OPSIZE_128)
    dwMult |= 16;
  
  if (bLastOp)
    {
      if (nArguments & arg_LAST_OPRND_OPSIZE_8)
	dwMult = 1;
      if (nArguments & arg_LAST_OPRND_OPSIZE_16)
	dwMult = 2;
      if (nArguments & arg_LAST_OPRND_DBL_OPSIZE)
	dwMult <<= 1;
      if (nArguments & arg_LAST_OPRND_HALF_OPSIZE)
	dwMult >>= 1;
    }
  assert (dwMult);
  return dwMult;
}

#define ARG_MULT_FROM_OPSIZE(arg,lastop,op) (assert(ArgMultFromOpsize(arg,lastop)<8),ArgMultFromOpsize(arg,lastop)*op)

void AddMemOtherSize (InstrArgInfo* pArgInfo,word nExtra)
{
  pArgInfo->AddMemOtherSize(OPSIZE_FROM_EXTRA(nExtra));
}

void AddArgTypes(const OpcodeInfo& info,dword dwOp,InstrArgInfo* pArgInfo,dword& dwArg,bool bLastOp,byte bSIDIMask = 0)
{
  switch (dwOp)
    {
    case arg_OP_IS_IMM:
      if (bLastOp && info.m_nArguments & (arg_LAST_OPRND_OPSIZE_8 | arg_LAST_OPRND_OPSIZE_16))
	{
	  if (info.m_nArguments & arg_LAST_OPRND_OPSIZE_8)
	    dwArg |= ARG_INFO_IMM_8;
	  else
	    {
	      assert (info.m_nArguments & arg_LAST_OPRND_OPSIZE_16);
	      dwArg |= ARG_INFO_IMM_16;
	    }
	}
      else if (info.m_nArguments & arg_LAST_OP_MEM_OR_IMM_FAR_PTR)
	dwArg |= ARG_INFO_IMM_HI16_16_32;
      else
	dwArg |= ARG_MULT_FROM_OPSIZE(info.m_nArguments,bLastOp,ARG_INFO_IMM_8);
      break;
    case arg_OP_CONST_1: 
      dwArg |= ARG_INFO_IMM_8;
      break;
      
    case arg_OP_AS_REG_EXTRA:
      dwArg |= ArgRegFromType (info.m_nExtra1);
      break;
    case arg_OP_AS_REG_EXTRA2:
      dwArg |= ArgRegFromType (info.m_nExtra2);
      break;
      
    case arg_OP_LOW_3BITS_REG:
      dwArg |= ARG_MULT_FROM_OPSIZE(info.m_nArguments,
				    bLastOp,ARG_INFO_REG_STD_8);
      break;
      
    case arg_OP_IS_MEM_DISP:
      dwArg |= ARG_MULT_FROM_OPSIZE(info.m_nArguments,
				    bLastOp,ARG_INFO_MEM_8);
      break;
      
    case arg_OP_IS_MODRM_REG:
      if (info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL)
	dwArg |= ArgRegFromType (info.m_nExtra1);
      else if (info.m_nArguments & arg_MODRM_RT_REG_MASK)
	dwArg |= ArgRegFromRT ((info.m_nArguments >> arg_MODRM_RT_REG_SHIFT) & 
			       arg_MODRM_RT_MASK);
      else
	dwArg |= ARG_MULT_FROM_OPSIZE(info.m_nArguments,
				      bLastOp,ARG_INFO_REG_STD_8);
      break;
      
    case arg_OP_IS_MODRM_MODRM:
      if (!(info.m_nArguments & arg_MODRM_ONLY_M_MODRM))
	{
	  if (info.m_nArguments & arg_MODRM_RT_RM_MASK)
	    dwArg |= ArgRegFromRT ((info.m_nArguments >> arg_MODRM_RT_RM_SHIFT) & arg_MODRM_RT_MASK);
	  else
	    dwArg |= ARG_MULT_FROM_OPSIZE(info.m_nArguments,
					  bLastOp,ARG_INFO_REG_STD_8);
	}
      if (!(info.m_nArguments & arg_MODRM_ONLY_R_MODRM))
	{
	  if (info.m_nArguments & arg_OPSIZE_IN_EXTRA2)
	    {
	      dwArg |= ARG_INFO_MEM_OTHER;
	      
	      AddMemOtherSize (pArgInfo,info.m_nExtra2);
	    }
	  else if (info.m_nArguments & arg_OPSIZE_IN_EXTRAS_16_32)
	    {
	      dwArg |= ARG_INFO_MEM_OTHER;
	      
	      AddMemOtherSize (pArgInfo,info.m_nExtra1);
	      AddMemOtherSize (pArgInfo,info.m_nExtra2);
	    }			
	  else
	    dwArg |= 
	      ArgMultFromOpsize(info.m_nArguments,bLastOp) * ARG_INFO_MEM_8;
	}
      break;
    case 0:
      if ((info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) &&
	  (info.m_nExtra2 & bSIDIMask))
	dwArg |= ARG_MULT_FROM_OPSIZE(info.m_nArguments,
				      bLastOp,ARG_INFO_MEM_8) | 
	  ARG_INFO_MEM_SI_DI;
      else
	dwArg |= ARG_INFO_NO_OPERAND;
      break;
    }
}	


/*
#define ARG_INFO_IMM_8             0x00000001
#define ARG_INFO_IMM_16            0x00000002
#define ARG_INFO_IMM_32			   0x00000004
#define ARG_INFO_IMM_16_32         (ARG_INFO_IMM_16 | ARG_INFO_IMM_32)
#define ARG_INFO_IMM_HI16_16_32    0x00000008
#define ARG_INFO_REG_STD_8         0x00000010
#define ARG_INFO_REG_STD_16        0x00000020
#define ARG_INFO_REG_STD_32        0x00000040
#define ARG_INFO_REG_STD_16_32     (ARG_INFO_REG_STD_16 | ARG_INFO_REG_STD_32)
#define ARG_INFO_REG_SEG           0x00000100
#define ARG_INFO_REG_DBG_CTRL_TST  0x00000200
#define ARG_INFO_REG_ST            0x00000400
#define ARG_INFO_REG_MM            0x00000800
#define ARG_INFO_REG_XMM           0x00001000
#define ARG_INFO_MEM_8             0x00010000
#define ARG_INFO_MEM_16            0x00020000
#define ARG_INFO_MEM_32            0x00040000
#define ARG_INFO_MEM_16_32         (ARG_INFO_MEM_16 | ARG_INFO_MEM_32)
#define ARG_INFO_MEM_64            0x00080000
#define ARG_INFO_MEM_128           0x00100000
#define ARG_INFO_MEM_OTHER         0x00200000
*/

void TestEncoderDecoderInstrs::AddArgTypes(const OpcodeInfo& info,word nInstr)
{
  // find someplace to put the info
  
  InstrArgInfo* pArgInfo;
  
  MapInstrArgInfo::iterator arg_info;
  
  if ((arg_info = mapInstrArgInfo.find(nInstr)) == mapInstrArgInfo.end())
    arg_info = 
      (mapInstrArgInfo.insert(MapInstrArgInfo::value_type(nInstr,
							  InstrArgInfo()))).first;
  
  pArgInfo = &(arg_info->second);
  
  // ok, so let's see what arguments this says that info.m_nInstruction can have
	
  dword dwDestOp = (info.m_nArguments & arg_OP_DEST_MASK) >> arg_OP_DEST_SHIFT;
  dword dwSrcOp  = (info.m_nArguments & arg_OP_SRC_MASK ) >> arg_OP_SRC_SHIFT;
  dword dwSrc2Op = (info.m_nArguments & arg_OP_SRC2_MASK) >> arg_OP_SRC2_SHIFT;
  
  ::AddArgTypes(info,dwDestOp,pArgInfo,pArgInfo->dwDestArg,
		false    ,(ext2_DEST_MEM_DI | ext2_DEST_MEM_SI));
  ::AddArgTypes(info,dwSrcOp ,pArgInfo,pArgInfo->dwSrcArg ,
		!dwSrc2Op,(ext2_SRC_MEM_DI  | ext2_SRC_MEM_SI ));
  ::AddArgTypes(info,dwSrc2Op,pArgInfo,pArgInfo->dwSrc2Arg,
		true);
}

void TestEncoderDecoderInstrs::AddArgTypes(const OpcodeInfo& info)
{
  if (!(info.m_nInstruction & instrgNOT_INSTR))
    {
      AddArgTypes(info,info.m_nInstruction);
      
      if (((info.m_nArguments & arg_OP_DEST_MASK) == arg_DEST_IS_MEM_DISP ||
	   ((info.m_nArguments & arg_OP_DEST_MASK) == arg_DEST_IS_MODRM_MODRM &&
	    !(info.m_nArguments & arg_MODRM_ONLY_R_MODRM))) && // tested first, as easier, and most likely false
	  IsValidLockInstruction (info.m_nInstruction)) // tested last 
	{
	  // if it may have lock prefix
	  // and can encode dest as memory reference 
	  
	  OpcodeInfo info2 = info;
	  
	  info2.m_nArguments |= arg_MODRM_ONLY_M_MODRM; 
	  
	  AddArgTypes(info2,
		      info2.m_nInstruction | instrgmINT_ARITH_MEMORY_LOCK);
	}
      else // can't have both lock and rep
	if ((info.m_nInstruction & instrgMASK) == instrgSTRING)
	  {
	    if (info.m_nInstruction == instrCMPS ||
		info.m_nInstruction == instrSCAS)
	      {
		AddArgTypes(info,info.m_nInstruction | instrgmSTRING_REPNE);
		AddArgTypes(info,info.m_nInstruction | instrgmSTRING_REPE);
	      }
	    else
	      AddArgTypes(info,info.m_nInstruction | instrgmSTRING_REP);
	  }
    }
  else if ((info.m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
    {
      int nGroupType = info.m_nInstruction & instrmGROUP_TYPE_MASK;
      
      if ((nGroupType & instrmGROUP_MOD_BOTH) ||
	  (nGroupType & instrmGROUP_MOD_MEM))
	for (int n = 0; n < 8; n++)
	  {
	    OpcodeInfo info2 = infoGroups[info.m_nExtra1][n];
	    info2.m_nArguments |= 
	      info.m_nArguments | ((nGroupType & instrmGROUP_MOD_BOTH) ? 
				   0 : arg_MODRM_ONLY_M_MODRM);
	    if (!(nGroupType & instrmGROUP_MOD_11))
	      info2.m_nExtra2    |= info.m_nExtra2;
	    AddArgTypes(info2);
	  }	
      if (nGroupType & instrmGROUP_MOD_11)
	for (int n = 0; n < 8; n++)
	  {
	    OpcodeInfo info2 = infoGroups[info.m_nExtra2][n];
	    info2.m_nArguments |= info.m_nArguments | arg_MODRM_ONLY_R_MODRM;
	    AddArgTypes(info2);
	  }	
    }
  else if (info.m_nInstruction == instrSUBGRP_COPROC)
    {
      for (int n = 0; n < 8; n++)
	{
	  OpcodeInfo info2 = infoGroups[info.m_nExtra1][n];
	  info2.m_nArguments |= info.m_nArguments;
	  info2.m_nExtra2    |= info.m_nExtra2;
	  AddArgTypes(info2);
	}
    }
  else if (info.m_nInstruction == instrPRE_MMX_SSE)
    {
      int nOffset = (info.m_nExtra1 << 1);
      int nNumPrefix = ((info.m_nExtra2 & ext2_SSE_PREFIX_F3_F2) ? 4 : 2);
      
      for (int nPrefix = 0 ; nPrefix < nNumPrefix; nPrefix++)
	{
	  OpcodeInfo info2 = infoMMX_SSE[nOffset+nPrefix];
	  AddArgTypes(info2);
	}
    }
  else if (info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
    {
      int nOffset = (info.m_nExtra1 << 1);
      int nNumPrefix = 4;
      
      for (int nPrefix = 0 ; nPrefix < nNumPrefix; nPrefix++)
	{
	  OpcodeInfo info2 = infoMMX_SSE[nOffset+nPrefix];
	  info2.m_nArguments |= 
	    (nPrefix & 0x02 ? arg_MODRM_ONLY_R_MODRM : arg_MODRM_ONLY_M_MODRM);
	  AddArgTypes(info2);
	}
    }
  else
    {
      assert (info.m_nInstruction == instrRESERVED ||
	      info.m_nInstruction == instrPREFIX ||
	      info.m_nInstruction == instrESCAPE2BYTE);
    }
}

// #include "gen/Instruction_Names.h" // dep_chk_unknown 
int compare_short (const void* p1,const void* p2);

void TestEncoderDecoderInstrs::DoTest()
{
  // loop through table, and find out (roughly) what kind of arguments
  // can be used with each kind of instruction.  (So we don't try
  // unusable ones) actually, we can generate som impossible too, to
  // make sure they are NOT accepted
  
  mapInstrArgInfo.clear();
  
  for (int nTwoByte = 0; nTwoByte < 2; nTwoByte++)
    for (int nByte = 0; nByte < 0x100; nByte++)
      AddArgTypes(infoByte[nTwoByte][nByte]);
  
  // loop through all possible instructions generate instruction with
  // some random arguments if we're able to encode it, we should also
  // be able to decode it
  
  AnyInstruction* pAnyInstr = new AnyInstruction;
  BranchInstr* pBranchInstr = new BranchInstr;
  
  for (word wInstr = 1; wInstr; wInstr++)
    {
      // is it existing instruction?
      
      // would we have a name for it?, and can we encode it
      
      InstrIdNameIndex* pNameIndex = 
	(InstrIdNameIndex*) bsearch (&wInstr,
				     indexInstrIdName,
				     NUMBER_OF_INSTR_NAMES,
				     sizeof (InstrIdNameIndex),
				     compare_short);
      
      InverseOpcodeInfo* pInfo = m_encoder.FindFirstInverseOpcodeInfo(wInstr);
      
      MapInstrArgInfo::iterator arg_info = mapInstrArgInfo.find(wInstr);
      
      assert (!pNameIndex == !pInfo && // if it's encodable <=> it's nameable
	      !pNameIndex == (arg_info == mapInstrArgInfo.end())); //  if it's encodable <=> we should know about possible operands
      
      if (!pNameIndex)
	continue;
      
      InstrArgInfo* pArgInfo = &(arg_info->second);
      
      // ok, so we've got an instruction to test
      // now we have to create some operands for it
      
      VectorInstrArgOperand vectOpDest;
      VectorInstrArgOperand vectOpSrc;
      VectorInstrArgOperand vectOpSrc2;
      
      // fill in possible operands into vectors
      
      pArgInfo->GenerateOperands(vectOpDest,vectOpSrc,vectOpSrc2,wInstr + 456); // 456 can be changed to change seed
      
      // loop through all combinations
      
      pAnyInstr->m_nInstruction = wInstr;
      pBranchInstr->m_nInstruction = wInstr;
      
      Instruction* pInstr = 
	((wInstr & instrgMASK) == instrgCTRL_TRANSFER) ? 
	(Instruction*) pBranchInstr : (Instruction*) pAnyInstr;
      
      {					
	char fmt3[64];
	
	pInstr->InstrAsmString(fmt3);
	
	printf ("%-9s: %3d %3d %3d\n",fmt3,vectOpDest.size(),vectOpSrc.size(),vectOpSrc2.size());
      }
      
      /*
		{FOR_EACH_ITER(VectorInstrArgOperand,vectOpDest,iter) // this loop first as it changes the least
		{					
			char fmt3[64];
			if (iter->m_pOperand)
				*(iter->m_pOperand->GenAsmString(fmt3)) = '\0';		
			else
				strcpy (fmt3,"NULL");
			printf ("dest: %s\n",fmt3);

		}}
		{FOR_EACH_ITER(VectorInstrArgOperand,vectOpSrc,iter) // this loop first as it changes the least
		{					
			char fmt3[64];
			if (iter->m_pOperand)
				*(iter->m_pOperand->GenAsmString(fmt3)) = '\0';		
			else
				strcpy (fmt3,"NULL");
			printf ("src: %s\n",fmt3);
		}}
		{FOR_EACH_ITER(VectorInstrArgOperand,vectOpSrc2,iter) // this loop first as it changes the least
		{					
			char fmt3[64];
			if (iter->m_pOperand)
				*(iter->m_pOperand->GenAsmString(fmt3)) = '\0';		
			else
				strcpy (fmt3,"NULL");
			printf ("src2: %s\n",fmt3);
		}}
*/

      bool bAnyEncoding = false;
      
      FOR_EACH_ITER(VectorInstrArgOperand,vectOpSrc2,src2) // this loop first as it changes the least
	{
	  pAnyInstr->m_pSrc2Op = src2->m_pOperand;
	  FOR_EACH_ITER(VectorInstrArgOperand,vectOpSrc,src)
	    {
	      pAnyInstr->m_pSrcOp = src->m_pOperand;
	      FOR_EACH_ITER(VectorInstrArgOperand,vectOpDest,dest)
		{
		  pAnyInstr->m_pDestOp = dest->m_pOperand;
		  pBranchInstr->m_pTargetOp = dest->m_pOperand;
		  
		  // get all possible encodings of this instruction,
		  // and make sure they decode to the same instruction
		  
		  VectorInstrCodeLength vectInstrCodeLength;
		  
		  if (m_encoder.Encode (pInstr,vectInstrCodeLength))
		    {
		      /*
			{					
			char fmt3[64];
			char fmt4[128];
		
							pInstr->InstrAsmString(fmt3);
							pInstr->OpAsmString(fmt4);
						
							printf ("%-9s %s\n",fmt3,fmt4);
						}

						if (1)
						{
							FOR_EACH_ITER(VectorInstrCodeLength,vectInstrCodeLength,instr)
							{
								printf ("                                ");
								int j;
								for (j = 0; j < instr->m_bLength; j++)
									printf (" %02x",*(instr->m_cCode+j));
//								while (j++ < 14)
//									printf ("   ");								
								printf ("\n");
							}
						}
*/

		      // because we should not be able to encode it
		      // with an wrong operand (either encoder faulty,
		      // or system that detect possible/wrong
		      // operands) there is of course no guarantee
		      // that possible operands will encode, just that
		      // wrong ones shouldn't
		      assert (!dest->m_bWrongOp); 
		      assert (!src->m_bWrongOp); 
		      assert (!src2->m_bWrongOp);
		      
		      FOR_EACH_ITER(VectorInstrCodeLength,
				    vectInstrCodeLength,instr)
			{
			  // so decode the bytes and see that they decode into what we created
			  
			  VerifyDecodeIntoSame (pInstr,
						instr->m_cCode,
						instr->m_bLength);
			}
		      bAnyEncoding = true;
		    }
		}
	    }
	}
      assert (bAnyEncoding); // if we have the instruction, it should be encodable with at least some set of arguments
    }
  pAnyInstr->m_pSrc2Op = NULL;
  pAnyInstr->m_pSrcOp = NULL;
  pAnyInstr->m_pDestOp = NULL;
  delete pAnyInstr;
  delete pBranchInstr;
}


//////////////////////////////////////////////////////////////////////
/*
int bit_errors(byte b1,byte b2)
{
	// there is probably some assembler for this...
	// counting number of differing bits

	b1 ^= b2;

	// all bits being 1 are errors

	b2 = (b1 & 0x55) + ((b1 & 0xaa) >> 1); // 0?0?0?0? + ?0?0?0?0 >> 1  =>  4x(?? ) ?? =0,1 or 4

	b1 = (b2 & 0x33) + ((b2 & 0xcc) >> 2); // 00??00?? + ??00??00 >> 2  =>  2x(0??? ) ?? =0,1,2,3 or 4

	return (b1 & 0x0f) + (b1 >> 4);
}
*/

bool TestDecoderEncoder::TestDecodeEncode(const byte** pEndDecode,
					  bool bOutputInstrCode)
{
  // now attempt decoding
  
  m_code.ResetCurrent();
  
  Instruction* pInstr = NULL;
  
  try
    {
      pInstr = m_decoder.Decode (m_code); // m_cCode
    }
  catch (...)
    {
      
    }
  
  *pEndDecode = (const byte*) m_code.GetCurrent();
  
  if (!pInstr)
    return false; // could not decode
  
  byte bLength = m_code.GetRVA();
  
  // debug print
  
  if (bOutputInstrCode)
    {
      int j;
      printf ("-> ");
      for (j = 0; j < bLength; j++)
	printf (" %02x",*(m_code.m_pStart+j));
      while (j++ < 14)
	printf ("   ");
      
      char fmt3[64];
      char fmt4[128];
      
      pInstr->InstrAsmString(fmt3);
      pInstr->OpAsmString(fmt4);
      
      printf ("  %-9s %s  \n",fmt3,fmt4);
    }
  
  // ok, so now recode the stuff
  
  /* Testing code has changed (see below), because It's not necessary
      that the encoder will find any encoding with the same length as
      the original instruction decoded, as it may have had an
      unnecessary prefix that has been simplified away, making it
      shorter. Or It may have increased the length of the displacement
      in the memory operand (to avoid an prefix)
      
      But: we should somehow be able to encode it...

      word wLengths = m_encoder.EncodeLengths (pInstr);

        assert (wLengths & (1 << (bLength - 1))); // or we can't encode it with the same length: encoding/decoding error

        // now make sure any encoding with that length is a match, and
        // that at least some differs by at most, say 7 bits, or an
        // entire byte is useless and there are no such as far as I
        // know...

        VectorInstrCode vectInstrCode;

        verify (m_encoder.Encode(pInstr,vectInstrCode,bLength)); // should work as EncodeLengths worked

//	int nMinWrongBits = bLength*8;

	FOR_EACH_ITER(VectorInstrCode,vectInstrCode,iter)
		{
			if (memcmp (iter->m_cCode,m_code.m_pStart,bLength) == 0)
				{
//					nMinWrongBits = 0;
					continue;
				}
			// so it wasn't perfect match

			int j;
			printf ("   ");
			for (j = 0; j < bLength; j++)
				printf (" %02x",*(iter->m_cCode+j));
			printf ("\n");

			// make sure it decodes to same instruction

			VerifyDecodeIntoSame(pInstr,iter->m_cCode,bLength);

//			int nWrongBits = 0;
//
//			for (const byte *p1 = m_code.m_pStart, *p2 = iter->m_cCode; p1 < *pEndDecode; p1++, p2++)
//				nWrongBits += bit_errors(*p1,*p2);
//
//			if (nWrongBits < nMinWrongBits)
//				nMinWrongBits = nWrongBits;
		}
// assert (nMinWrongBits <= 7); */

 word wLengths = m_encoder.EncodeLengths (pInstr);
 assert (wLengths); // somehow be able to encode it
 
 if (!(wLengths & ~(0xffff << bLength)))
   {
     // we must be able to encode at least as short as we decoded the
     // instruction, the only thing that could change is that the
     // encoder dislike prefixes, and might reorder base/index to
     // avoid segment prefix, this reordering will cause an extra SIB,
     // but then we got rid of the prefix, so the instruction still
     // has the same length we can also increase instruction size by
     // issuing to long displacement, but only if request to make long
     // instruction, the shortest possible should have no extra
     // displacement
     
     // however, DS:[EBP+??] can only be encoded as that or
     //             [no base+EBP+????????], which does not require prefix, but SIB and 4 extra bytes of displacement
     
     const OperandMemory* pOpMem;
     
     if (((pOpMem = dynamic_cast<const OperandMemory*>(pInstr->GetDestOp())) ||
	  (pOpMem = dynamic_cast<const OperandMemory*>(pInstr->GetSrcOp()))) &&
	 ((pOpMem->m_bBaseReg == regEBP && pOpMem->m_bIndexReg == 0) ||
	  (pOpMem->m_bIndexReg == regEBP && pOpMem->m_bBaseReg == 0)) &&
	 pOpMem->m_bScale == 0 &&
	 pOpMem->m_bSegmentReg == regDS &&
	 pOpMem->m_bDispBytes == 1)
       {
	 assert ((wLengths & ~(0xffff << (bLength+3)))); // +sib-prefix+disp4-disp1
       }
     else
       assert(false);
   }
 
 
 VectorInstrCodeLength vectInstrCodeLength;
 
 verify (m_encoder.Encode(pInstr,vectInstrCodeLength)); // should work as EncodeLengths worked
 
 //	int nMinWrongBits = bLength*8;
 
 FOR_EACH_ITER(VectorInstrCodeLength,vectInstrCodeLength,iter)
   {
     wLengths &= ~(1 << (iter->m_bLength-1));		
     
     if (iter->m_bLength == bLength &&
	 memcmp (iter->m_cCode,m_code.m_pStart,bLength) == 0)
       {
	 // perfect match
	 
	 //					nMinWrongBits = 0;
	 //			printf ("Match...\n");
	 continue;
       }
     // so it wasn't perfect match
     
     //		int j;
     //		printf ("   ");
     //		for (j = 0; j < iter->m_bLength; j++)
     //			printf (" %02x",*(iter->m_cCode+j));
     //		printf ("\n");
     
     // make sure it decodes to same instruction
     
     VerifyDecodeIntoSame(pInstr,iter->m_cCode,iter->m_bLength);
   }
 assert (!wLengths); // should have found all lengths he claimed to be able to generate
 
 delete pInstr;
 
 return true;
}

void TestDecoderEncoder::TestDispImm(byte* bCode,
				     const byte** pEndDecode,
				     bool bOutputInstrCode)
{
  // first attempt with disp and imm set to zero then with all one,
  // and then some pseudo-random
  
  // normally disp and imm should not affect success/failure, so that
  // should be equal for all
  
  const byte* pEndDecode1;
  
  byte* ptr;
  
  for (ptr = bCode; ptr < m_code.m_pEnd; ptr++)
    *ptr = 0;
  
  bool bSuccess0 = TestDecodeEncode(pEndDecode,bOutputInstrCode);
  
  for (ptr = bCode; ptr < *pEndDecode; ptr++)
    *ptr = 0xff;
  
  bool bSuccess1 = TestDecodeEncode(&pEndDecode1);
  
  assert (bSuccess0 == bSuccess1);     // should fail / success on equal
  assert (*pEndDecode == pEndDecode1); // should fail / success on equal
  
  if (!bSuccess0)
    return;
  
  if (*pEndDecode == bCode)
    return; // no need to try more variants as disp/imm is not used anyway
  
  // so generate some different imm/disp pairs and make sure
  
  // the only instructions that could care about contents of imm/disp
  // are some MMX/SSE compare instructions CMPSS et. al., that have
  // their predicate in the imm byte the decoder/encoder does not care
  // about if those are in range... (if it did, we'd have different
  // success here depending on value of imm
  
  Rnd rnd((*((dword*) m_cCode) & 0x7fffffff) | 0x10000); // make sure we're in range, and not zero
  /*
    byte bAdd = 0;
	byte bXor = 0;

	for (ptr = m_cCode; ptr < bCode; ptr++)
		{
			bAdd += *ptr;
			bXor ^= *ptr;
		}
*/
  for (int i = 2; i ; i--)
    {
      for (ptr = bCode; ptr < pEndDecode1; ptr++)
	*ptr = rnd.rbyte(); 
      
      bSuccess1 = TestDecodeEncode(&pEndDecode1);
      
      assert (bSuccess1);     
      assert (*pEndDecode == pEndDecode1); // should end on equal (should always need same amount of immediate / disp)
    }														
  // ok, so we're done with this
}



void TestDecoderEncoder::TestByte (byte* bCode)
{
  if (m_nTwoByte)
    *(bCode++) = 0x0f;
  *(bCode++) = (byte) m_nByte;
  
  // now make modrm, sib and disp/imm
  
  // if decoding was successful not using mod/rm, then do not generate
  // more of that and if sib was not used, do not make more sibs, etc
  
  const byte* pEndDecode;
  
  *(bCode) = 0;
  
  TestDispImm(bCode+1,&pEndDecode,true);
  
  if (pEndDecode <= bCode)
    return; // modrm was ignored, so no use looping through that
  
  Rnd rnd(m_nByte | (m_nTwoByte << 8) | 0x1000000); 
  
  for (int bModRM = 1; bModRM < 0x100; bModRM++)
    {
      *(bCode) = bModRM;
      
      if (m_bAddrSize32 && (bModRM & 0xc0) != 0xc0 && (bModRM & 0x07) == 0x04) // modrm suggest sib
	{
	  *(bCode+1) = 0;
	  
	  TestDispImm(bCode+2,&pEndDecode);
	  
	  assert (pEndDecode > bCode); // if it needed modrm to decide when it was zero, it should always need it, but we can't know exactly how much imm/disp
	  
	  if (pEndDecode <= bCode + 1)
	    continue; // sib was ignored, so no use looping through that
	  
	  for (int bSIB = 1; bSIB < 0x100; bSIB += rnd.rbyte())
	    {
	      *(bCode+1) = (byte) bSIB;
	      
	      TestDispImm(bCode+2,&pEndDecode);
	      assert (pEndDecode > bCode + 1); // if it needed sib to decide when it was zero, it should always need it, but we can't know exactly how much imm/disp
	    }
	}
      else
	{
	  TestDispImm(bCode+1,&pEndDecode);
	  assert (pEndDecode > bCode); // if it needed modrm to decide when it was zero, it should always need it, but we can't know exactly how much imm/disp
	}				
    }
}

#define MOD_FROM_MODRM(modrm) (((modrm) >> 6) & 0x03)
#define REG_FROM_MODRM(modrm) (((modrm) >> 3) & 0x07)
#define RM_FROM_MODRM(modrm) ((modrm) & 0x07)
#define NNN_FROM_MODRM(modrm) REG_FROM_MODRM(modrm)


void TestDecoderEncoder::TestMMXSSEPrefix (byte* bCode)
{
  TestByte(bCode  );
  if (bCode == m_cCode || *(bCode-1) != 0x66)
    *bCode = 0x66; TestByte(bCode+1);
  *bCode = 0xf2; TestByte(bCode+1);
  *bCode = 0xf3; TestByte(bCode+1);
}

void TestDecoderEncoder::TestOpSizePrefix(byte* bCode)
{
	         TestMMXSSEPrefix (bCode  );
  *bCode = 0x66; TestMMXSSEPrefix (bCode+1);
}

void TestDecoderEncoder::TestAddrSizePrefix (byte* bCode)
{
  m_bAddrSize32 =  m_encoder.m_bCodeSize32;                TestOpSizePrefix (bCode  );
  m_bAddrSize32 = !m_encoder.m_bCodeSize32; *bCode = 0x67; TestOpSizePrefix (bCode+1);
}

void TestDecoderEncoder::TestSegmentPrefix (byte* bCode)
{
                 TestAddrSizePrefix (bCode  );
  *bCode = 0x26; TestAddrSizePrefix (bCode+1);
  *bCode = 0x2e; TestAddrSizePrefix (bCode+1);
  *bCode = 0x36; TestAddrSizePrefix (bCode+1);
  *bCode = 0x3e; TestAddrSizePrefix (bCode+1);
  *bCode = 0x64; TestAddrSizePrefix (bCode+1);
  *bCode = 0x65; TestAddrSizePrefix (bCode+1);
}

void TestDecoderEncoder::TestLockRepPrefix (byte* bCode)
{
                 TestSegmentPrefix(bCode  );
  *bCode = 0xf0; TestSegmentPrefix(bCode+1);
//	*bCode = 0xf2; TestSegmentPrefix(bCode+1); // done by TestMMXSSEPrefix
//	*bCode = 0xf3; TestSegmentPrefix(bCode+1); // done by TestMMXSSEPrefix
}

TestDecoderEncoder::TestDecoderEncoder()
{
  m_code.SetSrc(m_cCode,m_cCode+sizeof(m_cCode));
}

void TestDecoderEncoder::DoTest ()
{
  // generate semi-random patterns and if we can decode it, we should
  // alse be able to encode it

  // instructions are represented by [prefixes][two-byte-prefix]byte[mod/rm][sib][disp][imm]
  // so generate by looping over     --------------------------------------------
  // and do some random variations on                                            -----------
  
  // prefixes: four groups segment:6+1 rep/lock: 3+1 opsize: 1+1 addrsize 1+1
  // +1 is for no prefix, giving 7*4*2*2 variations, and the possible
  // prefixing for mmx/sse: *4
  // two-byte-prefix: *2
  // byte: *256
  // mod/rm: *256
  // sib: *256 (/32  reduced, because only needed when mod/rm said so)
  // random disp/imm, some 10 combinations, I suggest based on the preceding bytes
  
  // multiply by four to change default op/addr size

  // => 7*4*2*2*4*2*256*256*256/32*10*2 = 9e9 which is a lot
  
  
  for (m_nByte = 0x00; m_nByte < 0x100; m_nByte++)
    for (m_nTwoByte = 0; m_nTwoByte < 2; m_nTwoByte++)
      {
	if (infoByte[m_nTwoByte][m_nByte].m_nInstruction == instrRESERVED ||
	    infoByte[m_nTwoByte][m_nByte].m_nInstruction == instrESCAPE2BYTE ||
	    infoByte[m_nTwoByte][m_nByte].m_nInstruction == instrPREFIX)
	  continue; // no use trying nonexistent, or prefix
	
	/* Now we could probably simplify away of rep/lock prefixes by
	   only issuing them when giving a suitable instruction, but
	   that sometimes requires knowledge of mod/rm part (for group
	   nnn), so we don't */
				
	TestLockRepPrefix (m_cCode);
      }
}

//////////////////////////////////////////////////////////////////////

void VerifyEncoderDecoder()
{
  // perform tests of our representation of the x86 instruction set
  // 
  // the decoder/encoder system is built around a (or actually three)
  // tables in InstructionDecoder_Info.cpp used for decoding
  // instructions with functions in InstructionDecoder, the encoder
  // uses the very same table with functions in InstructionEncoder,
  // and a automatically built inverse table to quickly find in what
  // ways an instruction possibly can be encoded
  //
  // this test is not all inclusive as it uses the table in
  // InstructionDecoder_Info.cpp to guide it's tests, so if the table
  // is wrong then we wont find that. But it should find most errors
  // in the encoding/decoding functions
  //
  // The idea is to loop through all possible instructions and
  // generate them in the internal representation. Then try to encode
  // the instruction. After that, decode the very same byte pattern
  // (which should consume exactly the same length of the pattern, and
  // check that the decoded instruction matches the encoded
  //
  // Another idea is to generate a pattern, have it decoded and if the
  // decoding was successful (i.e. the instruction was intelligible)
  // have it encoded and see that it's encoded in the same way this
  // approch have the trouble that certain instructions may be encoded
  // in different ways: very different ways, but that's taken care of
  // by encoder giving all different ways different because some bit
  // is silently ignored. This is a bit harder, but could be verified
  // be redecoding the similar stuff and if it's ok, then accept the
  // encoding
  // 
  // Bottom line: we're not testing everything, but should find most
  // bad things There is nothing like writing perfect code from start,
  // so good that bugs never appear, not even while testing! (This is
  // not on of those rare programs)
  
#define DO_TEST(classname) {classname test;test.Test();}
  
  DO_TEST(TestEncoderDecoderInstrs);
  //	DO_TEST(TestDecoderEncoder);
}
