/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// main.cpp: main
//
//////////////////////////////////////////////////////////////////////

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#ifdef WIN32
#include <io.h>
#include <conio.h>
#else
#include <unistd.h>
#endif

#include "GlobalDefs.h"
#include "PEFile.h" 
#include "ELF_File.h" 
#include "ErrorTracker.h" 
#include "ErrorContext.h"

#include "gen/GeneralError.h" 

void usage ()
{
  printf ("usage:\n");
  printf ("  CRACYGLUE filename\n");
}

#include "UrlFolder.h" 
#include "HtmlPage.h" 
#include "HtmlTags.h" 

#include "InstructionEncoder.h" 

string convert_parent_dots(const string& path)
{
  string str = path;
  
  string::size_type f;
  
  printf ("%s\n",str.c_str());
  
  while ((f = str.find("\\")) != str.npos || 
         (f = str.find("/")) != str.npos)
    {
      str.erase(0,f+1);
    }
  printf ("%s\n",str.c_str());
  return str;
}

typedef vector<ExecutableImage*> VectExecutableImage;

VectExecutableImage vectExecutableImages;

bool ReadFile (const char* szFileName)
{
#ifdef WIN32
  int file = open (szFileName,_O_BINARY | _O_RDONLY,_S_IREAD);
#else
  int file = open (szFileName,O_RDONLY);
#endif
  
  if (file == -1)
    {
      EMIT_ERROR(ERROR_GENERAL_FILE_READ_OPEN_FAILED(szFileName));
      return false;
    }
  
  PushErrorContext ec_read(CONTEXT_READING_FILE,szFileName);
  
  ExecutableImage* pImage = NULL;
  
  if (PEFile::IsType (file))
    {
      pImage = new PEFile(szFileName);
      
      printf ("\nReading %s as PE file.\n",szFileName);
    }
  else if (ELF_File::IsType (file))
    {
      pImage = new ELF_File(szFileName);
      
      printf ("\nReading %s as ELF file.\n",szFileName);
    }
  else
    EMIT_ERROR(ERROR_GENERAL_FILE_READ_UNKNOWN_TYPE(szFileName));
  
  if (!pImage)
    goto fail_read;
  
  {
    PushErrorContext ec_read_actual(CONTEXT_READING_ACTUAL);
    
    try
      {
        pImage->ReadFile (file);
      }
    catch (Error* pError)
      {
        EMIT_ERROR(pError);
        EMIT_ERROR(ERROR_GENERAL_FILE_READ_FATAL(szFileName));
        goto fail_read;
      }
  }
  
  {
    PushErrorContext ec_read_post(CONTEXT_READING_POST);
    
    try
      {
        pImage->PostReadFile ();
      }
    catch (Error* pError)
      {
        EMIT_ERROR(pError);
        EMIT_ERROR(ERROR_GENERAL_FILE_READ_FATAL(szFileName));
        goto fail_read;
      }
  }
  
  // Read was successful
  
  {
    vectExecutableImages.push_back(pImage);
    pImage = NULL;
  }
  close(file);
  return true;
  
 fail_read:
  delete pImage;
  close(file);
  return false;
}

#ifdef WIN32
#include <new.h>

int HandleOutOfMemory(size_t t)
{
  printf ("\n\nFATAL: memory allocation error!\n\n");
  exit(1);
}
#endif

int main (int argc,char* argv[])
{
#ifdef WIN32
  _set_new_handler(HandleOutOfMemory); // so we do not have to worry about such things
  _set_new_mode(1); // when malloc fails, we get the same treatment (this way, we don't have to check if we were successfull)
#else
  // On Linux it is seem like an abort is made if new fails and malloc
  // always seem to succeed; we'll get a gp fault or something later
  // on when we actually use the memory and the system can't store our
  // stuff.  Anyway, when memory request fail, we get an more or less
  // cryptic fatal error condition and are dead anyway
#endif
  
  PushErrorContext ec_general(CONTEXT_GENERAL);
  
  InstructionEncoder::InitInverseOpcodeInfo();
#ifndef NDEBUG
  //	VerifyEncoderDecoder();
#endif
  
  printf ("\nexmhcg 0.01\n\n");
  
  if (argc < 2 || !strcmp (argv[1],"-"))
    {
      /*
        usage ();
        return 1;
      */
      
      PushErrorContext ec_process(CONTEXT_WRITING_FILE);
      
      
      printf ("\nCreating and writing file...\n");
      
      ExecutableImage* pImageWrite = CreateSimpleElfExecutable ();
      
      if (!pImageWrite->PrepareWrite ())
        {
          EMIT_ERROR(ERROR_GENERAL_FILE_WRITE_FATAL(pImageWrite->m_strFileName.c_str()));				
        }
      else if (!pImageWrite->WriteFile ())
        {
          EMIT_ERROR(ERROR_GENERAL_FILE_WRITE_FATAL(pImageWrite->m_strFileName.c_str()));				
        }
      
      delete pImageWrite;
      
      return 1;
    }
  
  ReadFile(argv[1]);
  
  PushErrorContext ec_process(CONTEXT_PROCESSING);
  
  UrlFolder folder("report");
  
  FOR_EACH_ITER(VectExecutableImage,vectExecutableImages,image)
    {
      UrlFolder* pParent = 
	folder.AddFolder (convert_parent_dots((*image)->m_strFileName));
      
      if (!(*image)->ProcessFile (pParent))
        {
          EMIT_ERROR(ERROR_GENERAL_FILE_PROCESS_FATAL((*image)->m_strFileName.c_str()));
        }
      
      printf ("\nGenerating report...\n");
      
      UrlFile* pPageMain = pParent->AddIndexFile((*image)->m_strFileName);
      
      HtmlPage page(pPageMain);
      
      page.Header();
      
      (*image)->HTMLFileInfo (pParent);
      
      page <<	p("Contents");
      
      page.LinksToFiles(pParent);
    }
  
  ec_process.Pop();
  ec_process.Push(CONTEXT_TUNING_FILE);
  
  printf ("\nTuning files...\n");
  
  FOR_EACH_ITER(VectExecutableImage,vectExecutableImages,imagew)
    { 
      if (!(*imagew)->TuneImage())
        {
          EMIT_ERROR(ERROR_GENERAL_FILE_TUNE_FATAL((*imagew)->m_strFileName.c_str()));				
        }
    }
  
  ec_process.Pop();
  ec_process.Push(CONTEXT_WRITING_FILE);
  
  printf ("\nWriting files...\n");
  
  FOR_EACH_ITER(VectExecutableImage,vectExecutableImages,imagew)
    { 
      ExecutableImage* pImageWrite = 
	(*imagew)->GetWriteImage("processed.exe");
      
      if (!pImageWrite->PrepareWrite ())
        {
          EMIT_ERROR(ERROR_GENERAL_FILE_WRITE_FATAL(pImageWrite->m_strFileName.c_str()));				
        }
      else if (!pImageWrite->WriteFile ())
        {
          EMIT_ERROR(ERROR_GENERAL_FILE_WRITE_FATAL(pImageWrite->m_strFileName.c_str()));				
        }
      
      delete pImageWrite;
    }
  
  ec_process.Pop();
  
#ifdef WIN32
  printf ("\nFinished. Press key...\n");
  
  getchar ();
#else
  printf ("\nFinished...\n");
#endif
  
  return 0;
}

